/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef UDM_SQLDBMS_H
#define UDM_SQLDBMS_H


/* SQL Data types */
typedef enum
{
  UDM_SQLTYPE_UNKNOWN=       0,
  UDM_SQLTYPE_LONGVARBINARY= 1,
  UDM_SQLTYPE_LONGVARCHAR=   2,
  UDM_SQLTYPE_VARCHAR=       3,
  UDM_SQLTYPE_INT32=         4
} udm_sqltype_t;

const char *UdmSQLTypeToStr(udm_sqltype_t type);

#define UDM_SQL_NULL_DATA	(-1)
#define UDM_SQL_DATA_AT_EXEC	(-2)
#define UDM_SQL_NTS		(-3)

#define UDM_SQL_HAVE_GROUPBY       1
#define UDM_SQL_HAVE_TRUNCATE      2
#define UDM_SQL_HAVE_SUBSELECT     4
#define UDM_SQL_HAVE_LIMIT         8
#define UDM_SQL_HAVE_TOP          16  /* SELECT TOP 10 FROM t1              */
#define UDM_SQL_HAVE_BIND_BINARY  32  /* Bind BINARY/VARBINARY/BLOB         */
#define UDM_SQL_HAVE_0xHEX        64  /* INSERT INTO t1 VALUES (0xAABBCC)   */
#define UDM_SQL_HAVE_STDHEX      128  /* INSERT INTO t1 VALUES (X'AABBCC')  */
#define UDM_SQL_HAVE_ROWNUM      256  /* ROWNUM to limit result size        */
#define UDM_SQL_HAVE_GOOD_COMMIT 512  /* failing query doesn't stop trasaction */
#define UDM_SQL_IGNORE_ERROR    1024
#define UDM_SQL_DEBUG_QUERY     2048
#define UDM_SQL_HAVE_FIRST_SKIP 4096  /* SELECT FIRST 10 * FROM t1 */
#define UDM_SQL_HAVE_TRANSACT   8192  /* If transactions are  supported */
#define UDM_SQL_HAVE_RENAME         0x00004000 /* If can rename tables */
#define UDM_SQL_HAVE_CREATE_LIKE    0x00008000 /* If can CREATE TABLE LIKE */
#define UDM_SQL_HAVE_DROP_IF_EXISTS 0x00010000 /* If can DROP TABLE IF EXISTS */
#define UDM_SQL_HAVE_BLOB_AS_HEX    0x00200000 /* INSERT INTO t1 (blob1) VALUES ('AABBCC') */
#define UDM_SQL_HAVE_BIND_TEXT      0x00400000 /* Bind CHAR/VARCHAR/TEXT */

#define UDM_SQL_HAVE_BIND (UDM_SQL_HAVE_BIND_BINARY|UDM_SQL_HAVE_BIND_TEXT)

#define UDM_SQL_MAX_BIND_PARAM 64

struct udm_sql_st;

typedef struct
{
  char *sqlname;
  int sqltype;
  int sqllen;
} UDM_SQLFIELD;


typedef struct
{
  size_t nRows;
  size_t nCols;
  size_t curRow;
  UDM_SQLFIELD *Fields;
  UDM_STR *Items;	
  void *specific;
  struct udm_sql_st *db;
} UDM_SQLRES;


udm_rc_t UdmSQLResFreeGeneric(UDM_SQLRES *res);

typedef struct udm_sql_res_list_st
{
  size_t nitems;
  UDM_SQLRES *Item;
} UDM_SQLRESLIST;

void UdmSQLResListInit(UDM_SQLRESLIST *List);
udm_rc_t UdmSQLResListAdd(UDM_SQLRESLIST *List, UDM_SQLRES *Res);
void UdmSQLResListFree(UDM_SQLRESLIST *List);

void DecodeHexStr (const char *src, UDM_STR *dst, size_t size);

/****************************************************************/

/* Database type and driver type */
typedef enum
{
  UDM_DB_MYSQL=          2,
  UDM_DB_PGSQL=          3,
  UDM_DB_SOLID=          4,
  UDM_DB_VIRT=           6,
  UDM_DB_IBASE=          7,
  UDM_DB_ORACLE8=        8,
  UDM_DB_MSSQL=         10,
  UDM_DB_SAPDB=         11,
  UDM_DB_DB2=           12,
  UDM_DB_SQLITE=        13,
  UDM_DB_ACCESS=        14,
  UDM_DB_MIMER=         15,
  UDM_DB_CACHE=         16,
  UDM_DB_SYBASE=        17,
  UDM_DB_SQLITE3=       18,
  UDM_DB_MONETDB=       19
} udm_sqldbtype_t;


typedef enum
{
  UDM_DBAPI_ODBC=      1,
  UDM_DBAPI_MYSQL=     2,
  UDM_DBAPI_PGSQL=     3,
  UDM_DBAPI_IBASE=     7,
  UDM_DBAPI_ORACLE8=   8,
  UDM_DBAPI_CTLIB=    10,
  UDM_DBAPI_SQLITE=   13,
  UDM_DBAPI_SQLITE3=  18
} udm_sqldbapi_t;


typedef struct udm_sqldb_handler_st
{
  size_t   (*EscStr)(struct udm_sql_st *db, char *to, const char *from, size_t len);
  udm_rc_t (*Query)(struct udm_sql_st *db, UDM_SQLRES *R, const char *query);
  udm_rc_t (*Connect)(struct udm_sql_st *db);
  udm_rc_t (*Close)(struct udm_sql_st *db);
  udm_rc_t (*Begin)(struct udm_sql_st *db);
  udm_rc_t (*Commit)(struct udm_sql_st *db);
  udm_rc_t (*Prepare)(struct udm_sql_st *db, const char *query);
  udm_rc_t (*Bind)(struct udm_sql_st *db, int position, const void *data, int size, udm_sqltype_t type);
  udm_rc_t (*Exec)(struct udm_sql_st *db);
  udm_rc_t (*StmtFree)(struct udm_sql_st *db);
  udm_rc_t (*FetchRow)(struct udm_sql_st *db, UDM_SQLRES *R, UDM_STR *);
  udm_rc_t (*StoreResult)(struct udm_sql_st *db, UDM_SQLRES *R);
  udm_rc_t (*FreeResult)(struct udm_sql_st *db, UDM_SQLRES *R);
  udm_rc_t (*ExecDirect)(struct udm_sql_st *db, UDM_SQLRES *R, const char *q);
  udm_rc_t (*RenameTable)(struct udm_sql_st *db, const char *from, const char *to);
  udm_rc_t (*CopyStructure)(struct udm_sql_st *db, const char *from, const char *to);
  udm_rc_t (*LockOrBegin)(struct udm_sql_st *db, const char *param);
  udm_rc_t (*UnlockOrCommit)(struct udm_sql_st *db);
} UDM_SQLDB_HANDLER;


typedef struct udm_sql_st
{
  UDM_SQLDB_HANDLER handler;
  void *specific;
  void *ps;                  /* for prepared statements */
  int  errcode;
  char errstr[2048];
  UDM_VARLIST Vars;
  udm_sqldbtype_t DBType;
  udm_sqldbapi_t DBDriver;
  int version; /* vendor specific version */
  int DBSQL_IN;
  int flags;
  size_t res_limit;
  udm_bool_t connected;
} UDM_SQL;


extern const UDM_SQLDB_HANDLER udm_sqldb_ctlib_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_mysql_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_pgsql_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_ibase_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_sqlite_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_sqlite3_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_odbc_handler;
extern const UDM_SQLDB_HANDLER udm_sqldb_oracle_handler;

/*********************************************************************/

udm_rc_t UdmSQLEscDSTR(UDM_SQL *db, UDM_DSTR *dst, const char *src, size_t srclen);

size_t UdmSQLEscStrGeneric(UDM_SQL *db, char *to, const char *from, size_t len);
size_t UdmSQLEscStr(UDM_SQL *db, char *to, const char *from, size_t l);
char*  UdmSQLEscStrAlloc(UDM_SQL *db, const char *src, size_t srclen);
char*  UdmSQLEscStrSimple(UDM_SQL *db, char *to, const char *from, size_t l);
size_t UdmSQLBinEscStr(UDM_SQL *db, char *dst, size_t dstlen, const char *src, size_t srclen);
udm_rc_t UdmSQLTableTruncateOrDelete(UDM_SQL *db, const char *name);
udm_rc_t UdmSQLDropTableIfExists(UDM_SQL *db, const char *name);
udm_rc_t UdmSQLQueryOneRowInt(UDM_SQL *db, int *res, const char *qbuf);

udm_rc_t UdmSQLPrepareGeneric(UDM_SQL *db, const char *query);
udm_rc_t UdmSQLBindGeneric(UDM_SQL *db, int position, const void *data, int size, udm_sqltype_t type);
udm_rc_t UdmSQLExecGeneric(UDM_SQL *db);
udm_rc_t UdmSQLStmtFreeGeneric(UDM_SQL *db);
udm_rc_t UdmSQLLockOrBeginGeneric(UDM_SQL *, const char *param);
udm_rc_t UdmSQLUnlockOrCommitGeneric(UDM_SQL *);

#define UDM_SQL_TOP_BUF_SIZE 64

typedef struct udm_sql_top_st
{
  char rownum[UDM_SQL_TOP_BUF_SIZE];
  char limit[UDM_SQL_TOP_BUF_SIZE];
  char top[UDM_SQL_TOP_BUF_SIZE];
} UDM_SQL_TOP_CLAUSE;
void UdmSQLTopClause(UDM_SQL *db, size_t top_num, UDM_SQL_TOP_CLAUSE *Top);

const char *UdmSQLParamPlaceHolder(UDM_SQL *db, size_t i);
udm_sqltype_t UdmSQLLongVarCharBindType(UDM_SQL *db);

UDM_API(udm_rc_t) UdmSQLQuery(UDM_SQL *db, UDM_SQLRES *R, const char *query);
UDM_API(size_t) UdmSQLNumRows(UDM_SQLRES *res);
size_t		UdmSQLNumCols(UDM_SQLRES *res);
void UdmSQLValueToConstStr(UDM_CONST_STR *str, UDM_SQLRES *res, size_t row, size_t col);
size_t UdmSQLLen(UDM_SQLRES * res,size_t i,size_t j);
UDM_API(const char *) UdmSQLValue(UDM_SQLRES * res,size_t i,size_t j);
UDM_API(udm_rc_t) UdmSQLFetchRowSimple(UDM_SQL *,UDM_SQLRES *res, UDM_STR *pstr);
UDM_API(udm_rc_t) UdmSQLStoreResultSimple(UDM_SQL *,UDM_SQLRES *res);
UDM_API(udm_rc_t) UdmSQLFreeResultSimple(UDM_SQL *,UDM_SQLRES *res);
UDM_API(void) UdmSQLFree(UDM_SQLRES*);
udm_rc_t UdmSQLClose(UDM_SQL *db);
                                                
udm_rc_t UdmSQLBegin(UDM_SQL *db);
udm_rc_t UdmSQLCommit(UDM_SQL *db);
udm_rc_t UdmSQLExecDirect(UDM_SQL *db, UDM_SQLRES *R, const char *query);
udm_rc_t UdmSQLPrepare(UDM_SQL *db, const char *query);
udm_rc_t UdmSQLExecute(UDM_SQL *db);
udm_rc_t UdmSQLStmtFree(UDM_SQL *db);
udm_rc_t UdmSQLBindParameter(UDM_SQL *db, int pos, const void *data, int size, udm_sqltype_t type);
udm_rc_t UdmSQLRenameTable(UDM_SQL *db, const char *from, const char *to);
udm_rc_t UdmSQLCopyStructure(UDM_SQL *db, const char *from, const char *to);
udm_rc_t UdmSQLLockOrBegin(UDM_SQL *db, const char *param);
udm_rc_t UdmSQLUnlockOrCommit(UDM_SQL *db);

#define UDM_SQLMON_DISPLAY_FIELDS	1
#define UDM_SQLMON_DONT_NEED_SEMICOLON  2 /* Execute immediately, for --exec */

typedef enum
{
  udm_sqlmon_mode_batch= 0,
  udm_sqlmon_mode_interactive= 1
} udm_sqlmon_mode_t;

typedef enum
{
  udm_sqlmon_column_flag_raw= 0,
  udm_sqlmon_column_flag_binary= 1,
  udm_sqlmon_column_flag_semibinary= 2
} udm_sqlmon_column_flag_t;

typedef struct udm_sqlmon_param_st
{
  UDM_IOHANDLER iohandler;
  int flags;
  udm_sqlmon_column_flag_t colflags[10];
  size_t currdbnum;
  size_t nqueries;
  size_t ngood;
  size_t nbad;
  size_t lineno;
  udm_sqlmon_mode_t mode;
  void* user_data;
} UDM_SQLMON_PARAM;

UDM_API(udm_rc_t) UdmSQLMonitor(UDM_AGENT *A, UDM_ENV *E, UDM_SQLMON_PARAM *prm);

#endif
