/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_UTILS_H
#define _UDM_UTILS_H

#include <stdio.h>

/* for time_t */
#include <time.h>

/* for va_list */
#include <stdarg.h>

#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>

#include "udm_config.h"
#include "udm_uniconv.h" /* for UDM_CHARSET */
/*
  This is used in UdmTime_t2Str and in its callers.
  Note, on AIX strftime() can return time zone in this format:
  'GMT+01:00'. So timezone can be longer than 3,4 or five-letter
  abbreviation.
*/
#define UDM_MAXTIMESTRLEN	64

#if !defined(__i386__) || defined(_WIN64)
#define udm_put_int4(I, C) do { *((char *)(C))= (char)(I);\
                                *(((char *)(C)) + 1)= (char)((I) >> 8);\
                                *(((char *)(C)) + 2)= (char)((I) >> 16);\
                                *(((char *)(C)) + 3)= (char)((I) >> 24); } while (0)
#define udm_get_int4(C) (int4)(((int4)((unsigned char)(C)[0])) +\
                              (((int4)((unsigned char)(C)[1]) << 8)) +\
                              (((int4)((unsigned char)(C)[2]) << 16)) +\
                              (((int4)((unsigned char)(C)[3]) << 24)))
#else
#define udm_put_int4(I, C) *((int4 *)(C))= (int4)(I)
#define udm_get_int4(C) (*((const int4 *)(C)))
#endif

/* Some useful MACROs */
#define UDM_STREND(s)		(s+strlen(s))
#define UDM_FREE(x)		{if((x)!=NULL){UdmFree(x);x=NULL;}}
#define UDM_SKIP(s,set)		while((*s)&&(strchr(set,*s)))s++;
#define UDM_SKIPN(s,set)	while((*s)&&(!strchr(set,*s)))s++;
#define UDM_SQR(x)		((x)*(x))
#define UDM_CSTR_WITH_LEN(x)  (x), ((size_t) (sizeof(x) - 1))
#define UDM_BZERO(s, len)       bzero((void*) (s), (len))

/* NULL safe atoi*/
#define UDM_ATOI(x)		((x)?atoi(x):0)
#define UDM_ATOF(x)		((x)?atof(x):0.0)
#define UDM_ATOU(x)		((x)?(urlid_t)strtoul(x, (char**)NULL,10):0)
#define UDM_NULL2EMPTY(x)	((x)?(x):&udm_null_char)

/* The total size that malloc() consumes in the heap */
#define UDM_MALLOC_ALLIGNED_SIZE(sz)  ((((sz) + 15) / 16) * 16)

extern char udm_null_char;

/* Misc functions */
UDM_API(udm_rc_t) UdmInit(void);
char *UdmGetStrToken(char * s,char ** last);
char *UdmTrim(char * p, const char * delim);
char *UdmRTrim(char* p, const char * delim);
size_t UdmUnescapeCGIQuery(char *d, const char *s);
size_t UdmURLDecode(char *dst, const char *src, size_t length);
UDM_API(size_t) UdmURLEncode(char *d, const char *s, size_t length);
char *UdmEscapeURI(char *d,const char *s);
char *UdmRemove2Dot(char *path);
char *UdmBuildParamStr(char * dst,size_t len,const char * src,char ** argv,size_t argc);
char *UdmStrRemoveChars(char * str, const char * sep);
char *UdmStrRemoveDoubleChars(char * str, const char * sep);
size_t UdmStrRemoveDoubleSpaces(char *str);

/* This should convert Last-Modified time returned by webserver
 * to time_t (seconds since the Epoch). -kir
 */
time_t UdmHttpDate2Time_t(const char * date);

time_t UdmFTPDate2Time_t(char *date);

/***********************************************************
 * converts time_str to time_t (seconds)
 * time_str can be exactly number of seconds
 * or in the form 'xxxA[yyyB[zzzC]]'
 * (Spaces are allowed between xxx and A and yyy and so on)
 *   there xxx, yyy, zzz are numbers (can be negative!)
 *         A, B, C can be one of the following:
 *		s - second
 *		M - minute	
 *		h - hour
 *		d - day
 *		m - month
 *		y - year
 *	(these letters are as in strptime/strftime functions)
 *
 * Examples:
 * 1234 - 1234 seconds
 * 4h30M - 4 hours and 30 minutes (will return 9000 seconds)
 * 1y6m-15d - 1 year and six month minus 15 days (will return 45792000 s)
 * 1h-60M+1s - 1 hour minus 60 minutes plus 1 second (will return 1 s)
 */
time_t Udm_dp2time_t(const char * time_str);


/* This one for printing HTTP Last-Modified: header */
void UdmTime_t2HttpStr(time_t t, char * time_str, size_t time_str_size);
/* This one deals with timezone offset */
udm_rc_t UdmInitTZ(void);
UDM_API(udm_timer_t) UdmStartTimer(void);
double UdmStopTimer(udm_timer_t *ticks);

size_t UdmStrToSize(const char *str, char **endptr, int *error);

char *UdmStrStore (char *dest, const char *src);

/* Probably string missing functions */

#ifndef HAVE_BZERO
UDM_API(void) bzero(void *b, size_t len);
#endif

#ifndef HAVE_STRCASECMP
UDM_API(int) strcasecmp(const char *s1, const char *s2);
#endif

#ifndef HAVE_STRNCASECMP
int strncasecmp(const char *s1, const char *s2, size_t n);
#endif

#ifndef HAVE_STRCASESTR
char * strcasestr(register const char *s1, register const char *s2);
#endif

#ifndef HAVE_VSNPRINTF
int vsnprintf(char *str, size_t size, const char  *fmt,  va_list ap);
#endif

UDM_API(int) udm_snprintf(char *str, size_t size, const char *fmt, ...)
                          __attribute__((format(printf,3,4)));
#ifndef HAVE_SNPRINTF
#define snprintf udm_snprintf
#endif

int udm_strnncasecmp(const char *s, size_t slen,
                            const char *t, size_t tlen);

char *udm_strtok_r(char *s, const char *delim, char **save_ptr);
#ifndef HAVE_STRTOK_R
#define strtok_r udm_rstok_r
#endif

#if !defined(HAVE_STRNDUP) || defined(EFENCE)
char *udm_strndup(const char *str, size_t len);
#endif
#ifndef HAVE_STRNDUP
#define strndup udm_strndup
#endif

double udm_strntod(const char *s, size_t len);
int udm_strntoi(const char *s, size_t len);
udm_bool_t udm_strntobool(const char *s, size_t len);

const char *udm_memmem(const char *str, size_t str_length,
                       const char *substr, size_t substr_length);

UDM_API(int) UdmHex2Int(int h);
UDM_API(int) UdmInt2Hex(int i);


size_t UdmHexEncode(char *dst, const char *src, size_t len);
size_t UdmHexDecode(char *dst, const char *src, size_t len);

#define BASE64_LEN(len) (4 * (((len) + 2) / 3) +2)
UDM_API(size_t) udm_base64_encode (const char *s, char *store, size_t length);
UDM_API(size_t) udm_base64_decode (char * dst, const char * src, size_t len);
char *udm_rfc1522_decode(char * dst, const char *src);

/* Whether to allow multuple chunks 'AAA= BB== CC==' */
#define UDM_BASE64_ALLOW_MULTIPLE_CHUNKS 1
int UdmBase64Decode(const char *src, size_t len,
                    void *dst0, const char **end, int flags);

size_t udm_quoted_printable_decode(const char *src, size_t srclen,
                                   char *dst, size_t dstlen);

/* Build directory */
int UdmBuild(char * path, int mode);

/* SetEnv */
int UdmSetEnv(const char * name,const char * value);
void UdmUnsetEnv(const char * name);

size_t UdmUniRemoveDoubleSpaces(int *ustr);
void UdmUniPrint(int * ustr);

ssize_t UdmRecvall(int s, void *buf, size_t len);
ssize_t UdmSend(int s, const void *msg, size_t len, int flags);

void UdmWriteLock(int fd);
void UdmUnLock(int fd);
void UdmReadLock(int fd);
void UdmReadLockFILE(FILE *f);
UDM_API(void) UdmWriteLockFILE(FILE *f);
UDM_API(void) UdmUnLockFILE(FILE *f);

/**************************** Variable strings *******************************/
void UdmStrInit(UDM_STR *str);
void UdmStrFree(UDM_STR *str);
udm_rc_t UdmStrMemdup(UDM_STR *str, const char *src, size_t length);
udm_rc_t UdmStrNDup(UDM_STR *str, const char *src, size_t length);
/**************************** Constant string routines ***********************/
void UdmConstStrInit(UDM_CONST_STR *str);
void UdmConstStrSet(UDM_CONST_STR *str, const char *s, size_t len);
void UdmConstStrSetStr(UDM_CONST_STR *str, const char *s);
int UdmConstStrNCaseCmp(const UDM_CONST_STR *str, const char *s, size_t length);
char *UdmConstStrDup(const UDM_CONST_STR *str);
void UdmConstStrTrim(UDM_CONST_STR *str, const char *sep);
/************************ Token **********************/
typedef struct udm_const_token_st
{
  const char *str;
  const char *end;
} UDM_CONST_TOKEN;
#define UdmConstTokenStr(x)              ((x)->str)
#define UdmConstTokenEnd(x)              ((x)->end)
#define UdmConstTokenLength(x)           ((x)->end - (x)->str)
#define UdmConstTokenIsEmpty(x)          ((x)->str == (x)->end)
char *UdmConstTokenStrDup(const UDM_CONST_TOKEN *str);
void UdmConstTokenSet(UDM_CONST_TOKEN *dst, const char *str, size_t length);
/************************ Dynamic strings  ***********/
udm_rc_t UdmDSTRInit(UDM_DSTR *dstr, size_t size_page);
void UdmDSTRFree(UDM_DSTR *dstr);
const char *UdmDSTRPtr(const UDM_DSTR *dstr);
size_t UdmDSTRLength(const UDM_DSTR *dstr);
size_t UdmDSTRAppend (UDM_DSTR *dstr, const char *data, size_t size_data);
size_t UdmDSTRAppendHex(UDM_DSTR *dstr, const char *s, size_t slen);
size_t UdmDSTRAppendSTR (UDM_DSTR *dstr, const char *data);
size_t UdmDSTRAppendINT4(UDM_DSTR *s, int i);
size_t UdmDSTRAppendINT2LE(UDM_DSTR *s, int i);
size_t UdmDSTRAppendINT2BE(UDM_DSTR *s, int i);
udm_rc_t UdmDSTRAppendCoord(UDM_DSTR *s, uint4 coord);
udm_rc_t UdmDSTRAppendRemoveHiLight(UDM_DSTR *dstr, const char *src, size_t length);
udm_rc_t UdmDSTRAppendConv(UDM_DSTR *dstr, UDM_CONV *conv,
                           const char *src, size_t srclen, int flags);
int UdmDSTRAppendf (UDM_DSTR *dstr, const char *fmt, ...)
                    __attribute__((format(printf,2,3)));
void UdmDSTRReset (UDM_DSTR *dstr);
udm_rc_t UdmDSTRAlloc (UDM_DSTR *dstr, size_t size_data);
udm_rc_t UdmDSTRRealloc (UDM_DSTR *dstr, size_t size_data);
udm_rc_t UdmDSTRReserve(UDM_DSTR *dstr, size_t additional_size);
void UdmDSTRShrinkLast(UDM_DSTR *dstr);
double UdmDSTRToDouble(const UDM_DSTR *dstr);
int UdmDSTRToInt(const UDM_DSTR *dstr);
void UdmDSTRToConstStr(const UDM_DSTR *dstr, UDM_CONST_STR *to);
int UdmDSTRCmp(const UDM_DSTR *s1, const UDM_DSTR *s2);
int UdmDSTRCaseCmp(const UDM_DSTR *s1, const UDM_DSTR *s2);
int UdmDSTRWildCaseCmp(const UDM_DSTR *s1, const UDM_DSTR *s2);
udm_rc_t UdmDSTRReadFile(UDM_DSTR *dstr, int df, size_t nbytes);
udm_rc_t UdmDSTRAppendURLDecode(UDM_DSTR *dstr, const char *str, size_t length);
void UdmDSTRRemoveHighlightWithOffset(UDM_DSTR *dstr, UDM_CHARSET *cset, size_t offset);
udm_rc_t UdmDSTRAppendInflate(UDM_DSTR *dstr, const char *src, size_t length);
udm_rc_t UdmDSTRAppendDeflate(UDM_DSTR *dstr, const char *src, size_t length);
void UdmDSTRPCase(UDM_DSTR *dstr);
void UdmDSTRLCase(UDM_DSTR *dstr);
void UdmDSTRUCase(UDM_DSTR *dstr);
void UdmDSTRBCut(UDM_DSTR *dstr);
void UdmDSTRGiveValue(UDM_DSTR *dstr, UDM_STR *to);
udm_rc_t UdmDSTRURLDecode(UDM_DSTR *dstr);
udm_rc_t UdmDSTRHTMLEncode(UDM_DSTR *dstr);
udm_rc_t UdmDSTRAppendURLEncode(UDM_DSTR *dstr, const char *str, size_t length);
udm_rc_t UdmDSTRAppendHTMLEncode(UDM_DSTR *dstr, const char *str, size_t length);
udm_rc_t UdmDSTRAppendRegexReplace(UDM_DSTR *dst,
                                   const char *src, size_t srclen,
                                   const char *pattern, size_t patternlen,
                                   const char *replacement, size_t replacementlen);
udm_rc_t UdmDSTRAppendRegexCut(UDM_DSTR *dst,
                               const char *src, size_t srclen,
                               const char *pattern, size_t patternlen);
udm_rc_t UdmDSTRAppendBase64Encode(UDM_DSTR *dst, const char *str, size_t length);
udm_rc_t UdmDSTRAppendHighlight(UDM_DSTR *dstr,
                                const char *src, size_t srclen,
                                const char *beg, size_t blen,
                                const char *end, size_t elen);
/********************************** zint4 stuff ******************************/
typedef struct st_udm_zint4_state
{
  int prev;
  unsigned char *buf, *end;
  unsigned char bits_left;
} UDM_ZINT4_STATE;

void udm_zint4_init(UDM_ZINT4_STATE *state, char *buf);
void udm_zint4(UDM_ZINT4_STATE *state, int next);
void udm_zint4_finalize(UDM_ZINT4_STATE *state);
int udm_dezint4(const char *buf, int4 *array, int buf_len);
/**************************** end of zint4 stuff *****************************/

extern int udm_l1tolower[256];

#ifdef HAVE_DEBUG
void *UdmBSearch(const void *key, const void *base, size_t nmemb,
                         size_t size,
                         int (*compar)(const void *, const void *));
void UdmSort(void *base, size_t nmemb, size_t size,
                    int(*compar)(const void *, const void *));
#else
#define UdmBSearch bsearch
#define UdmSort   qsort
#endif

#define UDM_MIN(x,y)  ((x) < (y) ? (x) : (y))
#define UDM_MAX(x,y)  ((x) > (y) ? (x) : (y))

typedef enum
{
  UDM_DIRTYPE_CONF,
  UDM_DIRTYPE_SHARE,
  UDM_DIRTYPE_VAR,
  UDM_DIRTYPE_TMP
} udm_dirtype_t;

size_t UdmGetDir(char *d, size_t dlen, udm_dirtype_t type);
size_t UdmGetFileName(char *d, size_t dlen, udm_dirtype_t type,
                             const char *fname);

int UdmNormalizeDecimal(char *dst, size_t dstsize, const char *src);

size_t UdmProcessCurrentResidentSize(void);

/************************************/
typedef struct
{
  char *ptr;
} UDM_MEMROOT_PAGE;


typedef struct
{
  size_t nitems;
  size_t mitems;
  UDM_MEMROOT_PAGE *Item;
  size_t page_size;
  size_t last_page_used_size;
} UDM_MEMROOT;


void UdmMemrootInit(UDM_MEMROOT *root, size_t page_size);
void UdmMemrootFree(UDM_MEMROOT *root);
void UdmMemrootReset(UDM_MEMROOT *root);
size_t UdmMemrootAllocedMemory(const UDM_MEMROOT *root);
size_t UdmMemrootUsedMemory(const UDM_MEMROOT *root);
char *UdmMemrootAlloc(UDM_MEMROOT *root, size_t size);
char *UdmMemrootStrndup(UDM_MEMROOT *root, const char *str, size_t length);

/************************************/

#endif /* _UDM_UTILS_H */
