/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_ZLIB

#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#include "udm_common.h"
#include "udm_contentencoding.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_http.h"


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"
#endif
static void
UdmZInit(z_stream *z, char *dst, size_t dstlen, const char *src, size_t srclen)
{
  z->zalloc= Z_NULL;
  z->zfree= Z_NULL;
  z->opaque= Z_NULL;
  z->next_in= (Byte *) src;
  z->next_out= (Byte *) dst;
  z->avail_in= srclen;
  z->avail_out= dstlen;
}
#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif

size_t
UdmInflate(char *dst, size_t dstlen, const char *src, size_t srclen, udm_rc_t *rc)
{
  z_stream z;
  int zrc;

  *rc= UDM_OK;
  UdmZInit(&z, dst, dstlen, src, srclen);
  
  if (inflateInit2(&z, -MAX_WBITS) != Z_OK)
  {
    *rc= UDM_ERROR;
    return 0;
  }

  if ((zrc= inflate(&z, Z_FINISH)) != Z_OK && zrc != Z_STREAM_END)
    *rc= UDM_ERROR;

  if (inflateEnd(&z) != Z_OK)
  {
    *rc= UDM_ERROR;
    return 0;
  }
  return z.total_out;
}


size_t
UdmDeflate(char *dst, size_t dstlen, const char *src, size_t srclen, udm_rc_t *rc)
{
  z_stream z;
  int zrc;

  *rc= UDM_OK;
  UdmZInit(&z, dst, dstlen, src, srclen);
  
  if (deflateInit2(&z, 9, Z_DEFLATED, -MAX_WBITS, 9, Z_DEFAULT_STRATEGY) != Z_OK)
  {
    *rc= UDM_ERROR;
    return 0;
  }

  if ((zrc= deflate(&z, Z_FINISH)) != Z_OK && zrc != Z_STREAM_END)
    *rc= UDM_ERROR;

  if (deflateEnd(&z) != Z_OK)
  {
    *rc= UDM_ERROR;
    return 0;
  }
  return (size_t) z.total_out;
}


struct gztrailer
{
  int crc32;
  int zlen;
};


size_t
UdmUnGzip(char *dst, size_t dstlen, const char *src, size_t srclen, udm_rc_t *rc)
{
  static const unsigned char gzheader[10]= { 0x1f, 0x8b, Z_DEFLATED, 0, 0, 0, 0, 0, 0, 3 };
  const char *srcdata;
  z_stream zstream;
  int zrc;

  *rc= UDM_OK;

  /* check magic identificator */
  if (srclen < sizeof(gzheader) || memcmp(src, gzheader, 2))
  {
    *rc= UDM_ERROR;
    return 0;
  }

  srcdata= src + sizeof(gzheader);
  srclen-= sizeof(gzheader);

  if (src[3] & 4)
  { /* FLG.FEXTRA */
    size_t xlen= (((size_t) ((unsigned char) srcdata[1])) << 8) + (unsigned char) srcdata[0];
    srcdata+= xlen + 2;
    srclen-= xlen + 2;
  }
  if (src[3] & 8)
  { /* FLG.FNAME */
    while (*srcdata != '\0')
    {
      srcdata++;
      srclen--;
    }
    srcdata++;
    srclen--;
  }
  if (src[3] & 16)
  { /* FLG.FCOMMENT */
    while (*srcdata != '\0')
    {
      srcdata++;
      srclen--;
    }
    srcdata++;
    srclen--;
  }
  if (src[3] & 2)
  { /* FLG.FHCRC */
    srcdata += 2;
    srclen -= 2;
  }

  UdmZInit(&zstream, dst, dstlen, srcdata, srclen - sizeof(struct gztrailer));

  if (Z_OK != inflateInit2(&zstream, -MAX_WBITS))
  {
    *rc= UDM_ERROR;
    return 0;
  }

  if ((zrc= inflate(&zstream, Z_FINISH)) != Z_OK && zrc != Z_STREAM_END)
    *rc= UDM_ERROR;

  if (Z_OK != inflateEnd(&zstream))
  {
    *rc= UDM_ERROR;
    return 0;
  }
  /*
  printf("HERE: srclen=%d dstlen=%d=====\n%.*s\n===\n",
         (int) srclen, (int) zstream.total_out, (int) zstream.total_out, dst);
  */
  return zstream.total_out;
}


size_t
UdmUncompress(char *dst, size_t dstlen, const char *src, size_t srclen, udm_rc_t *rc)
{
  uLong Len= dstlen;
  int res= uncompress((Byte*) dst, &Len, (const Byte *) src, srclen);
  *rc= UDM_OK;
  if (res != Z_OK)
  {
    *rc= UDM_ERROR;
    return 0;
  }
  return (size_t) Len;
}


/*********************************************************/

udm_content_encoding_t UdmContentEncodingID(const char *ce)
{
  if (!strcasecmp(ce, "gzip") || !strcasecmp(ce, "x-gzip"))
    return UDM_CONTENT_ENCODING_GZIP;
  if (!strcasecmp(ce, "deflate"))
    return UDM_CONTENT_ENCODING_DEFLATE;
  if (!strcasecmp(ce, "zlib-compress"))
  {
    /*
      Not an RFC encoding method:
      ZLib's compress() is something different from Unix *.Z compress.
    */
    return UDM_CONTENT_ENCODING_ZLIB_COMPRESS;
  }
  if (!strcasecmp(ce, "identity"))
    return UDM_CONTENT_ENCODING_IDENTITY;
  return UDM_CONTENT_ENCODING_UNKNOWN;

}


udm_rc_t
UdmHTTPBufContentEncodingDecode(UDM_HTTPBUF *Buf, udm_content_encoding_t ce)
{
  udm_rc_t rc= UDM_OK;
  UDM_CONST_STR content;
  UDM_STR buf;

  if (ce == UDM_CONTENT_ENCODING_UNKNOWN)
    return UDM_ERROR;

  if (ce == UDM_CONTENT_ENCODING_IDENTITY)
    return UDM_OK;

  if (UdmHTTPBufContentToConstStr(Buf, &content) ||
      !(buf.str= (char*) UdmMalloc((buf.length= content.length))))
    return UDM_ERROR;

  UdmHTTPBufContentExport(Buf, buf.str);
  UdmHTTPBufTruncateContent(Buf);

  switch (ce)
  {
    case UDM_CONTENT_ENCODING_GZIP:
      rc= UdmHTTPBufUnGzipAppend(Buf, buf.str, buf.length);
      break;
    case UDM_CONTENT_ENCODING_DEFLATE:
      rc= UdmHTTPBufInflateAppend(Buf, buf.str, buf.length);
      break;
    case UDM_CONTENT_ENCODING_ZLIB_COMPRESS:
      rc= UdmHTTPBufUncompressAppend(Buf, buf.str, buf.length);
      break;
    case UDM_CONTENT_ENCODING_IDENTITY:
      rc= UDM_OK;
    case UDM_CONTENT_ENCODING_UNKNOWN:
      rc= UDM_ERROR;
  }
  UdmFree(buf.str);
  return rc; 

}


#endif
