/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_ARPA_NAMESER_H
#include <arpa/nameser.h>
#endif
#ifdef HAVE_RESOLV_H
#include <resolv.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif


#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_url.h"
#include "udm_sdp.h"
#include "udm_vars.h"
#include "udm_searchtool.h"
#include "udm_result.h"
#include "udm_log.h"
#include "udm_agent.h"
#include "udm_proto.h"
#include "udm_host.h"
#include "udm_hash.h"
#include "udm_doc.h"
#include "udm_services.h"
#include "udm_xmalloc.h"
#include "udm_store.h"
#include "udm_match.h"
#include "udm_word.h"


void UdmDBInit(UDM_DB *db)
{
  bzero((void*) db, sizeof(*db));
}


void UdmDBFree(UDM_DB *db)
{
  if (db->dbhandler)
    db->dbhandler->Close(db);
  return;
}


int
UdmDBIsActive(UDM_AGENT *A, size_t num_arg)
{
  size_t num= UdmVarListFindInt(&A->Conf->Vars, "DBLimit", 0);
  return num == 0 || num == num_arg + 1;
}


size_t
UdmEnvCopyErrMsgFromDB(UDM_ENV *Env, UDM_DB *db)
{
  size_t nbytes;
  db->dbhandler->Info(db, Env->errstr, sizeof(Env->errstr), &nbytes, UDM_DBINFO_ERRSTR);
  return nbytes;
}


static void
UdmDBLogError(UDM_AGENT *A, UDM_DB *db)
{
  char errstr[UDM_ERRSTR_SIZE];
  size_t nbytes;
  db->dbhandler->Info(db, errstr, sizeof(errstr), &nbytes, UDM_DBINFO_ERRSTR);
  if (nbytes)
    UdmLog (A, UDM_LOG_ERROR, "DB: %s", errstr);
}


size_t
UdmDBNumBySeed(UDM_ENV *Env, udmhash32_t seed)
{
  return seed * Env->DBList.nitems / 256;
}


/*
  There are a few ways to detect which database to work with:
  
- When passed through URL: http://hostname/search.cgi?cc=1&dbnum=2
- When passed from command line: indexer -Edumpdata -D1
- Autodetection from URL using hash

*/

UDM_API(udm_rc_t)
UdmDocAction(UDM_AGENT *A, UDM_DOCUMENT *D, udm_doccmd_t cmd)
{
  udm_rc_t rc= UDM_ERROR;
  size_t i, dbfrom = 0, dbto= A->Conf->DBList.nitems;

#ifdef USE_TRACE
  fprintf(A->TR, "[%d] URLAction: %d\n", A->handle, cmd);
#endif

  if (cmd == UDM_DOCCMD_GET_CACHED_COPY)
  {
    int dbnum= UdmVarListFindInt(&D->Sections, "dbnum", 0);
    if (dbnum < 0 || dbnum >= A->Conf->DBList.nitems) /* Safety */
      dbnum= 0;
    dbfrom= dbnum;
    dbto= dbnum + 1;
  }
  else if (D != NULL)
  {
    udmhash32_t url_id=UdmVarListFindInt(&D->Sections,"URL_ID", 0);
    udmhash32_t seed= (((url_id) ? url_id :
                       UdmStrHash32(UdmVarListFindStr(&D->Sections, "URL", ""))) & 0xFF);
    dbfrom= dbto= UdmDBNumBySeed(A->Conf, seed);
    dbto++;
  }

  for (i= dbfrom; i < dbto; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];

    UDM_GETLOCK(A, UDM_LOCK_DB);
    if (UDM_OK != (rc= db->dbhandler->DocumentAction(A, db, D, cmd)))
    {
      fflush(stdout); /* TODO34: move UdmLog to the caller level */
      UdmDBLogError(A, db);
    }
    UDM_RELEASELOCK(A, UDM_LOCK_DB);
    if (rc != UDM_OK) break;
  }
  
  return rc;
}


static const char *
UdmQueryActionName(udm_querycmd_t cmd)
{
  switch (cmd)
  {
    case UDM_QUERYCMD_INDEX:               return "Indexing";
    case UDM_QUERYCMD_EXPORT:              return "Exporting";
    case UDM_QUERYCMD_CLEAR:               return "Clearing";
    case UDM_QUERYCMD_TARGETS:             return "Searching targets";
    case UDM_QUERYCMD_EXPIRE:              return "Expiring";
    case UDM_QUERYCMD_WORDSTAT:            return "Calculating word statistics";
    case UDM_QUERYCMD_REFERERS:            return "Listing referrers";
    case UDM_QUERYCMD_DUMPDATA:            return "Dumping data";
    case UDM_QUERYCMD_REWRITE_URLDATA:     return "Rewriting URL data";
    case UDM_QUERYCMD_REWRITE_LIMITS:      return "Rewriting limits";
    case UDM_QUERYCMD_REWRITE_POPULARITY:  return "Rewriting popularity";
    case UDM_QUERYCMD_STATISTICS:
    case UDM_QUERYCMD_FINDWORDS:
    case UDM_QUERYCMD_SUGGEST:
    case UDM_QUERYCMD_DOCINFO:
    case UDM_QUERYCMD_TRACK:
    case UDM_QUERYCMD_CLONES:
    case UDM_QUERYCMD_WORDFORMS:
      break;
  }
  return "QueryActionUnknown";
}


UDM_API(udm_rc_t)
UdmQueryAction(UDM_AGENT *A, UDM_QUERY *Query, udm_querycmd_t cmd)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UDM_ERROR;
  size_t i;

  if (cmd != UDM_QUERYCMD_TARGETS && cmd != UDM_QUERYCMD_STATISTICS)
    UdmLog(A, UDM_LOG_ERROR, "%s", UdmQueryActionName(cmd));

  for (i= 0; i < A->Conf->DBList.nitems; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];
    if (!UdmDBIsActive(A, i))
      continue;
    UDM_GETLOCK(A, UDM_LOCK_DB);
    if (UDM_OK != (rc= db->dbhandler->QueryAction(A, db, Query, cmd)))
      UdmDBLogError(A, db);
    UDM_RELEASELOCK(A, UDM_LOCK_DB);
    if (rc != UDM_OK) break;
  }
  if (cmd != UDM_QUERYCMD_TARGETS && cmd != UDM_QUERYCMD_STATISTICS)
    UdmLog(A, UDM_LOG_ERROR, "%s done\t%.2f", UdmQueryActionName(cmd), UdmStopTimer(&ticks));
  return rc;
}


UDM_API(udm_rc_t)
UdmSrvAction(UDM_AGENT *A, UDM_SERVERLIST *S, udm_srvcmd_t cmd)
{
  udm_rc_t rc= UDM_ERROR;
  size_t i;

  for (i= 0; i < A->Conf->DBList.nitems; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];
    UDM_GETLOCK(A, UDM_LOCK_DB); 
    if (UDM_OK != (rc= db->dbhandler->ServerAction(A, db, S, cmd)))
      UdmDBLogError(A, db);
    UDM_RELEASELOCK(A, UDM_LOCK_DB);
    if (rc != UDM_OK) break;
  }
  return rc;
}


udm_rc_t
UdmDBSetAddr(UDM_DB *db, const char *dbaddr)
{
  udm_rc_t rc= UDM_OK;

  if (db->dbhandler)
    db->dbhandler->Close(db);

  rc= udm_dbhandler_searchd.Init(db, dbaddr);
#ifdef HAVE_SQL /* TODO34 */
  if (rc == UDM_NOTARGET)
    rc= udm_dbhandler_sql.Init(db, dbaddr);
#endif
  return rc;
}


/********************************************************/


void
UdmDBListInit(UDM_DBLIST *List)
{
  bzero((void *) List, sizeof(*List));
}


udm_rc_t
UdmDBListAdd(UDM_DBLIST *List, const char *addr,
             char *err, size_t errlen)
{
  UDM_DB  *db;
  udm_rc_t rc;
  db= List->Item= (UDM_DB*) UdmRealloc(List->Item, (List->nitems+1) * sizeof(UDM_DB));
  db+= List->nitems;
  UdmDBInit(db);
  if (UDM_OK == (rc= UdmDBSetAddr(db, addr)))
  {
    List->nitems++;
  }
  else if (rc == UDM_NOTARGET)
  {
    udm_snprintf(err, errlen, "Unsupported DBAddr");
    rc= UDM_ERROR;
  }
  else
  {
    udm_snprintf(err, errlen, "%s", "UdmDBSetAddr failed");
    UdmDBFree(db);
  }
  return rc;
}


void
UdmDBListFree(UDM_DBLIST *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
  {
    UdmDBFree(&List->Item[i]);
  }
  UDM_FREE(List->Item);
  UdmDBListInit(List);
}


/******************************************/
static const char *
UdmDBActionName(udm_dbcmd_t cmd)
{
  switch (cmd)
  {
    case UDM_DBCMD_CREATE:    return "Creating database";
    case UDM_DBCMD_DROP:      return "Dropping database";
    case UDM_DBCMD_DOCCOUNT:  return "Getting document count";
    case UDM_DBCMD_CRAWLER_FINALIZE:    return "Flushing word cache";
  }
  return "DBActionUnknown";
}


UDM_API(udm_rc_t)
UdmDBAction(UDM_AGENT *A, udm_dbcmd_t cmd)
{
  udm_rc_t rc= UDM_OK;
  size_t i;
  udm_timer_t ticks= UdmStartTimer();
  int loglevel= (cmd == UDM_DBCMD_CRAWLER_FINALIZE) ? UDM_LOG_EXTRA : UDM_LOG_ERROR;

  UdmLog(A, loglevel, "%s", UdmDBActionName(cmd));

  for (i= 0; i < A->Conf->DBList.nitems; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];
    if (!UdmDBIsActive(A, i))
      continue;
    UDM_GETLOCK(A, UDM_LOCK_DB);
    if (UDM_OK != (rc= db->dbhandler->DBAction(A, db, cmd)))
      UdmDBLogError(A, db);
    UDM_RELEASELOCK(A, UDM_LOCK_DB);
    if (rc != UDM_OK)
      break;
  }

  UdmLog(A, loglevel, "%s done\t%.2f", UdmDBActionName(cmd), UdmStopTimer(&ticks));
  return rc;
}
