/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_SQL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_log.h"
#include "udm_hash.h"
#include "udm_coords.h"
#include "udm_word.h"
#include "udm_db_int.h"
#include "udm_doc.h"
#include "udm_server.h"
#include "udm_vars.h"
#include "udm_store.h"
#include "udm_indexcache.h"


static udm_rc_t
UdmFindWordInIndexCachePart(UDM_AGENT *A, UDM_QUERY *Query,
                            UDM_FINDWORD_ARGS *args,
                            UDM_CONV *cnv,
                            const UDM_INVERTED_INDEX_CACHE_PART *part,
                            int flags)
{
  size_t i;
  UDM_UNIDATA *unidata= A->Conf->unidata;
  const UDM_WIDEWORD *WW= &Query->Res.WWList.Word[args->Word.Param.order];
  UDM_URLCRDLIST CoordList;
  UDM_URL_CRD CoordTemplate;
  UDM_URLID_LIST *urls= (flags & UDM_RAWBLOB_DELTA) ?
                         &args->live_update_active_urls : &args->urls;

  bzero((void*)&CoordList, sizeof(CoordList));
  bzero((void*)&CoordTemplate, sizeof(CoordTemplate));
  CoordTemplate.num= args->Word.Param.order;

  for (i= 0; i < part->nitems; i++)
  {
    UDM_INVERTED_INDEX_CACHE_ITEM *Item= &part->Item[i];    
    uint4 word_length= strlen(Item->ptr);
    /* TODO34: fuzzy search: accents, substrings, etc */
    if (!UdmStrCaseCmp2(unidata, cnv,
                       Item->ptr, word_length,
                       WW->Word.str, WW->Word.length,
                       UDM_RECODE_HTML))
    {
      CoordList.acoords+= Item->length - word_length - 1;
    }
  }
  if (!(CoordList.Coords= (UDM_URL_CRD *) UdmMalloc((CoordList.acoords) *
                                                    sizeof(UDM_URL_CRD))))
  {
    UdmLog(A, UDM_LOG_ERROR,
           "UdmFindWordInIndexCachePart: UdmMalloc failed");
    return UDM_ERROR;
  }
  for (i= 0; i < part->nitems; i++)
  {
    UDM_INVERTED_INDEX_CACHE_ITEM *Item= &part->Item[i];
    uint word_length= strlen(Item->ptr);
    unsigned const char *src= (unsigned char *) Item->ptr + word_length + 1;
    unsigned const char *srcend= (unsigned char*) Item->ptr + Item->length;
    size_t ncoords, nbytes= udm_coord_get(&ncoords, src, srcend);
    if (!nbytes || nbytes > srcend - src)
    {
      UdmLog(A, UDM_LOG_ERROR,
             "UdmFindWordInIndexCachePart: unexpected ncoords: %d/%d",
             (int) ncoords, (int) (srcend - src));
      return UDM_ERROR;
    }
    src+= nbytes;
    /* TODO34: fuzzy search: accents, substrings, etc */
    if (!UdmStrCaseCmp2(unidata, cnv,
                        Item->ptr, word_length,
                        WW->Word.str, WW->Word.length,
                        UDM_RECODE_HTML))
    {
      CoordTemplate.urlid_coord.url_id= Item->url_id;
      CoordTemplate.urlid_coord.coord.secno= Item->secno;
      /*
      fprintf(stderr, "[%d:%d]%s length=%d ncoords=%d\n",
              Item->url_id, Item->secno, Item->data.str,
              (int) Item->data.length, Item->data.str[word_length+1]);
      */
      UdmCoordListMultiUnpack(&CoordList, &CoordTemplate,
                              (const unsigned char*) src, srcend - src,
                              args->query_param.SaveSectionSize);
    }
  }  
  if (urls->nurls)
    UdmApplyFastLimit(&CoordList, urls);
  UdmLog(A, UDM_LOG_DEBUG, "Raw cache: %d coords found", (int) CoordList.ncoords);
  if (CoordList.ncoords)
  {
    /*
      We have to sort here, because DBMode=multi
      returns data unsorted.
    */
    UdmURLCRDListSortByURLThenSecnoThenPos(&CoordList);
    UdmURLCRDListListAddWithSort2(&args->SearchSectionListList,
                                  &Query->Res.WWList,
                                  &args->Word, &CoordList);
  }
  else
  {
    UdmFree(CoordList.Coords);
  }
  args->Word.Param.count+= CoordList.ncoords;

  return UDM_OK;
}


static udm_rc_t
UdmFindWordRawBlobInternal(UDM_AGENT *A, UDM_QUERY *Query,
                           UDM_FINDWORD_ARGS *args, int flags)
{
  UDM_CONV cnv;
  UDM_UNIDATA *unidata= A->Conf->unidata;
  unsigned int i, tmin, tmax;
  const UDM_WIDEWORD *WW= &Query->Res.WWList.Word[args->Word.Param.order];

  if (!args->IndexCache.nitems)
    return UDM_OK;

  UdmConvInit(&cnv, A->Conf->lcs, A->Conf->lcs);

  if (args->Word.Param.match_mode != UDM_MATCH_FULL)
  {
    /* Scan all cache parts for substring search */
    tmin= 0;
    tmax= args->IndexCache.nitems;
  }
  else
  {
    /* TODO34: StripAccents */
    udmhash32_t crc= cnv.to->cset->crc32lcase(unidata, cnv.to,
                                              WW->Word.str,
                                              WW->Word.length,
                                              UDM_RECODE_HTML);
    tmin= crc % args->IndexCache.nitems;
    tmax= tmin + 1;
  }

  for (i= tmin;  i < tmax; i++)
  {
    UDM_INVERTED_INDEX_CACHE_PART *part= &args->IndexCache.Item[i];
    UdmFindWordInIndexCachePart(A, Query, args, &cnv, part, flags);
  }
  return UDM_OK;
}


static udm_rc_t
UdmRawBlobInitSearch(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                     UDM_FINDWORD_ARGS *args)
{
  udm_rc_t rc;
  int ndocs= 0, nskip= 0;
  size_t i;
  char qbuf[4096];
  UDM_SQLRES SQLRes;
  UDM_STR row[2];
  UDM_INVERTED_INDEX_CACHE *cache= &args->IndexCache;
  size_t max_doc_size= UdmVarListFindInt(&A->Conf->Vars, "MaxDocSize", UDM_MAXDOCSIZE);

  UdmInvertedIndexCacheInit(cache, A->Conf);
  if (UDM_OK != UdmInvertedIndexCacheAllocParts(A, cache,
                                                INVERTED_INDEX_CACHE_PARTS))
    return UDM_ERROR;

  if (args->where[0])
  {
    udm_snprintf(qbuf, sizeof(qbuf),
                 "SELECT d.url_id,d.content FROM cachedcopy d, url%s "
                 "WHERE url.rec_id=d.url_id "
                 "AND ts>=%d "
                 "AND %s",
                 UdmSQLDBQueryFrom(Query),
                 (int) args->live_updates_ts, args->where);
  }
  else
  {
    udm_snprintf(qbuf, sizeof(qbuf),
                 "SELECT url_id,content FROM cachedcopy "
                 "WHERE ts>=%d",
                 (int) args->live_updates_ts);
  }
  if (UDM_OK != (rc= UdmDBSQLExecDirect(A, db, &SQLRes, qbuf)))
    goto end;
  for (ndocs= nskip= 0; 
       UDM_OK == UdmDBSQLFetchRow(A, db, &SQLRes, row);
       ndocs++)
  {
    UDM_DOCUMENT Doc;
    UDM_CONSTWORDLIST CWL;
    UDM_CONSTWORD_HASH_DATA data;
    UDM_CHARSET *doccs;
    udm_timer_t timer_unpack= 0;
    udm_timer_t timer_parse= 0;
    udm_timer_t timer_words= 0;

    if (UdmMemrootUsedMemory(&cache->coord_root) >= 10*1024*1024) /* TODO34: template variable */
    {
      nskip++;
      continue;
    }
    UdmDocInit(&Doc);
    UdmSpiderParamInit(&Doc.Spider);
    Doc.Spider.robots.follow= UDM_FALSE;
    UdmVarListReplaceLst(&Doc.Sections, &A->Conf->Sections, NULL, "*");
    
    if (UDM_OK != (rc= UdmDocSetFromCachedHTTPResponse(&Doc,
                                                       row[1].str,
                                                       row[1].length,
                                                       max_doc_size,
                                                       &timer_unpack)))
      goto end2;

    doccs= UdmDocDetectCachedCharset(A, &Doc, data.url_id);
    bzero(&data, sizeof(data));
    data.url_id= atoi(row[0].str);
    UdmConvInit(&data.cnv, doccs, A->Conf->lcs);
    data.cache= cache;
    cache->Words.hash.user_data= &data;
    UdmConstWordListInit(&CWL);
    if (UDM_OK != (rc= UdmDocToConstWordList(A, &Doc, &CWL,
                                             doccs,
                                             cache->param.cnvflags,
                                             cache->param.aggregate_section_flags,
                                             &timer_parse, &timer_words)))
      goto end2;

    if (CWL.nitems)
    {
      UDM_INVERTED_INDEX_COORD_LIST CL;
      if (UDM_OK != (rc= UdmConstWordListToInvertedIndexCoordList(cache, &CL, &CWL)))
        goto end2;
      UdmInvertedIndexCoordListSort(&CL);
      rc= UdmInvertedIndexCoordList2InvertedIndexCache(A, &CL, &CWL, &data, cache);
      UdmInvertedIndexCoordListFree(&CL);
    }
end2:
    UdmConstWordListFree(&CWL);
    UdmDocFree(&Doc);
    if (rc != UDM_OK)
      break;
  }
  UdmSQLFree(&SQLRes);
end:
  UdmLog(A, UDM_LOG_DEBUG,
         "RawBlob: %d cached documents loaded (%lld bytes)",
          ndocs, (long long int) UdmMemrootUsedMemory(&cache->coord_root));
  if (nskip)
    UdmLog(A, UDM_LOG_DEBUG,
           "RawBlob: %d documents were skipped", nskip);
  for (i= 0; i < cache->nitems; i++)
    UdmInvertedIndexCachePartSort(&cache->Item[i]);
  return rc;
}


udm_rc_t
UdmFindWordRawBlobDelta(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                        UDM_FINDWORD_ARGS *args)
{
  return UdmFindWordRawBlobInternal(A, Query, args, UDM_RAWBLOB_DELTA);
}


static udm_rc_t
UdmFindWordRawBlobSearch(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                         UDM_FINDWORD_ARGS *args)
{
  return UdmFindWordRawBlobInternal(A, Query, args, UDM_RAWBLOB_SEARCH);
}



const UDM_DBMODE_HANDLER udm_dbmode_handler_rawblob=
{
  "rawblob",
  NULL,                      /* StoreWords        */
  NULL,                      /* QueryAction       */
  NULL,                      /* DeleteWordFromURL */
  UdmFindWordRawBlobSearch,  /* FindWord          */
  NULL,                      /* DumpWordInfo      */
  UdmRawBlobInitSearch,
};

#endif /* HAVE_SQL */
