/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/
#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "udm_common.h"
#include "udm_utils.h"
#include "udm_getopt.h"


static const UDM_CMDLINE_OPT *
UdmCmdLineFindShortOption(const UDM_CMDLINE_OPT *options, const char *name)
{
  const UDM_CMDLINE_OPT *opt;
  for (opt= options ; opt->name; opt++)
  {
    if (opt->id && *name == opt->id)
      return opt;
  }
  return NULL;
}


static int
optend(const char *a)
{
  return *a == '=' || *a == '\0';
}

static int
optcmp(const char *a, const char *b)
{
  for ( ; !optend(a); a++, b++)
  {
    if (*a != *b)
      break;
  }
  return !(optend(a) && optend(b));
}


static const UDM_CMDLINE_OPT *
UdmCmdLineFindLongOption(const UDM_CMDLINE_OPT *options, const char *name)
{
  const UDM_CMDLINE_OPT *opt;
  if (!*name)
    return NULL;
  for (opt= options ; opt->name; opt++)
  {
    if (!optcmp(opt->name, name))
      return opt;
  }
  return NULL;
}


static const char *
udm_opt_value_str(const UDM_CMDLINE_OPT *opt, int is_long)
{
  if (is_long)
  return
    opt->type == UDM_OPT_INT ? "=#" :
    opt->type == UDM_OPT_STR ? "=name" :
    "";

  return
    opt->type == UDM_OPT_INT ? " #" :
    opt->type == UDM_OPT_STR ? " name" :
    "";
}


UDM_API(void)
UdmCmdLineOptionsPrint(const UDM_CMDLINE_OPT *options, FILE *file)
{
  const UDM_CMDLINE_OPT *opt;
  for (opt= options; opt->name; opt++)
  {
    char command[40]= "";
    const char *comment;
    size_t clen= 0;
    if (!opt->comment)
      continue;
    if (opt->type == UDM_OPT_TITLE)
    {
      fprintf(file, "%s\n", opt->comment);
      continue;
    }
    if ((opt->id >= 0x20) && (opt->id <= 0x7E))
      clen= udm_snprintf(command, 20, "-%c%s%s", 
                         opt->id,
                         udm_opt_value_str(opt, 0),
                         opt->name[0] ? ", " : "");
    if (opt->name[0])
      udm_snprintf(command + clen, sizeof(command) - clen, "--%s%s",
                   opt->name, udm_opt_value_str(opt, 1));
    fprintf(file, "  %-15s ", command);
    for (comment= opt->comment; *comment; comment++)
    {
      if (*comment == '\r')
        ;
      else if (*comment == '\n')
        fprintf(file, "\n                  ");
      else
        fprintf(file, "%c", *comment);
    }
    printf("\n");
  }
}


UDM_API(int)
UdmCmdLineOptionsGet(void *user_data, int argc, char **argv,
                     const UDM_CMDLINE_OPT *options,
                     int(*handle)(void *user_data,
                                  const UDM_CMDLINE_OPT *opt,
                                  const char *value),
                     size_t *n)
{
  int error= 0;
  size_t i;
  const char *cmd= argv[0];
  UDM_ASSERT(argc >= 1);
  UDM_ASSERT(argv);
  argc--;
  argv++;
  
  for (i= 0; i < (size_t) argc && !error; i++)
  {
    const char *av= argv[i];
    const UDM_CMDLINE_OPT *opt;
    if (av[0] != '-' || av[1] == '\0') /* av[1]=0 example: "cmd -"  */
      break;
    /*printf("%s\n", av);*/
    if (av[1] == '-')
    {
      const char *eq;
      av+= 2;
      if (!(opt= UdmCmdLineFindLongOption(options, av)))
      {
        printf("%s: unknown option '--%s'\n", cmd, av);
        error++;
        break;
      }
      if (!(eq= strchr(av, '=')) && opt->type != UDM_OPT_BOOL)
      {
        if (i < argc)
        {
          i++;
          handle(user_data, opt, argv[i]);
          continue;
        }
        else
        {
          printf("%s: option '--%s' requires an argument\n", cmd, av);
          error++;
          break;
        }
      }
      if (eq && opt->type == UDM_OPT_BOOL)
      {
        printf("%s: option '--%s' cannot have an argument\n", cmd, av);
        error++;
        break;
      }
      handle(user_data, opt, eq ? eq + 1 : opt->type == UDM_OPT_BOOL ? "TRUE" : NULL);
    }
    else
    {
      av+= 1;
      while ((opt= UdmCmdLineFindShortOption(options, av)))
      {
        switch (opt->type)
        {
          case UDM_OPT_BOOL:
            error= handle(user_data, opt, "TRUE");
            av++;
            continue; /* while */

          case UDM_OPT_INT:
          case UDM_OPT_STR:
            if (av[1])
              error= handle(user_data, opt, av + 1);
            else if (i + 1 < argc)
            {
              i++;
              error= handle(user_data, opt, (av= argv[i]));
            }
            else
            {
              printf("%s: option '-%c' requires an argument\n", cmd, *av);
              error++;
            }
            av= "";
            break;
        }
        break;
      }
      
      if (*av)
      {
        printf("%s: unknown option '-%c'\n", cmd, *av);
        error++;
        break;
      }
    }
  }
  *n= i + 1;
  return error;
}
