/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#ifdef WIN32
#include <time.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_ctype.h"
#include "udm_log.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_robots.h"
#include "udm_db.h"
#include "udm_url.h"
#include "udm_parser.h"
#include "udm_proto.h"
#include "udm_hrefs.h"
#include "udm_crc32.h"
#include "udm_hash.h"
#include "udm_xmalloc.h"
#include "udm_http.h"
#include "udm_host.h"
#include "udm_server.h"
#include "udm_alias.h"
#include "udm_word.h"
#include "udm_parsehtml.h"
#include "udm_parsexml.h"
#include "udm_spell.h"
#include "udm_execget.h"
#include "udm_agent.h"
#include "udm_match.h"
#include "udm_doc.h"
#include "udm_result.h"
#include "udm_parsedate.h"
#include "udm_unicode.h"
#include "udm_contentencoding.h"
#include "udm_vars.h"
#include "udm_guesser.h"
#include "udm_textlist.h"
#include "udm_id3.h"
#include "udm_wild.h"
#include "udm_query.h"
#ifdef HAVE_ZLIB
#include "udm_store.h"
#endif

/* This should be last include */
#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define UDM_THREADINFO(A,s,m)	if(A->Conf->ThreadInfo)A->Conf->ThreadInfo(A,s,m)

/***************************************************************************/

static size_t UdmDocPerSiteCached(UDM_AGENT *Indexer,
                                  const char *hostinfo, size_t hostinfo_length)
{
  size_t i, res= 0;
  UDM_HREFLIST *Hrefs;
  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_HREF_CACHE);
  for (Hrefs= &Indexer->Conf->Hrefs, i= 0; i < Hrefs->nhrefs; i++)
  {
    UDM_HREF *Href= &Hrefs->Href[i];
    if (Href->Param.stored && Href->Param.method != UDM_METHOD_DISALLOW &&
        !strncmp(Href->url, hostinfo, hostinfo_length))
      res++;
  }
  return res;
}

static udm_rc_t
UdmCheckDocPerSite(UDM_AGENT *Indexer, UDM_HREF *Href, size_t *res,
                   const char *hostinfo, size_t hostinfo_length)
{
  udm_rc_t rc= UDM_OK;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_HREF_CACHE);
  *res= UdmDocPerSiteCached(Indexer, hostinfo, hostinfo_length);

  if (*res < Href->Param.max_doc_per_site)
  {
    size_t doc_per_site_db;
    UDM_DOCUMENT Doc;
    UdmDocInit(&Doc);
    UdmVarListAddStr(&Doc.Sections, "Hostinfo", hostinfo);
    rc= UdmDocAction(Indexer, &Doc, UDM_DOCCMD_DOCPERSITE);
    doc_per_site_db= UdmVarListFindInt(&Doc.Sections, "DocPerSite", 0);
    UdmDocFree(&Doc);
    *res+= doc_per_site_db;
  }
  return rc;
}


static udm_rc_t
UdmHrefsCheckDocPerSite(UDM_AGENT *Indexer)
{
  size_t i;
  udm_rc_t rc= UDM_OK;
  UDM_HREFLIST *Hrefs;
  size_t doc_per_site= 0;
  size_t hostinfo_length= 0;
  char hostinfo[128]= "";
  
  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_HREF_CACHE);
  
  Hrefs= &Indexer->Conf->Hrefs;

  for (i= 0; i < Hrefs->nhrefs; i++)
  {
    UDM_HREF *H = &Hrefs->Href[i];
    if (!H->Param.stored)
    {
      if (H->Param.max_doc_per_site)
      {
        if (!hostinfo[0] || strncmp(hostinfo, H->url, hostinfo_length))
        {
          /* New site */
          UDM_URL url;
          UdmURLInit(&url);
          UdmURLParse(&url, H->url);
          hostinfo_length= udm_snprintf(hostinfo, sizeof(hostinfo),
                                       "%s://%s/", url.schema, url.hostinfo);
          rc= UdmCheckDocPerSite(Indexer, H, &doc_per_site,
                                 hostinfo, hostinfo_length);
          UdmLog(Indexer, UDM_LOG_DEBUG, "DocPerSite: %d/%d",
                 (int) doc_per_site, (int) H->Param.max_doc_per_site);
          UdmURLFree(&url);
      
          if (rc != UDM_OK)
            goto ret;
        }
        else
        {
          /* Same site */
          doc_per_site++;
        }
        if (doc_per_site > H->Param.max_doc_per_site)
        {
          UdmLog(Indexer, UDM_LOG_DEBUG,
                 "Too many docs (%d) per site, skip it", (int) doc_per_site);
          H->Param.method= UDM_METHOD_DISALLOW;
          H->Param.stored= UDM_TRUE;
        }
      }
    }
  }
ret:

  return rc;
}


static udm_rc_t
UdmHrefCopyNotStored(UDM_HREFLIST *dst, UDM_HREF *Href,
                     const char *url, int link_sources_to_follow)
{
  /* Don't put self links into the shared cache */
  if (Href->Param.method != UDM_METHOD_DISALLOW && strcmp(url, Href->url) &&
     (link_sources_to_follow & (1 << Href->Param.link_source)))
  {
    udm_rc_t rc;
    if (UDM_OK != (rc= UdmHrefListAdd(dst, Href)))
      return rc;
    Href->Param.stored= UDM_TRUE;
  }
  return UDM_OK;
}


static udm_rc_t
UdmHrefListCopyNotStored(UDM_HREFLIST *dst, UDM_HREFLIST *src,
                         const char *url,
                         int link_sources_to_follow)
{
  size_t i;
  for (i= 0; i< src->nhrefs; i++)
  {
    udm_rc_t rc;
    UDM_HREF  *Href= &src->Href[i];
    if (UDM_OK != (rc= UdmHrefCopyNotStored(dst, Href, url, link_sources_to_follow)))
      return rc;
  }
  return UDM_OK;
}


#define MAXHSIZE	1023*4	/* TUNE */


typedef struct
{
  size_t nitems;
  UDM_HREFLIST *Item;
} UDM_HREFLISTLIST;


static udm_rc_t
UdmHrefListListInit(UDM_HREFLISTLIST *List, size_t nitems)
{
  size_t i;
  List->nitems= 0;
  if (!(List->Item= (UDM_HREFLIST*) UdmMalloc(nitems * sizeof(UDM_HREFLIST))))
    return UDM_ERROR;
  for (i= 0; i <  nitems; i++)
    UdmHrefListInit(&List->Item[i]);
  List->nitems= nitems;
  return UDM_OK;
}


static void
UdmHrefListListFree(UDM_HREFLISTLIST *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
    UdmHrefListFree(&List->Item[i]);
  UdmFree(List->Item);
}


static udm_rc_t
UdmStoreHrefsDB(UDM_AGENT *Indexer, UDM_DB *db, UDM_HREFLIST *HrefList)
{
  size_t i;
  udm_rc_t rc;

  if (!HrefList->nhrefs)
    return UDM_OK;
  if (UDM_OK != (rc= db->dbhandler->HrefAction(Indexer, db, NULL, UDM_HREFCMD_START_BULK)))
    return rc;
  for (i= 0; i < HrefList->nhrefs; i++)
  {
    UDM_HREF *H= &HrefList->Href[i];
    if (UDM_OK != (rc= db->dbhandler->HrefAction(Indexer, db, H, UDM_HREFCMD_ADD)))
      return rc;
  }
  if (UDM_OK != (rc= db->dbhandler->HrefAction(Indexer, db, NULL, UDM_HREFCMD_STOP_BULK)))
    return rc;

  for (i= 0; i < HrefList->nhrefs; i++)
  {
    UDM_HREF *H= &HrefList->Href[i];
    if (UDM_OK != (rc= db->dbhandler->HrefAction(Indexer, db, H, UDM_HREFCMD_EXPORT)))
      return rc;
  }
  return UDM_OK;
}


udm_rc_t
UdmStoreHrefs(UDM_AGENT *Indexer)
{
  udm_rc_t rc= UDM_ERROR;
  UDM_HREFLISTLIST HrefListList;
  size_t i;

  /*
    Copy links to store into a separate temporary list,
    to avoid locking UDM_LOCK_HREF_CACHE for a long time.
  */
  if (UDM_OK != (rc= UdmHrefListListInit(&HrefListList, Indexer->Conf->DBList.nitems)))
    return rc;

  UDM_GETLOCK(Indexer, UDM_LOCK_HREF_CACHE);

  if (UDM_OK != (rc= UdmHrefsCheckDocPerSite(Indexer)))
    goto ret; 

  for (i= 0; i < Indexer->Conf->Hrefs.nhrefs; i++)
  {
    UDM_HREF *Href= &Indexer->Conf->Hrefs.Href[i];
    udmhash32_t seed= UdmStrHash32(Href->url) & 0xFF;
    size_t dbnum= UdmDBNumBySeed(Indexer->Conf, seed);
    UDM_HREFLIST *HrefList=  &HrefListList.Item[dbnum];
    if (!Href->Param.stored)
    {
      if (UDM_OK != (rc= UdmHrefListAdd(HrefList, Href)))
        goto ret;
      Href->Param.stored= UDM_TRUE;
    }
  }

  if (Indexer->Conf->Hrefs.nhrefs > MAXHSIZE)
    UdmHrefListFree(&Indexer->Conf->Hrefs);

ret:
  UDM_RELEASELOCK(Indexer, UDM_LOCK_HREF_CACHE);


  /* Now store the links from the temporary list into the database */
  if (rc == UDM_OK)
  {
    for (i= 0; i < Indexer->Conf->DBList.nitems; i++)
    {
      UDM_DB *db= &Indexer->Conf->DBList.Item[i];
      UDM_GETLOCK(Indexer, UDM_LOCK_DB);
      if (UDM_OK != (rc= UdmStoreHrefsDB(Indexer, db, &HrefListList.Item[i])))
        break;
      UDM_RELEASELOCK(Indexer, UDM_LOCK_DB);
    }
  }
  
  UdmHrefListListFree(&HrefListList);
  return rc;
}


static void
RelLink(const UDM_SPIDERPARAM *Spider,
        UDM_URL *curURL, UDM_URL *newURL, char **str)
{
  const char  *schema = newURL->schema ? newURL->schema : curURL->schema;
  const char  *hostinfo = newURL->hostinfo ? newURL->hostinfo : curURL->hostinfo;
  const char  *path = (newURL->path && newURL->path[0]) ? newURL->path : curURL->path;
  const char  *fname = ((newURL->filename && newURL->filename[0]) || 
                       (newURL->path && newURL->path[0])) ? 
                       newURL->filename : curURL->filename;
  char    *pathfile;

  if (newURL->filename && *newURL->filename == '?' && (! newURL->path || ! newURL->path[0]))
  {
    char *qm = strchr(UDM_NULL2EMPTY(curURL->filename), '?');
    char qf = 0;
    if (qm) { *qm = 0; qf = 1; }
    pathfile = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(path)) + strlen(UDM_NULL2EMPTY(fname)) + strlen(UDM_NULL2EMPTY(curURL->filename)) + 5);
    if (pathfile == NULL) return;
    sprintf(pathfile, "/%s%s%s",  UDM_NULL2EMPTY(path), UDM_NULL2EMPTY(curURL->filename), UDM_NULL2EMPTY(fname));
    if (qf) *qm = '?';
  }
  else
  {
    pathfile = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(path)) + strlen(UDM_NULL2EMPTY(fname)) + 5);
    if (pathfile == NULL) return;
    sprintf(pathfile, "/%s%s",  UDM_NULL2EMPTY(path), UDM_NULL2EMPTY(fname));
  }

  
  UdmURLNormalizePath(pathfile);

  if (!strcasecmp(UDM_NULL2EMPTY(schema), "mailto") || 
      !strcasecmp(UDM_NULL2EMPTY(schema), "javascript")) 
  {
    *str = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(UDM_NULL2EMPTY(newURL->specific)) + 4);
    sprintf(*str, "%s:%s", UDM_NULL2EMPTY(schema), UDM_NULL2EMPTY(newURL->specific));
  }
  else if(!strcasecmp(UDM_NULL2EMPTY(schema), "htdb"))
  {
    *str = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(pathfile) + 4);
    sprintf(*str, "%s:%s", UDM_NULL2EMPTY(schema), pathfile);
  }
  else
  {
    const char *anchor= newURL->anchor && newURL->anchor[1] == '!' ? 
                        newURL->anchor : "";
    size_t nbytes= strlen(UDM_NULL2EMPTY(schema)) + strlen(pathfile) +
                   strlen(UDM_NULL2EMPTY(hostinfo)) + strlen(anchor) + 4;
    *str= (char*) UdmMalloc(nbytes + 1);
    udm_snprintf(*str, nbytes, "%s://%s%s%s",
                 UDM_NULL2EMPTY(schema), UDM_NULL2EMPTY(hostinfo),
                 pathfile, Spider->ajax_links ? anchor : "");
  }
  
  if(!strncmp(*str, "ftp://", 6) && (strstr(*str, ";type=")))
    *(strstr(*str, ";type")) = '\0';
  UDM_FREE(pathfile);
}

static udm_rc_t
UdmDocBaseHref(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  const char  *basehref=UdmVarListFindStr(&Doc->Sections,"base.href",NULL);
  /*
    <BASE HREF="xxx"> stuff
    Check that URL is properly formed baseURL is just temporary variable
    If parsing  fails we'll use old base href, passed via CurURL
    Note that we will not check BASE if delete_no_server is unset
    This is  actually dirty hack. We must check that hostname is the same
  */

  if (basehref)
  {
    UDM_URL    baseURL;
    udm_rc_t parse_res;
    
    UdmURLInit(&baseURL);
    
    if (UDM_OK == (parse_res= UdmURLParse(&baseURL, basehref)) &&
        UDM_OK == (parse_res= (baseURL.schema ? UDM_OK : UDM_ERROR)))
    {
      UdmURLParse(&Doc->CurURL, basehref);
      UdmLog(Indexer, UDM_LOG_DEBUG, "BASE HREF '%s'", basehref);
    }
    else
    {
      UdmLog(Indexer, UDM_LOG_ERROR, "BASE HREF: %s: '%s'",
                                      UdmURLErrorStr(parse_res), basehref);
    }
    UdmURLFree(&baseURL);
  }
  return UDM_OK;
}


static udm_method_t
UdmFilterFind(UDM_FILTERLIST *L,const char *newhref,char *reason)
{
  const UDM_FILTER *F;

  UdmDebugEnter();
  
  if ((F= UdmFilterListFind(L, newhref)))
  {
    sprintf(reason,"%s %s %s '%s'",
            UdmMethodStr(F->method), UdmMatchModeStr(F->Match.Param.match_mode),
            UdmMatchCaseSensitivityStr(&F->Match),
            UdmMatchPatternConstStr(&F->Match));
  }
  else
  {
    sprintf(reason,"Allow by default");
  }
  UdmDebugReturn(F ? F->method : UDM_METHOD_GET);
}


static const UDM_SECTIONFILTER *
UdmSectionFilterListFindWithSubst(const UDM_SECTIONFILTERLIST *L,
                                  const UDM_DOCUMENT *Doc,
                                  size_t nparts, UDM_MATCH_PART *Parts)
{
  UDM_DSTR dstr;
  size_t i, srclen;
  UdmDSTRInit(&dstr, 128);
  for(i=0; i < L->nitems; i++)
  {
    UDM_SECTIONFILTER *F= &L->Item[i];
    const char *src;
    if (strchr(F->section, '$'))
    {
      UdmDSTRReset(&dstr);
      UdmDSTRParse(&dstr, F->section, &Doc->Sections);
      src= UdmDSTRPtr(&dstr);
      srclen= UdmDSTRLength(&dstr);
    }
    else
    {
      src= UdmVarListFindStr(&Doc->Sections, F->section, "");
      srclen= strlen(src);
    }
    if (!UdmMatchExec(&F->Filter.Match, src, srclen, src, nparts, Parts))
    {
      UdmDSTRFree(&dstr);
      return F;
    }
  }
  UdmDSTRFree(&dstr);
  return NULL;
}


udm_method_t
UdmSectionFilterListFindMethod(const UDM_SECTIONFILTERLIST *L,
                               const UDM_DOCUMENT *Doc,
                               char *reason, size_t reason_size)
{
  const UDM_SECTIONFILTER *F;

  UdmDebugEnter();

  if ((F= UdmSectionFilterListFindWithSubst(L, Doc, 0, NULL)))
  {
    udm_snprintf(reason, reason_size,
                 "%s %s %s %s '%s' '%s'", UdmMethodStr(F->Filter.method),
                 UdmMatchIsNegative(&F->Filter.Match) ? "NoMatch" : "Match",
                 UdmMatchModeStr(F->Filter.Match.Param.match_mode),
                 UdmMatchCaseSensitivityStr(&F->Filter.Match),
                 F->section,
                 UdmMatchPatternConstStr(&F->Filter.Match));
  }
  else
  {
    sprintf(reason,"Allow by default");
  }
  UdmDebugReturn(F ? F->Filter.method : UDM_METHOD_INDEX);
}


static udm_rc_t
UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL,
               UDM_SPIDERPARAM *Spider, UDM_HREF *Href)
{
  int            cascade;
  udm_rc_t       rc= UDM_OK, parse_res;
  UDM_URL        newURL;
  const UDM_REPLACE *Alias;
  size_t         nparts= 10;
  UDM_MATCH_PART Parts[10];
  UDM_SERVER     *Srv;
  const char     *val;
  char           *newhref= NULL;
  char           *alias= NULL;
  char           reason[1024]="";

  UdmDebugEnter();

  UdmURLInit(&newURL);

  if ((parse_res= UdmURLParse(&newURL, Href->url)))
  {
    UdmLog(Indexer, UDM_LOG_DEBUG, "%s: '%s'",
                                   UdmURLErrorStr(parse_res), Href->url);
    Href->Param.method= UDM_METHOD_DISALLOW;
    Href->Param.method_reason= UDM_COLLECT_LINKS_BAD;
    goto ret;
  }

  RelLink(Spider, CurURL, &newURL, &newhref);
  UdmLog(Indexer, UDM_LOG_DEBUG,"Link '%s' %s", Href->url, newhref);

  for (cascade= 0;
       ((Alias= UdmReplaceListFind(&Indexer->Conf->ReverseAliases,
                                   newhref, nparts, Parts))) &&
       (cascade < 1024);
       cascade++)
  {
    size_t aliassize= Alias->Replace.length +
                      UdmMatchPatternLength(&Alias->Match) +
                      strlen(newhref) + 8;
    if (!(alias= (char*) UdmRealloc(alias, aliassize)))
    {
      UdmLog(Indexer, UDM_LOG_ERROR, "No memory (%d bytes). %s line %d",
             (int) aliassize, __FILE__, __LINE__);
      goto ret;
    }
    UdmMatchApply(alias, aliassize, newhref,
                  Alias->Replace.str, Alias->Replace.length,
                  &Alias->Match, nparts, Parts);
    if (alias[0])
    {
      UdmLog(Indexer,UDM_LOG_DEBUG,"ReverseAlias%d: '%s'", cascade, alias);
      UDM_FREE(newhref);
      newhref= (char*)UdmStrdup(alias);
    }
    else
      break;
  }

  UdmURLParse(&newURL, newhref);

  Href->Param.method= UDM_METHOD_GET;
  Href->Param.method_reason= UDM_COLLECT_LINKS_INNER;
  if (!(Srv= UdmServerFind(Indexer, &Indexer->Conf->Servers, newhref, NULL)))
  {
    UdmLog(Indexer, UDM_LOG_DEBUG,"no Server, skip it");
    Href->Param.method=UDM_METHOD_DISALLOW;
    Href->Param.method_reason= UDM_COLLECT_LINKS_OUTER;
    goto rel_to_abs;
  }

  if (Srv->webspace == UDM_WEBSPACE_SITE &&
      Spider->webspace != UDM_WEBSPACE_URLLIST &&
      (strcasecmp(newURL.schema, CurURL->schema) ||
       strcasecmp(newURL.hostinfo, CurURL->hostinfo)))
  {
    UdmLog(Indexer,UDM_LOG_DEBUG, "Skip: sites differ: %s %s",
           newURL.hostinfo, CurURL->hostinfo);
    Href->Param.method= UDM_METHOD_DISALLOW;
    Href->Param.method_reason= UDM_COLLECT_LINKS_OUTER;
  }

  if (Href->Param.method == UDM_METHOD_DISALLOW)
    goto rel_to_abs;

  if (!strcasecmp(UDM_NULL2EMPTY(newURL.schema), "mailto") ||
      !strcasecmp(UDM_NULL2EMPTY(newURL.schema), "javascript"))
  {
    UdmLog(Indexer,UDM_LOG_DEBUG,"'%s' schema, skip it", newURL.schema);
    Href->Param.method=UDM_METHOD_DISALLOW;
    Href->Param.method_reason= UDM_COLLECT_LINKS_BADSCHEME;
    goto rel_to_abs;
  }

  if (Href->Param.hops > Srv->MaxHops)
  {
    UdmLog(Indexer, UDM_LOG_DEBUG, "too many hops (%d), skip it", Href->Param.hops);
    Href->Param.method= UDM_METHOD_DISALLOW;
    Href->Param.method_reason= UDM_COLLECT_LINKS_HOPS;
    goto rel_to_abs;
  }

  UdmLog(Indexer, UDM_LOG_DEBUG, " Server applied: site_id: %d URL: %s",
         Srv->site_id, UdmMatchPatternConstStr(&Srv->Filter.Match));

  /* Check Allow/Disallow/CheckOnly stuff */
  Href->Param.method= UdmFilterFind(&Indexer->Conf->Filters, newhref, reason);
  if (Href->Param.method == UDM_METHOD_DISALLOW)
  {
    UdmLog(Indexer,UDM_LOG_DEBUG,"%s, skip it",reason);
    Href->Param.method_reason= UDM_COLLECT_LINKS_FILTER;
    goto rel_to_abs;
  }
  else
  {
    UdmLog(Indexer,UDM_LOG_DEBUG,"%s",reason);
  }

  Href->Param.max_doc_per_site= UdmVarListFindInt(&Srv->Vars, "MaxDocPerSite", 0);

  if ((val= UdmVarListFindStr(&Srv->Vars, "SQLExportHref", NULL)))
    UdmVarListAddStr(&Href->HrefVars, "SQLExportHref", val);

rel_to_abs:
  /* FIXME: add MaxHops, Robots */
  UDM_FREE(Href->url);
  Href->url= (char*) UdmStrdup(newhref);
  Href->Param.server_id= Srv ? Srv->site_id : 0;
ret:  
  UDM_FREE(newhref);
  UDM_FREE(alias);
  UdmURLFree(&newURL);
  UdmDebugReturn(rc);
}


static udm_rc_t
UdmDocConvertHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  size_t  i;
  int     hops=UdmVarListFindInt(&Doc->Sections,"Hops",0);
  urlid_t url_id = (urlid_t)UdmVarListFindInt(&Doc->Sections, "ID", 0);
  uint4   maxhops = UdmVarListFindUnsigned(&Doc->Sections, "MaxHops", 255);
  const char *url= UdmVarListFindStr(&Doc->Sections, "url", "");

  for (i=0; i < Doc->Hrefs.nhrefs; i++)
  {
    UDM_HREF  *Href=&Doc->Hrefs.Href[i];
    Href->Param.hops= hops + 1;
    UdmConvertHref(Indexer, &Doc->CurURL, &Doc->Spider, Href);
    Href->Param.referrer= url_id;
    if (!(Doc->Spider.collect_links_destination & UDM_COLLECT_LINKS_PAGE))
    {
      if (!strcmp(Href->url, url)) /* Skip self link */
      {
        Href->Param.method= UDM_METHOD_DISALLOW;
        Href->Param.method_reason= UDM_COLLECT_LINKS_PAGE;
        Href->Param.stored= UDM_TRUE;
        continue;
      }
    }
    if (maxhops >= Href->Param.hops)
    {
      Href->Param.stored= UDM_FALSE;
    }
    else
    {
      Href->Param.stored= UDM_TRUE;
      Href->Param.method= UDM_METHOD_DISALLOW;
      Href->Param.method_reason= UDM_COLLECT_LINKS_HOPS;
    }
  }
  return UDM_OK;
}


udm_rc_t
UdmDocStoreHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc;
  const char *url= UdmVarListFindStr(&Doc->Sections, "URL", "");

  if(Doc->method==UDM_METHOD_HEAD)
    return UDM_OK;

  UdmDocBaseHref(Indexer,Doc); 
  UdmDocConvertHrefs(Indexer,Doc);

  UDM_GETLOCK(Indexer, UDM_LOCK_HREF_CACHE);  
  rc= UdmHrefListCopyNotStored(&Indexer->Conf->Hrefs, &Doc->Hrefs,
                               url, Doc->Spider.link_sources_to_follow);
  UDM_RELEASELOCK(Indexer, UDM_LOCK_HREF_CACHE);

  return rc;
}

/*********************** 'UrlFile' stuff (for -f option) *******************/

UDM_API(udm_rc_t)
UdmURLFile(UDM_AGENT *Indexer, const char *fname, udm_urlfilecmd_t action)
{
  FILE *url_file;
  char str[1024]="";
  char str1[1024]="";
  udm_rc_t rc= UDM_OK;
  udm_rc_t res;
  UDM_URL myurl;
  UDM_HREFPARAM HrefParam;
  UDM_QUERY Query;
  UdmDebugEnter();
  
  UdmURLInit(&myurl);
  UdmQueryInit(&Query);

  /* Read lines and clear/insert/check URLs                     */
  /* We've already tested in main.c to make sure it can be read */
  /* FIXME !!! Checking should be done here surely              */
  
  if(!strcmp(fname,"-"))
    url_file=stdin;
  else
    url_file=fopen(fname,"r");
  
  while(fgets(str1,sizeof(str1),url_file))
  {
    char *end;
    if(!str1[0])continue;
    end=str1+strlen(str1)-1;
    while((end>=str1)&&(*end=='\r'||*end=='\n')){
      *end=0;if(end>str1)end--;
    }
    if(!str1[0])continue;
    if(str1[0]=='#')continue;

    if(*end=='\\'){
      *end=0;strcat(str,str1);
      continue;
    }
    strcat(str,str1);
    strcpy(str1,"");

    switch(action)
    {

    case UDM_URL_FILE_REINDEX:
      UdmVarListReplaceStr(&Indexer->Conf->Vars, "u", str);
      if (UDM_OK != (rc= UdmQueryAction(Indexer, &Query, UDM_QUERYCMD_EXPIRE)))
        goto ex;
      UdmVarListDel(&Indexer->Conf->Vars, "u");
      UdmQueryClearLimits(&Query);
      break;

    case UDM_URL_FILE_CLEAR:
      UdmVarListReplaceStr(&Indexer->Conf->Vars, "u", str);
      if (UDM_OK != (rc= UdmQueryAction(Indexer, &Query, UDM_QUERYCMD_CLEAR)))
        goto ex;
      UdmVarListDel(&Indexer->Conf->Vars, "u");
      UdmQueryClearLimits(&Query);
      break;

    case UDM_URL_FILE_INSERT:
      UdmHrefParamInit(&HrefParam);
      HrefParam.link_source= UDM_LINK_SOURCE_URLFILE;
      UdmHrefListAddConst(&Indexer->Conf->Hrefs, &HrefParam, str);
      break;

    case UDM_URL_FILE_PARSE:
      if ((UDM_OK != (res= UdmURLParse(&myurl, str))) ||
          (UDM_OK != (res= myurl.schema ? UDM_OK : UDM_ERROR)))
      {
        UdmLog(Indexer,UDM_LOG_ERROR,"%s: '%s'", UdmURLErrorStr(res), str);
        rc= UDM_ERROR;
        goto ex;
      }
      break;
    }
    str[0]=0;
  }
  if(url_file!=stdin)
    fclose(url_file);
ex:
  UdmURLFree(&myurl);
  UdmQueryFree(&Query);
  UdmDebugReturn(rc);
}




/*******************************************************************/


static udm_rc_t
UdmDocAlias(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  const UDM_REPLACE  *Alias;
  UDM_MATCH_PART  Parts[10];
  size_t    alstrlen, nparts=10;
  const char  *alias_prog=UdmVarListFindStr(&Indexer->Conf->Vars,"AliasProg",NULL);
  char    *alstr;
  udm_rc_t rc= UDM_OK;
  const char  *url=UdmVarListFindStr(&Doc->Sections,"URL","");

  alstrlen = 256 + strlen(url);
  if ((alstr = (char*)UdmMalloc(alstrlen)) == NULL) return UDM_ERROR;
  alstr[0] = '\0';
  if(alias_prog)
  {
    rc= UdmAliasProg(Indexer, alias_prog, url, alstr, alstrlen - 1);
    UdmLog(Indexer,UDM_LOG_EXTRA,"AliasProg result: '%s'",alstr);
    if (rc != UDM_OK)
    {
      UDM_FREE(alstr);
      return rc;
    }
  }

  /* Find alias when aliastr is empty, i.e.     */
  /* when there is no alias in "Server" command */
  /* and no AliasProg                           */
  if (!alstr[0] &&
      (Alias= UdmReplaceListFind(&Indexer->Conf->Aliases, url, nparts, Parts)))
  {
    UdmMatchApply(alstr, alstrlen - 1, url,
                  Alias->Replace.str, Alias->Replace.length,
                  &Alias->Match, nparts, Parts);
  }
  if(alstr[0]){
    UdmVarListReplaceStr(&Doc->Sections,"Alias",alstr);
  }
  UDM_FREE(alstr);
  return UDM_OK;
}


static udm_rc_t
UdmDocIPRequestPerMinLimit(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  int limit= UdmVarListFindInt(&Indexer->Conf->Vars, "IPRequestPerMinLimit", 10);
  UDM_INADDR *addr;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_INADDR_CACHE);

  if ((addr= UdmInAddrListFind(&Indexer->Conf->InAddr, Doc->connp.sin.sin_addr)))
  {
    time_t now= time(0);
    time_t delta= now - addr->first_used + 1;
    int hits_per_minute= addr->hits * 60 / delta;
    addr->last_used= now;
    if (addr->hits > 10 && hits_per_minute > limit)
    {
      Doc->method= UDM_METHOD_VISITLATER;
      UdmVarListReplaceInt(&Doc->Sections, "Status",
                           UDM_HTTP_STATUS_BANDWIDTH_LIMIT_EXCEEDED);
      UdmLog(Indexer, UDM_LOG_ERROR,
             "Skip %s: IPRequestPerMinLimit exceeded (%d hits/min)",
             Doc->connp.hostname, hits_per_minute);
    }
    else
      addr->hits++;
  }
  else
  {
    UDM_INADDR Item;
    UdmInAddrInit(&Item);
    Item.addr= Doc->connp.sin.sin_addr;
    Item.first_used= time(0);
    if (UDM_OK != (rc= UdmInAddrListAdd(&Indexer->Conf->InAddr, &Item)))
      UdmLog(Indexer, UDM_LOG_ERROR, "UdmInAddrListAdd() failed");
  }
  return rc;
}


static udm_rc_t
UdmDocLookupConn(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  udm_rc_t u;
  udm_rc_t rc= UDM_OK;

#ifdef WIN32
  if (Doc->CurURL.schema && !strcasecmp(Doc->CurURL.schema, "file"))
	return UDM_OK; /* Indexing Win LAN, no needs to resolve */
#endif

  u= UdmHostLookupEx(Indexer, &Indexer->Conf->Hosts, &Doc->connp,
                     Doc->Spider.dns_cache_timeout);

  if (Doc->CurURL.hostname != NULL && *Doc->CurURL.hostname != '\0')
  {
    if (u)
    {
      UdmLog(Indexer,UDM_LOG_WARN,"Can't resolve host '%s'",Doc->connp.hostname);
      Doc->method = UDM_METHOD_VISITLATER;
      UdmVarListReplaceInt(&Doc->Sections, "Status", UDM_HTTP_STATUS_SERVICE_UNAVAILABLE);
      return UDM_OK;
    }
    else
    {
      UDM_GETLOCK(Indexer, UDM_LOCK_INADDR_CACHE);
      rc= UdmDocIPRequestPerMinLimit(Indexer, Doc);
      UDM_RELEASELOCK(Indexer, UDM_LOCK_INADDR_CACHE);
    }
  }

  return rc;
}



static udm_rc_t
UdmDocCheck(UDM_AGENT *Indexer,UDM_SERVER *CurSrv,UDM_DOCUMENT *Doc)
{
  char   reason[1024]="";
  int    hops=UdmVarListFindInt(&Doc->Sections,"Hops",0);
  const char *method= UdmMethodStr(CurSrv->Filter.method);
  const char *follow= UdmFollowStr(CurSrv->webspace);

  switch (CurSrv->Filter.Match.Param.match_mode)
  {
    case UDM_MATCH_WILD:
      UdmLog(Indexer,UDM_LOG_DEBUG, "Realm %s %s wild '%s'",
             follow, method, UdmMatchPatternConstStr(&CurSrv->Filter.Match));
      break;
    case UDM_MATCH_REGEX:
      UdmLog(Indexer,UDM_LOG_DEBUG, "Realm %s %s regex '%s'",
             follow, method, UdmMatchPatternConstStr(&CurSrv->Filter.Match));
      break;
    case UDM_MATCH_SUBNET:
      UdmLog(Indexer,UDM_LOG_DEBUG, "Subnet %s %s '%s'",
             follow, method, UdmMatchPatternConstStr(&CurSrv->Filter.Match));
      break;
    case UDM_MATCH_BEGIN:
    case UDM_MATCH_FULL:
      UdmLog(Indexer,UDM_LOG_DEBUG, "Server %s %s '%s'",
             follow, method, UdmMatchPatternConstStr(&CurSrv->Filter.Match));
      break;
    case UDM_MATCH_NUMERIC_LT:
    case UDM_MATCH_NUMERIC_GT:
    case UDM_MATCH_RANGE:
    case UDM_MATCH_SUBSTR:
    case UDM_MATCH_END:
      UdmLog(Indexer, UDM_LOG_ERROR, "Unexpected match_mode");
      break;
  }

  if ((Doc->method= UdmMethod(method)) == UDM_METHOD_GET)
  {
    /* Check Allow/Disallow/CheckOnly stuff */
    Doc->method= UdmFilterFind(&Indexer->Conf->Filters,
                               UdmVarListFindStr(&Doc->Sections,"URL",""), reason);
    UdmLog(Indexer, UDM_LOG_DEBUG, "%s", reason);
  }

  if (Doc->method == UDM_METHOD_DISALLOW)
    return UDM_OK;

  /* Check that hops is less than MaxHops */
  if (hops > Doc->Spider.maxhops)
  {
    UdmLog(Indexer, UDM_LOG_WARN, "Too many hops (%d)", hops);
    Doc->method= UDM_METHOD_DISALLOW;
    return UDM_OK;
  }

  return UDM_OK;
}


static void
UdmAppendTarget(UDM_AGENT *Indexer,
                const char *url, const char *lang, int hops, int parent)
{
  UDM_DOCUMENT *Doc;

  UDM_GETLOCK(Indexer, UDM_LOCK_TARGETS);
  
  if (Indexer->Conf->Targets.num_rows > 0)
  {
    UDM_DOCUMENT *First= Indexer->Conf->Targets.Doc;
    UDM_DOCUMENT *Last= First + Indexer->Conf->Targets.num_rows - 1;
    for (Doc= Last; Doc >= First; Doc--)
    {
      if (!strcasecmp(UdmVarListFindStr(&Doc->Sections, "URL", ""), url) &&
          !strcmp(UdmVarListFindStr(&Doc->RequestHeaders, "Accept-Language", ""), lang))
        goto ret;
    }
  }
  if (UDM_OK != UdmResultAppendAndInit(&Indexer->Conf->Targets))
    goto ret;
  Doc= &Indexer->Conf->Targets.Doc[Indexer->Conf->Targets.num_rows];
  UdmVarListAddStr(&Doc->Sections, "URL", url);
  UdmVarListAddInt(&Doc->Sections, "Hops", hops);
  UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(url));
  UdmVarListReplaceInt(&Doc->Sections, "Referrer-ID", parent);
  UdmDocAction(Indexer, Doc, UDM_DOCCMD_ADD);
  if (*lang != '\0') UdmVarListAddStr(&Doc->RequestHeaders, "Accept-Language", lang);
  
ret:
  UDM_RELEASELOCK(Indexer, UDM_LOCK_TARGETS);
  return;
}


/*
  Remove character set part,
  and other extended parts.
  Return pointer to new value.
*/
static const char*
UdmDocNormalizeContentType(UDM_DOCUMENT *Doc)
{
  const UDM_VAR *Var= UdmVarListFind(&Doc->Sections, "Content-Type");
  char *p, contenttype[120];

  if (!Var || !UdmVarStr(Var) || !UdmVarLength(Var))
    return NULL;
  /*
    TODO34: this should be rewritten to cover all possible cases,
    e.g. "Content-Type: ttt; xxx=xxx; yyy=xxx; zzz=zzz; CharSet=CS;"
  */
  udm_snprintf(contenttype, sizeof(contenttype), "%s", UdmVarStr(Var));
  if ((p= strstr(contenttype, "charset=")) != NULL)
  {
    const char *cs= UdmCharsetCanonicalName(p + 8);
    *p= '\0';
    UdmRTrim(contenttype, "; ");
    UdmVarListReplaceStr(&Doc->Sections, "Server-Charset", cs ? cs : (p + 8));
    UdmVarListReplaceStr(&Doc->Sections, "Content-Type", contenttype);
    return UdmVarListFindStrNonEmpty(&Doc->Sections, "Content-Type", NULL);
  }
  if ((p= strstr(contenttype, "generator=")) != NULL)
  {
    const char *end;
    for (p+= 10, end= p; end[0] ; end++)
    {
      if (udm_isspace(end[0]) || end[0] == ';')
        break;
    }
    UdmVarListReplaceStrn(&Doc->Sections, "Content-Type.generator", p, end - p);
  }
  if ((p= strchr(contenttype, ';')))
  {
    *p= '\0';
    UdmVarListReplaceStr(&Doc->Sections, "Content-Type", contenttype);
    return UdmVarListFindStrNonEmpty(&Doc->Sections, "Content-Type", NULL);
  }
  return UdmVarStr(Var);
}


static udm_rc_t
UdmDocProcessVaryLang(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  const char *vary= UdmVarListFindStr(&Doc->Sections, "Vary", NULL);
  const char *VaryLang;
  
  if (vary && strcasestr(vary, "accept-language") &&
      (VaryLang= UdmVarListFindStr(&Doc->Sections, "VaryLang", "")) &&
      VaryLang[0])
  {  
    UDM_URL newURL;
    const char *ourl;
    
    UdmURLInit(&newURL);
    UdmVarListReplaceStr(&Doc->Sections, "Status", "300");
    UdmURLParse(&newURL, ourl= UdmVarListFindStr(&Doc->Sections, "URL", ""));
    
    if (strcmp(UDM_NULL2EMPTY(newURL.filename), "robots.txt"))
    {
      const char *CL= UdmVarListFindStr(&Doc->Sections, "Content-Location", UDM_NULL2EMPTY(newURL.filename));
      size_t urlen= 128 + strlen(UDM_NULL2EMPTY(newURL.hostinfo)) + strlen(UDM_NULL2EMPTY(newURL.path)) + strlen(CL);
      char *url;
      
      if ((url= (char*)UdmMalloc(urlen)) != NULL)
      {
        char *tok, *lt;
        int parent= UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0);
        int hops= UdmVarListFindInt(&Doc->Sections, "Hops", 0);
        
        char tmp[80];
        udm_snprintf(tmp, sizeof(tmp), "%s", VaryLang);
        snprintf(url, urlen, "%s://%s%s%s",
                 UDM_NULL2EMPTY(newURL.schema),
                 UDM_NULL2EMPTY(newURL.hostinfo), 
                 UDM_NULL2EMPTY(newURL.path), CL);
        UdmAppendTarget(Indexer,  url, "", hops, parent);
        
        for (tok= udm_strtok_r(tmp, " ,\t", &lt) ;
             tok != NULL ;
             tok= udm_strtok_r(NULL, " ,\t", &lt))
        {
          UdmAppendTarget(Indexer, ourl, tok, hops, parent);
        }
        UdmFree(url);
      }
    }
    UdmURLFree(&newURL);
  }
  return UDM_OK;
}


udm_rc_t
UdmDocProcessContentResponseHeaders(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  const UDM_VAR *var;
  const char *content_type= UdmDocNormalizeContentType(Doc);

  if (!UdmVarListFindBool(&Indexer->Conf->Vars, "UseRemoteContentType", UDM_TRUE) ||
     content_type == NULL)
  {
    const UDM_REPLACE  *M;
    UDM_MATCH_PART  P[10];
    const char *fn= (Doc->CurURL.filename && Doc->CurURL.filename[0]) ? Doc->CurURL.filename : "index.html";
    
    if ((M= UdmReplaceListFind(&Indexer->Conf->MimeTypes, fn, 10, P)))
    {
      UdmVarListReplaceStr(&Doc->Sections, "Content-Type", M->Replace.str);
      content_type= UdmDocNormalizeContentType(Doc); 
    }
    
    if ((M= UdmReplaceListFind(&Indexer->Conf->Encodings, fn, 10, P)))
      UdmVarListReplaceStr(&Doc->Sections, "Content-Encoding", M->Replace.str);
  }

  if ((var=UdmVarListFind(&Doc->Sections,"Server")))
  {
    if(!strcasecmp("yes",UdmVarListFindStr(&Indexer->Conf->Vars,"ForceIISCharset1251","no")))
    {
      if (!UdmWildCaseCmp(UdmVarStr(var), "*Microsoft*") ||
          !UdmWildCaseCmp(UdmVarStr(var), "*IIS*"))
      {
        const char *cs;
        if((cs=UdmCharsetCanonicalName("windows-1251")))
          UdmVarListReplaceStr(&Doc->Sections, "Server-Charset", cs);
      }
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  const UDM_VAR *var;
  const int content_length= UdmVarListFindInt(&Doc->Sections, "Content-Length", 0);

  UdmDocProcessVaryLang(Indexer, Doc);

  if ((size_t)content_length > UdmHTTPBufAllocedSize(&Doc->Buf))
    UdmVarListReplaceInt(&Doc->Sections, "Status", UDM_HTTP_STATUS_PARTIAL_OK);



  UdmDocProcessContentResponseHeaders(Indexer, Doc);
  
  if (UdmVarListFindBool(&Indexer->Conf->Vars, "UseCookie", UDM_FALSE))
  {
    size_t i;
    UDM_GETLOCK(Indexer, UDM_LOCK_COOKIE_CACHE);
    for (i= 0; i < Doc->Sections.nvars; i++)
    {
      var= UdmVarListFindConstByIndex(&Doc->Sections, i);
      if (!strncmp(UdmVarName(var), "Set-Cookie.", 11))
        UdmVarListReplaceStr(&Indexer->Conf->Cookies,
                             UdmVarName(var) + 11, UdmVarStr(var));
    }
    UDM_RELEASELOCK(Indexer, UDM_LOCK_COOKIE_CACHE);
  }

  if(!UdmVarListFind(&Doc->Sections,"Content-Type"))
    UdmVarListAddStr(&Doc->Sections,"Content-Type","application/octet-stream");

  if ((var= UdmVarListFind(&Doc->Sections, "Location")))
  {
    udm_rc_t rc;
    UDM_URL newURL;
    UdmURLInit(&newURL);
    if (UDM_OK == (rc= UdmURLParse(&newURL, UdmVarStr(var))))
    {
      UDM_HREFPARAM HrefParam;
      UdmHrefParamInit(&HrefParam);
      if (!newURL.schema)
        UdmLog(Indexer, UDM_LOG_WARN, "Redirect URL is not absolute: %s", UdmVarStr(var));
      HrefParam.hops= UdmVarListFindInt(&Doc->Sections,"Hops",0)+1;
      HrefParam.referrer= UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0);
      HrefParam.server_id= UdmVarListFindInt(&Doc->Sections,"Server_id", 0);
      HrefParam.link_source= UDM_LINK_SOURCE_REDIRECT;
      UdmHrefListAddConst(&Doc->Hrefs, &HrefParam, UdmVarStr(var));
    }
    else
    {
      UdmLog(Indexer, UDM_LOG_ERROR, "Redirect: %s: '%s'",
                                     UdmURLErrorStr(rc), UdmVarStr(var));
    }
    UdmURLFree(&newURL);
  }
  return UDM_OK;
}


/*
  Create HTTP.Content variable if requested
*/

static udm_rc_t
UdmVarListAddContentStr(UDM_VARLIST *Sections, const char *name,
                        UDM_DOCUMENT *Doc)
{
  UDM_VAR *Sec;
  UDM_CONST_STR content;
  if (!UdmHTTPBufContentToConstStr(&Doc->Buf, &content) && content.length && /* TODO34: UdmHTTPBuf: UdmVarList */
      (Sec= UdmVarListFindVar(Sections, name)))
  {
    return UdmVarSetStrn(Sec, content.str, content.length);
  }
  return UDM_OK;
}


/*
  Create "HTTP.LocalCharsetContent" variable if requested
  Must be executed after character set guesser;
*/
static udm_rc_t
UdmVarListAddLocalCharsetContentStr(const char *name,
                                    UDM_DOCUMENT *Doc, UDM_CHARSET *doccs)
{
  UDM_VAR *Sec;
  UDM_CONST_STR content;
  if (!UdmHTTPBufContentToConstStr(&Doc->Buf, &content) && content.length &&   /* TODO34: UdmHTTPBuf: UdmVarList */
      (Sec= UdmVarListFindVar(&Doc->Sections, "HTTP.LocalCharsetContent")))
  {
    if (!doccs)
      doccs= &udm_charset_latin1;
    if (Doc->lcs == doccs)
    {
      return UdmVarSetStrn(Sec, content.str, content.length);
    }
    else
    {
      UDM_CONV conv;
      UdmConvInit(&conv, doccs, Doc->lcs);
      return UdmVarSetConv(Sec, &conv, UDM_RECODE_HTML_NONASCII,
                           content.str, content.length);
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmVarListLog(UDM_AGENT *A,UDM_VARLIST *V,int l,const char *pre){
  size_t h;
  if (UdmNeedLog(A->Conf, l))
  {
    for(h=0; h < V->nvars; h++)
    {
      const UDM_VAR *v= UdmVarListFindConstByIndex(V, h);
      const char *val= UdmVarStr(v);
      UdmLog(A, l, "%s.%s: %s", pre, UdmVarName(v), val ?  val : "<NULL>");
    }
  }
  return UDM_OK;
}


static int
UdmParseHTTPResponseAndHeaders(UDM_AGENT *A, UDM_DOCUMENT *Doc)
{
  int status;
  const UDM_VAR *var;
  UdmParseHTTPResponse(A, Doc);
  UdmDocProcessResponseHeaders(A, Doc);
  UdmVarListAddContentStr(&Doc->Sections, "HTTP.Content", Doc);
  UdmUserSectionsParse(A, &A->Conf->SectionHdrMatch, Doc, NULL);
  UdmVarListLog(A, &Doc->Sections, UDM_LOG_DEBUG, "Response");
  if ((Doc->Spider.use_robots & UDM_ROBOTS_HEADER) &&
      (var= UdmVarListFind(&Doc->Sections, "X-Robots-Tag")))
  {
    UDM_CONST_STR valuebuf;
    UdmRobotsParamParse(&Doc->Spider.robots, &Doc->Spider.robots,
                        UdmVarGetConstStr(var, &valuebuf));
  }
  status= UdmVarListFindInt(&Doc->Sections,"Status",0);
  UdmLog(A, UDM_LOG_EXTRA, "Status: %d %s", status, UdmHTTPErrMsg(status));
  return status;
}


static udm_rc_t
UdmDocContentDecode(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  const char  *url=UdmVarListFindStr(&Doc->Sections,"URL","");
  const char  *ce= UdmVarListFindStr(&Doc->Sections,"Content-Encoding", NULL);
  udm_rc_t rc;
  udm_content_encoding_t encoding;

  if (!ce || !ce[0])
    return UDM_OK;

  if (UDM_CONTENT_ENCODING_UNKNOWN == (encoding= UdmContentEncodingID(ce)))
  {
    UdmLog(Indexer,UDM_LOG_WARN,"Unsupported Content-Encoding: %s", ce);
    UdmVarListReplaceInt(&Doc->Sections,"Status",UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE);
    return UDM_OK;
  }
  UDM_THREADINFO(Indexer, "Decoding Content-Encoding", url);

  if (UDM_OK!= (rc= UdmHTTPBufContentEncodingDecode(&Doc->Buf, encoding)))
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "Content-Encoding processing failed");
    UdmVarListReplaceInt(&Doc->Sections,"Status", UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE);
    return UDM_OK;
  }
  UdmVarListReplaceInt(&Doc->Sections, "Content-Length",
                       UdmHTTPBufContentSize(&Doc->Buf));  
  return UDM_OK;
}


udm_rc_t
UdmDocInternalParserExec(UDM_AGENT *A, UDM_DOCUMENT *Doc,
                         udm_content_type_t ct)
{
  switch (ct)
  {
    case UDM_CONTENT_TYPE_TEXT_PLAIN: return UdmParseText(A, Doc);
    case UDM_CONTENT_TYPE_TEXT_HTML:  return UdmHTMLParse(A, Doc);
    case UDM_CONTENT_TYPE_TEXT_XML:
    {
      udm_rc_t rc;
      /* TODO34: move to somewhere */
      UdmVarListReplaceStr(&A->Conf->Vars, "XMLDefaultSection", "body");
      rc= UdmXMLParse(A, Doc);
      if (rc != UDM_OK)
      {
        UdmLog(A, UDM_LOG_ERROR, "Error: %s",
               UdmVarListFindStr(&Doc->Sections, "X-Reason", ""));
      }
      return rc;
    }
    case UDM_CONTENT_TYPE_MESSAGE_RFC822: return UdmMessageRFC822Parse(A, Doc);
    case UDM_CONTENT_TYPE_AUDIO_MPEG:     return UdmMP3Parse(A, Doc);
    case UDM_CONTENT_TYPE_DOCX:           return UdmDOCXParse(A, Doc);
    case UDM_CONTENT_TYPE_TEXT_RTF:       return UdmRTFParse(A, Doc);

    case UDM_CONTENT_TYPE_HTDB:           return UdmHTDBParse(A, Doc);
    case UDM_CONTENT_TYPE_UNKNOWN:        return UDM_ERROR;
  }
  return UDM_OK;
}


udm_rc_t
UdmDocParseContent(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc);

udm_rc_t
UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc)
{
  
#ifdef USE_PARSER
  UDM_PARSER  *Parser;
  int status= UdmVarListFindInt(&Doc->Sections, "Status", 0);
#endif
  const char  *real_content_type=NULL;
  const char  *ct=UdmVarListFindStr(&Doc->Sections,"Content-Type","");
  udm_rc_t rc= UDM_OK;
  udm_content_type_t content_type;

  if(!strcmp(UDM_NULL2EMPTY(Doc->CurURL.filename), "robots.txt")) return UDM_OK;

  if (Doc->method == UDM_METHOD_HEAD)
    return rc;

  if (UDM_OK != (rc= UdmDocContentDecode(Indexer, Doc)))
    return rc;

  if (UdmVarListFindInt(&Doc->Sections, "status", 0) == UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE)
    return UDM_OK;

#ifdef USE_PARSER
  /* Let's try to start external parser for this Content-Type */

  if ((Parser= UdmParserFind(&Indexer->Conf->Parsers, ct)))
  {
    UdmLog(Indexer,UDM_LOG_DEBUG,"Found external parser '%s' -> '%s'",
           Parser->from_mime ? Parser->from_mime : "NULL",
           Parser->to_mime ? Parser->to_mime : "NULL");
  }
  if (Parser && (UDM_OK == UdmParserExec(Indexer, Parser, Doc)))
  {
    if (status == UDM_HTTP_STATUS_OK)
    {
      const char *to_charset;
      real_content_type= Parser->to_mime ? Parser->to_mime : "unknown";
      UdmLog(Indexer,UDM_LOG_DEBUG,"Cached.Content-Type: %s",real_content_type);
      if ((to_charset= strstr(real_content_type, "charset=")))
      {
        const char *cs;
        char csbuf[64];
        udm_snprintf(csbuf, sizeof(csbuf), "%s", to_charset + 8);
        cs= UdmCharsetCanonicalName(UdmTrim(csbuf, " \t"));
        UdmVarListReplaceStr(&Doc->Sections, "Server-Charset", cs);
        UdmVarListReplaceStr(&Doc->Sections, "RemoteCharset", cs);
        UdmLog(Indexer,UDM_LOG_DEBUG, "Cached.Charset: %s", cs);
        UdmVarListAddStr(&Doc->Sections,"Cached.Charset",cs);
      }
#ifdef DEBUG_PARSER
      fprintf(stderr,"content='%.*s'\n",
              (int) UdmHTTPBufSize(&Doc->Buf), UdmHTTPBufPtr(&Doc->Buf));
#endif
    }
    else
    {
      UdmLog(Indexer, UDM_LOG_WARN, "Parser not executed, document status: %d", status);
    }
  }
#endif
  
  if(!real_content_type)real_content_type=ct;
  UdmVarListAddStr(&Doc->Sections,"Cached.Content-Type",real_content_type);
  
  if (!strcasecmp(real_content_type, "application/http") ||
      !strcasecmp(real_content_type, "message/http"))
  {
    UdmLog(Indexer, UDM_LOG_DEBUG, "Re-parsing headers");
    UdmHTTPBufShiftLeft(&Doc->Buf, UdmHTTPBufHeaderSize(&Doc->Buf));
    UdmParseHTTPResponseAndHeaders(Indexer, Doc);
    real_content_type= UdmVarListFindStrNonEmpty(&Doc->Sections,
                                                "Content-Type", real_content_type);
    status= UdmVarListFindInt(&Doc->Sections, "Status", 0);
    UdmVarListReplaceStr(&Doc->Sections, "Cached.Content-Type", real_content_type);
  }

  /* CRC32 without headers */
  UdmVarListReplaceInt(&Doc->Sections, "crc32", UdmHTTPBufCRC32(&Doc->Buf));

  /* Status could change in case of message/http */
  if (status != UDM_HTTP_STATUS_OK &&
      status != UDM_HTTP_STATUS_PARTIAL_OK)
    return rc;

  if (UDM_CONTENT_TYPE_UNKNOWN ==
      (content_type= UdmContentTypeByName(real_content_type)))
  {
    /* Unknown Content-Type  */
    UdmLog(Indexer,UDM_LOG_WARN, "Unsupported Content-Type '%s'", real_content_type);
    UdmVarListReplaceInt(&Doc->Sections, "Status", UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE);
    return UDM_OK;
  }

  /*
    Parse the content first, then the user defined sections.
    The order is important for crc.
  */
  UdmDocInternalParserExec(Indexer, Doc, content_type);
  if (content_type == UDM_CONTENT_TYPE_TEXT_PLAIN ||
      content_type == UDM_CONTENT_TYPE_TEXT_HTML)
    UdmUserSectionsParse(Indexer, &Indexer->Conf->SectionMatch, Doc, NULL);

  return UDM_OK;
}


static udm_rc_t
UdmDocAddDocExtraHeaders(UDM_DOCUMENT *Doc){
  /* Host Name for virtual hosts */
  if(Doc->CurURL.hostname != NULL) {
    char    arg[128]="";
    if(Doc->CurURL.port){
      sprintf(arg,"%s:%d",Doc->CurURL.hostname,Doc->CurURL.port);
      UdmVarListReplaceStr(&Doc->RequestHeaders,"Host",arg);
    }else{
      UdmVarListReplaceStr(&Doc->RequestHeaders,"Host",Doc->CurURL.hostname);
    }
  }
  return UDM_OK;
}



static udm_rc_t
UdmDocAddCookieHeaders(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  size_t    i;
  UDM_DSTR cookie;

  if (!Doc->CurURL.hostname || !Indexer->Conf->Cookies.nvars)
    return UDM_OK;

  UdmDSTRInit(&cookie, 1024);
  UDM_GETLOCK(Indexer, UDM_LOCK_COOKIE_CACHE);
  for (i= 0; i < Indexer->Conf->Cookies.nvars; i++)
  {
    const UDM_VAR *v= UdmVarListFindConstByIndex(&Indexer->Conf->Cookies, i);
    const char *domain, *path;
    size_t short_dlen, long_dlen;
    if (!(domain= strchr(UdmVarName(v), '@')))
      continue;
    domain++;
    if (!(path= strchr(domain, '/')))
    {
      continue;
    }
    else
    {
      size_t plen= strlen(path);
      size_t path_length= strlen(Doc->CurURL.path);
      if (memcmp(Doc->CurURL.path, path, udm_min(plen, path_length)))
        continue;
      if (plen > path_length && (!Doc->CurURL.filename ||
                                  memcmp(Doc->CurURL.filename,
                                         path + path_length,
                                         plen - path_length)))
        continue;
    }
    short_dlen= path-domain;
    long_dlen= strlen(Doc->CurURL.hostname);
    if (short_dlen > long_dlen)
      continue;
    if (strncasecmp(domain,
                    Doc->CurURL.hostname + long_dlen - short_dlen, short_dlen))
      continue;
    if (UdmDSTRLength(&cookie))
      UdmDSTRAppend(&cookie, "; ", 2);
    UdmDSTRAppend(&cookie, UdmVarName(v), domain - UdmVarName(v) - 1);
    UdmDSTRAppend(&cookie, "=", 1);
    UdmDSTRAppendSTR(&cookie, UdmVarStr(v));
  }
  UDM_RELEASELOCK(Indexer, UDM_LOCK_COOKIE_CACHE);
  if (UdmDSTRLength(&cookie));
  {
    UdmVarListReplaceStr(&Doc->RequestHeaders, "Cookie", UdmDSTRPtr(&cookie));
  }
  UdmDSTRFree(&cookie);
  return UDM_OK;
}


static udm_rc_t
UdmDocAddConfExtraHeaders(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  char    arg[128]="";
  const char  *lc;
  size_t    i;
  
  /* If LocalCharset specified, add Accept-Charset header */
  if((lc=UdmVarListFindStr(&Indexer->Conf->Vars,"LocalCharset",NULL))){
    snprintf(arg,sizeof(arg)-1,"%s;q=1.0, *;q=0.9, utf-8;q=0.8",lc);
    arg[sizeof(arg)-1]='\0';
    UdmVarListAddStr(&Doc->RequestHeaders,"Accept-Charset",arg);
  }
  
  for (i=0; i< Indexer->Conf->Vars.nvars; i++)
  {
    const UDM_VAR *v= UdmVarListFindConstByIndex(&Indexer->Conf->Vars, i);
    if (!strncmp(UdmVarName(v), "Request.",8))
      UdmVarListInsStr(&Doc->RequestHeaders, UdmVarName(v) + 8, UdmVarStr(v));
  }
  
  if (UdmVarListFindBool(&Indexer->Conf->Vars, "UseCookie", UDM_FALSE))
    UdmDocAddCookieHeaders(Indexer, Doc);
  
#ifdef HAVE_ZLIB
  UdmVarListInsStr(&Doc->RequestHeaders,"Accept-Encoding","gzip,deflate,compress");
#endif
  return UDM_OK;
}


static const UDM_URL *
UdmServerGetRandomProxy(UDM_SERVER *Server)
{
  if (!Server->ProxyList.nitems)
    return NULL;
  if (Server->ProxyList.nitems == 1)
    return &Server->ProxyList.Item[0];
  /* rand() is not reentrant. Let's use time(0). */
  return &Server->ProxyList.Item[time(0) % Server->ProxyList.nitems];
}


static udm_rc_t
UdmDocSetConn(UDM_DOCUMENT *Doc, const UDM_URL *proxy)
{
  if (proxy)
  {
    char tmp[256];
    Doc->connp.port= proxy->port ? proxy->port : 3128;
    udm_snprintf(tmp, sizeof(tmp), "http://%s%s%s:%d/",
                 proxy->auth ? "***" : "", proxy->auth ? "@" : "",
                 proxy->hostname, Doc->connp.port);
    UdmVarListReplaceStr(&Doc->RequestHeaders, "Proxy", tmp);
    Doc->connp.hostname= UdmStrdup(proxy->hostname);
    if (proxy->auth && proxy->auth[0])
    {
      udm_snprintf(tmp, sizeof(tmp), "Basic %s", proxy->auth);
      UdmVarListReplaceStr(&Doc->RequestHeaders, "Proxy-Authorization", tmp);
    }
  }
  else
  {
    if (Doc->CurURL.hostname)
    {
      Doc->connp.hostname= UdmStrdup(Doc->CurURL.hostname);
      Doc->connp.port= Doc->CurURL.port ? Doc->CurURL.port : Doc->CurURL.default_port;
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmDocAddServExtraHeaders(UDM_SERVER *Server, UDM_DOCUMENT *Doc)
{
  char  arg[128]="";
  size_t  i;
  const UDM_URL *Proxy= UdmServerGetRandomProxy(Server);

  UdmDocSetConn(Doc, Proxy);
  for (i= 0 ; i < Server->Vars.nvars ; i++)
  {
    const UDM_VAR *Hdr= UdmVarListFindConstByIndex(&Server->Vars, i);
    
    if (!strcasecmp(UdmVarName(Hdr), "AuthBasic"))
    {
      /* HTTP and FTP specific stuff */
      if (!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "http") ||
          !strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "https") ||
          !strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "ftp") ||
          !strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "https"))
      {
        udm_snprintf(arg, sizeof(arg), "Basic %s", UdmVarStr(Hdr));
        UdmVarListReplaceStr(&Doc->RequestHeaders, "Authorization", arg);
      }

      if (!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp") || 
          !strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "news"))
      {
        /* Auth if required                      */
        /* NNTPGet will parse this header        */
        /* We'll pass authinfo still in base64   */
        /* form to avoid plain user name in core */
        /* file on crashes if any                */
        if (UdmVarStr(Hdr) && UdmVarStr(Hdr)[0])
          UdmVarListReplaceStr(&Doc->RequestHeaders, "Authorization", UdmVarStr(Hdr));
      }
    }
    else
    {
      if (!strncmp(UdmVarName(Hdr), "Request.", 8))
        UdmVarListReplaceStr(&Doc->RequestHeaders,
                             UdmVarName(Hdr) + 8, UdmVarStr(Hdr));
    }
  }
  return UDM_OK;
}


static size_t
UdmNumTargets(UDM_AGENT *Indexer)
{
  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_TARGETS);
  return Indexer->Conf->Targets.num_rows - Indexer->Conf->Targets.cur_row;
}

static udm_rc_t
UdmNextTarget(UDM_AGENT * Indexer, UDM_QUERY *Query, UDM_DOCUMENT *Result)
{
  udm_rc_t rc= UDM_NOTARGET;

  UDM_GETLOCK(Indexer, UDM_LOCK_TARGETS);

  if (Indexer->Conf->url_number <= 0)
    goto unlock; /* UDM_NOTARGET */

  /* Load targets into memory cache */
  if (Indexer->Conf->Targets.cur_row >= Indexer->Conf->Targets.num_rows)
  {
    UdmResultFree(&Indexer->Conf->Targets);
    if(UDM_OK != (rc= UdmQueryAction(Indexer, Query, UDM_QUERYCMD_TARGETS)))
      goto unlock;
    if (Indexer->Conf->Targets.num_rows == 0)
    {
      rc= UDM_NOTARGET;
      goto unlock;
    }
  }

  /* Choose next target */
  if(Indexer->Conf->Targets.num_rows &&
     Indexer->Conf->Targets.cur_row < Indexer->Conf->Targets.num_rows)
  {
    UDM_DOCUMENT *Doc=&Indexer->Conf->Targets.Doc[Indexer->Conf->Targets.cur_row];

    UdmVarListReplaceLst(&Result->Sections,&Doc->Sections,NULL,"*");
    UdmVarListReplaceLst(&Result->Sections,&Indexer->Conf->Sections,NULL,"*");
    UdmVarListReplaceLst(&Result->RequestHeaders, &Doc->RequestHeaders, NULL, "*");

    Indexer->Conf->Targets.cur_row++;
    Indexer->Conf->url_number--;
    rc= UDM_OK;
  }
  else
    rc= UDM_NOTARGET;

unlock:
  UDM_RELEASELOCK(Indexer, UDM_LOCK_TARGETS);
  return rc;
}


static udm_rc_t
UdmVarList2Spider(UDM_SPIDERPARAM *S,UDM_VARLIST *V)
{
  int collect_links;
  S->period    =UdmVarListFindInt(V, "Period",    UDM_DEFAULT_REINDEX_TIME);
  S->maxhops    =UdmVarListFindInt(V, "MaxHops",  UDM_DEFAULT_MAX_HOPS);
  S->doc_per_site= UdmVarListFindInt(V, "MaxDocPerSite",  0);
  S->max_net_errors  =UdmVarListFindInt(V, "MaxNetErrors",  UDM_MAXNETERRORS);
  S->net_error_delay_time  =UdmVarListFindInt(V, "NetErrorDelayTime",UDM_DEFAULT_NET_ERROR_DELAY_TIME);
  S->read_timeout    =UdmVarListFindInt(V, "ReadTimeOut",  UDM_READ_TIMEOUT);
  S->doc_timeout    =UdmVarListFindInt(V, "DocTimeOut",  UDM_DOC_TIMEOUT);
  S->robots.index= UdmVarListFindBool(V, "Index", UDM_TRUE);
  S->robots.follow= UDM_TRUE;
  S->robots.archive= UDM_TRUE;
  S->use_robots= (udm_use_robots_t) UdmVarListFindInt(V, "Robots", UDM_ROBOTS_ALL);
  S->use_clones= UdmVarListFindBool(V, "DetectClones", UDM_FALSE);
  collect_links= UdmVarListFindInt(V, "CollectLinks", UDM_COLLECT_LINKS_DEFAULT);
  S->collect_links_destination= (udm_link_destination_t) (collect_links & UDM_COLLECT_LINKS_ALL_DST);
  S->collect_links_format= (udm_link_format_t) (collect_links & UDM_COLLECT_LINKS_FORMAT);
  S->crawl_delay= UdmVarListFindInt(V, "CrawlDelay", 0);
  S->dns_cache_timeout= UdmVarListFindInt(V, "DNSCacheTimeout", UDM_DNSCACHE_TIMEOUT);
  S->link_sources_to_follow= UdmVarListFindInt(V, "FollowLinks", UDM_LINK_SOURCES_YES);
  S->ajax_links= UdmVarListFindBool(V, "AjaxLinks", UDM_FALSE);
  return UDM_OK;
}


static udm_rc_t
UdmDocImportSections(UDM_AGENT *A, UDM_DOCUMENT *D, UDM_CHARSET *doccs)
{
  udm_rc_t rc= UDM_OK;
  if (doccs && doccs != D->lcs)
  {
    UDM_DOCUMENT IDoc;
    UdmDocInit(&IDoc);
    UdmVarListReplaceLst(&IDoc.Sections, &D->Sections, NULL, "*");
    if (UDM_OK != (rc= UdmDocAction(A, &IDoc, UDM_DOCCMD_SQLIMPORTSEC)))
      goto ex;
    if (IDoc.TextList.nitems)
    {
      size_t i;
      UDM_CONV conv;
      /*
         Convert imported sections from local
         charset to document charset
      */
      UdmConvInit(&conv, D->lcs, doccs);

      for (i= 0; i < IDoc.TextList.nitems; i++)
      {
        UDM_TEXTITEM *I= &IDoc.TextList.Item[i];
        size_t slen= strlen(I->str);
        char *newval= (char*) UdmMalloc(slen * 12 + 1);
        UdmConvHTML(&conv, newval, slen * 12 + 1, I->str, slen + 1);
        UDM_FREE(I->str);
        I->str= newval;
        I->Param.flags= UDM_TEXTLIST_FLAG_NONE;
        UdmTextListAdd(&D->TextList, I);
      }
    }
    UdmDocFree(&IDoc);
  }
  else
  {
    if (UDM_OK != (rc= UdmDocAction(A, D, UDM_DOCCMD_SQLIMPORTSEC)))
      goto ex;
  }
ex:
  return rc;
}



static void
UdmCustomLog(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  const char *logfmt;

  if ((logfmt= UdmVarListFindStr(&Indexer->Conf->Vars, "CustomLog", NULL)))
  {
    UDM_DSTR line;
    char buf[256];
    const char *datefmt;
    struct tm *tim;
    time_t t;
    UdmDSTRInit(&line, 1024);
    /* Lock mutex: localtime and fprintf are not reentrant */
    UDM_GETLOCK(Indexer,UDM_LOCK_LOG);
    t= time(NULL);
    tim= localtime(&t);
    datefmt= UdmVarListFindStr(&Indexer->Conf->Vars, "DateFormat", "%a, %d %b %Y, %X");
    strftime(buf, sizeof(buf), datefmt , tim);
    UdmVarListReplaceStr(&Doc->Sections, "CurrentTime", buf);
    UdmVarListReplaceInt(&Doc->Sections, "PID", (int) getpid());
    UdmDSTRParse(&line, logfmt, &Doc->Sections);
    fprintf(stdout, "%s\n", UdmDSTRPtr(&line));
    UDM_RELEASELOCK(Indexer,UDM_LOCK_LOG);
    UdmDSTRFree(&line);
  }
}


static udm_rc_t
UdmGetAndParseRobotFile(UDM_AGENT *Indexer,
                        UDM_SERVER *Server,
                        UDM_HREFLIST *Sitemap,
                        UDM_DOCUMENT *Doc,
                        const char *UseRobotsTxtURL)
{
  UDM_SERVER  *rServer;
  char    *rurl;
  UDM_DOCUMENT  rDoc;
  size_t          rurlen;
  udm_rc_t rc;
  int status;
  UDM_CONST_STR content;

  UdmDocInit(&rDoc);
  if (UDM_OK != (rc= UdmHTTPBufAlloc(&rDoc.Buf, UdmHTTPBufAllocedSize(&Doc->Buf) + 1)))
    return rc;
  rDoc.Spider= Doc->Spider;

  if (UseRobotsTxtURL)
  {
    rurl = UdmStrdup(UseRobotsTxtURL);
  }
  else
  {
    rurlen= 32 + strlen(UDM_NULL2EMPTY(Doc->CurURL.schema)) + 
                 strlen(UDM_NULL2EMPTY(Doc->CurURL.hostinfo));
    rurl = (char*)UdmMalloc(rurlen);
    snprintf(rurl, rurlen, "%s://%s/robots.txt", 
             UDM_NULL2EMPTY(Doc->CurURL.schema),
             UDM_NULL2EMPTY(Doc->CurURL.hostinfo));
  }

  UdmVarListAddStr(&rDoc.Sections,"URL",rurl);
  UdmVarListReplaceInt(&rDoc.Sections, "URL_ID", UdmStrHash32(rurl));
  UdmURLParse(&rDoc.CurURL,rurl);
  UdmLog(Indexer,UDM_LOG_INFO,"ROBOTS: %s",rurl);

  rServer = UdmServerFind(Indexer, &Indexer->Conf->Servers, rurl, NULL);
  UdmDocAddDocExtraHeaders(&rDoc);
  UdmDocAddConfExtraHeaders(Indexer, &rDoc);

  if (rServer == NULL)
    rServer = Server;
  UdmVarListReplaceLst(&rDoc.Sections, &rServer->Vars, NULL, "*");
  UdmDocAddServExtraHeaders(rServer, &rDoc);
   
  UdmAgentSetTask(Indexer,  "Resolving");
  UDM_THREADINFO(Indexer,"Resolving",rurl);
  UdmDocLookupConn(Indexer,&rDoc);
  UdmVarListLog(Indexer,&rDoc.RequestHeaders,UDM_LOG_DEBUG,"Request");
  UdmAgentSetTask(Indexer,  "Downloading");
  UDM_THREADINFO(Indexer,"Getting",rurl);
  rc= UdmGetURL(Indexer,&rDoc);
  UdmParseHTTPResponse(Indexer,&rDoc);
  UdmDocProcessResponseHeaders(Indexer,&rDoc);
  UdmDocContentDecode(Indexer,&rDoc);
  UdmVarListLog(Indexer,&rDoc.Sections,UDM_LOG_DEBUG,"Response");
  UDM_FREE(rurl);

  status= UdmVarListFindInt(&rDoc.Sections, "Status", 0);

  if (status >= 500)
  {
    /*
      Can't access to robots.txt file.
      Set method of the original URL to VISITLATER,
      and set status of the original URL to the status of robots.txt.
    */
    Doc->method= UDM_METHOD_VISITLATER;
    UdmVarListReplaceInt(&Doc->Sections, "Status", status);
  }
  else if (status == UDM_HTTP_STATUS_OK &&
           !UdmHTTPBufContentToConstStr(&rDoc.Buf, &content))
  {
    UdmAgentSetTask(Indexer, "Parsing robots.txt");
    UDM_GETLOCK(Indexer, UDM_LOCK_ROBOT_CACHE);
    rc= UdmRobotParse(Indexer, rServer, Sitemap,
                      content.str, content.length,
                      UDM_NULL2EMPTY(rDoc.CurURL.hostinfo));
    UDM_RELEASELOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
  }
  else
  {
    UdmAgentSetTask(Indexer, "Cleaning robots.txt");
    UDM_GETLOCK(Indexer, UDM_LOCK_ROBOT_CACHE);
    UdmRobotAddEmpty(&Indexer->Conf->Robots, UDM_NULL2EMPTY(Doc->CurURL.hostinfo));
    UDM_RELEASELOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
  }

  UdmDocFree(&rDoc);
  return rc;
}


/*
  Process robots.txt
  Download if has not been downloaded yet.
  Then check the current document.
*/

static udm_rc_t
UdmRobotsCheck(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_SERVER *Server)
{
  udm_rc_t rc= UDM_OK;
  const char *UseRobotsTxtURL= UdmVarListFindStr(&Server->Vars, "UseRobotsTxtURL", NULL);
  int RobotsDisallowAction= UdmVarListFindInt(&Server->Vars, "RobotsDisallowAction", 0);
  UDM_ROBOT_RULE  *rule;
  int    take_robots;
  size_t mutex_num;

  /* If neither "http" nor "https" nor UseRobotsTxtURL - nothing to do */
  if (strncasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "http", 4) &&
      !UseRobotsTxtURL)
    return UDM_OK;
  
  if (!(Doc->Spider.use_robots & UDM_ROBOTS_TXT))
  {
    UdmLog(Indexer,UDM_LOG_WARN, "robots.txt support is disallowed for '%s'", 
           UDM_NULL2EMPTY(Doc->CurURL.hostinfo));
    UDM_GETLOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
    rc= UdmRobotParse(Indexer, NULL, NULL, NULL, 0, UDM_NULL2EMPTY(Doc->CurURL.hostinfo));
    UDM_RELEASELOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
    return rc;
  }
  
  /*
    hostinfo can be NULL when UseRobotsTxtURL is given
    and the current schema is file:///
  */
  mutex_num= Doc->CurURL.hostinfo ? UdmStrCRC32(Doc->CurURL.hostinfo) : 0;
  mutex_num= UDM_LOCK_ROBOT_FIRST + (mutex_num % UDM_LOCK_ROBOT_COUNT);
  UdmAgentSetTask(Indexer,  "Robots");

  /*
    Get robots.txt
    Note: Don't use "goto" below until mutex_num is unlocked
  */
  UDM_GETLOCK(Indexer, mutex_num);
  UDM_GETLOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
  take_robots= !UdmRobotFind(&Indexer->Conf->Robots, UDM_NULL2EMPTY(Doc->CurURL.hostinfo));
  UDM_RELEASELOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
  if(take_robots)
  {
    size_t i;
    udm_bool_t UseSitemap= UdmVarListFindBool(&Server->Vars, "UseSitemap", UDM_TRUE);
    UDM_HREFLIST Sitemap;
    UdmHrefListInit(&Sitemap);
    rc= UdmGetAndParseRobotFile(Indexer, Server, &Sitemap, Doc,
                                UseRobotsTxtURL);
    for (i= 0; UseSitemap && i < Sitemap.nhrefs; i++)
      UdmHrefListAdd(&Doc->Hrefs, &Sitemap.Href[i]);
    UdmHrefListFree(&Sitemap);
  }
  UDM_RELEASELOCK(Indexer, mutex_num);
        
  if (rc != UDM_OK)
    return rc;

  /* Check whether URL is disallowed by robots.txt */
  UDM_GETLOCK(Indexer,UDM_LOCK_ROBOT_CACHE);
  if ((rule= UdmRobotRuleFind(&Indexer->Conf->Robots, &Doc->CurURL)))
  {
    UdmLog(Indexer, UDM_LOG_WARN, "robots.txt: '%s %s'",
           rule->method == UDM_METHOD_DISALLOW ? "Disallow" : "Allow",
           rule->path);
    if (rule->method == UDM_METHOD_DISALLOW)
    {
      if (RobotsDisallowAction)
      {
        UdmVarListReplaceInt(&Doc->Sections, "Status", RobotsDisallowAction);
        Doc->method= UDM_METHOD_VISITLATER;
      }
      else
        Doc->method= rule->method;
    }
  }
  UDM_RELEASELOCK(Indexer,UDM_LOCK_ROBOT_CACHE);

  return rc;
}


static udm_rc_t
UdmDocGetRemote(UDM_AGENT *A, UDM_DOCUMENT *Doc,
                const char *origurl, const char *aliasurl, const char *alias)
{
  udm_rc_t rc= UDM_OK;
  udm_mp3type_t mp3type= UDM_MP3_UNKNOWN;
  int state, start= Doc->method == UDM_METHOD_CHECKMP3 ||
                    Doc->method == UDM_METHOD_CHECKMP3ONLY ? 1 : 0;
  
  UdmHTTPBufReset(&Doc->Buf);
  if (Doc->method == UDM_METHOD_IMPORTONLY)
  {
    UdmHTTPBufAppendf(&Doc->Buf,
                     "HTTP/1.0 200 OK\r\n"
                     "Content-Type: text/plain; charset=%s\r\n"
                     "X-Method: ImportOnly\r\n\r\n"
                     " ", Doc->lcs->name);
    if(origurl != NULL)
    {
      UdmVarListReplaceStr(&Doc->Sections,"URL",origurl);
      UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(origurl));
      UdmURLParse(&Doc->CurURL,origurl);
    }
    
    UdmParseHTTPResponseAndHeaders(A, Doc);
    return UDM_OK;
  }
  
  for(state= start; state >= 0; state--)
  {
    int status;
    const char *hdr= NULL;
    
    if(state==1)hdr="bytes=0-256";
    if(mp3type==UDM_MP3_TAG)hdr="bytes=-128";
    
    UdmVarListReplaceInt(&Doc->Sections, "Status", UDM_HTTP_STATUS_UNKNOWN);
    
    if(aliasurl != NULL)
    {
      UdmVarListReplaceStr(&Doc->Sections,"URL",alias);
      UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(alias));
      UdmURLParse(&Doc->CurURL,alias);
    }
    
    UdmVarListLog(A, &Doc->RequestHeaders, UDM_LOG_DEBUG, "Request");
      
    if(hdr)
    {
      UdmVarListAddStr(&Doc->RequestHeaders, "Range", hdr);
      UdmLog(A, UDM_LOG_INFO, "Range: [%s]", hdr);
    }
        
    rc= UdmGetURL(A, Doc);
      
    if(hdr)
    {
      UdmVarListDel(&Doc->RequestHeaders, "Range");
    }
    
    if(origurl != NULL)
    {
      UdmVarListReplaceStr(&Doc->Sections,"URL",origurl);
      UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(origurl));
      UdmURLParse(&Doc->CurURL,origurl);
    }
    
    if(rc != UDM_OK)
      return rc;
    
    status= UdmParseHTTPResponseAndHeaders(A, Doc);
    if (status !=206 && status != 200)
      break;
    
    if(state == 1)  /* Needs guessing */
    {
      if(UDM_MP3_UNKNOWN != (mp3type= UdmMP3Type(Doc)))
      {
        UdmVarListReplaceStr(&Doc->Sections,"Content-Type","audio/mpeg");
        if(Doc->method == UDM_METHOD_CHECKMP3ONLY &&
           mp3type != UDM_MP3_TAG)
          break;
      }
      if (Doc->method == UDM_METHOD_CHECKMP3ONLY)
        break;
    }
  }
  return rc;
}


static udm_rc_t
UdmPrepareRegularAndRawSections(UDM_AGENT *A, UDM_DOCUMENT *D)
{
  udm_rc_t rc;
  UdmAgentSetTask(A, "Preparing sections");
  if (UDM_OK == (rc= UdmPrepareSections(A, D)))
    rc= UdmPrepareRawSections(A, D);
  return rc;
}


static udm_rc_t
UdmRegisterChild(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  const UDM_VAR *Sec;
  const char *parent= NULL, *references= NULL;
  int parent_id= 0;

  if ((Sec= UdmVarListFind(&Doc->Sections,"Header.References")) &&
      (references= UdmVarStr(Sec)))
  {
    /* 
      References contains all message IDs of my predecessors,
      space separated my direct parent is the last in the list.
    */
    if ((parent= strrchr(references, ' ')))
    {
      /* parent now points to the space character, skip it */
      ++parent;
    }
    else
    {
      /* there is only one entry in references, so this is my parent */
      parent= references;
    }  
  }
    
  /* get parent from database */
  if (parent && strlen(parent) && strchr(parent,'@'))
  {
    UDM_DOCUMENT Msg;
      
    UdmDocInit(&Msg);
    UdmVarListReplaceStr(&Msg.Sections,"Header.Message-ID",parent);
    rc= UdmDocAction(Indexer, &Msg, UDM_DOCCMD_FINDBYMSG);
    parent_id = UdmVarListFindInt(&Msg.Sections,"ID",0);
    UdmVarListReplaceInt(&Doc->Sections,"Header.Parent-ID",parent_id);
    UdmDocFree(&Msg);
  }
  
  /* Now register me with my parent  */
  if(parent_id)
    rc = UdmDocAction(Indexer, Doc, UDM_DOCCMD_REGCHILD);
  return rc;
}


static udm_rc_t
DocUpdate(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc)
{
  udm_rc_t   rc= UDM_OK;
  int        status=UdmVarListFindInt(&Doc->Sections,"Status",0);
  urlid_t    url_id = (urlid_t)UdmVarListFindInt(&Doc->Sections, "ID", 0);
  time_t     next_index_time= time(NULL) + Doc->Spider.period;
  char       dbuf[UDM_MAXTIMESTRLEN];
  int        use_newsext; 
  udm_doccmd_t action;

  use_newsext= !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"NewsExtensions","no"),"yes");

  if (Doc->method == UDM_METHOD_VISITLATER)
  {
    action= UDM_DOCCMD_SUPDATE;
    goto update;
  }
  
  /* First of all check that URL must be delated */
  if(Doc->method==UDM_METHOD_DISALLOW)
  {
    UdmLog(Indexer,UDM_LOG_ERROR,"Deleting %s", UdmVarListFindStr(&Doc->Sections, "URL", ""));
    rc= UdmDocAction(Indexer, Doc, UDM_DOCCMD_DELETE);
    return rc;
  }

  switch(status){
  
  case 0: /* No HTTP code */
    UdmHostErrorIncrement(Indexer, &Indexer->Conf->Hosts, Doc->connp.hostname);
    UdmLog(Indexer,UDM_LOG_ERROR,"No HTTP response status");
    UdmVarListReplaceInt(&Doc->Sections,"Status",UDM_HTTP_STATUS_INTERNAL_SERVER_ERROR);
    next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
    action= UDM_DOCCMD_SUPDATE;
    break;
  
  case UDM_HTTP_STATUS_OK:                            /* 200 */
  case UDM_HTTP_STATUS_PARTIAL_OK:                    /* 206 */
    if(!UdmVarListFind(&Doc->Sections,"Content-Type"))
    {
      UdmLog(Indexer,UDM_LOG_ERROR,"No Content-type header");
      next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
      UdmVarListReplaceInt(&Doc->Sections,"Status",UDM_HTTP_STATUS_INTERNAL_SERVER_ERROR);
      UdmHostErrorIncrement(Indexer, &Indexer->Conf->Hosts, Doc->connp.hostname);
      action= UDM_DOCCMD_SUPDATE;
      break;
    }
    else
    {
      UdmHostErrorReset(Indexer, &Indexer->Conf->Hosts, Doc->connp.hostname);

      /* Check clones */
      if(Doc->method == UDM_METHOD_GET && Doc->Spider.use_clones)
      {
        urlid_t    origin_id = 0;
        if (UDM_OK != (rc= UdmDocAction(Indexer, Doc, UDM_DOCCMD_FINDORIG)))
          return rc;
        origin_id= (urlid_t)UdmVarListFindInt(&Doc->Sections,"Origin-ID",0);
  
        if(origin_id && origin_id != url_id)
        {
          if (UdmNeedLog(Indexer->Conf, UDM_LOG_EXTRA))
            UdmLog(Indexer, UDM_LOG_EXTRA, "Duplicate Document %s with #%d", 
                   UdmVarListFindStr(&Doc->Sections, "URL", ""), origin_id);
          action= UDM_DOCCMD_UPDCLONE;
          break;
        }
      }
  
      /* Check that document wasn't modified since last indexing */
      if ((UdmVarListFindInt(&Doc->Sections,"crc32", 0) != 0) 
          &&  (UdmVarListFindInt(&Doc->Sections,"crc32old",0)==UdmVarListFindInt(&Doc->Sections,"crc32",0)) 
          &&  (!(Indexer->flags&UDM_FLAG_REINDEX)))
      {
        action= UDM_DOCCMD_SUPDATE;
        break;
      }
  
      /* For NEWS extension: get rec_id from my */
      /* parent out of db (if I have one...)    */
      if (use_newsext && (UDM_OK != (rc= UdmRegisterChild(Indexer, Doc))))
        return rc;

      action= UDM_DOCCMD_LUPDATE;
      break;
    }

  case UDM_HTTP_STATUS_NOT_MODIFIED:                  /* 304 */
    action= UDM_DOCCMD_SUPDATE;
    break;
  
  case UDM_HTTP_STATUS_MULTIPLE_CHOICES:              /* 300 */
  case UDM_HTTP_STATUS_MOVED_PARMANENTLY:             /* 301 */
  case UDM_HTTP_STATUS_MOVED_TEMPORARILY:             /* 302 */
  case UDM_HTTP_STATUS_SEE_OTHER:                     /* 303 */
  case UDM_HTTP_STATUS_USE_PROXY:                     /* 305 */
  case UDM_HTTP_STATUS_BAD_REQUEST:                   /* 400 */
  case UDM_HTTP_STATUS_UNAUTHORIZED:                  /* 401 */
  case UDM_HTTP_STATUS_PAYMENT_REQUIRED:              /* 402 */
  case UDM_HTTP_STATUS_FORBIDDEN:                     /* 403 */
  case UDM_HTTP_STATUS_NOT_FOUND:                     /* 404 */
  case UDM_HTTP_STATUS_METHOD_NOT_ALLOWED:            /* 405 */
  case UDM_HTTP_STATUS_NOT_ACCEPTABLE:                /* 406 */
  case UDM_HTTP_STATUS_PROXY_AUTHORIZATION_REQUIRED:  /* 407 */
  case UDM_HTTP_STATUS_REQUEST_TIMEOUT:               /* 408 */
  case UDM_HTTP_STATUS_CONFLICT:                      /* 409 */
  case UDM_HTTP_STATUS_GONE:                          /* 410 */
  case UDM_HTTP_STATUS_LENGTH_REQUIRED:               /* 411 */
  case UDM_HTTP_STATUS_PRECONDITION_FAILED:           /* 412 */
  case UDM_HTTP_STATUS_REQUEST_ENTITY_TOO_LARGE:      /* 413 */
  case UDM_HTTP_STATUS_REQUEST_URI_TOO_LONG:          /* 414 */  
  case UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE:        /* 415 */
  case UDM_HTTP_STATUS_NOT_IMPLEMENTED:               /* 501 */
  case UDM_HTTP_STATUS_BAD_GATEWAY:                   /* 502 */
  case UDM_HTTP_STATUS_NOT_SUPPORTED:                 /* 505 */  
    /*
      Remove words and outgoing links from database.
      Keep "urlinfo" and "url" values for DBMode=blob searches,
      as well as for cached search results. Information from
      these two tables will be deleted later, after "HoldBadHref"
    */
    action= UDM_DOCCMD_DUPDATE;
    break;
  
  case UDM_HTTP_STATUS_INTERNAL_SERVER_ERROR:         /* 500 */
  case UDM_HTTP_STATUS_SERVICE_UNAVAILABLE:           /* 503 */
  case UDM_HTTP_STATUS_GATEWAY_TIMEOUT:               /* 504 */
  
    /* Keep words in database                */
    /* We'll retry later, maybe host is down */
    UdmHostErrorIncrement(Indexer, &Indexer->Conf->Hosts, Doc->connp.hostname);
    next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
    action= UDM_DOCCMD_SUPDATE;
    break;
  
  default: /* Unknown status, retry later */
    UdmLog(Indexer,UDM_LOG_WARN,"HTTP %d We don't yet know how to handle it, skipped",status);
    action= UDM_DOCCMD_SUPDATE;
    break;
  }

update:

  UdmTime_t2HttpStr(next_index_time, dbuf, sizeof(dbuf));
  UdmVarListReplaceStr(&Doc->Sections,"Next-Index-Time",dbuf);  
  rc= UdmDocAction(Indexer, Doc, action);
  return rc;
}


/* Add document into cache */
static udm_rc_t
UdmDocCacheAdd(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DOC_CACHE);
  return UdmResultAppendDoc(&Indexer->Indexed, Doc);
}


static udm_rc_t
UdmDocCacheFlush(UDM_AGENT *Indexer)
{
  udm_rc_t rc= UDM_OK;
  size_t  docnum;
  UDM_RESULT  *I = &Indexer->Indexed;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DOC_CACHE);

  if (I->num_rows)
  {
    UdmLog(Indexer, UDM_LOG_EXTRA,
           "Flush %d document(s)", (int) I->num_rows);
  }

  for (docnum = 0; docnum < I->num_rows; docnum++)
  {
    /*
      Flush all hrefs from cache in-memory cache into database.
      Note, this must be done before call of StoreCrossWords
      because we need to know all new URL IDs
    */
    UDM_THREADINFO(Indexer, "Updating", UdmVarListFindStr(&I->Doc[docnum].Sections, "URL", ""));
    if (UDM_OK != (rc= DocUpdate(Indexer, &I->Doc[docnum])))
      goto ret;
  }
ret:
  if (Indexer->Indexed.num_rows)
    UdmResultFree(&Indexer->Indexed);
  Indexer->Indexed_memused= 0;
  return rc;
}


static udm_rc_t
UdmDocUpdate(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  size_t maxsize;
  udm_rc_t rc= UDM_OK;
  UDM_RESULT  *I= &Indexer->Indexed;

  UDM_GETLOCK(Indexer, UDM_LOCK_DOC_CACHE);
  maxsize = UdmVarListFindInt(&Indexer->Conf->Vars,"DocMemCacheSize",0) * 1024 * 1024;

  if (maxsize > 0 && Indexer->Indexed_memused > 0)
  {
    UdmLog(Indexer, UDM_LOG_EXTRA,
           "DocCacheSize: %d/%d", (int) Indexer->Indexed_memused, (int) maxsize);
  }

  Indexer->Indexed_memused+= UdmDocUsedMemory(Doc);

  if (Indexer->Indexed_memused >= maxsize || I->num_rows >= 1024)
  {    
    UDM_THREADINFO(Indexer, "Updating", UdmVarListFindStr(&Doc->Sections, "URL", ""));
    if (UDM_OK != (rc= DocUpdate(Indexer, Doc)))
      goto ret;
    rc= UdmDocCacheFlush(Indexer);
  }
  else
  {
    rc= UdmDocCacheAdd(Indexer, Doc);
  }
  UDM_RELEASELOCK(Indexer, UDM_LOCK_DOC_CACHE);
ret:
  return rc;
}


static udm_rc_t
UdmCrawlNextURL(UDM_AGENT *Indexer, UDM_QUERY *Query)
{
  udm_rc_t rc= UDM_OK, parse_res;
  int status= 0;
  UDM_DOCUMENT  Doc;
  UDM_CHARSET *doccs= NULL;
  const char  *url, *alias;
  char  *origurl= NULL, *aliasurl= NULL;
  UDM_SERVER  *Server = NULL;
  char reason[1024]= "";
  char state_param[128];
  char *alstr= NULL;
  const char *lang= NULL;
  size_t maxdocsize= 0;
  UdmDebugEnter();
  
  UdmDocInit(&Doc);
  
  UdmAgentSetTask(Indexer, "Selecting");
  UDM_THREADINFO(Indexer, "Selecting", "");
  /* Need UdmStoreHrefs() to store URLs from indexer.conf and command line */
  if (UDM_OK != (rc= UdmStoreHrefs(Indexer)) ||
      UDM_OK != (rc= Indexer->action) ||
      UDM_OK != (rc= UdmNextTarget(Indexer, Query, &Doc)))
    goto ret;
  
  UdmAgentSetTask(Indexer, "Preparing");
  url=UdmVarListFindStr(&Doc.Sections,"URL","");
  UdmVarListReplaceInt(&Doc.Sections,"crc32old",UdmVarListFindInt(&Doc.Sections,"crc32",0));
  UdmLog(Indexer,UDM_LOG_INFO,"URL: %s",url);
  udm_snprintf(state_param, sizeof(state_param) - 1, "%s", url);
  Indexer->State.param= state_param;
  
#ifdef HAVE_SETPROCTITLE
  /* To see the URL being indexed in "ps" output on FreeBSD */
  /* Do it if single thread version */
  if(!(Indexer->handle)) setproctitle("%s",url);
#endif
  
  
  /* Check that URL has valid syntax */
  if (UDM_OK != (parse_res= UdmURLParse(&Doc.CurURL,url)))
  {
    UdmLog(Indexer, UDM_LOG_WARN, "%s: %s", UdmURLErrorStr(parse_res), url);
    Doc.method= UDM_METHOD_DISALLOW;
    goto flush;
  }
  
  /* Don't index robots.txt */
  if ((Doc.CurURL.filename != NULL) &&
           (!strcmp(Doc.CurURL.filename, "robots.txt")))
  {
    Doc.method= UDM_METHOD_DISALLOW;
    goto flush;
  }
  
  
  /* Find correspondent Server */
  UdmAgentSetTask(Indexer, "Checking Server");
  if ((Server= UdmServerFind(Indexer, &Indexer->Conf->Servers, url, &alstr)))
  {
    UdmVarList2Spider(&Doc.Spider, &Server->Vars);
    Doc.Spider.webspace= Server->webspace;
    UdmDocAddServExtraHeaders(Server, &Doc);
    UdmVarListReplaceLst(&Doc.Sections, &Server->Vars,NULL,"*");
    UdmVarListReplaceInt(&Doc.Sections, "Server_id", Server->site_id); /* TODO: not re-entrant */
  }
  if (!Server)
  {
    UdmLog(Indexer,UDM_LOG_WARN,"No 'Server' command for url");
    Doc.method= UDM_METHOD_DISALLOW;
    goto flush;
  }


  /* Prepare for downloading: collec information from Conf */
  UdmAgentSetTask(Indexer, "Initializing");
  maxdocsize= (size_t) UdmVarListFindInt(&Indexer->Conf->Vars, "MaxDocSize", UDM_MAXDOCSIZE);
  Doc.lcs= Indexer->Conf->lcs;
  Doc.Hrefs.flags= Doc.Spider.collect_links_destination ?
                   UDM_HREFLIST_FLAG_NONUNIQ : 0;
  UdmDocAddConfExtraHeaders(Indexer, &Doc);

  UdmVarListReplaceInt(&Doc.Sections, "MaxHops", Doc.Spider.maxhops);
  UdmVarListReplaceInt(&Doc.Sections, "MaxDocPerSite", Doc.Spider.doc_per_site);

  if (!UdmHTTPBufPtr(&Doc.Buf))
  {
    UdmAgentSetTask(Indexer, "Allocating buffer");
    if (UDM_OK != UdmHTTPBufAlloc(&Doc.Buf, maxdocsize + 1))
    {
      UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory.");
      rc= UDM_ERROR;
      goto ret;
    }
  }
  /* Apply aliases */
  if (alstr != NULL)
  {
    /* Server Primary alias found */
    UdmVarListReplaceStr(&Doc.Sections,"Alias",alstr);
    UdmFree(alstr);
  }
  else
  {
    /* Apply non-primary alias */
    if (UDM_OK != (rc= UdmDocAlias(Indexer,&Doc)))
      goto flush;
  }


  /* Check hops, filters */
  UdmAgentSetTask(Indexer, "Checking filters");
  if (UDM_OK != (rc= UdmDocCheck(Indexer,Server,&Doc)))
    goto ret;

  if (Doc.method == UDM_METHOD_VISITLATER ||
      Doc.method == UDM_METHOD_DISALLOW)
    goto flush;


  /* Replace URL with alias, if exists */
  if((alias = UdmVarListFindStr(&Doc.Sections, "Alias", NULL)))
  {
    const char *u = UdmVarListFindStr(&Doc.Sections, "URL", NULL);
    origurl= (char*)UdmStrdup(u);
    aliasurl= (char*)UdmStrdup(alias);
    UdmLog(Indexer,UDM_LOG_EXTRA,"Alias: '%s'", alias);
    UdmVarListReplaceStr(&Doc.Sections, "URL", alias);
    UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(alias));
    UdmURLParse(&Doc.CurURL, alias);
  }


  /* Resolve */
  UdmAgentSetTask(Indexer,  "Resolving");
  UdmDocLookupConn(Indexer, &Doc);


  /* Check for too many errors on this server */
  if((Doc.connp.net_errors >= Doc.Spider.max_net_errors)&&
     (Doc.Spider.max_net_errors))
  {
    size_t next_index_time= time(NULL) + Doc.Spider.net_error_delay_time;
    char  buf[UDM_MAXTIMESTRLEN];
    UdmLog(Indexer,UDM_LOG_WARN,"Too many network errors (%d) for this server", Doc.connp.net_errors);
    UdmVarListReplaceInt(&Doc.Sections,"Status",UDM_HTTP_STATUS_SERVICE_UNAVAILABLE);
    UdmTime_t2HttpStr((int)next_index_time, buf, sizeof(buf));
    UdmVarListReplaceStr(&Doc.Sections,"Next-Index-Time",buf);
    Doc.method= UDM_METHOD_VISITLATER;
    goto flush;
  }
  
  
  /* Process CrawlDelay */
  {
    time_t now, next;
    if (Doc.Spider.crawl_delay &&
        (now= time(NULL)) <
         (next= Doc.connp.host_last_used + Doc.Spider.crawl_delay))
    {
      time_t diff= next - now;
      UdmLog(Indexer, UDM_LOG_DEBUG, "CrawlDelay %d/%d",
             (int) Doc.Spider.crawl_delay, (int) diff);
      UdmAgentSetTask(Indexer,  "CrawlDelay");
      UDMSLEEP(diff);
    }
  }
  
  
  if (Doc.method == UDM_METHOD_DISALLOW ||
      Doc.method == UDM_METHOD_VISITLATER)
    goto flush;
  
  
  if (Doc.method == UDM_METHOD_IMPORTONLY)
    goto download;

  /* Add "Hostname:" for virtual hosts */
  UdmDocAddDocExtraHeaders(&Doc);


  /*
    Download robots.txt it has not been downloaded before.
    Then check the current URL againts host robot rules.
  */
  if (UDM_OK != (rc= UdmRobotsCheck(Indexer, &Doc, Server)))
    goto ret;


  /* Restore original URL */
  if(origurl != NULL)
  {
    UdmVarListReplaceStr(&Doc.Sections, "URL", origurl);
    UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(origurl));
    UdmURLParse(&Doc.CurURL, origurl);
  }


download:

  if (Doc.method == UDM_METHOD_DISALLOW ||
      Doc.method == UDM_METHOD_VISITLATER)
    goto flush;


  if (!(Indexer->flags & UDM_FLAG_REINDEX))
  {
    const char *l= UdmVarListFindStrNonEmpty(&Doc.Sections,"Last-Modified",NULL);
    if (l) UdmVarListReplaceStr(&Doc.RequestHeaders, "If-Modified-Since", l);
  }

  UdmAgentSetTask(Indexer, "Downloading");
  UDM_THREADINFO(Indexer,"Getting",url);

  if (UDM_OK != (rc= UdmDocGetRemote(Indexer, &Doc,
                                         origurl, aliasurl, alias)))
    goto flush;

  status= UdmVarListFindInt(&Doc.Sections, "Status", 0);
  {
    UDM_CONST_STR content;
    if (UdmHTTPBufContentToConstStr(&Doc.Buf, &content) && status < 500)
    {
      UdmLog(Indexer, UDM_LOG_ERROR, "No data received");
      status= UDM_HTTP_STATUS_SERVICE_UNAVAILABLE;
      UdmVarListReplaceInt(&Doc.Sections,"Status",status);
    }
  }
  /* Increment indexer's download statistics */
  Indexer->nbytes+= UdmHTTPBufSize(&Doc.Buf) /*+
                    UdmVarListFindInt(&Doc.Sections, "HTDB-Content-Length", 0)*/;
  Indexer->ndocs++;


  /*
    Add URL from "Location:" header.
    This is to give a chance for a concurent thread to take it.
  */
  UdmAgentSetTask(Indexer, "Caching links");
  if (UDM_OK != (rc= UdmDocStoreHrefs(Indexer,&Doc)))
    goto ret;

  if (status != UDM_HTTP_STATUS_OK &&
      status != UDM_HTTP_STATUS_PARTIAL_OK)
    goto flush;
  
  
  /* Parse document */
  UdmAgentSetTask(Indexer, "Parsing");
  UDM_THREADINFO(Indexer,"Parsing",url);
  if (UDM_OK != (rc= UdmDocParseContent(Indexer, &Doc)))
    goto ret;

  /* Build langmap for guesser */
  if (Indexer->Conf->LangMaps.nmaps && Doc.method != UDM_METHOD_DISALLOW)
  {
    register size_t t;
    int flag= !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars, "LangMapUpdate", "no"), "yes");
    UdmAgentSetTask(Indexer, "LangMap");
    bzero((void*)Indexer->LangMap, sizeof(UDM_LANGMAP));
    for (t= 0; t <= UDM_LM_HASHMASK; t++)
      Indexer->LangMap->memb[t].index= t;
    for(t= 0; t < Doc.TextList.nitems; t++)
    {
      UDM_TEXTITEM *Item=&Doc.TextList.Item[t];
      UdmBuildLangMap(Indexer->LangMap, Item->str, strlen(Item->str), flag);
    }
  }


  UdmAgentSetTask(Indexer, "Guessing charset");
  {
    const char *csname;

    /* Guess charset */
    UdmGuessCharSet(Indexer, &Doc, &Indexer->Conf->LangMaps, Indexer->LangMap);
    UdmLog(Indexer, UDM_LOG_EXTRA, "Guesser: Lang: %s, Charset: %s",
           lang= UdmVarListFindStr(&Doc.Sections,"Content-Language",""),
           csname= UdmVarListFindStr(&Doc.Sections,"Charset",""));


    /*
      Convert Href variables into LocalCharset.
      Typically, for RSS indexing.
    */
    if ((doccs= UdmGetCharSet(csname)) && doccs != Doc.lcs)
    {
      size_t h;
      UDM_CONV conv;
      UdmConvInit(&conv, doccs, Doc.lcs);
      for (h= 0; h < Doc.Hrefs.nhrefs; h++)
      {
        UdmVarListConvert(&Doc.Hrefs.Href[h].HrefVars, &conv, UDM_RECODE_HTML);
      }
    }
  }


  /* Parse additional content */
  if (Doc.Spider.robots.index)
  {
    /*
      Doc.Spider.index can be unset e.g. in HTML parser by:
      <META NAME="ROBOTS" CONTENT="NOINDEX">
      Don't parse URL and headers in this case.
    */
    UdmAgentSetTask(Indexer, "Parsing URLText");
    UdmParseURLText(Indexer,&Doc);
    UdmAgentSetTask(Indexer, "Parsing HdrText");
    UdmParseHeaders(Indexer,&Doc);
  }


  /* Process "SQLImportSection" */
  UdmAgentSetTask(Indexer, "Importing sections");
  rc= UdmDocImportSections(Indexer, &Doc, doccs);
  if (UDM_OK != rc)
    goto flush;


  /* Create "HTTP.LocalCharsetContent" variable if requested */
  if (UDM_OK != (rc= UdmVarListAddLocalCharsetContentStr("HTTP.LocalCharsetContent", &Doc, doccs)))
    goto flush;


  /* Process "afterguesser" user-defined sections */
  UdmAgentSetTask(Indexer, "Parsing sections");
  UdmUserSectionsParse(Indexer, &Indexer->Conf->SectionGsrMatch, &Doc, doccs);


  if (Doc.method == UDM_METHOD_HREFONLY)
  {
    UdmHTTPBufReset(&Doc.Buf); /* Don't store cached copy */
    Doc.Spider.robots.index= UDM_FALSE;
    goto flush;
  }

  /* Process IndexIf, SkipIf */
  {
    size_t nfilters= Indexer->Conf->SectionFilters.nitems;

    if (nfilters)
    {
      UDM_VARLIST TmpList;
      udm_method_t sec_method;
      
       UdmAgentSetTask(Indexer, "Checking If-filters");
      /* Save original sections into TmpList */
      UdmVarListInit(&TmpList);
      UdmVarListReplaceLst(&TmpList, &Doc.Sections, NULL, "*");
      
      /* Now parse TextList into words and create sections */
      UdmPrepareRegularAndRawSections(Indexer,&Doc);

      /* Now check newly created sections agains IndexIf/SkipIf*/
      sec_method= UdmSectionFilterListFindMethod(&Indexer->Conf->SectionFilters,
                                                 &Doc, reason, sizeof(reason));
      
      if (sec_method != UDM_METHOD_INDEX)
      {
        /* Restore saved sections */
        UdmVarListFree(&Doc.Sections);
        UdmVarListReplaceLst(&Doc.Sections, &TmpList, NULL, "*");
        
        /* Remove unnesessary data */
        if (sec_method == UDM_METHOD_VISITLATER)
        {
          UdmHrefListFree(&Doc.Hrefs);
          Doc.method= sec_method;
        }
        Doc.Spider.robots.index= UDM_FALSE;
      }
      UdmVarListFree(&TmpList);
      UdmLog(Indexer,UDM_LOG_DEBUG,"SectionFilter: %s",reason);
    }
    else
    {
      /* Parse TextList into words and create sections */
      UdmPrepareRegularAndRawSections(Indexer,&Doc);
    }
  }


flush:

  UdmAgentSetTask(Indexer, "Storing links");
  /* Store links to the link cache, then to the database */
  if ((UDM_OK != (rc= UdmDocStoreHrefs(Indexer,&Doc))) ||
      (UDM_OK != (rc= UdmStoreHrefs(Indexer))))
    goto ret;


  UdmAgentSetTask(Indexer, "CustomLog");
  UdmCustomLog(Indexer, &Doc);
  
  /* Free unnecessary information */
  UdmVarListFree(&Doc.RequestHeaders);
  /* UdmTextListFree(&Doc.TextList); Needed for StoreWords */
  /* UdmHTTPBufFree(&Doc.Buf); Needed for new CachedCopy */


  /* Update document in the database */
  UdmAgentSetTask(Indexer, "Updating");
  rc= UdmDocUpdate(Indexer, &Doc);

  /*
    Check action.
    The main thread may have requested to return something else than UDM_OK.
  */
  UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
  if (rc == UDM_OK)
    rc= Indexer->action;
  UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);


ret:
  /* Free all allocated memory */
  UDM_FREE(origurl);
  UDM_FREE(aliasurl);
  UdmDocFree(&Doc);
  Indexer->State.param= NULL;
  UdmAgentSetTask(Indexer, "Unknown");
  UdmDebugReturn(rc);
}


static void
UdmAgentSleep(UDM_AGENT *Indexer, time_t seconds)
{
#ifndef WIN32
      Indexer->nsleepsecs+= seconds - UDMSLEEP(seconds);
#else
      Indexer->nsleepsecs+= seconds;
      UDMSLEEP(seconds);
#endif
}


#include "udm_signals.h"
#include "udm_env.h"

unsigned int seconds= 0; /* To sleep between documents    */
int flags=            0; /* For indexer            */
int total_threads=    0; /* Total threads number         */
int sleep_threads=    0; /* Number of sleepping threads      */
int max_index_time=  -1;
int maxthreads=       1;
UDM_CRAWLER *ThreadCrawlers= NULL;
int thd_errors= 0;

#define UDM_NOTARGETS_SLEEP 10

#ifdef  WIN32
unsigned int __stdcall UdmCrawlerMain(void *arg)
{
  char *str_buf;
#else
void * UdmCrawlerMain(void *arg)
{
#endif
  UDM_CRAWLER *Crawler= (UDM_CRAWLER *) arg;
  UDM_AGENT *Indexer= &Crawler->Agent;
  UDM_QUERY *Query= Crawler->Query;
  udm_rc_t rc= UDM_OK;
  int done= 0;
  int i_sleep= 0;

  while (!done)
  {
    if (max_index_time>=0)
    {
      time_t now;
      time(&now);
      if ((now-Indexer->start_time)>max_index_time)
        break;
    }

    UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
    if (have_sighup)
    {
      /* UdmReloadEnv(Indexer); */
      have_sighup=0;
    }

    if (have_sigint || have_sigterm)
    {
      int z;
      UdmLog(Indexer, UDM_LOG_ERROR, "%s received. Terminating. Please wait...", (have_sigint) ? "SIGINT" : "SIGTERM");
      for (z = 0 ; z < maxthreads; z++)
      {
        if (ThreadCrawlers[z].Agent.handle)
          UdmAgentSetAction(&ThreadCrawlers[z].Agent, UDM_TERMINATED);
      }
      UdmAgentSetAction(Indexer, UDM_TERMINATED);
      have_sigint = have_sigterm = 0;
    }

    if (have_sigusr1)
    {
      UdmEnvIncLogLevel(Indexer->Conf);
      have_sigusr1 = 0;
    }

    if (have_sigusr2)
    {
      UdmEnvDecLogLevel(Indexer->Conf);
      have_sigusr2 = 0;
    }
    UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);

    if (done)break;

    if (rc == UDM_OK || rc == UDM_NOTARGET)
    { /* Possible after bad startup */
      rc= UdmCrawlNextURL(Indexer, Query);
    }

    switch (rc)
    {
      case UDM_OK:
        if (i_sleep)
        {
          UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
          sleep_threads--;
          UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
          i_sleep= 0;
        }
        break;


      case UDM_NOTARGET:
#ifdef HAVE_PTHREAD
        /*
         in multi-threaded environment we
         should wait for a moment when every thread
         has nothing to do
        */
        UDM_GETLOCK(Indexer, UDM_LOCK_DOC_CACHE);
        UdmDocCacheFlush(Indexer);
        UDM_RELEASELOCK(Indexer, UDM_LOCK_DOC_CACHE);

        if (!i_sleep)
        {
          UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
          sleep_threads++;
          UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
          i_sleep= 1;
        }

        UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
        done= (sleep_threads >= total_threads);
        UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
        break;
#else
        done= 1;
        break;
#endif
      case UDM_ERROR:
        thd_errors++;
#ifdef WIN32
        str_buf= (char *)UdmMalloc(1024);
        udm_snprintf(str_buf, 1024, "Error: %s",  UdmEnvErrMsg(Indexer->Conf));
        UdmShowThreadInfo(0, NULL, str_buf);
        UDM_FREE(str_buf);
        UdmShowThreadInfo(Indexer, "Error", UdmEnvErrMsg(Indexer->Conf));
#endif
      case UDM_TERMINATED:
#ifdef WIN32
        UdmShowThreadInfo(Indexer, "Aborted", "");
#endif
      default:
#ifdef HAVE_PTHREAD
        /*
         in multi-threaded environment we
         should wait for a moment when every thread
         has nothing to do
        */
        if (!i_sleep)
        {
          if (rc == UDM_ERROR)
            UdmLog(Indexer,UDM_LOG_ERROR,"Error: '%s'",UdmEnvErrMsg(Indexer->Conf));
          UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
          sleep_threads++;
          UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
          i_sleep=1;
        }
        UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
        done= (sleep_threads>=total_threads);
        UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
        break;
#else
        if (res == UDM_ERROR)
          UdmLog(Indexer,UDM_LOG_ERROR,"Error: '%s'",UdmEnvErrMsg(Indexer->Conf));
        done=1;
#endif
        break;
    }

    if ((seconds)&&(!done))
    {
      UdmAgentSetTask(Indexer, "Sleeping");
      UdmLog(Indexer,UDM_LOG_DEBUG,"Sleeping %d second(s)",seconds);
      UdmAgentSleep(Indexer, seconds);
    }
  
    if ((i_sleep) && (!done))
    {
      size_t n;
      UdmLog(Indexer, UDM_LOG_ERROR, "%s, sleeping %d seconds (%d threads, %d sleeping)",
             (rc == UDM_NOTARGET) ? "No targets" : 
             ((rc == UDM_TERMINATED) ? "Terminating" : "An error occured"),
             UDM_NOTARGETS_SLEEP, total_threads, sleep_threads);
      for (n= 0; n < UDM_NOTARGETS_SLEEP && !done; n++)
      {
        UdmAgentSleep(Indexer, 1);
        UDM_GETLOCK(Indexer, UDM_LOCK_TARGETS);
        if (UdmNumTargets(Indexer) == 0 && sleep_threads >= total_threads)
          done= 1;
        UDM_RELEASELOCK(Indexer, UDM_LOCK_TARGETS);
      }
    }
    UdmAgentSetTask(Indexer, "Unknown");
  }

  if (rc != UDM_ERROR)
  {
    time_t now, sec;
    double M= 0.0, K= 0.0;

    UDM_GETLOCK(Indexer, UDM_LOCK_DOC_CACHE);
    UdmDocCacheFlush(Indexer);
    UDM_RELEASELOCK(Indexer, UDM_LOCK_DOC_CACHE);
    UdmDBAction(Indexer, UDM_DBCMD_CRAWLER_FINALIZE);

    time(&now);
    sec= now - Indexer->start_time - Indexer->nsleepsecs;
    if (sec > 0)
    {
      /* Convert to int64 - conversion from uint64 to double doesn't work on windows */
      M= ((udm_timer_t)Indexer->nbytes) / 1048576.0 / sec;
      if (M < 1.0) K = ((udm_timer_t)Indexer->nbytes) / 1024.0 / sec;
    }
    UdmLog(Indexer, UDM_LOG_ERROR,
           "Done (%d seconds, %u documents, %llu bytes, %5.2f %cbytes/sec.)",
           (int) sec, (unsigned int) Indexer->ndocs, (unsigned long long) Indexer->nbytes,
           (M < 1.0) ? K : M, (M < 1.0) ? 'K' : 'M' );
#if !defined(WIN32) && defined(HAVE_PTHREAD)
    /* wake up sleeping threads*/
    /* 
       This was wrong, because thread may have already exited when
       pthread_kill() is called, which causes undefined behaviour
       and typically crashes.
       This should be rewritten to use pthread_cond_wait()
       and pthread_cond_signal() eventually.
    */
    /*
    {
      int z;
      for (z= 0 ; z < maxthreads; z++)
      {
        UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
        if (threads[z])
          pthread_kill(threads[z], SIGALRM);
        UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
      }
    }
    */
#endif
  }
  UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
  total_threads--;
  UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);

  UdmAgentFree(Indexer);

#ifdef WIN32
  return(0);
#else
  return(NULL);
#endif
}
