/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_lex.h"
#include "udm_ctype.h"


const char *UdmLex2str(udm_lex_t lex)
{
  switch (lex)
  {
    case UDM_LEX_EOF:       return "EOF";
    case UDM_LEX_STRING :   return "STRING";
    case UDM_LEX_ESCAPED_STRING :   return "ESTRING";
    case UDM_LEX_CHAR_LITERAL:      return "CHAR";
    case UDM_LEX_IDENT:     return "IDENT";
    case UDM_LEX_EQ:        return "'='";
    case UDM_LEX_LT:        return "'<'";
    case UDM_LEX_GT:        return "'>'";
    case UDM_LEX_EQ_EQ:     return "'=='";
    case UDM_LEX_NOT_EQ:    return "'!='";
    case UDM_LEX_LE:        return "'<='";
    case UDM_LEX_GE:        return "'>='";
    case UDM_LEX_LB:        return "'['";
    case UDM_LEX_RB:        return "']'";
    case UDM_LEX_SLASH:     return "'/'";
    case UDM_LEX_TEXT:      return "TEXT";
    case UDM_LEX_UINT:      return "UINT";
    case UDM_LEX_QUESTION:  return "'?'";
    case UDM_LEX_EXCLAM:    return "'!'";
    case UDM_LEX_LP:        return "'('";
    case UDM_LEX_RP:        return "')'";
    case UDM_LEX_PLUS:      return "'+'";
    case UDM_LEX_MINUS:     return "'-'";
    case UDM_LEX_ASTERISK:  return "'*'";
    case UDM_LEX_HASH:      return "'#'";
    case UDM_LEX_DOLLAR:    return "'$'";
    case UDM_LEX_PERCENT:   return "'%'";
    case UDM_LEX_AMPERSAND: return "'%'";
    case UDM_LEX_COMMA:     return "','";
    case UDM_LEX_DOT:       return "'.'";
    case UDM_LEX_COLON:     return "':'";
    case UDM_LEX_SEMICOLON: return "';'";
    case UDM_LEX_AT:        return "'@'";
    case UDM_LEX_CARET:     return "'^'";
    case UDM_LEX_LCB:       return "'{'";
    case UDM_LEX_RCB:       return "'}'";
    case UDM_LEX_VBAR:      return "'|'";
    case UDM_LEX_TILDE:     return "'~'";
    case UDM_LEX_INC:       return "++";
    case UDM_LEX_DEC:       return "--";
    case UDM_LEX_MUL_EQ:    return "*=";
    case UDM_LEX_DIV_EQ:    return "/=";
    case UDM_LEX_REM_EQ:    return "%=";
    case UDM_LEX_INC_EQ:    return "+=";
    case UDM_LEX_DEC_EQ:    return "+=";
    case UDM_LEX_AND_EQ:    return "&=";
    case UDM_LEX_OR_EQ:     return "|=";
    case UDM_LEX_XOR_EQ:    return "&=";
    case UDM_LEX_BOOL_OR:   return "||";
    case UDM_LEX_BOOL_AND:  return "||";
    case UDM_LEX_LSHIFT:    return "<<";
    case UDM_LEX_RSHIFT:    return ">>";
    case UDM_LEX_LSHIFT_EQ: return "<<=";
    case UDM_LEX_RSHIFT_EQ: return ">>=";
    case UDM_LEX_IF:        return "if";
    case UDM_LEX_ELSE:      return "else";
    case UDM_LEX_WHILE:     return "while";
    case UDM_LEX_DO:        return "do";
    case UDM_LEX_FOR:       return "for";
    case UDM_LEX_BREAK:     return "break";
    case UDM_LEX_CONTINUE:  return "continue";
    case UDM_LEX_SIZEOF:    return "sizeof";

    case UDM_LEX_AUTO:      return "auto";
    case UDM_LEX_CASE:      return "case";
    case UDM_LEX_CONST:     return "const";
    case UDM_LEX_DEFAULT:   return "default";
    case UDM_LEX_ENUM:      return "enum";
    case UDM_LEX_EXTERN:    return "extern";
    case UDM_LEX_GOTO:      return "goto";
    case UDM_LEX_REGISTER:  return "register";
    case UDM_LEX_RETURN:    return "return";
    case UDM_LEX_STATIC:    return "static";
    case UDM_LEX_STRUCT:    return "struct";
    case UDM_LEX_SWITCH:    return "switch";
    case UDM_LEX_TYPEDEF:   return "typedef";
    case UDM_LEX_UNION:     return "union";
    case UDM_LEX_VOLATILE:  return "volatile";

    case UDM_LEX_CHAR:      return "char";
    case UDM_LEX_DOUBLE:    return "double";
    case UDM_LEX_FLOAT:     return "float";
    case UDM_LEX_INT:       return "int";
    case UDM_LEX_LONG:      return "long";
    case UDM_LEX_SHORT:     return "short";
    case UDM_LEX_SIGNED:    return "signed";
    case UDM_LEX_UNSIGNED:  return "unsigned";

    case UDM_LEX_OPERATOR:  return "operator";

    case UDM_LEX_COUT:      return "cout";
    case UDM_LEX_UNKNOWN:   return "UNKNOWN";
    case UDM_LEX_COMMENT:   return "COMMENT";
  }
  return "UNKNOWN";
}


void UdmLexScannerInit(UDM_LEX_SCANNER *s, const char *str, size_t length)
{
  s->cur= str;
  UdmConstTokenSet(&s->content, str, length);
}


void
UdmLexScannerSkipSpaces(UDM_LEX_SCANNER *s)
{
  for ( ; !UdmLexScannerEOF(s) ; UdmLexScannerShift(s))
  {
    if (!UdmLexScannerIsSpace(UdmLexScannerCur(s)[0]))
      break;
  }
}


static inline udm_bool_t
UdmIsIdentStart(int x)
{
  return ((x >= 'a' && x <= 'z') || (x >= 'A' && x <= 'Z') || x == '_');
}


static inline udm_bool_t
UdmIsIdentBody(int x)
{
  return UdmIsIdentStart(x) || (x >= '0' && x <= '9');
}



static udm_bool_t
UdmLexScannerScanKeywordOrOperator(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a,
                                   const char *str, size_t length,
                                   udm_lex_t type,
                                   udm_bool_t is_keyword)
{
  if (UdmLexScannerCur(s) + length + 1 > UdmLexScannerEnd(s) ||
      memcmp(UdmLexScannerCur(s), str, length) ||
      (is_keyword && UdmIsIdentBody(UdmLexScannerCur(s)[length])))
    return UDM_TRUE;
  for ( ; length; length--)
    UdmLexScannerShift(s);
  a->token.end= UdmLexScannerCur(s);
  a->type= type;
  return UDM_FALSE;  
}


static udm_bool_t
UdmLexScannerScanOperator2(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a,
                           const char *str, size_t length,
                           udm_lex_t type)
{
  return UdmLexScannerScanKeywordOrOperator(s, a, str, length, type, UDM_FALSE);
}


static udm_bool_t
UdmLexScannerScanOneKeyword(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a,
                         const char *str, size_t length,
                         udm_lex_t type)
{
  return UdmLexScannerScanKeywordOrOperator(s, a, str, length, type, UDM_TRUE);
}


/*
  Used keywords:
  --------------
  break
  continue
  do
  else
  for
  if
  sizeof
  while

  C remaining keywords:
  --------------------
  char
  double
  float
  int
  long
  short
  signed
  unsigned
  void

  Unused keywords
  ---------------
  auto
  case
  const
  default
  enum
  extern
  goto
  register
  return
  static
  struct
  switch
  typedef
  union
  volatile

  C++ used keywords:
  ------------------
  operator
  
  C++ remaining keywords:
  -----------------------

  alignas (since C++11)
  alignof (since C++11)
  and
  and_eq
  asm
  auto(1)
  bitand
  bitor
  bool
  case
  catch
  char
  char16_t (since C++11)
  char32_t (since C++11)
  class
  compl
  concept (concepts TS)
  const
  constexpr (since C++11)
  const_cast
  decltype (since C++11)
  default(1)
  delete(1)
  double
  dynamic_cast
  enum
  explicit
  export(1)
  extern
  false
  float
  friend
  goto
  inline
  int
  long
  mutable
  namespace
  new
  noexcept (since C++11)
  not
  not_eq
  nullptr (since C++11)
  or
  or_eq
  private
  protected
  public
  register
  reinterpret_cast
  requires (concepts TS)
  return
  short
  signed
  static
  static_assert (since C++11)
  static_cast
  struct
  switch
  template
  this
  thread_local (since C++11)
  throw
  true
  try
  typedef
  typeid
  typename
  union
  unsigned
  using(1)
  virtual
  void
  volatile
  wchar_t
  xor
  xor_eq 
*/
static udm_bool_t
UdmLexScannerScanKeyword(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("if"), UDM_LEX_IF))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("else"), UDM_LEX_ELSE))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("while"), UDM_LEX_WHILE))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("do"), UDM_LEX_DO))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("for"), UDM_LEX_FOR))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("break"), UDM_LEX_BREAK))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("sizeof"), UDM_LEX_SIZEOF))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("continue"), UDM_LEX_CONTINUE))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("cout"), UDM_LEX_COUT))
    return UDM_FALSE;

  /* Unused keywords */
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("auto"), UDM_LEX_AUTO))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("case"), UDM_LEX_CASE))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("const"), UDM_LEX_CONST))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("default"), UDM_LEX_DEFAULT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("enum"), UDM_LEX_ENUM))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("extern"), UDM_LEX_EXTERN))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("goto"), UDM_LEX_GOTO))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("register"), UDM_LEX_REGISTER))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("return"), UDM_LEX_RETURN))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("static"), UDM_LEX_STATIC))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("struct"), UDM_LEX_STRUCT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("switch"), UDM_LEX_SWITCH))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("typedef"), UDM_LEX_TYPEDEF))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("union"), UDM_LEX_UNION))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("volatile"), UDM_LEX_VOLATILE))
    return UDM_FALSE;

  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("char"), UDM_LEX_CHAR))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("double"), UDM_LEX_DOUBLE))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("float"), UDM_LEX_FLOAT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("int"), UDM_LEX_INT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("long"), UDM_LEX_LONG))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("short"), UDM_LEX_SHORT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("signed"), UDM_LEX_SIGNED))
    return UDM_FALSE;
  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("unsigned"), UDM_LEX_UNSIGNED))
    return UDM_FALSE;

  if (!UdmLexScannerScanOneKeyword(s, a, UDM_CSTR_WITH_LEN("operator"), UDM_LEX_OPERATOR))
    return UDM_FALSE;

  return UDM_TRUE;
}


udm_bool_t
UdmLexScannerScanCOperator(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("=="), UDM_LEX_EQ_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("!="), UDM_LEX_NOT_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("++"), UDM_LEX_INC))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("--"), UDM_LEX_DEC))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("*="), UDM_LEX_MUL_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("/="), UDM_LEX_DIV_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("%="), UDM_LEX_REM_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("+="), UDM_LEX_INC_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("-="), UDM_LEX_DEC_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("&="), UDM_LEX_AND_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("|="), UDM_LEX_OR_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("||"), UDM_LEX_BOOL_OR))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("&&"), UDM_LEX_BOOL_AND))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("^="), UDM_LEX_XOR_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("<<="), UDM_LEX_LSHIFT_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("<<"), UDM_LEX_LSHIFT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN(">>="), UDM_LEX_RSHIFT_EQ))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN(">>"), UDM_LEX_RSHIFT))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("<="), UDM_LEX_LE))
    return UDM_FALSE;
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN(">="), UDM_LEX_GE))
    return UDM_FALSE;
  if (!UdmLexScannerScanKeyword(s, a))
    return UDM_FALSE;
  return UDM_TRUE;
}


/*
  Simple tokens, consisting on one character.
*/
udm_bool_t
UdmLexScannerScanPunctuation(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  switch (UdmLexScannerCur(s)[0])
  {
  case UDM_LEX_EXCLAM:        /*  '!'   0x21 */
  case UDM_LEX_HASH:          /*  '#'   0x23 */
  case UDM_LEX_DOLLAR:        /*  '$'   0x24 */
  case UDM_LEX_PERCENT:       /*  '%'   0x25 */
  case UDM_LEX_AMPERSAND:     /*  '&'   0x26 */
  case UDM_LEX_LP:            /*  '('   0x28 */
  case UDM_LEX_RP:            /*  ')'   0x29 */
  case UDM_LEX_ASTERISK:      /*  '*'   0x2A */
  case UDM_LEX_PLUS:          /*  '+'   0x2B */
  case UDM_LEX_COMMA:         /*  ','   0x2C */
  case UDM_LEX_MINUS:         /*  '-'   0x2D */
  case UDM_LEX_DOT:           /*  '.'   0x2E */
  case UDM_LEX_SLASH:         /*  '/'   0x2F */
  case UDM_LEX_COLON:         /*  ':'   0x3A */
  case UDM_LEX_SEMICOLON:     /*  ';'   0x3B */
  case UDM_LEX_LT:            /*  '<'   0x3C */
  case UDM_LEX_EQ:            /*  '='   0x3D */
  case UDM_LEX_GT:            /*  '>'   0x3E */
  case UDM_LEX_QUESTION:      /*  '?'   0x3F */
  case UDM_LEX_AT:            /*  '@'   0x40 */
  case UDM_LEX_LB:            /*  '['   0x5B */
  case UDM_LEX_RB:            /*  ']'   0x5D */
  case UDM_LEX_CARET:         /*  '^'   0x5E */
  case UDM_LEX_LCB:           /*  '{'   0x7B */
  case UDM_LEX_VBAR:          /*  '|'   0x7C */
  case UDM_LEX_RCB:           /*  '}'   0x7D */
  case UDM_LEX_TILDE:         /*  '~'   0x7E */
    UdmLexScannerShift(s);
    a->token.end= UdmLexScannerCur(s);
    a->type= (udm_lex_t) a->token.str[0];
    return UDM_FALSE;
  default:
    break;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmLexScannerScanStringInternal(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a,
                                udm_bool_t escaped)
{
  UDM_ASSERT(!UdmLexScannerEOF(s));
  if (UdmLexScannerCur(s)[0] != '"' /*&& UdmLexScannerCur(s)[0] != '\''*/)
    return UDM_TRUE;

  a->type= UDM_LEX_STRING;
  UdmLexScannerShift(s);
  for( ; !UdmLexScannerEOF(s); UdmLexScannerShift(s))
  {
    if (escaped && UdmLexScannerCur(s)[0] == '\\')
    {
      UdmLexScannerShift(s);
      a->type= UDM_LEX_ESCAPED_STRING;
      continue;
    }
    if (UdmLexScannerCur(s)[0] == a->token.str[0])
      break;
  }
  a->token.end= UdmLexScannerCur(s);
  if (a->token.str[0] == UdmLexScannerCur(s)[0])
    UdmLexScannerShift(s);
  a->token.str++;
  return UDM_FALSE;
}


udm_bool_t
UdmLexScannerScanString(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  return UdmLexScannerScanStringInternal(s, a, UDM_FALSE);
}


udm_bool_t
UdmLexScannerScanEscapedString(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  return UdmLexScannerScanStringInternal(s, a, UDM_TRUE);
}

udm_bool_t
UdmLexScannerScanChar(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  UDM_ASSERT(!UdmLexScannerEOF(s));
  if (UdmLexScannerCur(s)[0] != '\'')
    return UDM_TRUE;

  UdmLexScannerShift(s);
  for( ; !UdmLexScannerEOF(s); UdmLexScannerShift(s))
  {
    if (UdmLexScannerCur(s)[0] == '\\')
    {
      UdmLexScannerShift(s);
      continue;
    }
    if (UdmLexScannerCur(s)[0] == a->token.str[0])
      break;
  }
  a->token.end= UdmLexScannerCur(s);
  if (a->token.str[0] == UdmLexScannerCur(s)[0])
    UdmLexScannerShift(s);
  a->token.str++;
  a->type= UDM_LEX_CHAR_LITERAL;
  return UDM_FALSE;
}


udm_bool_t
UdmLexScannerScanIdentifier(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  UDM_ASSERT(!UdmLexScannerEOF(s));

  if (!UdmIsIdentStart(UdmLexScannerCur(s)[0]))
    return UDM_TRUE;

  for (UdmLexScannerShift(s) ; !UdmLexScannerEOF(s) ; UdmLexScannerShift(s))
  {
    if (!UdmIsIdentBody(UdmLexScannerCur(s)[0]))
      break;
  }
  a->token.end= UdmLexScannerCur(s);
  a->type= UDM_LEX_IDENT;
  return UDM_FALSE;
}


udm_bool_t
UdmLexScannerScanUnsignedNumber(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  UDM_ASSERT(!UdmLexScannerEOF(s));

  if (!udm_isdigit(UdmLexScannerCur(s)[0]))
    return UDM_TRUE;

  for (UdmLexScannerShift(s) ; !UdmLexScannerEOF(s) ; UdmLexScannerShift(s))
  {
    if (!udm_isdigit(UdmLexScannerCur(s)[0]))
      break;
  }
  a->token.end= UdmLexScannerCur(s);
  a->type= UDM_LEX_UINT; /* TODO34: add ulonglong */
  return UDM_FALSE;
}


udm_bool_t
UdmLexScannerScanCComment(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("//"), UDM_LEX_COMMENT))
  {
    for ( ; !UdmLexScannerEOF(s) ; UdmLexScannerShift(s))
    {
      if (UdmLexScannerCur(s)[0] == '\n')
        break;
    }
    a->token.end= UdmLexScannerCur(s);
    return UDM_FALSE;
  }
  if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("/*"), UDM_LEX_COMMENT))
  {
    for ( ;  !UdmLexScannerEOF(s) ; UdmLexScannerShift(s))
    {
      if (!UdmLexScannerScanOperator2(s, a, UDM_CSTR_WITH_LEN("*/"), UDM_LEX_COMMENT))
        break;
    }
    a->token.end= UdmLexScannerCur(s);
    return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*** XML *******************************************************************/

/*
 NameChar ::= Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender
*/
static inline udm_bool_t
UdmIsXMLNameBody(int x)
{
  return ((x >= 'a' && x <= 'z') ||
          (x >= 'A' && x <= 'Z') ||
          (x >= '0' && x <= '9') ||
           x == '.' ||
           x == '-' ||
           x == '_' ||
           x == ':' ||
           x < 0);  
}


static inline udm_bool_t
UdmIsXMLNameStart(int x)
{
  return ((x >= 'a' && x <= 'z') ||
          (x >= 'A' && x <= 'Z') ||
           x == '_' ||
           x == ':' ||
           x < 0);  
}


/*
  Name ::= (Letter | '_' | ':') (NameChar)*
*/
udm_bool_t
UdmLexScannerScanXMLIdentifier(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a)
{
  UDM_ASSERT(!UdmLexScannerEOF(s));

  if (!UdmIsXMLNameStart(UdmLexScannerCur(s)[0]))
    return UDM_TRUE;

  for (UdmLexScannerShift(s) ; !UdmLexScannerEOF(s) ; UdmLexScannerShift(s))
  {
    if (!UdmIsXMLNameBody(UdmLexScannerCur(s)[0]))
      break;
  }
  a->token.end= UdmLexScannerCur(s);
  a->type= UDM_LEX_IDENT;
  return UDM_FALSE;
}


/*
  Scan until "str" is found in the content.
  Returns UDM_FALSE if the searched string was not found.
  scanner->token.type is set to "on_error" (usually UDM_LEX_EOF).
  Returns UDM_TRUE if the searched string was found.
  scanner->token.type is set to "on_success" (usually UDM_LEX_TEXT).
*/
udm_bool_t
UdmLexScannerScanUntil(UDM_LEX_SCANNER *scanner,
                       const UDM_CONST_STR *str,
                       udm_lex_t on_success,
                       udm_lex_t on_error)
{
  scanner->token.token.str= scanner->cur;
  for ( ; scanner->cur < scanner->content.end; scanner->cur++)
  {
    size_t length= scanner->content.end - scanner->cur;
    if (length >= str->length && !memcmp(scanner->cur, str->str, str->length))
    {
      scanner->token.type= on_success;
      scanner->token.token.end= scanner->cur;
      scanner->cur+= str->length;
      return UDM_TRUE;
    }
  }
  scanner->token.type= on_error;
  scanner->token.token.end= scanner->cur;
  return UDM_FALSE;
}
