/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_log.h"


UDM_API(void)
UdmSetRefProc(UDM_ENV * Conf, 
              void (*_RefProc)(int code,const char *url, const char *ref))
{
  Conf->RefInfo= _RefProc;
}

UDM_API(void)
UdmSetThreadProc(UDM_ENV * Conf,
                 void (*_ThreadInfo)(UDM_AGENT *A,const char *state, const char* str))
{
  Conf->ThreadInfo=_ThreadInfo;
}


UDM_API(void)
UdmShowThreadInfo(UDM_AGENT *A, const char *state, const char *str)
{
  if (A->Conf->ThreadInfo)
    A->Conf->ThreadInfo(A, state, str);
}


typedef struct _code
{
  const char *c_name;
  int          c_val;
} CODE;


#if defined HAVE_SYSLOG_H && defined USE_SYSLOG
static const CODE facilitynames[]={
#ifdef LOG_AUTH
  {"auth",     LOG_AUTH},
#endif
#ifdef LOG_AUTHPRIV
  {"authpriv", LOG_AUTHPRIV},
#endif
#ifdef LOG_CRON
  {"cron",     LOG_CRON},
#endif
#ifdef LOG_DAEMON
  {"daemon",   LOG_DAEMON},
#endif
#ifdef LOG_FTP
  {"ftp",      LOG_FTP},
#endif
#ifdef LOG_KERN
  {"kern",     LOG_KERN},
#endif
#ifdef LOG_LPR
  {"lpr",      LOG_LPR},
#endif
#ifdef LOG_MAIL
  {"mail",     LOG_MAIL},
#endif
#ifdef LOG_NEWS
  {"news",     LOG_NEWS},
#endif
#ifdef LOG_SYSLOG
  {"syslog",   LOG_SYSLOG},
#endif
#ifdef LOG_USER
  {"user",     LOG_USER},
#endif
#ifdef LOG_UUCP
  {"uucp",     LOG_UUCP},
#endif
#ifdef LOG_LOCAL0
  {"local0",   LOG_LOCAL0},
#endif
#ifdef LOG_LOCAL1
  {"local1",   LOG_LOCAL1},
#endif
#ifdef LOG_LOCAL2
  {"local2",   LOG_LOCAL2},
#endif
#ifdef LOG_LOCAL3
  {"local3",   LOG_LOCAL3},
#endif
#ifdef LOG_LOCAL4
  {"local4",   LOG_LOCAL4},
#endif
#ifdef LOG_LOCAL5
  {"local5",   LOG_LOCAL5},
#endif
#ifdef LOG_LOCAL6
  {"local6",   LOG_LOCAL6},
#endif
#ifdef LOG_LOCAL7
  {"local7",   LOG_LOCAL7},
#endif
  {"none",     UDM_LOG_FACILITY_NONE},
  {NULL,        0},
};


#define UDM_LOG_FACILITY_DEFAULT UDM_LOG_FACILITY_NONE

static int
syslog_facility(const char *f)
{
  const CODE *fn= facilitynames;
  if (!f || !f[0])
    return UDM_LOG_FACILITY_DEFAULT;
  while (fn->c_name != NULL)
  {
    if (strcasecmp(f, fn->c_name) == 0)
      return fn->c_val;
    fn++;
  }
  fprintf(stderr, "Config file error: unknown facility given: '%s'\n\r", f);
  fprintf(stderr, "Will continue with default facility\n\r");
  return UDM_LOG_FACILITY_DEFAULT;
}

#endif /* HAVE_SYSLOG_H */


UDM_API(void)
UdmEnvSetLogLevel(UDM_ENV *Env, int level)
{
  Env->Log.level= level;
}


UDM_API(void)
UdmEnvIncLogLevel(UDM_ENV *Env)
{
  if (Env->Log.level < UDM_LOG_DEBUG)
    Env->Log.level++;
}


UDM_API(void)
UdmEnvDecLogLevel(UDM_ENV *Env)
{
  if (Env->Log.level > 0)
    Env->Log.level--;
}


int
UdmOpenLog(const char * appname,UDM_ENV *Env, int log2stderr)
{
#if defined HAVE_SYSLOG_H && defined USE_SYSLOG
  int openlog_flag= LOG_PID;
  
  Env->Log.flags= UdmVarListFindInt(&Env->Vars, "LogFlags", 0);
  Env->Log.facility= syslog_facility(UdmVarListFindStr(&Env->Vars,"SyslogFacility",""));
  
#ifdef LOG_PERROR
  /* LOG_PERROR supported by 4.3BSD Reno releases and later */
  if (log2stderr)
  {
    if (Env->Log.facility != UDM_LOG_FACILITY_NONE)
    {
      openlog_flag|= LOG_PERROR;
      log2stderr= 0;
    }
  }
#endif

  if (Env->Log.facility != UDM_LOG_FACILITY_NONE)
    openlog(appname ? appname : "<NULL>", openlog_flag, Env->Log.facility);


#endif /* HAVE_SYSLOG_H */

  if (log2stderr)
    Env->Log.logFD= stderr;

  Env->Log.is_log_open= 1;
  return 0;
}


UDM_API(int)
UdmNeedLog(const UDM_ENV *Env, int level)
{
  if (Env->Log.level < level)
    return 0;
  return 1;
}


static int
udm_logger(UDM_ENV *Env, int handle, int level, const char *fmt, va_list ap)
{
  char buf[UDM_LOG_BUF_LEN + 1];
  size_t maxlen= sizeof(buf) - 4; /* for "...\0" */
  int i= 0;

  if (!Env->Log.is_log_open)
    UdmOpenLog("search.cgi", Env, Env->Log.logFD != NULL);

  strcpy(buf + maxlen, "...");
  if (!(Env->Log.flags & UDM_LOG_FLAG_SKIP_PID))
  {
    if (handle)
      i= snprintf(buf, maxlen, "[%d]{%02d} ", (int) getpid(), handle);
    else
      i= snprintf(buf, maxlen,"[%d]{--} ", (int) getpid());
  }
  if (i < 0) /* Extra safety */
    i= 0;
  maxlen-= (size_t) i;

#ifdef WIN32
  _vsnprintf(buf + i, maxlen, fmt, ap);
  if(Env->ThreadInfo)
    Env->ThreadInfo(0,"",buf);
  else
    printf("%s\n",buf);
#else
  vsnprintf(buf + i, maxlen, fmt, ap);
#endif

#if defined HAVE_SYSLOG_H && defined USE_SYSLOG
  if (Env->Log.facility != UDM_LOG_FACILITY_NONE)
    syslog((level != UDM_LOG_ERROR) ? LOG_INFO : LOG_ERR, "%s", buf);
#endif
  if (Env->Log.logFD)
    fprintf(Env->Log.logFD,"%s\n",buf); 
  return 1;
}


void
UdmLog(UDM_AGENT *Agent, int level, const char *fmt, ...)
{
  UDM_GETLOCK(Agent, UDM_LOCK_LOG);
  if (UdmNeedLog(Agent->Conf, level))
  {
    va_list ap;
    va_start(ap,fmt);
    udm_logger(Agent->Conf, Agent->handle,level,fmt,ap);
    va_end(ap);
  }
  UDM_RELEASELOCK(Agent, UDM_LOCK_LOG);
  return;
}


void
UdmEnvLog(UDM_ENV * Env, int level, const char *fmt, ...)
{
  va_list ap;
  if (UdmNeedLog(Env, level))
  {
    va_start(ap,fmt);
    udm_logger(Env, 0, level, fmt, ap);
    va_end(ap);
  }
  return;
}
