/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <fcntl.h>

#ifdef WIN32
#include <io.h>
#endif


#include "udm_common.h"
#include "udm_proto.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_log.h"
#include "udm_http.h"
#include "udm_hash.h"

#ifdef NAME_MAX
#define MIRROR_NAME_MAX NAME_MAX
#else
#define MIRROR_NAME_MAX 255
#endif

typedef struct 
{
  const char *data;
  const char *hdrs;
  size_t name_max;
  int period;
} UDM_MIRROR_PARAM;


static void
UdmMirrorParamInit(UDM_MIRROR_PARAM *M, UDM_AGENT *A, UDM_DOCUMENT *D)
{
  M->name_max= (int) UdmVarListFindInt(&A->Conf->Vars, "MirrorNameMax", MIRROR_NAME_MAX);
  M->period= UdmVarListFindInt(&D->Sections, "MirrorPeriod", -1);
  M->data= UdmVarListFindStr(&D->Sections,"MirrorRoot", NULL);
  M->hdrs= UdmVarListFindStr(&D->Sections,"MirrorHeadersRoot", NULL);
}


static size_t
udm_mirror_filename(char *str, size_t size, size_t name_max,
                    const char *prefix, const UDM_URL *url,
                    const char *suffix1, const char *suffix2)
{
  size_t len= udm_snprintf(str, size, "%s"UDMSLASHSTR"%s"UDMSLASHSTR"%s%s%s%s",
                           prefix,
                           UDM_NULL2EMPTY(url->schema),
                           UDM_NULL2EMPTY(url->hostname), 
                           UDM_NULL2EMPTY(url->path), suffix1, suffix2);
  if (name_max < len)
  {
    udmhash64_t hash= UdmHash64(suffix1, strlen(suffix1));
    unsigned int hi= hash >> 32;
    unsigned int lo= hash & 0xFFFFFFFFULL;
    len= udm_snprintf(str, size, "%s"UDMSLASHSTR"%s"UDMSLASHSTR"%s%s%08X%08X%s",
                      prefix,
                      UDM_NULL2EMPTY(url->schema),
                      UDM_NULL2EMPTY(url->hostname), 
                      UDM_NULL2EMPTY(url->path),
                      hi, lo, suffix2);
  }
  return len;
}


static size_t
udm_mirror_dirname(char *str, size_t size,
                   const char *prefix, const UDM_URL *url)
{
  return udm_snprintf(str, size,"%s"UDMSLASHSTR"%s"UDMSLASHSTR"%s%s",
                      prefix,
                      UDM_NULL2EMPTY(url->schema),
                      UDM_NULL2EMPTY(url->hostname),
                      UDM_NULL2EMPTY(url->path));
}


static size_t
udm_mirror_urlpart_length(const UDM_URL *url)
{
  return strlen(UDM_NULL2EMPTY(url->schema)) +
         strlen(UDM_NULL2EMPTY(url->hostname)) +
         strlen(UDM_NULL2EMPTY(url->path));
}


UDM_API(udm_rc_t)
UdmMirrorGET(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_URL *url)
{
  udm_rc_t rc= UDM_OK;
  int fbody, fheader, have_headers= 0;
  char *str, *estr;
  struct stat sb;
  time_t nowtime= time(NULL);
  size_t str_len, estr_len;
  UDM_MIRROR_PARAM Mirror;

  UdmMirrorParamInit(&Mirror, Indexer, Doc);

  if (Mirror.period <= 0)
    return UDM_MIRROR_NOT_FOUND;

  /* MirrorRoot is not specified, nothing to do */
  if (!Mirror.data)
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "MirrorGet: MirrorRoot is not set");
    return UDM_MIRROR_NOT_FOUND;
  }

  str_len= 128 + strlen(Mirror.data) +
           ((Mirror.hdrs) ? strlen(Mirror.hdrs) : 0) +
           udm_mirror_urlpart_length(url) +
           (estr_len= (url->filename && url->filename[0]) ?
                       3 * strlen(url->filename) + 1 : 16);

  if ((str= (char*) UdmMalloc(str_len)) == NULL)
    return UDM_MIRROR_NOT_FOUND;

  if ((estr= (char*) UdmMalloc(estr_len)) == NULL)
  {
    UdmFree(str);
    return UDM_MIRROR_NOT_FOUND;
  }

  
  UdmURLEncode(estr, str, udm_snprintf(str, str_len, "%s",
                                       url->filename && strlen(url->filename) ?
                                       url->filename : "index.html"));

  udm_mirror_filename(str, str_len, Mirror.name_max, Mirror.data, url, estr, ".body");
  if ((fbody= open(str, O_RDONLY|UDM_BINARY)) == -1)
  {
    UdmLog(Indexer, UDM_LOG_EXTRA, "Mirror file %s not found", str);
    rc= UDM_MIRROR_NOT_FOUND;
    goto end;
  }

  /* Check on file mtime > days ? return */
  if (fstat(fbody, &sb))
  {
    rc= UDM_MIRROR_NOT_FOUND;
    goto end;
  }

  if (nowtime > sb.st_mtime + Mirror.period)
  {
    close(fbody);
    UdmLog(Indexer, UDM_LOG_EXTRA, "%s is older then %d secs", str, Mirror.period);
    rc= UDM_MIRROR_EXPIRED;
    goto end;
  }

  if (Mirror.hdrs)
  {
    udm_mirror_filename(str, str_len, Mirror.name_max, Mirror.hdrs, url, estr, ".header");
    if ((fheader= open(str, O_RDONLY|UDM_BINARY)) >= 0)
    {
      UdmHTTPBufReset(&Doc->Buf);
      UdmHTTPBufAppendFromFile(&Doc->Buf, fheader);
      close(fheader);
      have_headers= 1;
    }
  }

  if (!have_headers)
  {
    /* header file not found, get body as file method */
    UdmHTTPBufReset(&Doc->Buf);
    UdmHTTPBufAppend(&Doc->Buf, UDM_CSTR_WITH_LEN("HTTP/1.0 200 OK\r\n\r\n"));
  }

  rc= UdmHTTPBufAppendFromFile(&Doc->Buf, fbody);
  close(fbody);

end:
  UdmFree(estr);
  UdmFree(str);
  return rc;
}


static udm_rc_t
UdmMirrorWriteToFile(UDM_AGENT *Indexer,
                     char *str, size_t str_len, size_t name_max,
                     const char *prefix,
                     const UDM_URL *url,
                     const char *suffix1,
                     const char *suffix2,
                     const char *src, size_t srclen)
{
  int fd;
  size_t dirlen; /* Can't assign here */
  udm_mirror_dirname(str, str_len, prefix, url);
  if (UdmBuild(str, 0755) != 0) /* TODO: fix UdmBuild() not to change "str" */
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "Can't create dir %s", str);
    return UDM_MIRROR_CANT_BUILD;
  }

  dirlen=  strlen(str); /* Note, UdmBuild() removes trailing slash from "str" */
  UDM_ASSERT(dirlen < str_len);
  if (name_max < (dirlen + udm_snprintf(str + dirlen, str_len - dirlen,
                                        "%s%s%s", UDMSLASHSTR,
                                        suffix1, suffix2)))
    udm_mirror_filename(str, str_len, name_max, prefix, url, suffix1, suffix2);

  if ((fd= open(str, O_CREAT|O_WRONLY|UDM_BINARY,UDM_IWRITE)) == -1)
  {
    UdmLog(Indexer, UDM_LOG_EXTRA, "Can't open mirror file %s", str);
    return UDM_MIRROR_CANT_OPEN;
  }
  write(fd, src, srclen);
  close(fd);
  return UDM_OK;
}


UDM_API(udm_rc_t)
UdmMirrorPUT(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_URL *url)
{
  char *str, *estr;
  size_t str_len, estr_len;
  size_t content_offset;
  udm_rc_t rc= UDM_OK;
  UDM_MIRROR_PARAM Mirror;

  UdmMirrorParamInit(&Mirror, Indexer, Doc);

  if (!Mirror.data)
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "MirrorPUT: MirrorRoot is not set");
    return UDM_ERROR;
  }

  /* Find HTTP header end */
  if (!(content_offset= UdmHTTPBufFindContent(&Doc->Buf)))
    return UDM_ERROR;

  str_len= 128 + strlen(Mirror.data) +
           (Mirror.hdrs ? strlen(Mirror.hdrs) : 0) +
           udm_mirror_urlpart_length(url) +
           (estr_len= (url->filename && url->filename[0]) ?
            3 * strlen(url->filename) + 1 : 16);

  if ((str= (char*)UdmMalloc(str_len)) == NULL)
    return UDM_MIRROR_CANT_BUILD;

  if ((estr= (char*) UdmMalloc(estr_len)) == NULL)
  {
    UdmFree(str);
    return UDM_MIRROR_CANT_BUILD;
  }

  
  UdmURLEncode(estr, str, udm_snprintf(str, str_len, "%s",
                                       (url->filename && strlen(url->filename)) ?
                                       url->filename : "index.html"));

  /* Put Content if MirrorRoot is specified */
  if (Mirror.data)
  {
    UDM_CONST_STR content;
    if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
    {
      rc= UDM_ERROR;
      goto end;
    }
    if (UDM_OK != (rc= UdmMirrorWriteToFile(Indexer, str, str_len,
                                            Mirror.name_max, Mirror.data,
                                            url, estr, ".body",
                                            content.str, content.length)))
      goto end;
  }

  /* Put Headers if MirrorHeadersRoot is specified */
  if (Mirror.hdrs)
  {
    const char *src= UdmHTTPBufPtr(&Doc->Buf);
    size_t srclen= UdmHTTPBufHeaderSize(&Doc->Buf);
    if (UDM_OK != (rc= UdmMirrorWriteToFile(Indexer, str, str_len,
                                            Mirror.name_max, Mirror.hdrs,
                                            url, estr, ".header",
                                            src, srclen)))
      goto end;
  }

end:
  UdmFree(estr);
  UdmFree(str);
  return rc;
}
