/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#include "udm_common.h"
#include "udm_utils.h" /* for bzero */
#include "udm_mutex.h"

#ifdef WIN32
#define UdmThreadSelf              GetCurrentThreadId
#define udm_mutex_t                CRITICAL_SECTION
#define InitMutex(x)               InitializeCriticalSection(x)
#define DestroyMutex(x)            DeleteCriticalSection(x)
#define UDM_MUTEX_LOCK(x)          EnterCriticalSection(x)
#define UDM_MUTEX_UNLOCK(x)        LeaveCriticalSection(x)
#else   /* not WIN32 */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTR_H */

#ifdef HAVE_PTHREAD
#define UdmThreadSelf              pthread_self
#define udm_mutex_t                pthread_mutex_t
#define InitMutex(x)               pthread_mutex_init(x,NULL)
#define DestroyMutex(x)            pthread_mutex_destroy(x)
#define UDM_MUTEX_LOCK(x)          pthread_mutex_lock(x)
#define UDM_MUTEX_UNLOCK(x)        pthread_mutex_unlock(x)
#else
#define udm_mutex_t                int
#define InitMutex(x)               *(x)=0
#define DestroyMutex(x)
#define UDM_MUTEX_LOCK(x)
#define UDM_MUTEX_UNLOCK(x)
#endif /* HAVE_PTHREAD */

#endif /* WIN32 */


typedef struct
{
  udm_thread_t thread;
  udm_mutex_t mutex;
  udm_threadid_t handle;
  int count;
  const char *file;
  int line;
} UDM_MUTEX;

/* #define DEBUG_LOCK */


#ifdef USE_TRACE
#define UDM_TRACE_STREAM(A)		((A)->TR)
#define UDM_TRACE_STREAM_FFLUSH(A)	fflush((A)->TR)
#else
#define UDM_TRACE_STREAM(A)		(stderr)
#define UDM_TRACE_STREAM_FFLUSH(A)
#endif


size_t
UdmMutexNamePrint(char *str, size_t length, udm_mutexno_t id)
{
  static const char *name[UDM_LOCK_ROBOT_FIRST]=
  {
    "dummy#00", "cnf#01", "win#02", "thd#03",
    "sql#04",   "seg#05", "db#06",  "log#07",
    "hrefs#08", "trg#09", "host#0A", "ip#0B",
    "doc#0C", "rbt#0D", "cook#0E"
  };
  if (id < UDM_LOCK_ROBOT_FIRST)
    return udm_snprintf(str, length, "%s", name[id]);
  if (id <= UDM_LOCK_ROBOT_LAST)
    return udm_snprintf(str, length, "rbt#%02X", id);
  if (id <= UDM_LOCK_HOST_LAST)
    return udm_snprintf(str, length, "nsl#%02X", id);
  return udm_snprintf(str, length, "unk#%02X\n", id);
}


static UDM_MUTEX MuMu[UDM_LOCK_MAX];


#ifdef WIN32
__C_LINK udm_mutex_t ThreadProcLock; /* TODO34 WIN: remove this, use MuMu[0] */
#endif


UDM_API(void)
UdmInitMutexes(void)
{
  int i;
#ifdef WIN32
  InitMutex(&ThreadProcLock);
#endif
  for(i=0;i<UDM_LOCK_MAX;i++)
  {
    bzero((void*)&MuMu[i], sizeof(MuMu[i]));
    InitMutex(&MuMu[i].mutex);
  }
}


UDM_API(void)
UdmDestroyMutexes(void)
{
  int i;
  for(i=0;i<UDM_LOCK_MAX;i++)
  {
    DestroyMutex(&MuMu[i].mutex);
  }
#ifdef WIN32
  DestroyMutex(&ThreadProcLock);
#endif
}


static void
UdmMutexLock(UDM_AGENT *A, UDM_MUTEX *M, const char *file, int line)
{
  UDM_ASSERT(M != &MuMu[UDM_LOCK_INTERNAL]);
  UDM_ASSERT(M->thread != UdmThreadSelf());
  UDM_MUTEX_LOCK(&M->mutex);
  UDM_MUTEX_LOCK(&MuMu[UDM_LOCK_INTERNAL].mutex);
  M->thread= UdmThreadSelf();
  M->file= file;
  M->line= line;
  M->count++;
  M->thread= UdmThreadSelf();
  M->handle= A->handle;
  UDM_MUTEX_UNLOCK(&MuMu[UDM_LOCK_INTERNAL].mutex);
}


static void
UdmMutexUnlock(UDM_MUTEX *M)
{
  UDM_ASSERT(M != &MuMu[UDM_LOCK_INTERNAL]);
  UDM_MUTEX_LOCK(&MuMu[UDM_LOCK_INTERNAL].mutex);
  M->count--;
  M->thread= 0;
  M->handle= 0;
  UDM_MUTEX_UNLOCK(&MuMu[UDM_LOCK_INTERNAL].mutex);
  UDM_MUTEX_UNLOCK(&M->mutex);
}


static size_t
UdmMutexStatePrintInternal(char *str, size_t length,
                           const char *state, udm_mutexno_t number,
                           const char *file, int line)
{
  char name[32];
  UdmMutexNamePrint(name, sizeof(name), number);
  return udm_snprintf(str, length, "%s%s%s(%s:%s:%d)%s%s",
                      state[0] == 'O' ? "<b>" : "",
                      number >=  UDM_LOCK_ROBOT_FIRST ? "<font color=\"#DDDDDD\">" : "",
                      state, name, file, line,
                      number >=  UDM_LOCK_ROBOT_FIRST ? "</font>" : "",
                      state[0] == 'O' ? "</b>" : ""); 
}

size_t
UdmMutexStatePrint(char *str, size_t length,
                   UDM_AGENT *A, udm_mutexno_t number)
{
  UDM_MUTEX *M= &MuMu[number];
  UDM_MUTEX_LOCK(&MuMu[UDM_LOCK_INTERNAL].mutex);
  if (M->handle == A->handle)
  {
    length= UdmMutexStatePrintInternal(str, length,
                                       "Owner", number, M->file, M->line);
  }
  else
  {
    length= 0;
  }
  UDM_MUTEX_UNLOCK(&MuMu[UDM_LOCK_INTERNAL].mutex);
  return length;
}


/* CALL-BACK Locking function */
UDM_API(void)
UdmLockProc(UDM_AGENT *A,
            udm_mutexcmd_t command, udm_mutexno_t number,
            const char *fn, int ln)
{
#ifdef DEBUG_LOCK
  if (number != UDM_LOCK_THREAD)
  fprintf(UDM_TRACE_STREAM(A),
          "[%d]{%02d} %2d Try %s\t%s\t%d\n",
          (int) getpid(), A ? A->handle : -1, number,
          (command==UDM_LOCK) ? "lock\t" : "unlock", fn, ln);
  UDM_TRACE_STREAM_FFLUSH(A);
#endif

  switch(command)
  {
    case UDM_LOCK:
      /*fprintf(stderr, "[%d].count=%d\n", type, MuMu[type].count);*/
      {
#ifdef HAVE_DEBUG
        char state_extra[128]= "";
        if (MuMu[number].thread)
        {
          UdmMutexStatePrintInternal(state_extra, sizeof(state_extra),
                                     "Waiting", number, fn, ln);
          A->State.extra= state_extra;
        }
#endif
        UdmMutexLock(A, &MuMu[number], fn, ln);
#ifdef HAVE_DEBUG
        if (state_extra[0])
          A->State.extra= NULL;
#endif
      }
      break;
    case UDM_UNLOCK:
      UdmMutexUnlock(&MuMu[number]);
      break;
    case UDM_CKLOCK:
#ifdef HAVE_DEBUG
      UDM_ASSERT_MUTEX_OWNER(&MuMu[number]);
#endif
      break;
  }

#ifdef DEBUG_LOCK
  if (number != UDM_LOCK_THREAD)
  fprintf(UDM_TRACE_STREAM(A),
         "[%d]{%02d} %2d %s\t%s\t%d\n", (int) getpid(),
         A ? A->handle : -1, number,
         (command==UDM_LOCK) ? "locked\t" : "unlocked\t", fn, ln);
  UDM_TRACE_STREAM_FFLUSH(A);
#endif

}


#if defined(HAVE_PTHREAD) && !defined(WIN32)
udm_rc_t
UdmThreadJoin(void *thd)
{
  return pthread_join((pthread_t) thd, NULL) ? UDM_ERROR : UDM_OK;
}        
#endif


#ifdef HAVE_PTHREAD
udm_rc_t
UdmThreadCreate(void *thd, udm_thread_routine_t start_routine, void *arg)
{
#ifdef WIN32
  {
    int res;
    res= _beginthreadex(NULL, 0, start_routine, arg, 0, NULL);
    UDM_ASSERT(res != -1);
    return res <= 0 ? UDM_ERROR : UDM_OK; /* res must be positive on success */
  }
#else
  {
    int res;
    pthread_attr_t attr;
    size_t stksize= 1024 * 512;
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, stksize);
    res= pthread_create((pthread_t *) thd, &attr, start_routine, arg);
    pthread_attr_destroy(&attr);
    return res ? UDM_ERROR : UDM_OK; /* res must be 0 on success */
  }
#endif
}
#endif
