/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udm_common.h"
#include "udm_textlist.h"
#include "udm_parsehtml.h"
#include "udm_utils.h"
#include "udm_url.h"
#include "udm_match.h"
#include "udm_log.h"
#include "udm_xmalloc.h"
#include "udm_server.h"
#include "udm_hrefs.h"
#include "udm_word.h"
#include "udm_spell.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_uniconv.h"
#include "udm_sgml.h"
#include "udm_guesser.h"
#include "udm_crc32.h"
#include "udm_vars.h"
#include "udm_searchtool.h"
#include "udm_http.h"
#include "udm_stopwords.h"
#include "udm_robots.h"
#include "udm_env.h"

/*

HTML tags:

[+5] New in HTML5
[-5] Deprecated in HTML5


Basic
=====
<body>       Defines the document's body  STF
<br/>        Inserts a single line break  STF
<!--...-->   Defines a comment  STF
<!DOCTYPE>   Defines the document type  STF
<hr/>        Defines a horizontal line  STF (a thematic change in the content)
<html>       Defines an HTML document  STF
<style>      Defines style information for a document  STF

  
Formatting      
==========
<abbr>            Defines an abbreviation or an acronym STF
<acronym>    [-5] Use <abbr> instead. Defines an acronym STF
<address>         Defines contact information for the author/owner of a document  STF
<b>               Defines bold text  STF
<blockquote>      Defines a long quotation  STF
<cite>            Defines a citation  STF
<code>            Defines a piece of computer code  STF
<del>             Defines text that has been deleted from a document  STF
<dfn>             Defines a definition term  STF
<em>              Defines emphasized text   STF
<i>               Defines italic text  STF
<ins>             Defines text that has been inserted into a document  STF
<kbd>             Defines keyboard input  STF
<mark>       [+5] Defines marked/highlighted text
<pre>             Defines preformatted text  STF
<q>               Defines a short quotation  STF
<samp>            Defines sample output from a computer program  STF
<s>               Deprecated. Defines strikethrough text  TF
<small>           Defines smaller text  STF
<strike>     [-5] Deprecated. Use <del> instead. Defines strikethrough text  TF
<strong>          Defines strong text  STF
<sub>             Defines subscripted text  STF
<sup>             Defines superscripted text  STF
<u>               Deprecated. Defines underlined text  TF
<var>             Defines a variable  STF
<wbr>        [+5] Defines a possible line-break
<xmp>             Deprecated. Defines preformatted text   


Fortmatting that is deprecated, use CSS instead
===============================================
<basefont>   [-5] Specifies a default color, size, and font for all text in a document
<big>        [-5] Defines big text STF
<center>     [-5] Defines centered text  TF
<font>       [-5] Defines font, color, and size for text  TF
<tt>         [-5] Defines teletype text  STF


Forms      
=====
<form>            Defines an HTML form for user input  STF
<input/>          Defines an input control  STF
<textarea>        Defines a multiline input control (text area)  STF
<button>          Defines a clickable button  STF
<select>          Defines a drop-down list  STF
<optgroup>        Defines a group of related options in a drop-down list  STF
<option>          Defines an option in a drop-down list  STF
<label>           Defines a label for an <input> element  STF
<fieldset>        Groups related elements in a form  STF
<legend>          Defines a caption for a <fieldset> element  STF
<keygen>     [+5] Defines a key-pair generator field (for forms)
<datalist>   [+5] Specifies a list of pre-defined options for input controls
<dialog>     [+5] Defines a dialog box or window
<menu>            Defines a menu list  TF
<menuitem>   [+5] Defines a command/menu item that the user can invoke from a popup menu
<output>     [+5] Defines the result of a calculation
<meter>      [+5] Defines a scalar measurement within a known range (a gauge)
<progress>   [+5] Represents the progress of a task


Frames      
======
<frame/>     [-5] Defines a window (a frame) in a frameset  F
<frameset>   [-5] Defines a set of frames  F
<noframes>   [-5] Defines an alternate content for users that do not support frames  TF
<iframe>     Defines an inline frame  TF


Images      
======
<img/>       Defines an image  STF
<map>        Defines an image-map   STF
<area/>      Defines an area inside an image-map  STF


Links      
=====
<a>          Defines a hyperlink  STF
<link/>      Defines the relationship between a document and an external resource  STF


Lists      
=====
<ul>         Defines an unordered list  STF
<ol>         Defines an ordered list  STF
<li>         Defines a list item  STF
<dir>   [-5] Use <ul> instead. Defines a directory list TF
<dl>         Defines a definition list  STF
<dt>         Defines an item in a definition list  STF
<dd>         Defines a description of an item in a definition list  STF


Tables      
======
<table>      Defines a table  STF
<caption>    Defines a table caption  STF
<th>         Defines a header cell in a table  STF
<tr>         Defines a row in a table  STF
<td>         Defines a cell in a table  STF
<thead>      Groups the header content in a table  STF
<tbody>      Groups the body content in a table  STF
<tfoot>      Groups the footer content in a table  STF
<col/>       Defines attribute values for one or more columns in a table  STF
<colgroup>   Defines a group of columns in a table for formatting  STF


Sections
========
<article>    [+5] Defines an article
<aside>      [+5] Defines content aside from the page content
<details>    [+5] Defines additional details that the user can view or hide
<div>             Defines a section in a document STF
<figcaption> [+5] Defines a caption for a <figure> element
<figure>     [+5] Specifies self-contained content
<footer>     [+5] Defines a footer for a document or section
<h1> to <h6>      Defines HTML headings  STF
<header>     [+5] Defines a header for a document or section
<hgroup>     [+5] Defines a group of headings
<main>       [+5] Specifies the main content of a document
<nav>        [+5] Defines navigation links
<p>               Defines a paragraph  STF
<section>    [+5] Defines a section in a document
<span>            Defines a section in a document STF
<summary>    [+5] Defines a visible heading for a <details> element


Meta Info      
=========
<head>       Defines information about the document  STF
<title>      Defines the document title  STF
<meta>       Defines metadata about an HTML document  STF
<base/>      Specifies the base URL/target for all relative URLs in a document  STF


Programming      
===========
<canvas>     [+5] Used to draw graphics, on the fly, via scripting (usually JavaScript)
<embed>      [+5] Defines a container for an external (non-HTML) application
<script>          Defines a client-side script  STF
<noscript>        Defines an alternate content for users that do not support client-side scripts  STF
<applet>     [-5] Use <object> instead. Defines an embedded applet. TF
<object>          Defines an embedded object  STF
<param/>          Defines a parameter for an object  STF


Multi-media and other content
=============================
<audio>      [+5] Defines sound content
<video>      [+5] Defines a video or movie
<source>     [+5] Defines multiple media resources for media elements (<video> and <audio>)
<track>      [+5] Defines text tracks for media elements (<video> and <audio>)


Misc
====
<bdi>        [+5] Isolates a part of text that might be formatted in a different direction from other text outside it
<bdo>             Overrides the current text direction  STF
<rp>         [+5] Defines what to show in browsers that do not support ruby annotations
<rt>         [+5] Defines an explanation/pronunciation of characters (for East Asian typography)
<ruby>       [+5] Defines a ruby annotation (for East Asian typography)
<time>       [+5] Defines a date/time

*/


/****************************************************************/

static UDM_CONST_STR spcs= {" ", 1};


static void
UdmSectionGetSeparator(UDM_CONST_STR *str,
                       const UDM_ENV *Env, const char *name)
{
  char sepname[128];
  udm_snprintf(sepname, sizeof(sepname), "separator.%s", name);
  UdmVarListFindConstStr(str, &Env->Vars, sepname, &spcs);
}


static udm_rc_t
UdmSectionAppendSeparator(UDM_SECTION *Sec, const UDM_ENV *Env, const char *name)
{
  if (UdmSectionLength(Sec))
  {
    UDM_CONST_STR sep;
    UdmSectionGetSeparator(&sep, Env, name);
    return UdmSectionAppendStrnWithLimitIfFits(Sec, Env, sep.str, sep.length);
  }
  return UDM_OK;
}


static udm_rc_t
UdmMakeRawSection(UDM_AGENT *A, UDM_DOCUMENT *Doc, const UDM_TEXTITEM *Item)
{
  char secname[128];
  UDM_VAR *Var;
  udm_snprintf(secname, sizeof(secname)-1, "Raw.%s", Item->section_name);
  if ((Var= UdmVarListFindVar(&Doc->Sections, secname)))
  {
    UDM_SECTION *Sec= UdmVarToSection(Var);
    if (Sec && !UdmVarIsReadOnly(Var))
    {
      size_t srclen= strlen(Item->str);
      if (UDM_OK != UdmSectionAppendSeparator(Sec, A->Conf, UdmVarName(Var)) ||
          UDM_OK != UdmSectionAppendStrnWithLimitFittingPart(Sec, A->Conf,
                                                             Item->str, srclen))
        return UDM_ERROR;
    }
  }
  return UDM_OK;
}


udm_rc_t
UdmPrepareRawSections(UDM_AGENT *A, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  size_t i;
  UDM_TEXTLIST *L= &Doc->TextList;
  udm_bool_t raw_sections= UDM_TEST(UdmVarListFindByPrefix(&Doc->Sections, "Raw.", 4) != NULL);
  if (!raw_sections)
    return UDM_OK;
  for (i= 0; i < L->nitems ; i++)
  {
    if (UDM_OK != (rc= UdmMakeRawSection(A, Doc, &L->Item[i])))
      break;
  }
  return rc;
}


static udm_rc_t
UdmPrepareWordsFromUniStr(UDM_AGENT *Indexer,
                          UDM_WORDLIST *Words,
                          int **ustr,
                          UDM_CONV *uni_lcs,
                          int strip_accents,
                          int segmenter,
                          udm_secno_t secno)
{
  int *lt, *tok;
  UDM_UNIDATA *unidata= Indexer->Conf->unidata;
  char lcsword[256];

  /*
    We need to check Doc->Spider.index, because
    user-defined sections must respect 
    <META NAME="Robots" CONTENT="NOINDEX">
  */
  if (secno)
  {
    if (strip_accents)
      UdmUniStrStripAccents(unidata, *ustr);
    UdmUniStrToLower(unidata, *ustr);
    *ustr= UdmUniSegmentByType(Indexer, *ustr, segmenter, ' ');

    for(tok= UdmUniGetToken(unidata, *ustr,&lt);
        tok ;
        tok= UdmUniGetToken(unidata, NULL, &lt))
    {
      size_t cnvlen, tlen= lt - tok;  /* Word length */ 

      /*
        Word lengths are checked later in indexer.c.
        We cannot just skip this block here because
        wordpos should be incremented even if we skip
        a word. Otherwise we have phrase search problem.
      */
      cnvlen= UdmConvHTML(uni_lcs, lcsword, sizeof(lcsword) - 1,
                          (char*) tok, sizeof(*tok) * tlen);
      lcsword[cnvlen]= '\0';

      if (UDM_OK != UdmWordListAdd(Words, lcsword, secno))
        return UDM_ERROR;
    }
  }
  return UDM_OK;
}


static void
NormDec(UDM_CONST_STR *res,
        UDM_VAR *Sec, UDM_TEXTITEM *Item, char *decimal, size_t decimal_size)
{
  if (Sec && (UdmVarFlags(Sec) & UDM_VARFLAG_DECIMAL))
  {
    UdmNormalizeDecimal(decimal, decimal_size, Item->str);
    res->length= strlen(decimal);
    res->str= decimal;
  }
  else
  {
    res->length= strlen(Item->str);
    res->str= Item->str;
  }
}


static void
UdmWordListMarkBadWords(UDM_AGENT *A, UDM_WORDLIST *Words, const char *lang)
{
  size_t wordnum;
  for(wordnum= 0; wordnum < Words->nwords; wordnum++)
  {
    const char  *w= Words->Word[wordnum].word;
    size_t    wlen=strlen(w);
    
    if(wlen > A->Conf->WordParam.max_word_len ||
       wlen < A->Conf->WordParam.min_word_len ||
       UdmStopListListFind(&A->Conf->StopWord, w, lang ) != NULL)
    {
      Words->Word[wordnum].coord.secno= 0;
    }  
  }
}


udm_rc_t
UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc, UDM_WORDLIST *Words)
{
  size_t    i;
  UDM_CHARSET  *doccs;
  UDM_CHARSET  *loccs;
  UDM_CHARSET  *sys_int;
  UDM_CONV  dc_uni;
  UDM_CONV  uni_lc;
  UDM_TEXTLIST  *tlist=&Doc->TextList;
  udm_rc_t rc= UDM_OK;
  UDM_VAR *Sec;
  const char *lang= UdmVarListFindStr(&Doc->Sections, "Content-Language", "");
  const char *seg=  UdmVarListFindStr(&Indexer->Conf->Vars, "Segmenter", NULL);
  int segmenter= UdmUniSegmenterFind(Indexer->Conf, lang, seg);
  int strip_accents= UdmVarListFindBool(&Indexer->Conf->Vars, "StripAccents", UDM_FALSE);

  doccs= UdmVarListFindCharset(&Doc->Sections, "Cached.Charset", NULL);
  if (!doccs) doccs= UdmVarListFindCharset(&Doc->Sections, "Charset", NULL);
  if (!doccs)doccs= UdmVarListFindCharset(&Doc->Sections,"RemoteCharset", NULL);
  if (!doccs)doccs= &udm_charset_latin1;
  if (!(loccs= Doc->lcs))
    loccs= &udm_charset_latin1;
  sys_int= &udm_charset_sys_int;

  UdmConvInit(&dc_uni, doccs, sys_int);
  UdmConvInit(&uni_lc, sys_int, loccs);

  /* Now convert everything to UNICODE format and calculate CRC32 */ 
  for(Sec= NULL, i= 0; i< tlist->nitems && rc == UDM_OK; i++)
  {
    size_t dstlen, cnvlen;
    int    *ustr = NULL;
    UDM_TEXTITEM  *Item=&tlist->Item[i];
    char decimal[128];
    UDM_CONST_STR str;

    if (!Sec || strcasecmp(UdmVarName(Sec), Item->section_name))
      Sec= UdmVarListFindVar(&Doc->Sections, Item->section_name);

    NormDec(&str, Sec, Item, decimal, sizeof(decimal));

    dstlen= (3 * (str.length + 1))*sizeof(int);  /* with '\0' */

    if ((ustr= (int*)UdmMalloc(dstlen)) == NULL)
    {
      UdmLog(Indexer, UDM_LOG_ERROR, "%s:%d Can't alloc %u bytes",
                                     __FILE__, __LINE__, (int) dstlen);
      rc= UDM_ERROR;
      goto ex;
    }

    cnvlen= UdmConv(&dc_uni, (char*) ustr, dstlen, str.str, str.length,
                    (Item->Param.flags & UDM_TEXTLIST_FLAG_HTML) ?
                     UDM_RECODE_HTML_IN : 0) / 4;
    ustr[cnvlen]= 0;
    
    /*
      We need to check Doc->Spider.index, because
      user-defined sections must respect 
      <META NAME="Robots" CONTENT="NOINDEX">
    */
    if (Doc->Spider.robots.index && Item->Param.secno)
    {
      rc= UdmPrepareWordsFromUniStr(Indexer, Words,
                                    &ustr, &uni_lc,
                                    strip_accents,
                                    segmenter, Item->Param.secno);
    }    
    UDM_FREE(ustr);
  }

ex:
  UdmWordListMarkBadWords(Indexer, Words, lang);
  return rc;
}


udm_rc_t
UdmPrepareSections(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
{
  size_t i;
  UDM_CHARSET  *doccs;
  UDM_CHARSET  *loccs;
  UDM_CHARSET  *sys_int;
  UDM_CONV  dc_uni;
  UDM_CONV  uni_lc;
  UDM_TEXTLIST  *tlist=&Doc->TextList;
  udm_rc_t rc= UDM_OK;
  int    crc32=0;
  UDM_VAR *Var;

  doccs= UdmVarListFindCharset(&Doc->Sections, "Cached.Charset", NULL);
  if (!doccs) doccs= UdmVarListFindCharset(&Doc->Sections, "Charset", NULL);
  if (!doccs)doccs= UdmVarListFindCharset(&Doc->Sections,"RemoteCharset", NULL);
  if (!doccs)doccs= &udm_charset_latin1;
  if (!(loccs= Doc->lcs))
    loccs= &udm_charset_latin1;
  sys_int= &udm_charset_sys_int;

  UdmConvInit(&dc_uni, doccs, sys_int);
  UdmConvInit(&uni_lc, sys_int, loccs);


  /* Now convert everything to UNICODE format and calculate CRC32 */ 
  for(Var= NULL, i= 0; i< tlist->nitems && rc == UDM_OK; i++)
  {
    size_t dstlen, cnvlen, ustr_char_len;
    char   *dst;
    int    *ustr = NULL;
    UDM_TEXTITEM  *Item=&tlist->Item[i];
    char decimal[128];
    UDM_CONST_STR str;
    UDM_SECTION *Sec;

    if (!Var || strcasecmp(UdmVarName(Var), Item->section_name))
      Var= UdmVarListFindVar(&Doc->Sections, Item->section_name);

    if (!Var || !(Sec= UdmVarToSection(Var))) /* Not a section variable */
      continue;

    NormDec(&str, Var, Item, decimal, sizeof(decimal));

    dstlen= (3 * (str.length + 1))*sizeof(int);  /* with '\0' */
    
    if ((ustr = (int*)UdmMalloc(dstlen)) == NULL)
    {
      UdmLog(Indexer, UDM_LOG_ERROR,
             "%s:%d Can't alloc %u bytes", __FILE__, __LINE__, (int) dstlen);
      rc= UDM_ERROR;
      goto ex;
    }

    dst=(char*)ustr;
    cnvlen= UdmConv(&dc_uni, dst, dstlen, str.str, str.length,
                    (Item->Param.flags & UDM_TEXTLIST_FLAG_HTML) ? 
                    UDM_RECODE_HTML_IN : 0) / 4;
    ustr[cnvlen]= 0;
    
    ustr_char_len= UdmUniRemoveDoubleSpaces(ustr);

    /*
      Append section text.
      Note, we need to check ustr_char_len,
      because if the original string consisting of entities like "&nbsp;",
      it can turn into an empty string after UdmUniRemoveDoubleSpaces().
      No needs to call UdmReallocSection() in this case,
      to avoid unnecessary extra space delimiter.
    */
    if (Var && !UdmVarIsReadOnly(Var) && ustr_char_len &&
        !(Item->Param.flags & UDM_TEXTLIST_FLAG_SKIP_ADD_SECTION))
    {
      size_t maxlen= UdmEnvSectionMaxLengthEx(Indexer->Conf, UdmVarSecno(Var),
                                              Var->header.name);
      if (UDM_OK != UdmSectionAppendSeparator(Sec, Indexer->Conf, UdmVarName(Var)))
        return UDM_ERROR;
      if (UDM_OK != UdmVarAppendConv(Var, maxlen, &uni_lc, UDM_RECODE_HTML,
                                     (const char *) ustr,
                                     ustr_char_len * sizeof(int)))
        return UDM_ERROR;
    }
    
    
    /*
    TODO for clones detection:
    Replace any separators into space to ignore 
    various pseudo-graphics, commas, semicolons
    and so on to improve clone detection quality
    */
    
    if (Doc->Spider.robots.index &&
        (!Var || !(UdmVarFlags(Var) & UDM_VARFLAG_NOCLONE)))
    {
      crc32= UdmCRC32UpdateUnicode(crc32, ustr, ustr_char_len);
    }
    
    UDM_FREE(ustr);
  }

  UdmVarListReplaceInt(&Doc->Sections,"crc32",crc32);

ex:
  return rc;
}


/**************************** Built-in Parsers ***************************/

/**
  Unescape CGI and convert to HTML
*/
static udm_rc_t
UdmTextListAddWithConversion(UDM_DOCUMENT *Doc, 
                             const UDM_CONST_STR *name,
                             const UDM_CONST_STR *src,
                             size_t secno,
                             UDM_CONV *cnv)
{
  size_t dstlen= src->length * 8 + 1;
  char *tmp= (char*) UdmMalloc(src->length + 1);
  char *dst= (char*) UdmMalloc(dstlen + 1);
  
  if (tmp != NULL && dst != NULL)
  {
    int convres;
    UDM_CONST_TEXTITEM Item;
    UDM_TEXT_PARAM TextParam;
    UdmUnescapeCGIQuery(tmp, src->str); /* TODO34: return len */
    convres= UdmConvHTML(cnv, dst, dstlen, tmp, strlen(tmp) + 1);
    UdmConstTextItemInit(&Item);
    UdmConstStrSet(&Item.section_name, name->str, name->length);
    UdmConstStrSet(&Item.text, dst, convres > 0 ? (size_t) convres : 0);
    UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, secno);
    UdmTextListAddConst(&Doc->TextList, &Item, &TextParam);
  }
  UDM_FREE(tmp);
  UDM_FREE(dst);
  return UDM_OK;  
}


udm_rc_t
UdmParseURLText(UDM_AGENT *A,UDM_DOCUMENT *Doc)
{
  UDM_CONST_TEXTITEM Item;
  const UDM_VAR *Sec;
  UDM_CONV      cnv;
  UDM_VARLIST   *V= &Doc->Sections;
  UDM_CHARSET   *l1= &udm_charset_latin1;
  UDM_CHARSET   *rcs= UdmVarListFindCharset(V, "RemoteCharset", l1);
  UDM_CHARSET   *fscs= UdmVarListFindCharset(V, "RemoteFileNameCharset", rcs);
  UDM_CHARSET   *doccs= UdmVarListFindCharset(V, "CharSet", l1);
  
  UdmConstTextItemInit(&Item);
  
  if ((Sec=UdmVarListFind(&Doc->Sections,"url.proto")))
  {
    UDM_TEXT_PARAM TextParam;
    UdmConstStrSet(&Item.section_name, UDM_CSTR_WITH_LEN("url.proto"));
    UdmConstStrSetStr(&Item.text, UDM_NULL2EMPTY(Doc->CurURL.schema));
    UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_NONE, UdmVarSecno(Sec));
    UdmTextListAddConst(&Doc->TextList, &Item, &TextParam);
  }
  if ((Sec=UdmVarListFind(&Doc->Sections,"url.host")))
  {
    UDM_TEXT_PARAM TextParam;
    UdmConstStrSet(&Item.section_name, UDM_CSTR_WITH_LEN("url.host"));
    UdmConstStrSetStr(&Item.text, UDM_NULL2EMPTY(Doc->CurURL.hostname));
    UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_NONE, UdmVarSecno(Sec));
    UdmTextListAddConst(&Doc->TextList, &Item, &TextParam);
  }
  
  UdmConvInit(&cnv, fscs, doccs);
  if((Sec=UdmVarListFind(&Doc->Sections,"url.path")))
  {
    UDM_CONST_STR src, name;
    UdmConstStrSetStr(&name, UdmVarName(Sec));
    UdmConstStrSetStr(&src, UDM_NULL2EMPTY(Doc->CurURL.path));
    UdmTextListAddWithConversion(Doc, &name, &src, UdmVarSecno(Sec), &cnv);
  }
  if((Sec=UdmVarListFind(&Doc->Sections,"url.file")))
  {
    UDM_CONST_STR src, name;
    UdmConstStrSetStr(&name, UdmVarName(Sec));
    UdmConstStrSetStr(&src, UDM_NULL2EMPTY(Doc->CurURL.filename));
    UdmTextListAddWithConversion(Doc, &name, &src, UdmVarSecno(Sec), &cnv);
  }
  return UDM_OK;
}


udm_rc_t
UdmParseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  size_t i;
  UDM_CONST_TEXTITEM Item;
  
  UdmConstTextItemInit(&Item);
  for (i= 0; i < Doc->Sections.nvars; i++)
  {
    char  secname[128];
    const UDM_VAR *Var= UdmVarListFindConstByIndex(&Doc->Sections, i);
    const UDM_VAR *Sec;
    size_t secnamelen= udm_snprintf(secname, sizeof(secname),
                                    "header.%s", UdmVarName(Var));
    if ((Sec= UdmVarListFind(&Doc->Sections, secname)))
    {
      UDM_TEXT_PARAM TextParam;
      UdmConstStrSet(&Item.section_name, secname, secnamelen);
      UdmVarGetConstStr(Sec, &Item.text);
      UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_NONE, UdmVarSecno(Sec));
      UdmTextListAddConst(&Doc->TextList, &Item, &TextParam);
    }
  }
  return UDM_OK;
}


static const char *
UdmGetLine(const char *s, const char *e, const char **lt)
{
  const char *se;
  if (s == NULL && (s= *lt) == NULL)
    return NULL;

  while (s < e && (*s == '\r' || *s == '\n'))
    s++;

  if (s == e)
  {
    *lt= NULL;
    return NULL;
  }

  for (se= s; se < e; se++)
  {
    if (*se == '\r' || *se == '\n')
      break;
  }
  *lt= se;
  return s;
}


udm_rc_t
UdmParseText(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  UDM_CONST_TEXTITEM  ConstItem;
  UDM_TEXT_PARAM TextParam;
  const UDM_VAR *BSec= UdmVarListFind(&Doc->Sections, "body");
  UDM_CONST_STR content;

  UdmConstTextItemInit(&ConstItem);
  UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_NONE, 0);

  if (BSec && Doc->Spider.robots.index &&
      !UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
  {
    const char *lt= NULL; /* Make "gcc -O3" happy */
    const char *content_end= content.str + content.length;
    TextParam.secno= UdmVarSecno(BSec);
    ConstItem.section_name.str= "body";
    ConstItem.section_name.length= 4;
    ConstItem.text.str= UdmGetLine(content.str, content_end, &lt);
    ConstItem.text.length= lt - ConstItem.text.str;
    while (ConstItem.text.str)
    {
      /*printf("'%.*s'\n", (int) ConstItem.text.length, ConstItem.text.str);*/
      UdmTextListAddConst(&Doc->TextList, &ConstItem, &TextParam);
      ConstItem.text.str= UdmGetLine(NULL, content_end, &lt);
      ConstItem.text.length= lt - ConstItem.text.str;
    }
  }
  return UDM_OK;
}


/*********** Generic HTML Parser *********/
static char spacemap[256]=
{
  0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0  
};

static inline int IsSpaceChar(int x)
{
  return (int) spacemap[x];
}


static int /* Inlining makes GCC prodice a warning in -O3 */
UdmHTMLTagIsOpening(const UDM_HTML_TAG *tag)
{
  UDM_ASSERT(tag->ntoks > 0 && tag->toks[0].name.length > 0);
  return tag->toks[0].name.str[0] != '/';
}


void
UdmHTMLTagInit(UDM_HTML_TAG *tag)
{
  tag->ntoks= 0;
}


size_t
UdmHTMLTagParse(UDM_HTML_TAG *t, const char *src, size_t length)
{
  const char *b, *srcend= src + length;
  
  UDM_ASSERT(length > 0 && *src == '<');

  for (b= src + 1; b < srcend && *b; )
  {
    const char *e;
    const char *valbeg= NULL;
    const char *valend= NULL;
    size_t nt= t->ntoks;

    /* Skip leading spaces */
    for ( ; b < srcend && IsSpaceChar((unsigned char)*b) ; b++);
    if (b == srcend)
      goto ret;

    if (*b == '>') /* End of tag found: <> */
    {
      b++;
      goto ret;
    }

    if (b == srcend || *b == '<') /* Broken tag: << */
      goto ret;

    /* Skip non-spaces, i.e. name */
    for (e= b; e < srcend && *e && !strchr(" =>\t\r\n", *e); e++);
    if (e == b) /* Bad syntax */
      goto ret;

    if (t->ntoks < UDM_MAXTAGVAL)
    {    
      UdmConstStrSet(&t->toks[t->ntoks].name, b, e - b);
      UdmConstStrInit(&t->toks[t->ntoks].value);
      t->ntoks++;
    }

    if (e == srcend || !*e) /* End of input found */
    {
      b= e;
      goto ret;
    }

    if (*e == '>') /* Tag end found */
    {
      b= e + 1;
      goto ret;
    }

    /* Skip spaces */
    for( ; e < srcend && IsSpaceChar((unsigned char) *e) ; e++);
    if (e == srcend)
    {
      b= e;
      goto ret;
    }

    if (*e != '=') /* Attribute without value: <tag attr> */
    {
      b= e;
      /* *lt= b; TODO34 */    /* Don't hang on broken pages */
      continue;
    }

    /* Skip spaces */
    for(b= e + 1; b < srcend && IsSpaceChar((unsigned char) *b); b++);
    if (b == srcend)
      goto ret;

    if (*b == '"') /* Attribute value: double quoted string */
    {
      b++;
      valbeg= b;
      for (e= b; e < srcend && *e && *e != '"'; e++);
      valend= e;
      b= e;
      if (e < srcend && *b == '"') b++;
    }
    else if (*b == '\'') /* Attribute value: quoted string */ 
    {
      b++;
      valbeg= b;
      for (e= b; e < srcend && *e && *e != '\''; e++);
      valend= e;
      b= e;
      if (e < srcend && *b == '\'')b++;
    }
    else /* Attribute value, non-quoted: attr=value */
    {
      valbeg= b;
      for (e= b; e < srcend && *e && !strchr(" >\t\r\n", *e); e++);
      valend= e;
      b= e;
    }
    if (nt < UDM_MAXTAGVAL)
      UdmConstStrSet(&t->toks[nt].value, valbeg, valend - valbeg);
  }

ret:
  return b - src;
}


static int
commentcmp(const char *s1, size_t slen1, const char *s2, size_t slen2)
{
  if (slen1 != slen2)
    return 1;
  return strncasecmp(s1, s2, slen1);
}


static size_t
UdmHTMLCommentParse(UDM_HTML_PARSER_STATE *state,
                    const char *src, const char *srcend)
{
  const char *e;

  for (e= src; e < srcend && *e ; e++)
  {
    if (e + 3 < srcend && e[0] == '-' && e[1] == '-' && e[2] == '>')
    {
      e+= 3;
      if (!commentcmp(src, e - src, UDM_CSTR_WITH_LEN("<!--UdmComment-->")) ||
          !commentcmp(src, e - src, UDM_CSTR_WITH_LEN("<!--noindex-->")) ||
          !commentcmp(src, e - src, UDM_CSTR_WITH_LEN("<!--X-BotPNI-->")))
        state->comment= 1;
      else if(!commentcmp(src, e - src, UDM_CSTR_WITH_LEN("<!--/UdmComment-->")) ||
              !commentcmp(src, e - src, UDM_CSTR_WITH_LEN("<!--/noindex-->")) ||
              !commentcmp(src, e - src, UDM_CSTR_WITH_LEN("<!--X-BotPNI-End-->")))
      state->comment= 0;
      return e - src;
    }
  }

  return e - src; /* Closing --> not found */
}


static size_t
UdmHTMLTextParse(UDM_HTML_PARSER *parser, const char *src, const char *srcend)
{
  const char *e;
  
  /*
    Special case when script body is not commented:    
    <script> x="<"; </script> 
    We should find </script>
  */

  for (e= src; e < srcend && *e; e++)
  {
    if (*e == '<')
    {
      if (parser->state.script)
      {
        if (e + 9 < srcend && !strncasecmp(e, "</script>", 9))
        {
          /* This is when script body  */
          /* is not hidden using <!--  */
          break;
        }
        else if (e + 4 < srcend && !strncmp(e, "<!--", 4))
        {
          /*
            Script body is hidden in a comment but there are
            some spaces between <SCRIPT> and <!--
          */
          break;
        }
      }
      else
      {
        break;
      }
    }
  }
  return e - src;
}


static size_t
UdmHTMLPIParse(UDM_HTML_PARSER *parser, const char *src, const char *srcend)
{
  const char *src0= src;
  for ( ; src + 1 < srcend; src++)
  {
    if (src[0] == '?' && src[1] == '>')
      return src - src0 + 2;
  }
  return src - src0;
}


static void
TagToState(const UDM_HTML_TAG *tag, UDM_HTML_PARSER_STATE *state)
{
  if (tag->ntoks > 0)
  {
    int is_opening= UdmHTMLTagIsOpening(tag);
    const char *beg= tag->toks[0].name.str + (is_opening ? 0 : 1);
    if (!strncasecmp(beg, UDM_CSTR_WITH_LEN("script")))
      state->script= is_opening;
    else if (!strncasecmp(beg, UDM_CSTR_WITH_LEN("noindex")))
      state->comment= is_opening;
    else if (!strncasecmp(beg, UDM_CSTR_WITH_LEN("style")))
      state->style= is_opening;
    else if (!strncasecmp(beg, UDM_CSTR_WITH_LEN("body")))
      state->body= is_opening;
    else if (!strncasecmp(beg, UDM_CSTR_WITH_LEN("title")))
      state->title= is_opening;
  }
}


/*
  TODO: extend UdmHTMLToken() to unserstand ISO-2022-XX sequences, which
  can have '<' and '>' as part of the text when in multi-byte mode, and
  which don't mean the beginning and the end of the tag corrspondingly.
  See uconv-eucjp.c for more details.
*/
static enum udm_html_token
UdmHTMLToken(UDM_HTML_PARSER *parser, const char *src, const char *srcend)
{
  enum udm_html_token toktype;

  if (!src)
    src= parser->lasttok.str + parser->lasttok.length;

  if (src >= srcend)
    return UDM_HTML_TOKEN_UNKNOWN;

  if (*src == '<')
  {
    if (src + 1 < srcend && src[1] == '?')
    {
      toktype= UDM_HTML_TOKEN_PI;
    }
    else
    {
      const char *tmp= src + 1;
      toktype= (*tmp++ == '!' && *tmp++ == '-' && *tmp++== '-') ?
                UDM_HTML_TOKEN_COMMENT : UDM_HTML_TOKEN_TAG;
    }
  }
  else
  {
    if (!*src)
      return UDM_HTML_TOKEN_UNKNOWN;
    toktype= UDM_HTML_TOKEN_TEXT;
  }

  parser->lasttok.str= src;

  switch (toktype)
  {
    case UDM_HTML_TOKEN_TAG:
      UdmHTMLTagInit(&parser->tag);
      parser->lasttok.length= UdmHTMLTagParse(&parser->tag, src, srcend - src);
      if (parser->tag.ntoks == 0 || parser->tag.toks[0].name.length == 0)
      {
        /*
          '<' found, but this is note a real tag.
          This is possible when '<' is not escaped using &amp;
          For example, inside <pre>..</pre>:
            <pre> a < > b </pre>
            <pre> a <= b  </pre>
        */
        parser->lasttok.length= 1;
        if (parser->text_handler)
          parser->text_handler(parser);
        toktype= UDM_HTML_TOKEN_TEXT;
        break;
      }
      TagToState(&parser->tag, &parser->state);
      if (parser->tag_handler)
        parser->tag_handler(parser);
      break;

    case UDM_HTML_TOKEN_COMMENT: /* comment */
      parser->lasttok.length= UdmHTMLCommentParse(&parser->state, src, srcend);
      if (parser->comment_handler)
        parser->comment_handler(parser);
      break;

    case UDM_HTML_TOKEN_TEXT: /* text */
      parser->lasttok.length= UdmHTMLTextParse(parser, src, srcend);
      if (parser->text_handler)
        parser->text_handler(parser);
      break;

    case UDM_HTML_TOKEN_PI:
      parser->lasttok.length= UdmHTMLPIParse(parser, src, srcend);
      if (parser->pi_handler)
        parser->pi_handler(parser);
      break;
    
    case UDM_HTML_TOKEN_UNKNOWN:
      parser->lasttok.length= 0;
      break;
  }
  return toktype;
}


udm_rc_t
UdmHTMLParserExec(UDM_HTML_PARSER *parser, const char *src, size_t srclen)
{
  enum udm_html_token htok;
  const char *srcend= src + srclen;
  for (htok= UdmHTMLToken(parser, src, srcend);
       htok;
       htok= UdmHTMLToken(parser, NULL, srcend))
  {
  }
  return UDM_OK;
}


udm_rc_t
UdmHTMLParserInit(UDM_HTML_PARSER *tag)
{
  bzero((void*)tag, sizeof(*tag));
  return UDM_OK;
}


void
UdmHTMLParserSetUserData(UDM_HTML_PARSER *parser, void *data)
{
  parser->user_data= data;
}


void
UdmHTMLParserSetTextHandler(UDM_HTML_PARSER *p, udm_rc_t (*action)(UDM_HTML_PARSER*))
{
  p->text_handler= action;
}


void
UdmHTMLParserSetCommentHandler(UDM_HTML_PARSER *p, udm_rc_t (*action)(UDM_HTML_PARSER*))
{
  p->comment_handler= action;
}


void
UdmHTMLParserSetTagHandler(UDM_HTML_PARSER *p, udm_rc_t (*action)(UDM_HTML_PARSER*))
{
  p->tag_handler= action;
}


void
UdmHTMLParserSetPIHandler(UDM_HTML_PARSER *p, udm_rc_t (*action)(UDM_HTML_PARSER*))
{
  p->pi_handler= action;
}


/********************************************/

static inline int
tagcmp(const UDM_CONST_STR *str, const char *name)
{
  return strncasecmp(str->str, name, str->length);
}

/*
  Detect tag type by tag name.
  Set flags:
    tag->body
    tag->comment
    tag->style
    tag->script
*/
static enum udm_html_tag
UdmHTMLTagGetType(UDM_HTML_TAG *tag)
{
  UDM_CONST_STR name;

  if (!tag->ntoks || !tag->toks[0].name.str)
    goto ex;

  name= tag->toks[0].name;
  if (name.str[0] == '/') /* Closing tag */
  {
    name.str++;
    name.length--;
  }

  switch (name.length)
  {
    case 1:
      if (!tagcmp(&name, "a"))
        return UDM_HTML_TAG_A;
      if (!tagcmp(&name, "p"))
        return UDM_HTML_TAG_P;
      break;

    case 3:
      if (!tagcmp(&name, "img"))
        return UDM_HTML_TAG_IMG;
      break;

    case 4:
      if (!tagcmp(&name, "html"))
        return UDM_HTML_TAG_HTML;
      if (!tagcmp(&name, "body"))
        return UDM_HTML_TAG_BODY;
      if (!tagcmp(&name, "meta"))
        return UDM_HTML_TAG_META;
      if (!tagcmp(&name, "base"))
        return UDM_HTML_TAG_BASE;
      if (!tagcmp(&name, "link"))
        return UDM_HTML_TAG_LINK;
      if (!tagcmp(&name, "area"))
        return UDM_HTML_TAG_AREA;
      /* if (!tagcmp(&name, "font"))
        return UDM_HTML_TAG_FONT; */ /* TODO34: non-breaking */
      break;

    case 5:
      if (!tagcmp(&name, "title"))
        return UDM_HTML_TAG_TITLE;
      if (!tagcmp(&name, "frame"))
        return UDM_HTML_TAG_FRAME;
      if (!tagcmp(&name, "style"))
        return UDM_HTML_TAG_STYLE;

    case 6:
      if (!tagcmp(&name, "iframe"))
        return UDM_HTML_TAG_IFRAME;
      if (!tagcmp(&name, "script"))
        return UDM_HTML_TAG_SCRIPT;
      break;

    case 7:
      if (!tagcmp(&name, "noindex"))
        return UDM_HTML_TAG_NOINDEX;
      break;
  }

ex:
  return UDM_HTML_TAG_UNKNOWN;
}


/************ HTML parser processing routines for indexing ***********/

static void
UdmDocAddHrefConstStr(UDM_DOCUMENT *Doc,
              const UDM_CONST_STR *href,
              const UDM_CONST_STR *rel,
              const char *linktext,
              udm_link_source_t link_source)
{
  UDM_HREFPARAM HrefParam;
  UDM_VARLIST HrefVars;

  UdmHrefParamInit(&HrefParam);
  UdmVarListInit(&HrefVars);
  /*
    referrer, hops and site_id will be set later,
    in the end of UdmHTMLParse()
  */
  if (linktext)
  {
    urlid_t url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
    UdmVarListAddStr(&HrefVars, "LinkText", linktext);
    UdmVarListAddInt(&HrefVars, "Referrer-ID", url_id);
  }
  if (rel && rel->length)
    UdmVarListAddStrn(&HrefVars, "Rel", rel->str, rel->length);
  HrefParam.link_source= link_source;
  UdmHrefListAddConstStrSGMLUnescape(&Doc->Hrefs, &HrefParam, &HrefVars, href);
  UdmVarListFree(&HrefVars);
}


typedef struct
{
  UDM_DOCUMENT      *Doc;
  UDM_CONST_STR     lasthref;
  UDM_DSTR          linktext;
  size_t            parano;
  UDM_ROBOTSPARAM   orig_robots; /* State before processing robots meta tags */
  udm_bool_t        skip_attribute_sections;
  udm_link_source_t link_source_a;
  UDM_SECTION_PARAM SectionParamBody;
  UDM_SECTION_PARAM SectionParamNoBody;
  UDM_SECTION_PARAM SectionParamTitle;
  UDM_SECTION_PARAM SectionParamBodyOLinkText; /* <a href="xx">Text</a> */
} UDM_HTML_PARSER_PARAM;


static void
UdmHTMLParserParamInit(UDM_HTML_PARSER_PARAM *param, UDM_DOCUMENT *Doc)
{
  const UDM_VAR *BSec= UdmVarListFind(&Doc->Sections, "body");
  const UDM_VAR *NBSec= UdmVarListFind(&Doc->Sections, "nobody");
  const UDM_VAR *TSec= UdmVarListFind(&Doc->Sections, "title");
  const UDM_VAR *OSec= UdmVarListFind(&Doc->Sections, "body.olinktext");
  param->Doc= Doc;
  UdmConstStrInit(&param->lasthref);
  UdmDSTRInit(&param->linktext, 128);
  UdmSectionParamInitFromVar(&param->SectionParamBody, BSec);
  UdmSectionParamInitFromVar(&param->SectionParamNoBody, NBSec);
  UdmSectionParamInitFromVar(&param->SectionParamTitle, TSec);
  /* body.olinktext uses body parameters, if not specified explicitly */
  UdmSectionParamInitFromVar(&param->SectionParamBodyOLinkText,
                             OSec ? OSec : BSec);
  param->skip_attribute_sections=
    UDM_TEST(UdmVarListFindByPrefix(&Doc->Sections, "attribute.", 10) == NULL);
  param->link_source_a= strcmp(UdmVarListFindStr(&Doc->Sections,
                                                "Content-Type.generator", ""),
                              "dirlist") ?
                              UDM_LINK_SOURCE_A_HREF : UDM_LINK_SOURCE_DIR;
  param->parano= 0;
  param->orig_robots= Doc->Spider.robots;
}


static void
UdmHTMLParserParamFree(UDM_HTML_PARSER_PARAM *param)
{
  if (param->lasthref.str)
  {
    /*
      The last "<a href="..">" tag was not closed properly:
      <html>
      <body>
      <a href="..">
      </body>
      </html>
      Add the link without link text.
    */
    UdmDocAddHrefConstStr(param->Doc, &param->lasthref, NULL, NULL,
                          param->link_source_a);
  }

  /* Set site_id, referrer and hops */
  if (param->Doc->Hrefs.nhrefs)
  {
    UDM_DOCUMENT *Doc= param->Doc;
    size_t i;
    UDM_HREF Tmp;
    Tmp.Param.referrer= UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0);
    Tmp.Param.hops= 1 + UdmVarListFindInt(&Doc->Sections,"Hops",0);

    for (i= 0; i < Doc->Hrefs.nhrefs; i++)
    {
      UDM_HREF *Href= &Doc->Hrefs.Href[i];
      Href->Param.referrer= Tmp.Param.referrer;
      Href->Param.hops= Tmp.Param.hops;
    }
  }
  UdmVarListReplaceInt(&param->Doc->Sections, "Parano", param->parano);
  UdmDSTRFree(&param->linktext);
}


typedef struct
{
  UDM_CONST_STR href;
  UDM_CONST_STR rel;
} UDM_HTMLTOK_HELPER;


static void
UdmHTMLParseTagTokens(const UDM_HTML_TAG *tag,
                      UDM_HTMLTOK_HELPER *tokhelper)
{
  size_t i;
  bzero(tokhelper, sizeof(*tokhelper));
  for (i= 0; i < tag->ntoks; i++)
  {
    UDM_CONST_STR name= tag->toks[i].name;
    UDM_CONST_STR value= tag->toks[i].value;

    switch (name.length)
    {
      case 4:
        if (!tagcmp(&name, "href"))
        {
          /* A, LINK, AREA*/
          UdmConstStrTrim(&value, " \t\r\n");
          tokhelper->href= value;
        }
        break;

      case 3:
        if (!tagcmp(&name, "rel"))
        {
          UdmConstStrTrim(&value, " \t\r\n");
          tokhelper->rel= value;
        }
        else if (!tagcmp(&name, "src"))
        {
          /* IMG, FRAME */
          UdmConstStrTrim(&value, " \t\r\n");
          tokhelper->href= value;
        }
    }
  }
}


static void
UdmMetaRobotsParse(UDM_HTML_PARSER_PARAM *param, UDM_CONST_STR *content)
{
  UDM_DOCUMENT *Doc= param->Doc;
  UdmRobotsParamParse(&Doc->Spider.robots, &param->orig_robots, content);
  if (!Doc->Spider.robots.index)
    UdmTextListFree(&Doc->TextList);
}


/*
  Create attribute.xxx sections from a tag.
*/
static void
UdmHTMLParseCreateAttributeSections(UDM_HTML_TAG *tag, UDM_DOCUMENT *Doc)
{
  size_t i;
  UDM_CONST_TEXTITEM Item;
  UDM_TEXT_PARAM TextParam;
  UdmConstTextItemInit(&Item);
  UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, 0);
  for (i= 0; i < tag->ntoks; i++)
  {
    const UDM_VAR *Sec;
    char secname[128];
    size_t secnamelen= udm_snprintf(secname, sizeof(secname), "attribute.%.*s",
                                    (int) tag->toks[i].name.length,
                                    tag->toks[i].name.str);
    if ((Sec= UdmVarListFind(&Doc->Sections, secname)) &&
        Doc->Spider.robots.index)
    {
      UdmConstStrSet(&Item.section_name, secname, secnamelen);
      UdmConstStrSet(&Item.text, tag->toks[i].value.str, tag->toks[i].value.length);
      TextParam.secno= UdmVarSecno(Sec);
      UdmTextListAddConst(&Doc->TextList, &Item, &TextParam);
    }
  }
}


static void
UdmDocAddLang(UDM_DOCUMENT *Doc, UDM_CONST_STR *lang)
{
  char buf[64], *s;
  UDM_CONST_STR tmp= *lang;
  UdmConstStrTrim(&tmp, " \t\r\n");
  if (tmp.length)
  {
    udm_snprintf(buf, sizeof(buf), "%.*s", (int) tmp.length, tmp.str);
    for (s= buf; *s; *s= tolower(*s), s++);
    UdmVarListReplaceStr(&Doc->Sections, "Meta-Language", buf);
  }
}


UDM_CONST_STR *
UdmHTMLTagFindAttrByName(UDM_HTML_TAG *tag, const char *name, size_t namelen)
{
  size_t i;
  for (i= 1; i < tag->ntoks; i++)
  {
    if (!UdmConstStrNCaseCmp(&tag->toks[i].name, name, namelen))
      return &tag->toks[i].value;
  }
  return NULL;
}


UDM_CONST_STR *
UdmHTMLTagFindAttrMetaName(UDM_HTML_TAG *tag)
{
  size_t i;
  for (i= 1; i < tag->ntoks; i++)
  {
    if (!UdmConstStrNCaseCmp(&tag->toks[i].name, UDM_CSTR_WITH_LEN("name")) ||
        !UdmConstStrNCaseCmp(&tag->toks[i].name, UDM_CSTR_WITH_LEN("http-equiv")) ||
        !UdmConstStrNCaseCmp(&tag->toks[i].name, UDM_CSTR_WITH_LEN("property")))
      return &tag->toks[i].value;
  }
  return NULL;
}


static void
UdmHTMLTagParseMeta(UDM_HTML_PARSER_PARAM *param,
                    UDM_CONST_STR *name,
                    UDM_CONST_STR *content,
                    UDM_CONST_STR *href)
{
  UDM_DOCUMENT *Doc= param->Doc;
  if (!UdmConstStrNCaseCmp(name, UDM_CSTR_WITH_LEN("Content-Type")))
  {
    char *p, buf[128];
    udm_snprintf(buf, sizeof(buf), "%.*s", (int) content->length, content->str);
    if ((p= strstr(buf, "charset=")))
    {
      const char *cs= UdmCharsetCanonicalName(UdmTrim(p + 8, "' \t"));
      UdmVarListReplaceStr(&Doc->Sections, "Meta-Charset", cs ? cs : p + 8);
    }
  }
  else if (!UdmConstStrNCaseCmp(name, UDM_CSTR_WITH_LEN("Content-Language")) ||
           !UdmConstStrNCaseCmp(name, UDM_CSTR_WITH_LEN("DC.Language")))
  {
    UdmDocAddLang(Doc, content);
  }
  else if (!UdmConstStrNCaseCmp(name, UDM_CSTR_WITH_LEN("refresh")))
  {
    /* Format: "10; Url=http://something/" */
    /* URL can be written in different     */
    /* forms: URL, url, Url and so on      */
    const char *p;

    if ((p= (const char *) memchr(content->str, '=', content->length)))
    {
      if ((p >= content->str + 3) &&
          (!strncasecmp(p - 3, "URL=",4)))
      {
        size_t length= p + 1 - content->str;
        *href= *content;
        href->str+= length;
        href->length-= length;
      }
    }
  }
  else if (!UdmConstStrNCaseCmp(name, UDM_CSTR_WITH_LEN("robots")) &&
           (Doc->Spider.use_robots & UDM_ROBOTS_META) && content->str)
  {
    UdmMetaRobotsParse(param, content);
  }
}


static void
UdmHTMLTagProcessMeta(UDM_HTML_PARSER_PARAM *param, UDM_HTML_TAG *tag,
                      UDM_HTML_PARSER_STATE *state)
{
  UDM_CONST_STR *metaname, *content, *charset;
  if ((charset= UdmHTMLTagFindAttrByName(tag, UDM_CSTR_WITH_LEN("charset"))))
  {
    /* HTML5: <meta charset="character_set">  */
    char buf[64];
    const char *cs;
    udm_snprintf(buf, sizeof(buf), "%.*s", (int) charset->length, charset->str);
    cs= UdmCharsetCanonicalName(UdmTrim(buf, "' \t"));
    UdmVarListReplaceStr(&param->Doc->Sections, "Meta-Charset", cs ? cs : buf);
  }
  if ((metaname= UdmHTMLTagFindAttrMetaName(tag)) &&
      (content= UdmHTMLTagFindAttrByName(tag, UDM_CSTR_WITH_LEN("content"))))
  {
    UDM_CONST_STR refresh;
    const UDM_VAR *Sec;
    char secname[128];
    size_t secnamelen= udm_snprintf(secname, sizeof(secname), "meta.%.*s",
                                   (int) metaname->length, metaname->str);
    if(!state->comment &&
       (Sec= UdmVarListFind(&param->Doc->Sections, secname)) && 
       param->Doc->Spider.robots.index)
    {
      /*
        We can't call UdmSGMLUnescape(metacont); here,
        because it works only with dcs=lati1.
        Un-SGML will happen later, in Unicode format.
      */
      UDM_CONST_TEXTITEM Item;
      UDM_TEXT_PARAM TextParam;
      UdmConstTextItemInit(&Item);
      UdmConstStrSet(&Item.section_name, secname, secnamelen);
      UdmConstStrSet(&Item.text, content->str, content->length);
      UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, UdmVarSecno(Sec));
      UdmTextListAddConst(&param->Doc->TextList, &Item, &TextParam);
    }
    UdmConstStrInit(&refresh);
    UdmHTMLTagParseMeta(param, metaname, content, &refresh);
    if (refresh.str && param->Doc->Spider.robots.follow)
    {
      /* <meta name="refresh" content="10; URL=http://www.mnogosearch.org/"> */
      UdmDocAddHrefConstStr(param->Doc, &refresh, NULL, NULL, UDM_LINK_SOURCE_META_REFRESH);
    }
  }
}


static void
UdmHTMLTagProcessHref(UDM_HTML_TAG *tag,
                      UDM_HTML_PARSER_STATE *state,
                      UDM_HTML_PARSER_PARAM *param)
{
  UDM_HTMLTOK_HELPER tokhelper;
  UdmHTMLParseTagTokens(tag, &tokhelper);
  if (UdmHTMLTagIsOpening(tag))
  {
    int rel_nofollow= (param->Doc->Spider.use_robots & UDM_ROBOTS_REL) ?
                      !UdmConstStrNCaseCmp(&tokhelper.rel, UDM_CSTR_WITH_LEN("nofollow")) :
                      0;

    if (param->lasthref.str && param->Doc->Spider.robots.follow)
    {
      /* The previous '<a href="..">' was not closed properly. */
      UdmDocAddHrefConstStr(param->Doc, &param->lasthref, NULL, NULL, param->link_source_a);
    }
    UdmDSTRReset(&param->linktext);

    state->ahref= tokhelper.href.str ? UDM_TRUE : UDM_FALSE;

    if (tokhelper.href.str && !rel_nofollow && param->Doc->Spider.robots.follow)
    {
      param->lasthref= tokhelper.href;
    }
    else
    {
      UdmConstStrInit(&param->lasthref);
    }
  }
  else
  {
    if (param->lasthref.str)
      UdmDocAddHrefConstStr(param->Doc, &param->lasthref, &tokhelper.rel,
                            UdmDSTRLength(&param->linktext) ?
                            UdmDSTRPtr(&param->linktext) : NULL,
                            param->link_source_a);
    UdmDSTRReset(&param->linktext);
    UdmConstStrInit(&param->lasthref);
    state->ahref= UDM_FALSE;
  }
}


/*
  Add non-a link.
  These link types do not have link text,
  so we can add the link right now.
  Add the link only if there's no 'nofollow'.
*/
static void
UdmHTMLTagAddLink(UDM_HTML_PARSER_PARAM *param, UDM_HTML_TAG *tag,
                  udm_link_source_t link_source)
{
  UDM_HTMLTOK_HELPER href;
  UdmHTMLParseTagTokens(tag, &href);
  if (UdmHTMLTagIsOpening(tag) && href.href.str &&
      param->Doc->Spider.robots.follow &&
      UdmConstStrNCaseCmp(&href.rel, UDM_CSTR_WITH_LEN("nofollow")))
    UdmDocAddHrefConstStr(param->Doc, &href.href, &href.rel, NULL, link_source);
}


static udm_rc_t
UdmHTMLProcessTag(UDM_HTML_PARSER *parser)
{
  UDM_HTML_PARSER_PARAM *param= (UDM_HTML_PARSER_PARAM *) parser->user_data;
  enum udm_html_tag type= UdmHTMLTagGetType(&parser->tag);

  if (parser->tag.ntoks == 0) /* Bad syntax? */
    return UDM_ERROR;

  if (UdmHTMLTagIsOpening(&parser->tag) && !param->skip_attribute_sections)
    UdmHTMLParseCreateAttributeSections(&parser->tag, param->Doc);    

  switch (type)
  {
  case UDM_HTML_TAG_A:
    UdmHTMLTagProcessHref(&parser->tag, &parser->state, param);
    break;
  case UDM_HTML_TAG_P:
    param->parano++;
    break;

  case UDM_HTML_TAG_HTML:
    {
      UDM_CONST_STR *lang= UdmHTMLTagFindAttrByName(&parser->tag, UDM_CSTR_WITH_LEN("lang"));
      if (lang)
        UdmDocAddLang(param->Doc, lang);
    }
    break;

  case UDM_HTML_TAG_BODY:
    {
      UDM_CONST_STR *lang= UdmHTMLTagFindAttrByName(&parser->tag, UDM_CSTR_WITH_LEN("lang"));
      if (lang)
        UdmDocAddLang(param->Doc, lang);
    }
    break;

  case UDM_HTML_TAG_META:
    UdmHTMLTagProcessMeta(param, &parser->tag, &parser->state);
    break;

  case UDM_HTML_TAG_BASE:
    {
      UDM_HTMLTOK_HELPER href;
      UdmHTMLParseTagTokens(&parser->tag, &href);
      if (href.href.str && href.href.length)
      {
        /*
          Do not add BASE HREF itself into database.
          It will be used only to compose relative links.
        */
        UdmVarListReplaceStrn(&param->Doc->Sections,"base.href",
                              href.href.str, href.href.length);
      }
    }
    break;

  case UDM_HTML_TAG_LINK:
    /*
      <link rel="alternate" type="application/rss+xml"
       title="Link title" href="http://www.mnogosearch.org/"> 
    */
    UdmHTMLTagAddLink(param, &parser->tag, UDM_LINK_SOURCE_LINK_HREF);
    break;

  case UDM_HTML_TAG_AREA:
    /* <area shape="rect" coords="0,0,82,126" href="test.htm" alt="test"> */
    UdmHTMLTagAddLink(param, &parser->tag, UDM_LINK_SOURCE_AREA_HREF);
    break;

  case UDM_HTML_TAG_IMG:
    /* <img src=""> */
    /* TODO34: indexer.conf option to collect images */
    /*UdmHTMLTagAddLink(param, &parser->tag, UDM_LINK_SOURCE_IMG_SRC);*/
    break;

  case UDM_HTML_TAG_IFRAME:
    /* <iframe src=""> */
    UdmHTMLTagAddLink(param, &parser->tag, UDM_LINK_SOURCE_IFRAME_SRC);
    break;

  case UDM_HTML_TAG_FRAME:
    /* <frame src=""> */
    UdmHTMLTagAddLink(param, &parser->tag, UDM_LINK_SOURCE_FRAME_SRC);
    break;

  case UDM_HTML_TAG_SCRIPT:
    /* <script src=""> */
    /* TODO34: indexer.conf option to collect <script src> links */
    /*UdmHTMLTagAddLink(param, &parser->tag, UDM_LINK_SOURCE_SCRIPT_SRC);*/
    break;
    
  case UDM_HTML_TAG_TITLE:
  case UDM_HTML_TAG_STYLE:
  case UDM_HTML_TAG_NOINDEX:
  case UDM_HTML_TAG_FONT:
  case UDM_HTML_TAG_UNKNOWN:
    break;
  }
  return UDM_OK;
}




static UDM_CONST_STR scb= {UDM_CSTR_WITH_LEN("body")};
static UDM_CONST_STR scnb= {UDM_CSTR_WITH_LEN("nobody")};
static UDM_CONST_STR sct= {UDM_CSTR_WITH_LEN("title")};
static UDM_CONST_STR sco= {UDM_CSTR_WITH_LEN("body.olinktext")};

static udm_rc_t
UdmDSTRAppendLinkText(UDM_DSTR *dstr, const char *src, size_t length)
{
  const char *srcend= src + length;
  char prev= 'x';
  for ( ; src < srcend ; src++) /* Skip characters 0x00..0x20 */
  {
    if (*src < 0 || *src > ' ')
      break;
  }

  for ( ; src < srcend; src++)
  {
    char ch= (*src > '\0' && *src < ' ') ? ' ' : *src;
    if (ch != ' ')
    {
      if (prev == ' ')
        UdmDSTRAppend(dstr, " ", 1);
      UdmDSTRAppend(dstr, &ch, 1);
    }
    prev= ch;
  }
  return UDM_OK;
}


static udm_bool_t
good_section(UDM_SECTION_PARAM *Param)
{
  return UDM_TEST(Param->secno && !(Param->flags & UDM_VARFLAG_USERDEF));
}

static udm_rc_t
UdmHTMLProcessText(UDM_HTML_PARSER *parser)
{
  UDM_HTML_PARSER_PARAM *param= (UDM_HTML_PARSER_PARAM *) parser->user_data;
  UDM_CONST_TEXTITEM Item;
  UDM_TEXT_PARAM TextParam;

  if (parser->state.comment || parser->state.style ||
      parser->state.script  || !param->Doc->Spider.robots.index)
    return UDM_OK;

  UdmConstTextItemInit(&Item);
  UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, 0);
  Item.text= parser->lasttok;
  UdmConstStrTrim(&Item.text, " \r\n\t");
  if (Item.text.length == 0)
    return UDM_OK;

  if (param->lasthref.str)
  {
    /*
      Protection against unlimited linktext growth
      in case of a malformet HTML, when the closing
      tag </a> is forgotten.
    */
    if (UdmDSTRLength(&param->linktext) < 256) /* TODO34: what should be the limit? */
    {
      if (UdmDSTRLength(&param->linktext))
        UdmDSTRAppend(&param->linktext, " ", 1);
      UdmDSTRAppendLinkText(&param->linktext, Item.text.str, Item.text.length);
    }
  }

  if (parser->state.body)
  {
    UDM_SECTION_PARAM *Sec= parser->state.ahref ?
                            &param->SectionParamBodyOLinkText :
                            &param->SectionParamBody;
    if (good_section(Sec))
    {
      Item.href= param->lasthref;
      TextParam.secno= Sec->secno;
      Item.section_name= parser->state.ahref &&
                         Sec->secno != param->SectionParamBody.secno ?
                         sco : scb;
#if NOT_READY_YET
      if (tag.nonbreaking)
        UdmTextListAppendConst(&param->Doc->TextList, &Item);
      else
#endif
        UdmTextListAddConst(&param->Doc->TextList, &Item, &TextParam);
    }
  }
  else if (parser->state.title)
  {
    if (good_section(&param->SectionParamTitle))
    {
      TextParam.secno= param->SectionParamTitle.secno;
      Item.section_name= sct;
      UdmTextListAddConst(&param->Doc->TextList, &Item, &TextParam);
    }
  }
  else /* Text outside body and title */
  {
    if (good_section(&param->SectionParamNoBody))
    {
      Item.href= param->lasthref;
      TextParam.secno= param->SectionParamNoBody.secno;
      Item.section_name= (param->SectionParamNoBody.secno ==
                          param->SectionParamBody.secno) ? scb : scnb;
      UdmTextListAddConst(&param->Doc->TextList, &Item, &TextParam);
    }
  }
  return UDM_OK;
}



udm_rc_t
UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  UDM_HTML_PARSER parser;
  UDM_HTML_PARSER_PARAM param;
  UDM_CONST_STR content;

  /*
    We cannot use BSec and TSec later, as they are
    modified by UdmVarListReplace from UdmHTMLParseTag
  */

  UdmDebugEnter();
  if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
    UdmDebugReturn(UDM_ERROR);

  UdmHTMLParserParamInit(&param, Doc);

  UdmHTMLParserInit(&parser);
  UdmHTMLParserSetUserData(&parser, &param);
  UdmHTMLParserSetTagHandler(&parser, UdmHTMLProcessTag);
  UdmHTMLParserSetTextHandler(&parser, UdmHTMLProcessText);

  UdmHTMLParserExec(&parser, content.str, content.length);

  UdmHTMLParserParamFree(&param);
  UdmDebugReturn(UDM_OK);
}
