/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_uniconv.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"
#include "udm_stopwords.h"
#include "udm_word.h"
#include "udm_vars.h"
#include "udm_db.h"
#include "udm_hash.h"
#include "udm_parsehtml.h"
#include "udm_store.h"
#include "udm_doc.h"
#include "udm_conf.h"
#include "udm_result.h"
#include "udm_log.h"
#include "udm_sgml.h"
#include "udm_chinese.h"
#include "udm_synonym.h"


typedef struct udm_stack_parser_state_st
{
  int secno;
  int secno_match_type;
  int secno_flags;
  udm_bool_t use_numeric_operators;
  udm_bool_t use_range_operators;
  udm_bool_t dehyphenate;
  udm_bool_t strip_accents;
  int nphrasecmd;
  udm_match_mode_t match_mode;
  udm_match_mode_t next_match_mode_type;
  const char *lang;
  int auto_phrase;
  int phrpos;
  int user_weight;
  size_t SubstringMatchMinWordLength;
  int range_command;
  size_t range_nwords;
} UDM_STACK_PARSER_STATE;


static void
UdmQueryParserStateInit(UDM_AGENT *A, UDM_STACK_PARSER_STATE *state)
{
  state->secno= 0;
  state->secno_match_type= 0;
  state->secno_flags= 0;
  state->use_numeric_operators= UdmVarListFindBool(&A->Conf->Vars, "UseNumericOperators", UDM_FALSE);
  state->use_range_operators= UdmVarListFindBool(&A->Conf->Vars, "UseRangeOperators", UDM_FALSE);
  state->dehyphenate= UdmVarListFindBool(&A->Conf->Vars, "Dehyphenate", UDM_FALSE);
  state->strip_accents= UdmVarListFindBool(&A->Conf->Vars, "StripAccents", UDM_FALSE);
  state->nphrasecmd= 0;
  state->match_mode= UdmMatchMode(UdmVarListFindStr(&A->Conf->Vars, "wm", "wrd"));
  state->next_match_mode_type= state->match_mode;
  state->lang= UdmVarListFindStr(&A->Conf->Vars, "g", NULL);
  state->auto_phrase= 0;
  state->phrpos= 0;
  state->user_weight= UDM_DEFAULT_USER_WORD_WEIGHT;
  state->SubstringMatchMinWordLength=
    (size_t) UdmVarListFindInt(&A->Conf->Vars, "SubstringMatchMinWordLength", 0);
  state->range_command= 0;
  state->range_nwords= 0;
}


/*
  Set phrlen of the last phrase
*/
static void
UdmWWListSetPhrlen(UDM_WIDEWORDLIST *WWL, size_t phrlen)
{
  UDM_WIDEWORD *W= &WWL->Word[WWL->nwords - 1];
  for (; W >= WWL->Word && W->Param.phrlen == 0; W--)
    W->Param.phrlen= phrlen;
}


static udm_rc_t
UdmBoolExprAddCmd(UDM_QUERY *Query,
                  UDM_STACK_PARSER_STATE *state,
                  int *lex, size_t length)
{
  udm_rc_t rc= UDM_OK;
  size_t i;
  
  for (i = 0; i < length; i++)
  {
    UDM_BOOLEXPR_ITEM item;
    bzero((void*) &item, sizeof(item));
     switch(lex[i])
     {
       case '&':
       case '+':
         item.cmd= UDM_BOOLCMD_AND;
         state->next_match_mode_type= state->match_mode;
         break;
       case '|':
         item.cmd= UDM_BOOLCMD_OR;
         state->next_match_mode_type= state->match_mode;
         break;
       case '~':
         item.cmd= UDM_BOOLCMD_NOT;
         state->next_match_mode_type= state->match_mode;
         break;
       case '(':
         item.cmd= UDM_BOOLCMD_LEFT;
         state->next_match_mode_type= state->match_mode;
         break;
       case ')':
         item.cmd= UDM_BOOLCMD_RIGHT;
         state->next_match_mode_type= state->match_mode;
         break;
       case '"':
         item.cmd= UDM_BOOLCMD_PHRASE;
         state->next_match_mode_type= state->match_mode;
         state->nphrasecmd++;         
         if (!(state->nphrasecmd % 2)) /* phrase has just closed */
         {
           if (state->phrpos > 0)
           {
             /*
               It's possible when the closing quot sign is not
               immeidately after the word: "aa bb "
               TODO: get rid of duplicate UdmWWListSetPhrlen()
               call in UdmBoolExprAddWord().
             */
             UdmWWListSetPhrlen(&Query->Res.WWList, state->phrpos);
             state->phrpos= 0;
           }
         }
         break;
       case '<':
         state->next_match_mode_type= state->use_numeric_operators ?
                                      UDM_MATCH_NUMERIC_LT : state->match_mode;
         continue;
       case '>':
         state->next_match_mode_type= state->use_numeric_operators ? 
                                      UDM_MATCH_NUMERIC_GT : state->match_mode;
         continue;
       case '{':
       case '[':
         state->range_command= lex[i];
         state->range_nwords= Query->Res.WWList.nwords;
         continue;
       case '}':
       case ']':
         /*
           fprintf(stderr, "nwords=%d-%d %d-%d\n",
                   state->range_nwords, Res->WWList.nwords,
                   state->range_command, lex[i]);
         */
         if (state->use_range_operators && state->range_command != 0 &&
             Query->Res.WWList.nwords - state->range_nwords == 3)
         {
           UDM_WIDEWORD *W= &Query->Res.WWList.Word[state->range_nwords];
           if (!strcmp(W[1].Word.str, "to"))
           {
             UdmWideWordListMakeRange(&Query->Res.WWList, state->range_command, lex[i]);
             UDM_ASSERT(Query->BoolExpr.nitems > 2);
             if (state->phrpos >= 2)
               state->phrpos-= 2;
             Query->BoolExpr.nitems-= 2;
           }
         }
         state->range_command= 0;
         continue;
       default:
         if (state->auto_phrase && !UdmAutoPhraseChar(lex[i]))
         {
           int quot= '"';
           item.cmd= UDM_BOOLCMD_PHRASE;
           item.arg= state->secno_match_type;
           UdmBoolExprAddCmd(Query, state, &quot, 1);
           state->auto_phrase= 0;
         }
         continue;
     }
     /* Ignore all operators if we are in a phrase,  except phrase end. */
    if (!(state->nphrasecmd % 2) || item.cmd == UDM_BOOLCMD_PHRASE)
    {
      item.arg= state->secno_match_type;
      rc= UdmBoolExprAddItem(&Query->BoolExpr, &item);
      Query->BoolExpr.ncmds++;
    }
  }
  return rc;
}


static udm_match_mode_t
UdmWordMatchApplyLength(UDM_STACK_PARSER_STATE *S, size_t wlen)
{
  return (wlen < S->SubstringMatchMinWordLength &&
          (S->next_match_mode_type == UDM_MATCH_BEGIN ||
           S->next_match_mode_type == UDM_MATCH_SUBSTR ||
           S->next_match_mode_type == UDM_MATCH_END)) ?
  UDM_MATCH_FULL : S->next_match_mode_type;
}


static udm_rc_t
UdmAllFormsMulDB(UDM_AGENT *A, UDM_QUERY *Query)
{
  UDM_DB *db= &A->Conf->DBList.Item[0];
  UDM_ASSERT(A->Conf->DBList.nitems);
  return db->dbhandler->QueryAction(A, db, Query, UDM_QUERYCMD_WORDFORMS);
}


static udm_rc_t
UdmBoolExprAddWord(UDM_AGENT *A,
                   UDM_QUERY *Query,
                   UDM_STACK_PARSER_STATE *state,
                   size_t length,
                   /* Pointers to the next token and the query end*/
                   int *lt, int *ltend, 
                   char *wrd)
{
  UDM_WIDEWORD OWord;
  UDM_WIDEWORDLIST Forms;
  UDM_BOOLEXPR_ITEM item;
  udm_rc_t rc;
  udm_wordorigin_t origin;
  int lt_auto_phrase= UdmAutoPhraseChar(lt[0]);
  size_t phrlen= 0;
  int end_of_phrase= 0;
  udm_match_mode_t final_match_mode= UdmWordMatchApplyLength(state, length);

  /*
    Ignore AutoPhraseChar in case when we have a sequence:
    - a word
    - followed by a AutoPhraseChar
    - followed by a separator
    For example: B. Abama
    Notice space character after dot.
  */
  if (&lt[1] < ltend &&
      UdmUniCType(udm_unidata_default, lt[1]) == UDM_UNI_SEPAR)
    lt_auto_phrase= 0;

  if (Query->Res.WWList.nuniq >= UDM_MAXWORDPERQUERY-1)
    return UDM_OK;

  if (state->nphrasecmd % 2) /* in phrase */
  {
    if (((state->auto_phrase && !lt_auto_phrase) || lt[0] == '"'))
    {
      /* End of auto- or non-auto-phrase*/
      phrlen= state->phrpos + 1;
      end_of_phrase= 1;
    }
  }
  else /* not in phrase */
  {
    if (lt_auto_phrase)
    {
      /* Start of auto-phrase */
      int quot= '"';
      state->auto_phrase= 1;
      item.cmd= UDM_BOOLCMD_PHRASE;
      item.arg= 0;
      UdmBoolExprAddCmd(Query, state, &quot, 1);
      phrlen= 0;
    }
    else
      phrlen= 1; /* Single word */
  }
  
  item.cmd= UDM_BOOLCMD_WORD;
  item.arg= Query->Res.WWList.nuniq;
  rc= UdmBoolExprAddItem(&Query->BoolExpr, &item);


  /*
    Check stopword only when full word.
    Substring searches should not exclude them.
  */
  if(final_match_mode == UDM_MATCH_FULL &&
     (UdmStopListListFind(&A->Conf->StopWord, wrd, state->lang) ||
      A->Conf->WordParam.min_word_len > length ||
      A->Conf->WordParam.max_word_len < length))
  {
    origin= UDM_WORD_ORIGIN_STOP;
    Query->BoolExpr.items[Query->BoolExpr.nitems - 1].cmd= UDM_BOOLCMD_STOP;
  }
  else
  {
    origin= UDM_WORD_ORIGIN_QUERY;
  }

  if (phrlen > 1)
  {
    /* Phrase end found: set phrlen for all words in the same phrase */
    UdmWWListSetPhrlen(&Query->Res.WWList, phrlen);
  }

  /* Add the original word form */
  bzero((void*) &OWord, sizeof(OWord));
  OWord.Word.str= wrd;
  OWord.Word.length= strlen(wrd);
  OWord.Param.order= Query->Res.WWList.nuniq;
  OWord.Param.phrpos= state->phrpos;
  OWord.Param.phrlen= phrlen;
  OWord.Param.count= 0;
  OWord.Param.origin= origin;
  OWord.Param.match_mode= final_match_mode;
  OWord.Param.secno= state->secno;
  OWord.Param.phrwidth= state->dehyphenate ? 1 : 0;
  OWord.Param.user_weight= state->user_weight;
  
  UdmWideWordListAdd(&Query->Res.WWList, &OWord);
  if (state->nphrasecmd % 2)
    state->phrpos++;
  state->user_weight= UDM_DEFAULT_USER_WORD_WEIGHT;
  
  if (origin == UDM_WORD_ORIGIN_STOP)
    return UDM_OK;
  
  /* Add generated word forms: synonyms, stemming, etc. */
  UdmWideWordListInit(&Forms);
  if (UDM_OK != (rc= UdmAllForms(A, &Forms, &OWord)))
    return rc;

  if (UDM_OK != (rc= UdmAllFormsMulDB(A, Query)))
    return rc;

  {
    UDM_WIDEWORD_PARAM Param;
    size_t frm;
    UdmWideWordParamInit(&Param);
    for (frm= 0; frm < Forms.nwords ; frm++)
    {
      UDM_CONST_STR FWord;
      FWord.str= Forms.Word[frm].Word.str; /* TODO34: set word in a single shot */
      FWord.length= Forms.Word[frm].Word.length;
      Param.order= Query->Res.WWList.nuniq;
      Param.phrpos= OWord.Param.phrpos; /* state->phrpos is already changed here */
      Param.phrlen= phrlen;
      Param.count= 0;
      Param.origin= Forms.Word[frm].Param.origin;
      Param.match_mode= state->next_match_mode_type;
      Param.secno= state->secno;
      Param.user_weight= OWord.Param.user_weight;
      UdmWideWordListAddLikeConstStr(&Query->Res.WWList, &Param, FWord);
/*    UdmLog(query, UDM_LOG_DEBUG, "Word form: [%d] %s", FWord.origin, wrd);*/
    }
  }
  UdmWideWordListFree(&Forms);
  if (end_of_phrase)
    state->phrpos= 0;
  Query->Res.WWList.nuniq++;
  return rc;
}


static udm_rc_t
UdmParseSearchQuery(UDM_AGENT *A,
                    UDM_QUERY *Query,
                    UDM_STACK_PARSER_STATE *state,
                    UDM_CONV *uni_lc, int *ustr, int *uend)
{
  UDM_UNIDATA *unidata= A->Conf->unidata;
  int *lex, *lt, ctype;
  int *(*UniGetSepToken)(UDM_UNIDATA *unidata,
                         int *str, int *strend, int **last, int *ctype0)=
       state->dehyphenate ? UdmUniGetSepToken2 :UdmUniGetSepToken;

  for (lex= UniGetSepToken(unidata, ustr, uend, &lt , &ctype) ;
       lex;
       lex= UniGetSepToken(unidata, NULL, uend, &lt, &ctype))
  {
    char wrd[UDM_MAXWORDSIZE * sizeof("&#123456;") + 1];
    size_t len, wlen= lt - lex;
    len= UdmConvHTML(uni_lc, wrd, sizeof(wrd), (char *) lex, sizeof(lex[0]) * wlen);
    wrd[len]= '\0';
    UdmTrim(wrd, " \t\r\n");
/*
fprintf(stderr, "ctype=%d lex[0]=%c wlen=%d wrd='%s'\n", ctype, lex[0], wlen, wrd);
*/
    if (ctype == UDM_UNI_SEPAR)
    {
      UdmBoolExprAddCmd(Query, state, lex, wlen);
    } 
    else
    {
      const UDM_VAR *Section;
      udm_rc_t rc;

      if (lt[0] == ':' || lt[0] == '=')
      {
        if ((Section= UdmVarListFind(&A->Conf->Sections, wrd)))
        {
          state->secno= UdmVarSecno(Section);
          state->secno_match_type= lt[0];
          state->secno_flags= UdmVarFlags(Section);
          continue;
        }
        if (wlen > 5 && !strncmp(wrd, "secno", 5))
        {
          state->secno= atoi(wrd + 5);
          state->secno_match_type= lt[0];
          Section= UdmVarListFindBySecno(&A->Conf->Sections, state->secno);
          state->secno_flags= Section ? UdmVarFlags(Section) : 0;
          continue;
        }
        else if (wlen > 10 && !strncmp(wrd, "importance", 10) &&
                 wrd[10] >= '0' && wrd[10] <= '9')
        {
          state->user_weight= atoi(wrd + 10);
          continue;
        }
      }
      if ((state->secno_flags & UDM_VARFLAG_DECIMAL) &&
          (wrd[0] >= '0' && wrd[0] <= '9'))
      {
        char *wend= wrd + wlen;
        if (lt[0] == '.' && wend < wrd + sizeof(wrd) - 1)
        {
          *wend++= '.';
          lt++;
        }
        for (; lt < uend && wend < wrd + sizeof(wrd) - 1; )
        {
          if (*lt < '0' || *lt > '9')
            break;
          *wend++= *lt++;
        }
        *wend= '\0';
        UdmNormalizeDecimal(wrd, sizeof(wrd), wrd);
      }
      if (UDM_OK != (rc= UdmBoolExprAddWord(A, Query, state,
                                            wlen, lt, uend, wrd)))
        return rc;
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmEnvDBListCheck(UDM_ENV *Env)
{
  if (Env->DBList.nitems == 0)
  {
    udm_snprintf(Env->errstr, sizeof(Env->errstr), "No DBAddr defined");
    return UDM_ERROR;
  }
  return UDM_OK;
}



udm_rc_t
UdmQueryPrepare(UDM_AGENT * query, UDM_QUERY *Query)
{
  UDM_VARLIST *Vars= &query->Conf->Vars;
  UDM_CHARSET *browser_cs, *local_cs;
  udm_rc_t rc;
  int *ustr, *uend;
  size_t ulen;
  const char * txt = UdmVarListFindStr(&query->Conf->Vars,"q","");
  const char * qprev = UdmVarListFindStr(&query->Conf->Vars,"qprev","");
  const char   *seg=  UdmVarListFindStr(&query->Conf->Vars, "Segmenter", NULL);
  char *ltxt;
  size_t llen, obytes;
  int segmenter;
  UDM_CONV uni_lc, bc_uni, bc_lc;
  const char *lang;
  UDM_STACK_PARSER_STATE state;
  UDM_UNIDATA *unidata= query->Conf->unidata;

  if (UdmEnvDBListCheck(query->Conf))
    return UDM_ERROR;
  
  UdmQueryParserStateInit(query, &state);

  if (!(browser_cs= query->Conf->bcs))
    browser_cs= &udm_charset_latin1;
  
  if(!(local_cs= query->Conf->lcs))
    local_cs= &udm_charset_latin1;
  
  UdmConvInit(&bc_uni, browser_cs, &udm_charset_sys_int);
  UdmConvInit(&uni_lc, &udm_charset_sys_int, local_cs);
  UdmConvInit(&bc_lc, browser_cs, local_cs);
  
  ulen= strlen(txt);
  ustr=(int*)(UdmMalloc((sizeof(int))*(ulen+1)));
  obytes= UdmConvHTML(&bc_uni,(char*)ustr,sizeof(ustr[0])*(ulen+1),txt,ulen+1);
  obytes-= sizeof(int);
  
  /* Create copy of query, converted into LocalCharset (for UdmTrack) */
  llen = ulen * 14 + 1;
  ltxt=(char*)UdmMalloc(llen);
  obytes= UdmConvHTML(&uni_lc, ltxt, llen, (char*) ustr, obytes);
  ltxt[obytes]='\0';
  UdmVarListReplaceStr(&query->Conf->Vars,"q",ltxt);  /* "q-lc" was here */
  UDM_FREE(ltxt);
  
  llen = strlen(qprev);
  ltxt=(char*)UdmMalloc(llen*14+1);
  obytes= UdmConvHTML(&bc_lc,ltxt,llen*14+1,qprev,llen);
  ltxt[obytes]='\0';
  UdmVarListReplaceStr(&query->Conf->Vars,"qprev",ltxt);
  UDM_FREE(ltxt);
  
  /* Parse query and build boolean search stack*/
  if (state.strip_accents)
    UdmUniStrStripAccents(unidata, ustr);
  UdmUniStrToLower(unidata, ustr);
  switch(browser_cs->family)
  {
    case UDM_CHARSET_CHINESE_SIMPLIFIED:
    case UDM_CHARSET_CHINESE_TRADITIONAL: lang = "zh"; break;
    case UDM_CHARSET_JAPANESE: lang = "ja"; break;
    case UDM_CHARSET_THAI: lang = "th"; break;
    default: lang = "";
  }

  if (seg && !strcasecmp(seg, "cjk") &&
      UDM_MODE_PHRASE == UdmSearchMode(UdmVarListFindStr(&query->Conf->Vars, "m", NULL)))
    seg= "cjk-phrase";
  
  if ((segmenter= UdmUniSegmenterFind(query->Conf, lang, seg)))
  {
    ustr= UdmUniSegmentByType(query, ustr, segmenter, (int) ' ');
    ulen= UdmUniLen(ustr); /* TODO: get rid of it */
    llen= ulen * 14 + 1;
    ltxt= (char*) UdmMalloc(llen);
    obytes= UdmConvHTML(&uni_lc, ltxt, llen, (const char*) ustr, ulen * sizeof(int));
    ltxt[obytes]= '\0';
    UdmVarListReplaceStr(&query->Conf->Vars, "SEGMENTER.QUERY_STRING", ltxt);
    UDM_FREE(ltxt);
  }
  uend= ustr + UdmUniLen(ustr); /* TODO: get rid of it */

  if (UDM_OK != (rc= UdmParseSearchQuery(query, Query, &state, &uni_lc,
                                         ustr, uend)))
  {
    UdmFree(ustr);
    return rc;
  }

  if (state.nphrasecmd & 1)
  {
    UDM_BOOLEXPR_ITEM item;
    item.cmd= UDM_BOOLCMD_PHRASE;
    item.arg= 0;
    UdmBoolExprAddItem(&Query->BoolExpr, &item);
    Query->BoolExpr.ncmds++;
  }
  UDM_FREE(ustr);
  
  if (UdmVarListFindBool(Vars, "ComplexSynonyms", UDM_FALSE))
    UdmComplexSynonyms(query, &Query->Res.WWList);
  
  Query->Res.WWList.match_mode= state.match_mode;
  Query->Res.WWList.strip_noaccents= state.strip_accents;
  return UDM_OK;
}



/*
  Applying fast limit to UDLCRDLIST.
  Used in:
  - UdmFindAlwaysFoundWord()
  - UdmFindMultiWord()
  - UdmFindWordSingle()
  - UdmFindCrossWord()
  - UdmMultiAddCoords()
*/
udm_rc_t
UdmApplyFastLimit(UDM_URLCRDLIST *Coord, UDM_URLID_LIST *urls)
{
  UDM_URL_CRD *dst= Coord->Coords;
  UDM_URL_CRD *src= Coord->Coords;
  UDM_URL_CRD *srcend= Coord->Coords + Coord->ncoords;
  
  if (urls->exclude)
  {
    for ( ; src < srcend; src++)
    {
      if (!UdmBSearch(&src->urlid_coord.url_id,
                      urls->urls, urls->nurls, sizeof(urlid_t),
                      (udm_qsort_cmp)UdmCmpURLID))
      {
        *dst= *src;
        dst++;
      }
    }
  }
  else
  {
    for ( ; src < srcend; src++)
    {
      if (UdmBSearch(&src->urlid_coord.url_id,
                     urls->urls, urls->nurls, sizeof(urlid_t),
                     (udm_qsort_cmp)UdmCmpURLID))
      {
        *dst= *src;
        dst++;
      }
    }
  }
  Coord->ncoords= dst- Coord->Coords;
  return UDM_OK;
}
