/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_proto.h"
#include "udm_url.h"
#include "udm_hrefs.h"
#include "udm_server.h"
#include "udm_xmalloc.h"
#include "udm_host.h"
#include "udm_vars.h"
#include "udm_wild.h"
#include "udm_match.h"
#include "udm_db.h"
#include "udm_hash.h"

#define DEFAULT_PROXY_PORT  3128
#define ERRSTRSIZ 1000


static void
UdmMatchCopyLocal(UDM_MATCH *dst, const UDM_MATCH *src, const char *urlstr)
{
  UdmMatchSetPattern(dst, urlstr);
  dst->Param= src->Param;
  dst->specific= src->specific;
}


static udm_rc_t
UdmServerCopy(UDM_ENV *Conf, UDM_SERVER *dst, UDM_SERVER *src,
              const char *urlstr)
{
  char regerrstr[ERRSTRSIZ]= "";
  udm_rc_t rc;

  dst->command= src->command;
  dst->ordre= src->ordre;
  dst->weight= src->weight;
  dst->MaxHops= src->MaxHops;
  dst->webspace= src->webspace;
  dst->Filter.method= src->Filter.method;

  UdmMatchCopyLocal(&dst->Filter.Match, &src->Filter.Match, urlstr);
  src->Filter.Match.specific= NULL;

  if (UDM_OK != (rc= UdmMatchComp(&dst->Filter.Match,
                                  regerrstr, sizeof(regerrstr))))
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr), "'%s:' %s",
                 UdmMatchPatternConstStr(&dst->Filter.Match), regerrstr);
    return rc;
  }

  UdmVarListReplaceLst(&dst->Vars, &src->Vars, NULL, "*");
  
  if (UDM_OK != (rc= UdmURLListCopy(&dst->ProxyList, &src->ProxyList)))
    return rc;

  return UDM_OK;
}

/* return values: 0 on success, non-zero on error */

UDM_API(udm_rc_t)
UdmServerAdd(UDM_AGENT *A, UDM_SERVER *src, int flags)
{
  udm_rc_t       rc= UDM_OK;
  int            add= 1;
  char           *urlstr= NULL;
  size_t         i;
  UDM_SERVER     *dst= NULL;
  UDM_SERVERLIST S;
  UDM_ENV        *Conf = A->Conf;

  if (src->Filter.Match.Param.match_mode == UDM_MATCH_BEGIN)
  {
    UDM_URL from;
    size_t len;
    
    /* Copy URL to temp string    */
    /* to keep srv->url unchanged */
    len= 3 * strlen(UdmMatchPatternConstStr(&src->Filter.Match)) + 4;
    if ((urlstr= (char*)UdmMalloc(len)) == NULL) return UDM_ERROR;
    UdmURLCanonize(UdmMatchPatternConstStr(&src->Filter.Match), urlstr, len);
    
    UdmURLInit(&from);
    
    /* Check whether valid URL is passed */
    if ((rc= UdmURLParse(&from, urlstr)))
    {
      sprintf(Conf->errstr, "%s", UdmURLErrorStr(rc));
      UDM_FREE(urlstr);
      UdmURLFree(&from);
      return UDM_ERROR;
    }
    if((from.hostinfo) && (from.filename == NULL))
    {
      /* Add trailing slash                    */
      /* http://localhost -> http://localhost/ */
      udm_snprintf(urlstr, len, "%s://%s%s", from.schema, from.hostinfo, UDM_NULL2EMPTY(from.path));
    }
    
    switch (src->webspace)
    {
      char * s, * anchor;
      case UDM_WEBSPACE_PATH:
        /* Cut before '?' and after last '/' */
        if((anchor=strchr(urlstr,'?')))
          *anchor='\0';
        if((s=strrchr(urlstr,'/')))
          *(s+1)='\0';
        break;

      case UDM_WEBSPACE_SITE:
        if (from.hostinfo != NULL)
        {
          /* Cut after hostinfo */
          udm_snprintf(urlstr, len, "%s://%s/", UDM_NULL2EMPTY(from.schema), from.hostinfo);
        }
        else
        {
          /* Cut after first '/' */
          if((s=strchr(urlstr,'/')))
            *(s+1)='\0';
        }
        break;
      
      case UDM_WEBSPACE_WORLD:
      case UDM_WEBSPACE_URLLIST:
      default:
        break;
    }
    if (!strcmp(UDM_NULL2EMPTY(from.schema), "news"))
    {
      char *c, *cc;
      /* Cat server name to remove group names */
      /* This is because group names do not    */
      /* present in message URL                */
      c=urlstr+7;
      cc=strchr(c,'/');
      if(cc)*(cc+1)='\0';
    }
    UdmURLFree(&from);
  }
  else if (src->Filter.Match.Param.match_mode == UDM_MATCH_REGEX)
  {
    urlstr= UdmStrdup(UdmMatchPatternConstStr(&src->Filter.Match));
  }
  else
  {
    urlstr= UdmStrdup(UdmMatchPatternConstStr(&src->Filter.Match));
  }
  
  if (!urlstr)return UDM_ERROR; 
  
  for (i = 0; i < Conf->Servers.nservers; i++)
  {
    if (!strcmp(UdmMatchPatternConstStr(&Conf->Servers.Server[i].Filter.Match),
                urlstr))
    {
      add= 0;
      dst= &Conf->Servers.Server[i];
      UdmMatchFreeAndInit(&dst->Filter.Match);
      break;
    }
  }
  
  if (add)
  {
    if(Conf->Servers.nservers>=Conf->Servers.mservers)
    {
      Conf->Servers.mservers+=16;
      Conf->Servers.Server=(UDM_SERVER *)UdmXrealloc(Conf->Servers.Server,Conf->Servers.mservers*sizeof(UDM_SERVER));
    }
    dst= &Conf->Servers.Server[Conf->Servers.nservers];
    Conf->Servers.nservers++;
    UdmServerInit(dst);
  }

  
  UdmServerCopy(Conf, dst, src, urlstr);
  
  S.Server= dst;
  if (!(flags & UDM_FLAG_DONT_ADD_TO_DB))
    rc= UdmSrvAction(A, &S, UDM_SRV_ACTION_ADD);
  src->site_id = dst->site_id;
  
  UDM_FREE(urlstr);
  return rc;
}

void UdmServerFree(UDM_SERVER *Server)
{
  UdmMatchFree(&Server->Filter.Match);
  UdmVarListFree(&Server->Vars);
  UdmURLListFree(&Server->ProxyList);
}

void UdmServerListFree(UDM_SERVERLIST *List)
{
  size_t i;
  
  for(i=0;i<List->nservers;i++)
    UdmServerFree(&List->Server[i]);
  
  List->nservers=List->mservers=0;
  UDM_FREE(List->Server);
}

/* This fuction finds Server entry for given URL         */
/* and return Alias in "aliastr" if it is not NULL       */
/* "aliastr" must be big enough to store result          */
/* not more than UDM_URLSTR bytes are written to aliastr */

UDM_SERVER *UdmServerFind(UDM_AGENT *A, UDM_SERVERLIST *List, const char *url, char **aliastr)
{
#define NS 10
  size_t      i;
  UDM_SERVER  *Res= NULL;
  char        net[32];
  size_t      urllen= strlen(url);

  net[0] = '\0';
  for(i=0;i<List->nservers;i++)
  {
    UDM_SERVER  *srv=&List->Server[i];
    UDM_MATCH_PART  P[10];
    const char  *alias=UdmVarListFindStr(&srv->Vars,"Alias",NULL);

    if (srv->Filter.Match.Param.match_mode == UDM_MATCH_SUBNET && *net == '\0')
    {
      UDM_CONN conn;
      UDM_URL  URL;
      
      UdmURLInit(&URL);
      if(UdmURLParse(&URL, url))
      {
        UdmURLFree(&URL);
        continue;
      }
      conn.hostname = URL.hostname;
      conn.port=80;
      if (!UdmHostLookup2(A, &A->Conf->Hosts, &conn))
      {
        unsigned char * h;
        h=(unsigned char*)(&conn.sin.sin_addr);
        snprintf(net, sizeof(net) - 1, "%d.%d.%d.%d", h[0], h[1], h[2], h[3]);
      }
      UdmURLFree(&URL);
    }
     
    if(srv->webspace == UDM_WEBSPACE_WORLD ||
       !UdmMatchExec(&srv->Filter.Match, url, urllen, net, 10, P))
    {
      Res=srv;
      if((aliastr != NULL) && (alias != NULL))
      {
        size_t aliaslen= strlen(alias);
        size_t aliastrlen= 128 + urllen + aliaslen +
                           strlen(UdmMatchPatternConstStr(&srv->Filter.Match));
        *aliastr = (char*)UdmMalloc(aliastrlen);
        if (*aliastr != NULL)
          UdmMatchApply(*aliastr, aliastrlen, url, alias, aliaslen,
                        &srv->Filter.Match, 10, P);
      }
      break;
    }
  }
  return(Res);
}

#if 0
static int cmpserver(const void *s1,const void *s2)
{
  int res;
  
  if(!(res=strlen(((const UDM_SERVER*)s2)->url)-strlen(((const UDM_SERVER*)s1)->url)))
    res=(((const UDM_SERVER*)s2)->rec_id)-(((const UDM_SERVER*)s1)->rec_id);
  return(res);
}
void UdmServerListSort(UDM_SERVERLIST *List)
{
  /*  Long name should be found first    */
  /*  to allow different options         */
  /*  for server and it's subdirectories */
  UdmSort(List->Server,List->nservers,sizeof(UDM_SERVER),cmpserver);
}
#endif


void
UdmSpiderParamInit(UDM_SPIDERPARAM *Spider)
{
  Spider->period=UDM_DEFAULT_REINDEX_TIME;
  Spider->max_net_errors=UDM_MAXNETERRORS;
  Spider->read_timeout=UDM_READ_TIMEOUT;
  Spider->doc_timeout=UDM_DOC_TIMEOUT;
  Spider->maxhops=UDM_DEFAULT_MAX_HOPS;
  Spider->robots.index= UDM_TRUE;
  Spider->robots.follow= UDM_TRUE;
  Spider->robots.archive= UDM_TRUE;
  Spider->webspace= UDM_WEBSPACE_DEFAULT;
  Spider->use_robots= UDM_ROBOTS_ALL;
  Spider->use_clones= UDM_TRUE;
  Spider->net_error_delay_time=UDM_DEFAULT_NET_ERROR_DELAY_TIME;
  Spider->crawl_delay= 0;
  Spider->dns_cache_timeout= UDM_DNSCACHE_TIMEOUT;
  Spider->link_sources_to_follow= UDM_LINK_SOURCES_YES;
}


UDM_API(void)
UdmServerInit(UDM_SERVER * srv)
{
  bzero((void*)srv, sizeof(*srv));
  srv->Filter.Match.Param.match_mode= UDM_MATCH_BEGIN;
  srv->weight = 1; /* default ServerWeight */
  srv->MaxHops = 255; /* default MaxHops value */
  srv->webspace= UDM_WEBSPACE_DEFAULT;
  srv->Filter.method= UDM_METHOD_DEFAULT;
  srv->enabled= UDM_TRUE;
}
