/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_db_int.h"
#include "udm_vars.h"
#include "udm_searchtool.h"
#include "udm_log.h"
#include "udm_hash.h"
#include "udm_word.h"


#ifdef HAVE_SQL

/********************  Suggest ************************************/

static float
UdmUCharArrayCorrelation(unsigned char *a, unsigned char *b, size_t size)
{
  float Asum= 0;
  float Bsum= 0;
  float ABsum= 0;
  size_t i;

  for (i= 0; i < size; i++)
  {
    Asum+=  (float) a[i] * (float) a[i];
    Bsum+=  (float) b[i] * (float) b[i];
    ABsum+= (float) a[i] * (float) b[i];
  }
  return ABsum / sqrt(Asum) / sqrt(Bsum);
}


static void
UdmUCharArrayFillByteStatistics(unsigned char *a, size_t size,
                                const unsigned char *src, size_t srclen)
{
  bzero((void*) a, size);
  for ( ; srclen ; src++, srclen--)
  {
    a[((unsigned char) *src) % size]++;
  }
}


static int
WordProximity(UDM_CHARSET *cs,
              const char *s1, size_t len1,
              const char *s2, size_t len2, int scale)
{
  size_t max= len1 > len2 ? len1 : len2;
  size_t min= len1 < len2 ? len1 : len2;
  if ((max - min)*3 > max)  /* Not more than 1/3 of length difference */
    return 0;
  {
    unsigned char a[256];
    unsigned char b[256];
    UdmUCharArrayFillByteStatistics(a, sizeof(a), (unsigned const char *) s1, len1);
    UdmUCharArrayFillByteStatistics(b, sizeof(b), (unsigned const char *) s2, len2);
    return (size_t) (UdmUCharArrayCorrelation(a, b, sizeof(a)) * scale);
  }
  return (1000 * min) / (max ? max : 1);
}


static udm_rc_t
UdmWWListSuggestOneWord(UDM_AGENT *A, UDM_DB *db,
                        UDM_WIDEWORDLIST *WWList,
                        UDM_WIDEWORD *W)
{
  udm_rc_t rc;

  if (W->Param.origin == UDM_WORD_ORIGIN_QUERY && !W->Param.count)
  {
    char snd[32];
    char qbuf[128];
    UDM_SQL_TOP_CLAUSE Top;
    UDM_SQLRES SQLRes;
    UDM_WIDEWORD_PARAM Param;
    size_t row, rows;
    size_t Word_order= W->Param.order;
    size_t phrpos= W->Param.phrpos;
    size_t count_sum;
    UDM_CONST_STR cword;

    cword.str= W->Word.str;
    cword.length= W->Word.length; /* W is not valid after UdmWideWordListAdd() */

    UdmDBSQLTopClause(A, db, 100, &Top);
    
    UdmSoundex(A->Conf->lcs, snd, sizeof(snd), W->Word.str, W->Word.length);
    UdmLog(A, UDM_LOG_DEBUG, "Suggest for '%s': '%s'", W->Word.str, snd);
    udm_snprintf(qbuf, sizeof(qbuf),
                 "SELECT %sword, cnt FROM wrdstat WHERE snd='%s'%s"
                 " ORDER by cnt DESC%s",
                  Top.top, snd, Top.rownum, Top.limit);
    
    if (UDM_OK!= (rc= UdmDBSQLQuery(A, db, &SQLRes, qbuf)))
      return rc;
    
    rows=UdmSQLNumRows(&SQLRes);
    UdmLog(A, UDM_LOG_DEBUG, "%d suggestions found", (int) rows);
    if (rows > 0)
    {
      UdmLog(A, UDM_LOG_DEBUG, "<%s>: <%s>/<%s>/<%s>/<%s>", 
             "word", "count", "count_weight", "proximity_weight",
             "final_weight");
    }
    UdmWideWordParamInit(&Param);
    
    for (count_sum= 0, row= 0; row < rows; row++)
    {
      count_sum+= UDM_ATOI(UdmSQLValue(&SQLRes, row, 1));
    }
    if (!count_sum)
      count_sum= 1;

    for(row=0 ; row < rows ; row++)
    {
      size_t nbytes, proximity_weight, count_weight, final_weight;
      int tmp[128];
      int word_count_factor= 63;
      int word_proximity_factor= 256 - word_count_factor;
      UDM_CONST_STR sg2;
      
      sg2.str= UdmSQLValue(&SQLRes, row, 0);
      sg2.length= UdmSQLLen(&SQLRes, row, 0);
      Param.count= UDM_ATOI(UdmSQLValue(&SQLRes, row, 1));

      count_weight= (word_count_factor * Param.count) / count_sum;
      proximity_weight= WordProximity(A->Conf->lcs,
                                      cword.str, cword.length,
                                      sg2.str, sg2.length,
                                      word_proximity_factor);
      final_weight= (count_weight + proximity_weight);
      UdmLog(A, UDM_LOG_DEBUG, "%s: %d/%d/%d/%d", 
             sg2.str, (int) Param.count, (int) count_weight,
             (int) proximity_weight, (int) final_weight);
      Param.count= final_weight;

      nbytes= (sg2.length + 1) * sizeof(int);
      if (nbytes < sizeof(tmp))
      {
        Param.order= Word_order;
        Param.phrpos= phrpos;
        Param.origin= UDM_WORD_ORIGIN_SUGGEST;
        UdmWideWordListAddLikeConstStr(WWList, &Param, sg2);/* W is not valid after this */
      }
    }
    UdmSQLFree(&SQLRes);
  }
  return UDM_OK;
}


udm_rc_t
UdmQuerySuggest(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc= UDM_OK;
  size_t i, nwords;
  UDM_WIDEWORDLIST *WWList= &Query->Res.WWList;
  
  UdmLog(A, UDM_LOG_DEBUG, "Generating suggestion list");

  for (i=0, nwords= WWList->nwords; i < nwords; i++)
  {
    UDM_WIDEWORD *W= &WWList->Word[i];
    if (UDM_OK != (rc= UdmWWListSuggestOneWord(A, db, WWList, W)))
      break;
  }
  return rc;
}

#endif
