/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_template.h"
#include "udm_server.h"
#include "udm_env.h"
#include "udm_conf.h"
#include "udm_db.h"
#include "udm_http.h"
#include "udm_parsehtml.h"
#include "udm_host.h"
#include "udm_contentencoding.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_wild.h"
#include "udm_searchtool.h"
#include "udm_doc.h"
#include "udm_match.h"
#include "udm_sgml.h"
#include "udm_proto.h"
#include "udm_ctype.h"
#include "udm_lex.h"
#include "udm_prog.h"
#include "udm_agent.h"

/******* Template initialization and freeing routines ********************/

static udm_rc_t
UdmTemplateItemInit(UDM_TEMPLATE_ITEM *Item, size_t lineno, const char *name,
                    const char *str, size_t length)
{
  Item->lineno= lineno;
  if (!(Item->name= UdmStrdup(name)) ||
      UdmDSTRInit(&Item->Value, 256) ||
      UdmDSTRAppend(&Item->Value, str, length))
    return UDM_ERROR;
  return UDM_OK;
}


static void
UdmTemplateItemFree(UDM_TEMPLATE_ITEM *Item)
{
  UdmFree(Item->name);
  UdmDSTRFree(&Item->Value);
}


UDM_API(void)
UdmTemplateInit(UDM_TEMPLATE *List)
{
  UDM_BZERO(List, sizeof(*List));
}


UDM_API(void)
UdmTemplateFree(UDM_TEMPLATE *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
    UdmTemplateItemFree(&List->Item[i]);
  UdmFree(List->Item);
}


static udm_rc_t
UdmTemplateAdd(UDM_TEMPLATE *List, size_t lineno, const char *name,
                             const char *str, size_t length)
{
  if (List->nitems >= List->mitems)
  {
    size_t mitems= List->mitems + 16;
    UDM_TEMPLATE_ITEM *Item;
    if (!(Item= (UDM_TEMPLATE_ITEM*) UdmRealloc(List->Item,
                                                mitems * sizeof(List->Item[0]))))
      return UDM_ERROR;
    List->Item= Item;
    List->mitems= mitems;
  }
  return UdmTemplateItemInit(&List->Item[List->nitems++],
                             lineno, name, str, length);
}


/******************* Template functions ********************/


static size_t
out_strn(FILE *stream, UDM_DSTR *dstr, const char * src, size_t srclen)
{
  if (src)
  {
    if (stream)
    {
      fwrite(src, srclen, 1, stream);
      fputs("", stream);
    }
    if (dstr)
      UdmDSTRAppend(dstr, src, srclen);
  }
  return(0);
}


size_t out_string(FILE *stream, UDM_DSTR *dstr, const char *src)
{
  return out_strn(stream, dstr, src, strlen(src));
}


static size_t
out_dstr(FILE *stream, UDM_DSTR *dstr, const UDM_DSTR *src)
{
  return out_strn(stream, dstr, UdmDSTRPtr(src), UdmDSTRLength(src));
}


static char *UdmConvDup(const char *src,
                        UDM_CHARSET *from, 
                        UDM_CHARSET *to, int fl)
{
  UDM_CONV cnv;
  size_t len;
  char *dst;
  len= strlen(src);
  dst= (char*) UdmMalloc(len * 12 + 1);
  UdmConvInit(&cnv, from, to);
  UdmConv(&cnv, dst, len * 12 + 1, src, len + 1, fl);
  return dst;
}


typedef enum
{
  UDM_PRINT_FMT_AUTO=            '(',
  UDM_PRINT_FMT_URL=             '%',
  UDM_PRINT_FMT_HTML=            '@',
  UDM_PRINT_FMT_HTML_HIGHLIGHT=  '&',
  UDM_PRINT_FMT_TEXT=            '#',
  UDM_PRINT_FMT_HTML_CACHED=     '^',
  UDM_PRINT_FMT_BASE64=           0xB64
} udm_print_fmt_t;



static size_t
UdmTemplatePrintHighlight(FILE *stream, UDM_DSTR *dstr,
                          const char *src, size_t srclen,
                          const char *beg, size_t blen,
                          const char *end, size_t elen)
{
  size_t res;
  UDM_DSTR tmp;
  UdmDSTRInit(&tmp, 128);
  UdmDSTRAppendHighlight(&tmp, src, srclen, beg, blen, end, elen);
  res= out_dstr(stream, dstr, &tmp);
  UdmDSTRFree(&tmp);
  return res;
}


static size_t
UdmTemplatePrintVar(UDM_ENV *Env, 
                    FILE *stream, UDM_DSTR *dstr,
                    const char *value, udm_print_fmt_t type,
                    const char *HlBeg,
                    const char *HlEnd)
{
  size_t res= 0;

  switch (type)
  {
    case UDM_PRINT_FMT_TEXT:
      {
        char *eval= NULL;
#ifndef NO_ADVANCE_CONV
        eval= UdmRemoveHiLightDup(value);
        res= out_string(stream, dstr, eval);
#else
        res= out_string(stream, dstr, value);
#endif
        UDM_FREE(eval);
      }
      break;
    case UDM_PRINT_FMT_HTML_HIGHLIGHT:
      {
#ifndef NO_ADVANCE_CONV
        char *cval= UdmConvDup(value, Env->bcs, Env->bcs, UDM_RECODE_HTML);
#else
        char *cval= UdmConvDup(value, Env->lcs, Env->bcs, UDM_RECODE_HTML);
#endif
        res= UdmTemplatePrintHighlight(stream, dstr, cval, strlen(cval),
                                       HlBeg, strlen(HlBeg),
                                       HlEnd, strlen(HlEnd));
        UDM_FREE(cval);
      }
      break;
    case UDM_PRINT_FMT_AUTO: 
    case UDM_PRINT_FMT_HTML:
      {
        char *cval= UdmConvDup(value, Env->bcs, Env->bcs, UDM_RECODE_HTML|UDM_RECODE_HTML_OUT_SPECIAL);
        res= UdmTemplatePrintHighlight(stream, dstr, cval, strlen(cval),
                                       "", 0, "", 0);
        UdmFree(cval);
      }
      break;
    case UDM_PRINT_FMT_HTML_CACHED:
      res= UdmTemplatePrintHighlight(stream, dstr, value, strlen(value),
                                     HlBeg, strlen(HlBeg),
                                     HlEnd, strlen(HlEnd));
      break;
    case UDM_PRINT_FMT_URL:
      {
        size_t valuelen= strlen(value);
        UDM_DSTR tmp;
        UdmDSTRInit(&tmp, 128);
        UdmDSTRAppendURLEncode(&tmp, value, valuelen);
        res= out_dstr(stream, dstr, &tmp);
        UdmDSTRFree(&tmp);
      }
      break;
    case UDM_PRINT_FMT_BASE64:
      {
        size_t len= strlen(value);
        char *eval= (char*)UdmMalloc(BASE64_LEN(len));
        len= udm_base64_encode(value, eval, len);
        res= out_strn(stream, dstr, eval, len);
        UDM_FREE(eval);
      }
  }
  return res;
}


/*
  Html-aware LEFT() function:
  Finds byte length of "pos" characters in a safe way,
  not to break entities like &#1234; or &acute; in the middle.
*/
static size_t
UdmHtmlStrLeft(const char *str, size_t length, size_t pos)
{
  const char *p= str + pos;
  const char *b;
  
  if (pos > length)
    return length;
  
  /*
    '&' '#' alnum* ';'
  */
  
  for (b= (p > str && *p == ';') ? p - 1 : p ;
       udm_isalnum(*b) && b > str ; b--) /* nop */;
  
  /* Check if we found a valid entity beginning */
  if (*b == '&' || (*b == '#' && b > str && *--b == '&'))
  {
    const char *strend= str + length;
    const char *e;
    for (e= (p < strend && (*p == '#' || *p == '&')) ? p + 1 : p ;
         e < strend && udm_isalnum(*e) ; e++) /* nop */;
    if (e < strend && *e == ';')
      p= b;
  }
  return p - str;
}


/*
  Find format length in a variable name:
    $(name:10)

  Ignore ':' if followed by a letter, e.g. $&(og:description)

  Note: $(og:description:4) is still recognized.
*/
static const char *find_format(const char *str, const char *end)
{
  for ( ; str < end ; str++)
  {
    if (str[0] == ':' && !udm_isalpha(str[1]))
      return str;
  }
  return NULL;
}


static udm_bool_t
UdmTemplateScanVariableName(UDM_CONST_TOKEN *name,
                            udm_print_fmt_t *type,
                            const char *s)
{
  int pcount= 0;

  if (!strncmp(s,"$(",2))
  {
    name->str= s + 2;
    *type= UDM_PRINT_FMT_AUTO;
  }
  else if (!strncmp(s, "$%(",3))
  {
    name->str= s + 3;
    *type= UDM_PRINT_FMT_URL;
  }
  else if(!strncmp(s, "$&(", 3))
  {
    name->str= s + 3;
    *type= UDM_PRINT_FMT_HTML_HIGHLIGHT;
  }
  else if (!strncmp(s, "$@(", 3))
  {
    name->str= s + 3;
    *type= UDM_PRINT_FMT_HTML;
  }
  else if (!strncmp(s, "$#(", 3))
  {
    name->str= s + 3;
    *type= UDM_PRINT_FMT_TEXT;
  }
  else if(!strncmp(s,"$^(",3))
  {
    name->str= s + 3;
    *type= UDM_PRINT_FMT_HTML_CACHED;
  }
  else if (!strncasecmp(s, "$base64(", 8))
  {
    name->str= s + 8;
    *type= UDM_PRINT_FMT_BASE64;
  }
  else
  {
    name->str= s;
    name->end= s;
    *type= UDM_PRINT_FMT_AUTO;
    return UDM_TRUE;
  }

  for (name->end= s; name->end[0]; name->end++)
  {
    if (name->end[0]== '(')
      pcount++;
    if (name->end[0]== ')' && (--pcount) == 0)
      return UDM_FALSE;
  }
  return UDM_TRUE; /* Closing ')' not found */
}


static char *MakeCutValue(const char *value,
                          size_t curlen, int neg, size_t maxlen)
{
  char *newvalue;
  if (!(newvalue= (char*) UdmMalloc(maxlen + 8)))
    return NULL;

  if (!neg)
  {
    char *c2, *c3;
    size_t pos= UdmHtmlStrLeft(value, curlen, maxlen);
    strncpy(newvalue, value, pos);
    newvalue[pos] = '\0';
    c2 = strrchr(newvalue, '\2');
    c3 = strrchr(newvalue, '\3');
    if ((c2 != NULL) && ((c3 == NULL) || (c2 > c3)))
    {
      strcpy(newvalue + pos, "\3...");
    }
    else
    {
      strcpy(newvalue + pos, "...");
    }
  }
  else
  {
    char *c2, *c3;
    size_t pos= UdmHtmlStrLeft(value, curlen, curlen - maxlen);
    size_t len= curlen - pos;
    size_t offs= 0;
    c2= strchr(value + pos, '\2');
    c3= strchr(value + pos, '\3');
    if (c3 && (!c2 || c2 > c3))
    {
      *newvalue= '\2';
      offs= 1;
    }
    memcpy(newvalue + offs, value + pos, len);
    newvalue[offs + len]= '\0';
  }
  return newvalue;
}


typedef struct
{
  UDM_CONST_STR name;
  size_t maxlen;
  int neg;
  udm_print_fmt_t type;
  const char *end;
} UDM_VARREF;


static udm_bool_t
UdmTemplateScanVariableReference(UDM_VARREF *ref, const char *s)
{
  UDM_CONST_TOKEN name;
  const char *sem;

  if (UdmTemplateScanVariableName(&name, &ref->type, s))
    return UDM_TRUE;
  ref->name.str=    UdmConstTokenStr(&name);
  ref->name.length= UdmConstTokenLength(&name);
  ref->end= name.end;

  if ((sem= find_format(ref->name.str, ref->end)))
  {
    char number[64];
    ref->name.length= sem - ref->name.str;
    sem++;
    udm_snprintf(number, sizeof(number), "%.*s", (int) (ref->end - sem), sem);
    ref->maxlen= atoi((ref->neg= (number[0] == '-')) ? number + 1 : number);
  }
  else
  {
    ref->neg= 0;
    ref->maxlen= 0;
  }
  
  return UDM_FALSE;
}


static UDM_VAR *
UdmVarListFindVarByNameConstStr(UDM_VARLIST *Vars, const UDM_CONST_STR *str)
{
  char name[128];
  udm_snprintf(name, sizeof(name), "%.*s", (int) str->length, str->str);
  return UdmVarListFindVar(Vars, name);
}


size_t PrintTextTemplate(UDM_ENV *Env, FILE *stream,
                         UDM_DSTR *dstr, 
                         UDM_VARLIST *vars,
                         const char *templ,
                         const char *HlBeg,
                         const char *HlEnd,
                         udm_bool_t auto_escape)
{
  const char * s;
  size_t dlen=0;
  
  for(s= templ; *s ; s++)
  {
    UDM_VARREF ref;
    UDM_VAR *Var;
    char *value= NULL;
    size_t curlen=0;
    const char *tmp;
    
    if (*s != '$' || UdmTemplateScanVariableReference(&ref, s))
    {
      if  (stream && (*s))
        fputc(*s, stream);
      if (dstr)
        UdmDSTRAppend(dstr, s, 1);
      continue;
    }
    
    s= ref.end;

    if (ref.name.length == 5 && !strncasecmp(ref.name.str, "ndocs", 5))
    {
      UDM_AGENT A;
      UdmAgentInit(&A, Env, 0);
      UdmDBAction(&A, UDM_DBCMD_DOCCOUNT);
      UdmVarListReplaceInt(vars, "ndocs", A.doccount);
      UdmAgentFree(&A);
    }
    
    if (!(Var= UdmVarListFindVarByNameConstStr(vars, &ref.name)) ||
        !(tmp= UdmVarStr(Var)))
      continue;

    if (ref.type == UDM_PRINT_FMT_AUTO)
    {
      if (auto_escape && (UdmVarSecno(Var) == UDM_VARSRC_QSTRING ||
                          UdmVarSecno(Var) == UDM_VARSRC_ENV))
        ref.type= UDM_PRINT_FMT_HTML_HIGHLIGHT;      
      else if (UdmVarFlags(Var) & UDM_VARFLAG_FORCETXT)
        ref.type= UDM_PRINT_FMT_TEXT;
    }
    
    switch (ref.type)
    {
      case '&':
      case '^': 
        if (!(value= UdmStrdup(tmp)))
          continue;
        break;
      default:
        if (!(value= UdmRemoveHiLightDup(tmp)))
          continue;
    }
    curlen= strlen(value);
    if (curlen > ref.maxlen && ref.maxlen > 0)
    {
      char *newvalue;
      if ((newvalue= MakeCutValue(value, curlen, ref.neg, ref.maxlen)))
      {
        UDM_FREE(value);
        value= newvalue;
      }
    }
    
    dlen+= UdmTemplatePrintVar(Env, stream, dstr, 
                               value, ref.type, HlBeg, HlEnd);
    UDM_FREE(value);
  }
  return dlen;
}


void
UdmProgFixJumpInRange(UDM_PROG *prg, size_t start, size_t end)
{
  size_t i;
  /*
  fprintf(stderr, "fixing jump %04X-%04X to %04X\n",
        (int)start, (int) end, (int) prg->nitems);
  */
  /*
    Now we know the address of <!ENDIF>
    We should pass through <!IF> and all <!ELSEIF>
    and set correct JMP addresses.
  */
  for (i= start; i < end; i++)
  {
    if (prg->Items[i].cmdarg.jump == 0 &&
        (prg->Items[i].cmdnum == UDM_PROG_JMP ||
         prg->Items[i].cmdnum == UDM_PROG_JE  ||
         prg->Items[i].cmdnum == UDM_PROG_JNE))
      prg->Items[i].cmdarg.jump= prg->nitems; 
  }
}


static udm_rc_t
PrintHtmlTemplate(UDM_PROG_EXECUTOR *tmpl,
                  UDM_ENV *Env,
                  const UDM_TEMPLATE_ITEM *item,
                  udm_bool_t check_pi_count,
                  int *progrc)
{
  UDM_CONST_STR str;
  UDM_VARLIST Vars;
  UDM_PROG_COMPILER compiler;
  UDM_PROG prg;
  udm_rc_t rc;
  const char *DebugProg= getenv("DebugProg");

  UdmDSTRReset(&tmpl->state.stack);
  UdmProgInit(&prg);
  UdmVarListInit(&Vars);  
  UdmDSTRToConstStr(&item->Value, &str);
  UdmProgCompilerInit(&compiler, item->lineno, &prg, str.str, str.length);

  if (check_pi_count)
    rc= UdmCompilePIProgram(&compiler, &str);
  else
    rc= UDM_OK;
  if (rc != UDM_OK)
  {
    if (tmpl->state.globals.stdout)
      udm_snprintf(Env->errstr, sizeof(Env->errstr), "%s", compiler.errstr);
    goto ret;
  }

  if (DebugProg && udm_strntobool(DebugProg, strlen(DebugProg)))
    UdmProgPrint(&prg);

  if (check_pi_count && !compiler.pi_count)
    goto ret;
  
  if (compiler.pi_count)
    *progrc= UdmProgExecutorExec(tmpl, &prg);

ret:
  UdmVarListFree(&Vars);
  UdmProgFree(&prg);
  UdmProgCompilerFree(&compiler);
  return rc;
}


static udm_rc_t
UdmTemplatePrintWithArgs(int argc, char **argv,
                         UDM_ENV *Env,
                         FILE *stream, UDM_DSTR *dstr,
                         UDM_TEMPLATE *tm)
{
  size_t  t;
  UDM_PROG_EXECUTOR tmpl;
  udm_rc_t rc= UDM_OK;

  UdmProgExecutorInit(&tmpl, stream, dstr);
  tmpl.state.globals.argc= argc;
  tmpl.state.globals.argv= argv;
  tmpl.state.globals.THDHandler= Env->THDHandler;
  
  for (t= 0; t < tm->nitems; t++)
  {
    UDM_TEMPLATE_ITEM *Var= &tm->Item[t];
    /*
      For security purposes, print template with empty name only
      if there are some native processing instructions found.
    */
    rc= PrintHtmlTemplate(&tmpl, Env, Var, UDM_TRUE, &tm->rc);
    break;
  }
  UdmProgExecutorFree(&tmpl);
  return rc;
}


UDM_API(udm_rc_t)
UdmTemplatePrintToStreamWithArgs(int argc, char **argv,
                                 UDM_ENV *Env,
                                 FILE *stream,
                                 UDM_TEMPLATE *tm)
{
  return UdmTemplatePrintWithArgs(argc, argv, Env, stream, NULL, tm);
}


/* Load template  */
udm_rc_t
UdmTemplateLoad(UDM_TEMPLATE *tmpl, UDM_ENV *Env, const char *tname)
{
  UDM_AGENT Agent;
  UDM_CFG Cfg;
  UDM_SERVER Srv;
  FILE    *file;
  char    str[1024];
  udm_rc_t rc= UDM_OK;
  UDM_DSTR dstr_nosection;
  
  UdmDSTRInit(&dstr_nosection, 16*1024);

  UdmServerInit(&Srv);
  bzero((void*)&Cfg, sizeof(Cfg));

  UdmAgentInit(&Agent, Env, 0);
  Cfg.Indexer= &Agent;
  Cfg.level = 0;
  Cfg.flags = UDM_FLAG_LOAD_LANGMAP + UDM_FLAG_SPELL;
  Cfg.Srv = &Srv;

  if(!(file=fopen(tname,"r")))
  {
    udm_snprintf(Env->errstr,sizeof(Env->errstr)-1,"Unable to open template '%s': %s",tname,strerror(errno));
    rc= UDM_ERROR;
    goto ret;
  }

  for (; fgets(str,sizeof(str)-1,file); )
  {
    str[sizeof(str)-1]='\0';
    UdmDSTRAppend(&dstr_nosection, str, strlen(str));
  }
  fclose(file);
  UdmTemplateAdd(tmpl, 0, "",
                 UdmDSTRPtr(&dstr_nosection), UdmDSTRLength(&dstr_nosection));

  UdmVarListReplaceLst(&Env->Vars, &Srv.Vars, NULL, "*");
  UdmServerFree(&Srv);
    
  if (UDM_OK != UdmEnvPrepare(Env))
  {
    rc= UDM_ERROR;
    goto ret;
  }

ret:  
  UdmDSTRFree(&dstr_nosection);
  UdmAgentFree(&Agent);
  return rc;
}


