/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udm_common.h"
#include "udm_wild.h"
#include "udm_utils.h"


int UdmWildCmp(const char *str, const char *wexp)
{
    int x, y;

    for (x = 0, y = 0; wexp[y]; ++y, ++x) {
	if ((!str[x]) && (wexp[y] != '*'))return -1;
	if (wexp[y] == '*') {
	    while (wexp[++y] == '*');
	    if (!wexp[y])return 0;
	    while (str[x]) {
		int ret;
		if ((ret = UdmWildCmp(&str[x++], &wexp[y])) != 1)return ret;
	    }
	    return -1;
	}else
	if ((wexp[y] != '?') && (str[x] != wexp[y]))return 1;
    }
    return (str[x] != '\0');
}


UDM_API(int)
UdmWildCaseCmp(const char *str, const char *wexp)
{
  for ( ; *wexp; ++str, ++wexp)
  {
    if (!*str && *wexp != '*')return -1;
    if (*wexp == '*')
    {
      while (*++wexp == '*');
      if (!*wexp)return 0;
      while (*str)
      {
        int ret;
        if ((ret= UdmWildCaseCmp(str++, wexp)) != 1)return ret;
      }
      return -1;
    }
    else if ((*wexp != '?') &&
             (udm_l1tolower[(unsigned char)*str] !=
              udm_l1tolower[(unsigned char)*wexp]))
      return 1;
  }
  return (*str != '\0');
}


#define UDM_MATCH_OK                          0
#define UDM_MATCH_FAILURE_WITH_WILDCARDS     -1
#define UDM_MATCH_FAILURE_WITHOUT_WILDCARDS   1

/*
  Robots.txt syntax supported by some bots include two meta characters:
  *  -  stands for zero or more characters
  $  -  stands for end-of-line

  To convert robots.txt pattern to wild pattern:
  - if the robots.txt pattern has no $ at the end, then add *
  - if the robots.txt pattern has $ at the end, then just remove $.

  @param matchend - If matchend is NULL, then full match is performend.
                    If matchend is not NULL, then prefix match is performed
                    and UDM_MATCH_OK is returned if a matching prefix was found.
                    Additionally, the position where the patched prefix
                    ended is returned in matchend[0].
*/
static int
UdmWildCmpExInternal(const char *str, const char *strend,
                     const char *wexp, const char *wexpend,
                     const UDM_WILD_PARAM *param,
                     const char **matchend)
{
  for ( ; wexp < wexpend; ++str, ++wexp)
  {
    if (str == strend)
    {
      if (*wexp == param->eol && wexp + 1 == wexpend)
      {
        if (matchend)
          *matchend= str;
        return UDM_MATCH_OK;
      }
      if (*wexp != param->many)
        return UDM_MATCH_FAILURE_WITH_WILDCARDS;
    }
    if (*wexp == param->many)
    {
      while (++wexp < wexpend && *wexp == param->many);
      if (wexp == wexpend)
      {
        if (matchend)
          *matchend= str;
        return UDM_MATCH_OK;
      }
      while (str < strend)
      {
        int ret;
        if ((ret= UdmWildCmpExInternal(str++, strend, wexp, wexpend,
                                       param, matchend)) != 
            UDM_MATCH_FAILURE_WITHOUT_WILDCARDS)
          return ret;
      }
      return UDM_MATCH_FAILURE_WITH_WILDCARDS;
    }
    UDM_ASSERT(str < strend);
    if (*wexp != param->one && *str != *wexp)
      return UDM_MATCH_FAILURE_WITHOUT_WILDCARDS;
  }
  if (matchend)
  {
    *matchend= str;
    return UDM_OK;
  }
  /* If param->eol is set, then we don't require '*' at the end */
  return str < strend ?
         (param->eol ? UDM_MATCH_OK : UDM_MATCH_FAILURE_WITHOUT_WILDCARDS) :
         UDM_MATCH_OK;
}


/**
  Case sensitive wild comparison.
*/
int
UdmWildCmpEx(const char *str, const char *strend,
             const char *wexp, const char *wexpend,
             const UDM_WILD_PARAM *param)
{
  return UdmWildCmpExInternal(str, strend, wexp, wexpend, param, NULL);
}


/**
  Case sensitive wild substring search.
  Part->beg and Part->end are set to offsets of the matched fragment.
*/
int
UdmWildCmpSubstr(const char *str, const char *strend,
                 const char *wexp, const char *wexpend,
                 const UDM_WILD_PARAM *param,
                 UDM_MATCH_PART *Part)
{
  const char *matchend= NULL;
  const char *matchbeg;
  Part->beg= 0;
  Part->end= 0;
  for (matchbeg= str; matchbeg < strend; matchbeg++)
  {
    if (UDM_MATCH_OK == UdmWildCmpExInternal(matchbeg, strend, wexp, wexpend,
                                             param, &matchend))
    {
      Part->beg= matchbeg - str;
      Part->end= matchend ? matchend - str : Part->beg;
      return UDM_MATCH_OK;
    }
  }
  return UDM_MATCH_FAILURE_WITHOUT_WILDCARDS;
}
