/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_SQL


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>

#ifdef WIN32
#include <time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_spell.h"
#include "udm_robots.h"
#include "udm_db.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_url.h"
#include "udm_log.h"
#include "udm_proto.h"
#include "udm_conf.h"
#include "udm_hash.h"
#include "udm_xmalloc.h"
#include "udm_boolean.h"
#include "udm_coords.h"
#include "udm_searchtool.h"
#include "udm_server.h"
#include "udm_stopwords.h"
#include "udm_doc.h"
#include "udm_result.h"
#include "udm_vars.h"
#include "udm_agent.h"
#include "udm_store.h"
#include "udm_hrefs.h"
#include "udm_word.h"
#include "udm_db_int.h"
#include "udm_match.h"
#include "udm_indexer.h"
#include "udm_textlist.h"
#include "udm_parsehtml.h"
#include "udm_wild.h"
#include "udm_http.h"
#include "udm_contentencoding.h"
#include "udm_indexcache.h"
#include "udm_query.h"
#include "udm_env.h"


static inline UDM_SQL *
UdmSQL(UDM_DB *db)
{
  return &((UDM_SQLDB*) db->specific)->sql;
}

UDM_SQLDB *
UdmSQLDB(UDM_DB *db)
{
  return (UDM_SQLDB*) db->specific;
}

udm_sqldbtype_t
UdmSQLDBType(UDM_DB *db)
{
  return UdmSQL(db)->DBType;
}

udm_sqldbapi_t
UdmSQLDBDriver(UDM_DB *db)
{
  return UdmSQL(db)->DBDriver;
}

udm_sqldbmode_t
UdmSQLDBMode(UDM_DB *db)
{
  return UdmSQLDB(db)->DBMode;
}

int
UdmSQLDBVersion(UDM_DB *db)
{
  return UdmSQL(db)->version;
}

int
UdmSQLDBConnected(UDM_DB *db)
{
  return UdmSQL(db)->connected;
}

UDM_SQLDB_HANDLER *
UdmSQLDBHandler(UDM_DB *db)
{
  return &(UdmSQL(db)->handler);
}

static const UDM_DBMODE_HANDLER *
UdmSQLDBModeHandler(UDM_DB *db)
{
  return UdmSQLDB(db)->dbmode_handler;
}

static int
UdmSQLDBHaveIn(UDM_DB *db)
{
  return UdmSQL(db)->DBSQL_IN;
}

int
UdmSQLDBFlags(UDM_DB *db)
{
  return UdmSQL(db)->flags;
}

const char *
UdmSQLDBQueryFrom(UDM_QUERY *query)
{
  return query->from;
}

UDM_VARLIST *
UdmSQLDBVars(UDM_DB *db)
{
  return &UdmSQL(db)->Vars;
}


char *
UdmDBSQLError(UDM_DB *db)
{
  return UdmSQL(db)->errstr;
}


size_t
UdmDBSQLErrorSize(UDM_DB *db)
{
  return sizeof(UdmSQL(db)->errstr);
}


const char *
UdmDBSQLParamPlaceHolder(UDM_DB *db, size_t i)
{
  return UdmSQLParamPlaceHolder(UdmSQL(db), i);
}


static inline udm_rc_t
UdmDBSQLTrace(UDM_AGENT *A, UDM_DB *db, udm_rc_t rc, udm_timer_t start,
              const char *func)
{
  if (UdmSQL(db)->flags & UDM_SQL_DEBUG_QUERY)
  {
    fprintf(stderr, "%.2f %s\n", UdmStopTimer(&start), func);
  }
  return rc;
}


static inline udm_rc_t
UdmDBSQLTrace1(UDM_AGENT *A, UDM_DB *db, udm_rc_t rc, udm_timer_t start,
               const char *func, const char *param)
{
  if (UdmSQL(db)->flags & UDM_SQL_DEBUG_QUERY)
  {
    fprintf(stderr, "%.2f %s %s\n", UdmStopTimer(&start), func, param);
  }
  return rc;
}


static inline udm_rc_t
UdmDBSQLTrace2(UDM_AGENT *A, UDM_DB *db, udm_rc_t rc, udm_timer_t start,
               const char *func, const char *from, const char *to)
{
  if (UdmSQL(db)->flags & UDM_SQL_DEBUG_QUERY)
  {
    fprintf(stderr, "%.2f %s %s %s\n", UdmStopTimer(&start), func, from, to);
  }
  return rc;
}


udm_rc_t
UdmDBSQLQuery(UDM_AGENT *A, UDM_DB *db, UDM_SQLRES *Res, const char *buf)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLQuery(UdmSQL(db), Res, buf);
  return UdmDBSQLTrace1(A, db, rc, ticks, "SQL", buf);
}


udm_rc_t
UdmDBSQLDropTableIfExists(UDM_AGENT *A, UDM_DB *db, const char *name)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLDropTableIfExists(UdmSQL(db), name);
  return UdmDBSQLTrace1(A, db, rc, ticks, "DropTableIfExists", name);
}


udm_rc_t
UdmDBSQLTableTruncateOrDelete(UDM_AGENT *A, UDM_DB *db, const char *name)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLTableTruncateOrDelete(UdmSQL(db), name);
  return UdmDBSQLTrace1(A, db, rc, ticks, "TableTruncateOrDelete", name);
}


udm_rc_t
UdmDBSQLCopyStructure(UDM_AGENT *A, UDM_DB *db, const char *from, const char *to)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLCopyStructure(UdmSQL(db), from, to);
  return UdmDBSQLTrace2(A, db, rc, ticks, "CopyStructure", from, to);
}


udm_rc_t
UdmDBSQLRenameTable(UDM_AGENT *A, UDM_DB *db, const char *from, const char *to)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLRenameTable(UdmSQL(db), from, to);
  return UdmDBSQLTrace2(A, db, rc, ticks, "RenameTable", from, to);
}


udm_rc_t
UdmDBSQLBegin(UDM_AGENT *A, UDM_DB *db)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLBegin(UdmSQL(db));
  return UdmDBSQLTrace(A, db, rc, ticks, "Begin");
}


udm_rc_t
UdmDBSQLCommit(UDM_AGENT *A, UDM_DB *db)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLCommit(UdmSQL(db));
  return UdmDBSQLTrace(A, db, rc, ticks, "Commit");
}


udm_rc_t
UdmDBSQLExecDirect(UDM_AGENT *A, UDM_DB *db, UDM_SQLRES *R, const char *query)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLExecDirect(UdmSQL(db), R, query);
  return UdmDBSQLTrace1(A, db, rc, ticks, "ExecDirect", query);
}


udm_rc_t
UdmDBSQLPrepare(UDM_AGENT *A, UDM_DB *db, const char *query)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLPrepare(UdmSQL(db), query);
  return UdmDBSQLTrace1(A, db, rc, ticks, "Prepare", query);
}


udm_rc_t
UdmDBSQLExecute(UDM_AGENT *A, UDM_DB *db)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLExecute(UdmSQL(db));
  return UdmDBSQLTrace(A, db, rc, ticks, "Execute");
}


udm_rc_t
UdmDBSQLFetchRow(UDM_AGENT *A, UDM_DB *db, UDM_SQLRES *R, UDM_STR *p)
{
  return UdmSQLDBHandler(db)->FetchRow(UdmSQL(db), R, p);
}


udm_rc_t
UdmDBSQLStmtFree(UDM_AGENT *A, UDM_DB *db)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLStmtFree(UdmSQL(db));
  return UdmDBSQLTrace(A, db, rc, ticks, "StmtFree");
}


static void
UdmSQLPrintParameter(FILE *file, const void *data, int size)
{
  const unsigned char *s= (const unsigned char*) data;
  const unsigned char *e= s + size;
  fprintf(stderr, "'");
  for ( ; s < e; s++)
  {
    if (*s >= 0x20 && *s <= 0x7E)
      fprintf(file, "%c", *s);
    else
      fprintf(file, "\\x%02X", (int) *s);
  }
  fprintf(stderr, "'");
}


udm_rc_t
UdmDBSQLBindParameter(UDM_AGENT *A, UDM_DB *db,
                      int pos, const void *data, int size, udm_sqltype_t type)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLBindParameter(UdmSQL(db), pos, data, size, type);
  UDM_ASSERT(size == 4 || type != UDM_SQLTYPE_INT32);
  if (UdmSQL(db)->flags & UDM_SQL_DEBUG_QUERY)
  {
    fprintf(stderr, "%.2f BindParameter[%d] %s(%d)",
            UdmStopTimer(&ticks), pos, UdmSQLTypeToStr(type), size);
    UdmSQLPrintParameter(stderr, data, size);
    fprintf(stderr, "\n");
  }
  return rc;
}


udm_rc_t
UdmDBSQLLockOrBegin(UDM_AGENT *A, UDM_DB *db, const char *param)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLLockOrBegin(UdmSQL(db), param);
  return UdmDBSQLTrace1(A, db, rc, ticks, "LockOrBegin", param);
}


udm_rc_t
UdmDBSQLUnlockOrCommit(UDM_AGENT *A, UDM_DB *db)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLUnlockOrCommit(UdmSQL(db));
  return UdmDBSQLTrace(A, db, rc, ticks, "UnlockOrCommit");
}


udm_rc_t
UdmDBSQLQueryOneRowInt(UDM_AGENT *A, UDM_DB *db, int *res, const char *qbuf)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmSQLQueryOneRowInt(UdmSQL(db), res, qbuf);
  return UdmDBSQLTrace1(A, db, rc, ticks, "QueryOneRowInt", qbuf);
}


char *
UdmDBSQLEscStrAlloc(UDM_AGENT *A, UDM_DB *db, const char *src, size_t srclen)
{
  return UdmSQLEscStrAlloc(UdmSQL(db), src, srclen);
}


char *
UdmDBSQLEscStrSimple(UDM_AGENT *A, UDM_DB *db, char *to, const char *from, size_t l)
{
  return UdmSQLEscStrSimple(UdmSQL(db), to, from, l);
}


size_t
UdmDBSQLBinEscStr(UDM_AGENT *A, UDM_DB *db, char *dst, size_t dstlen, const char *src, size_t srclen)
{
  return UdmSQLBinEscStr(UdmSQL(db), dst, dstlen, src, srclen);
}

size_t
UdmDBSQLEscStr(UDM_AGENT *A, UDM_DB *db, char *to, const char *from, size_t l)
{
  return UdmSQLEscStr(UdmSQL(db), to, from, l);
}

void
UdmDBSQLTopClause(UDM_AGENT *A, UDM_DB *db, size_t top_num, UDM_SQL_TOP_CLAUSE *Top)
{
  UdmSQLTopClause(UdmSQL(db), top_num, Top);
}

static const char*
UdmSQLDBTypeToStr(udm_sqldbtype_t dbtype)
{
  switch(dbtype)
  {
    case UDM_DB_MYSQL:   return "mysql";
    case UDM_DB_PGSQL:   return "pgsql";
    case UDM_DB_IBASE:   return "ibase";
    case UDM_DB_MSSQL:   return "mssql";
    case UDM_DB_ORACLE8: return "oracle";
    case UDM_DB_SQLITE:  return "sqlite";
    case UDM_DB_SQLITE3: return "sqlite";
    case UDM_DB_MIMER:   return "mimer";
    case UDM_DB_VIRT:    return "virtuoso";
    case UDM_DB_ACCESS:  return "access";
    case UDM_DB_DB2:     return "db2";
    case UDM_DB_CACHE:   return "cache";
    case UDM_DB_SYBASE:  return "sybase";
    case UDM_DB_MONETDB: return "monetdb";
    case UDM_DB_SOLID:   return "solid";
    case UDM_DB_SAPDB:   return "sapdb";
  }
  return "unknown_dbtype";
}


static const char*
UdmSQLDBModeToStr(udm_sqldbmode_t dbmode)
{
  switch (dbmode) 
  {
    case UDM_SQLDBMODE_SINGLE:  return "single";
    case UDM_SQLDBMODE_MULTI:   return "multi";
    case UDM_SQLDBMODE_BLOB:    return "blob";
    case UDM_SQLDBMODE_RAWBLOB: return "rawblob";
  }
  return "unknown_dbmode";
}


static udm_sqldbmode_t
UdmStr2DBMode(const char * str1, udm_bool_t *error)
{
  *error= UDM_FALSE;
  if (!strncasecmp(str1,"single",6)) return UDM_SQLDBMODE_SINGLE;
  if (!strncasecmp(str1,"multi",5))  return UDM_SQLDBMODE_MULTI;
  if (!strncasecmp(str1,"blob",4))   return UDM_SQLDBMODE_BLOB;
  if (!strncasecmp(str1,"rawblob",7))return UDM_SQLDBMODE_RAWBLOB;
  *error= UDM_TRUE;
  return UDM_SQLDBMODE_BLOB;
}


/************** some forward declarations ********************/
static udm_rc_t UdmDeleteURL(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db);
static udm_rc_t UdmDeleteBadHrefs(UDM_AGENT *Indexer, 
                                  UDM_DOCUMENT *Doc,
                                  UDM_DB *db,
                                  urlid_t url_id);
static udm_rc_t UdmDeleteWordFromURL(UDM_AGENT *A, UDM_DOCUMENT *D, UDM_DB *db);


/*********************** helper functions **********************/

static size_t
WhereConditionAddAnd(char *where, const char *add)
{
  if (add[0])
  {
    const char *joiner= where[0] ? " AND " : "";
    return sprintf(where + strlen(where), "%s%s", joiner, add);
  }
  return 0;
}


static void
WhereConditionDSTRAddAnd(char *where, UDM_DSTR *add)
{
  if (UdmDSTRLength(add))
  {
    if (where[0])
      strcat(where, " AND ");
    strcat(where, UdmDSTRPtr(add));
  }
}


/* Prepare to add a new condition into a class */
static udm_rc_t
UdmSQLWhereAddJoiner(UDM_DSTR *dstr, const char *joiner)
{
  if (UdmDSTRLength(dstr))  /* Second (or more) condition of the same class */
  {
    UdmDSTRShrinkLast(dstr);
    UdmDSTRAppendSTR(dstr, joiner);
  }
  else                 /* First condition of this class */
  {
    UdmDSTRAppend(dstr, "(", 1);
  }
  return UDM_OK;
}


static udm_rc_t
UdmSQLWhereIntParam(UDM_DB *db,
                    UDM_DSTR *dstr, const char *sqlname, const char *val)
{
  const char *range= strchr(val, '-');
  UdmDSTRRealloc(dstr, UdmDSTRLength(dstr) + strlen(val) + 50);
  if (UdmSQL(db)->DBSQL_IN && !range) /* Single value */
  {
    if (!UdmDSTRLength(dstr)) /* First parameter */
    {
      UdmDSTRAppendf(dstr, " %s IN (%d)", sqlname, atoi(val));
    }
    else /* Second or higher parameter */
    {
      UdmDSTRShrinkLast(dstr);
      UdmDSTRAppendf(dstr, ",%d)", atoi(val));
    }
  }
  else /* Range */
  {
    int first, second;
    UdmSQLWhereAddJoiner(dstr, " OR ");
    if (range && 2 == sscanf(val, "%d-%d", &first, &second))
    {
      UdmDSTRAppendf(dstr,
                     "%s>=%d AND %s<=%d)",
                     sqlname, first, sqlname, second);
    }
    else
    {
      UdmDSTRAppendf(dstr, "%s=%d)", sqlname, atoi(val));
    }
  }
  return UDM_OK;
}


#define UDM_ADD_PARAM_NEG                1  /* If reverse condition      */
#define UDM_ADD_PARAM_WITH_TAIL_PERCENT  2  /* If the trailing % needed  */
#define UDM_ADD_PARAM_CHECK_URL_SCHEMA   4  /* Detect URL prefix         */

#define UDM_ADD_PARAM_URL     (UDM_ADD_PARAM_CHECK_URL_SCHEMA|UDM_ADD_PARAM_WITH_TAIL_PERCENT)
#define UDM_ADD_PARAM_URL_NEG (UDM_ADD_PARAM_NEG|UDM_ADD_PARAM_URL)


static udm_rc_t
UdmSQLWhereStrParam(UDM_DB *db, UDM_DSTR *dstr,
                    const char *sqlname, const char *val, int flag)
{
  const char *first= "";
  const char *last= (flag & UDM_ADD_PARAM_WITH_TAIL_PERCENT) ? "%" : "";
  const char *joiner= (flag & UDM_ADD_PARAM_NEG) ? " AND " : " OR ";
  const char *notx= (flag & UDM_ADD_PARAM_NEG) ? "NOT " : "";

  if (flag & UDM_ADD_PARAM_CHECK_URL_SCHEMA)
  {
    UDM_URL URL;
    UdmURLInit(&URL);
    UdmURLParse(&URL,val);
    /* Check if URL prefix is not given / given */
    first= (URL.schema == NULL) ? "%" : "";
    UdmURLFree(&URL);
  }

  UdmSQLWhereAddJoiner(dstr, joiner);
  UdmDSTRAppendf(dstr, "%s %sLIKE '%s%s%s')", sqlname, notx, first, val, last);
  return UDM_OK;
}


typedef struct udm_date_param_st
{
  int dt;
  int dx;
  int dm;
  int dy;
  int dd;
  time_t dp;
  int DB;
  int DE;
  int dstmp;
} UDM_DATE_PARAM;


static void
UdmDateParamInit(UDM_DATE_PARAM *d)
{
  d->dt= UDM_DT_UNKNOWN;
  d->dx= 1;
  d->dm= 0;
  d->dy= 1970;
  d->dd= 1;
  d->dp= (time_t) 0;
  d->DB= 0;
  d->DE= time(NULL);
  d->dstmp= 0;
}


static udm_rc_t
UdmCheckDateParam(UDM_DATE_PARAM *d,
                  const char *var, const char *val)
{
  int intval= atoi(val);
  int longval= atol(val);

  if (!strcmp(var, "dt"))
  {
    if(!strcasecmp(val, "back")) d->dt= UDM_DT_BACK;
    else if (!strcasecmp(val, "er")) d->dt= UDM_DT_ER;
    else if (!strcasecmp(val, "range")) d->dt= UDM_DT_RANGE;
  }
  else if (!strcmp(var, "dx"))
  {
    if (intval == 1 || intval == -1) d->dx= intval;
    else d->dx= 1;
  }
  else if (!strcmp(var, "dm"))
  {
    d->dm= intval; /* 0=Jan, 1=Feb,..., 11=Dec.*/
  }
  else if (!strcmp(var, "dy"))
  {
    d->dy= (intval) ? intval : 1970;
  }
  else if (!strcmp(var, "dd"))
  {
    d->dd= (intval) ? intval : 1;
  }
  else if (!strcmp(var, "dstmp"))
  {
    d->dstmp= longval ? longval : 0;
  }
  else if (!strcmp(var, "dp"))
  {
    d->dp= Udm_dp2time_t(val);
  }
  else if (!strcmp(var, "db"))
  {
    struct tm tm;
    bzero((void*) &tm, sizeof(tm));
    sscanf(val, "%d/%d/%d", &tm.tm_mday, &tm.tm_mon, &tm.tm_year);
    tm.tm_year -= 1900; tm.tm_mon--;
    d->DB= mktime(&tm);
  }
  else if (!strcmp(var, "de"))
  {
    struct tm tm;
    bzero((void*) &tm, sizeof(tm));
    sscanf(val, "%d/%d/%d", &tm.tm_mday, &tm.tm_mon, &tm.tm_year);
    tm.tm_year -= 1900; tm.tm_mon--;
    d->DE= mktime(&tm) + 86400; /* Including the given date */
  }
  else
    return UDM_ERROR;
  return UDM_OK;
}


static void
UdmSQLWhereDateParam(UDM_DSTR *cond, UDM_DATE_PARAM *d)
{
  switch(d->dt)
  {
    case UDM_DT_BACK:
      if (d->dp)
        UdmDSTRAppendf(cond, "url.last_mod_time >= %li",
                       (long int) time(NULL) - d->dp);
      break;
    case UDM_DT_ER:
      {
      struct tm tm;
        bzero((void*) &tm, sizeof(tm));
        tm.tm_mday= d->dd;
        tm.tm_mon= d->dm;
        tm.tm_year= d->dy - 1900;
        UdmDSTRAppendf(cond, "url.last_mod_time %s %li",
                       (d->dx == 1) ? ">=" : "<=",
                       (long int) (d->dstmp ? d->dstmp : mktime(&tm)));
      }
      break;
    case UDM_DT_RANGE:
      UdmDSTRAppendf(cond,
                     "url.last_mod_time >= %li AND url.last_mod_time <= %li",
                     (long int) d->DB, (long int) d->DE);
      break;
    case UDM_DT_UNKNOWN:
    default:
      break;
  }
}

typedef struct udm_search_param_st
{
  const char *name;
} UDM_SEARCH_PARAM;


static const UDM_SEARCH_PARAM search_params[]=
{
  {"ul"},
  {"ue"},
  {"u"},
  {"tag"},
  {"t"},
  {"lang"},
  {"g"},
  {"type"},
  {"typ"},
  {"sl.*"},
  {NULL}
};


static const UDM_SEARCH_PARAM*
UdmFindStringParam(const char *name)
{
  const UDM_SEARCH_PARAM *param;
  for (param= search_params; param->name; param++)
  {
    if (!UdmWildCaseCmp(name, param->name))
      return param;
  }
  return NULL;
}


typedef struct
{
  UDM_DSTR from;
  UDM_DSTR lang;
  UDM_DSTR seed;
  UDM_DSTR server;
  UDM_DSTR site;
  UDM_DSTR status;
  UDM_DSTR tag;
  UDM_DSTR timecond;
  UDM_DSTR type;
  UDM_DSTR ue;
  UDM_DSTR url;
  UDM_DSTR urlinfo;
  UDM_DATE_PARAM datep;
  int fromserver;
  int fromurlinfo_lang;
  int fromurlinfo_type;
  int fromurlinfo;
} UDM_SQL_CONDITION_PARAM;


static void
UdmSQLConditionParamInit(UDM_SQL_CONDITION_PARAM *param)
{
  UdmDSTRInit(&param->from, 64);
  UdmDSTRInit(&param->lang, 64);
  UdmDSTRInit(&param->seed, 64);
  UdmDSTRInit(&param->server, 64);
  UdmDSTRInit(&param->site, 64);
  UdmDSTRInit(&param->status, 64);
  UdmDSTRInit(&param->tag, 64);
  UdmDSTRInit(&param->timecond, 64);
  UdmDSTRInit(&param->type, 64);
  UdmDSTRInit(&param->ue, 64);
  UdmDSTRInit(&param->url, 64);
  UdmDSTRInit(&param->urlinfo, 64);
  UdmDateParamInit(&param->datep);
  param->fromserver= 1;
  param->fromurlinfo_lang= 1;
  param->fromurlinfo_type= 1;
  param->fromurlinfo= 1;
}


static void
UdmSQLConditionParamFree(UDM_SQL_CONDITION_PARAM *param)
{
  UdmDSTRFree(&param->from);
  UdmDSTRFree(&param->lang);
  UdmDSTRFree(&param->seed);
  UdmDSTRFree(&param->server);
  UdmDSTRFree(&param->site);
  UdmDSTRFree(&param->status);
  UdmDSTRFree(&param->tag);
  UdmDSTRFree(&param->timecond);
  UdmDSTRFree(&param->type);
  UdmDSTRFree(&param->ue);
  UdmDSTRFree(&param->url);
  UdmDSTRFree(&param->urlinfo);
}


static size_t
UdmSQLConditionParamTotalLength(const UDM_SQL_CONDITION_PARAM *CondParam)
{
  return
    UdmDSTRLength(&CondParam->from)     +
    UdmDSTRLength(&CondParam->lang)     +
    UdmDSTRLength(&CondParam->seed)     +
    UdmDSTRLength(&CondParam->server)   +
    UdmDSTRLength(&CondParam->site)     +
    UdmDSTRLength(&CondParam->status)   +
    UdmDSTRLength(&CondParam->tag)      +
    UdmDSTRLength(&CondParam->timecond) +
    UdmDSTRLength(&CondParam->type)     +
    UdmDSTRLength(&CondParam->ue)       +
    UdmDSTRLength(&CondParam->url)      +
    UdmDSTRLength(&CondParam->urlinfo);
}


static udm_rc_t
UdmSQLConditionParamAdd(UDM_SQL_CONDITION_PARAM *CondParam,
                        UDM_AGENT *A, UDM_DB *db,
                        const char *var, size_t varlen,
                        const char *val, size_t vallen)
{
  char varbuf[64];
  char valbuf[128 + 1];

  if (!vallen || varlen > (int) sizeof(varbuf))
    return UDM_OK;

  if (vallen > (int) sizeof(valbuf) - 1)
  {
    vallen= sizeof(valbuf) - 1;
    if (UdmFindStringParam(var))
    {
      udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db),
                   "Limit is too long: %.*s=%.*s",
                   (int) varlen, var, (int) vallen, val);
      return UDM_ERROR;
    }
  }

  /* Protection against SQL injection */
  var= UdmDBSQLEscStrSimple(A, db, varbuf, var, varlen);
  val= UdmDBSQLEscStrSimple(A, db, valbuf, val, vallen);

  if (!strcmp(var, "status"))
    UdmSQLWhereIntParam(db, &CondParam->status, "url.status", val);
  
  if (!strcmp(var, "seed"))
    UdmSQLWhereIntParam(db, &CondParam->seed, "url.seed", val);
  
  if (!strcmp(var, "site") && val[0] != 0)
    UdmSQLWhereStrParam(db, &CondParam->site, "url.url", val, UDM_ADD_PARAM_WITH_TAIL_PERCENT);
  
  if(!strcmp(var,"ul"))
    UdmSQLWhereStrParam(db, &CondParam->url, "url.url", val, UDM_ADD_PARAM_URL);

  if(!strcmp(var,"ue"))
    UdmSQLWhereStrParam(db, &CondParam->ue, "url.url", val, UDM_ADD_PARAM_URL_NEG);
  
  if(!strcmp(var,"u"))
    UdmSQLWhereStrParam(db, &CondParam->url, "url.url", val, 0);
  
  if(!strcmp(var,"tag") || !strcmp(var,"t"))
  {
    UdmSQLWhereStrParam(db, &CondParam->tag, "s.tag", val, 0);
    if (CondParam->fromserver)
    {
      CondParam->fromserver= 0;
      UdmDSTRAppendSTR(&CondParam->from, ", server s");
      UdmDSTRAppendSTR(&CondParam->server, " AND s.rec_id=url.server_id");
    }
  }

  if(!strcmp(var,"lang") || !strcmp(var,"g"))
  {
    UdmSQLWhereStrParam(db, &CondParam->lang, "il.sval", val, 0);
    if (CondParam->fromurlinfo_lang)
    {
      CondParam->fromurlinfo_lang= 0;
      UdmDSTRAppendSTR(&CondParam->from, ", urlinfo il");
      UdmDSTRAppendSTR(&CondParam->server, " AND il.url_id=url.rec_id AND il.sname='Content-Language'");
    }
  }

  if(!strncmp(var, "sl.", 3))
  {
    UdmSQLWhereAddJoiner(&CondParam->urlinfo, " AND ");
    UdmDSTRAppendf(&CondParam->urlinfo,"isl%d.sname='%s' AND isl%d.sval LIKE '%s')",
                   CondParam->fromurlinfo, var + 3, CondParam->fromurlinfo, val);
    UdmDSTRAppendf(&CondParam->from, ", urlinfo isl%d", CondParam->fromurlinfo);
    UdmDSTRAppendf(&CondParam->server, " AND isl%d.url_id=url.rec_id",
                   CondParam->fromurlinfo);
    CondParam->fromurlinfo++;
  }

  if (!strcmp(var,"type") || !strcmp(var, "typ"))
  {
    /* 
       "type" is a reserved word in ASP,
       so "typ" is also added as a workaround
    */
    UdmSQLWhereStrParam(db, &CondParam->type, "it.sval", val, 0);
    if (CondParam->fromurlinfo_type)
    {
      CondParam->fromurlinfo_type= 0;
      UdmDSTRAppendSTR(&CondParam->from, ", urlinfo it");
      UdmDSTRAppendSTR(&CondParam->server, " AND it.url_id=url.rec_id AND it.sname='Content-Type'");
    }
  }
  UdmCheckDateParam(&CondParam->datep, var, val);
  return UDM_OK;
}


static udm_rc_t
UdmSQLConditionParamPopulate(UDM_SQL_CONDITION_PARAM *CondParam,
                             UDM_AGENT *A, UDM_DB *db,
                             const UDM_VARLIST *Vars)
{
  size_t i;
  for (i= 0; i < Vars->nvars; i++)
  {
    const UDM_VAR *Var= UdmVarListFindConstByIndex(Vars, i);
    const char *var= UdmVarName(Var) ? UdmVarName(Var) : "";
    UDM_CONST_STR valbuf, *val= UdmVarGetConstStr(Var, &valbuf);
    size_t varlen= strlen(var);
    udm_rc_t rc;

    if (UDM_OK != (rc= UdmSQLConditionParamAdd(CondParam, A, db,
                                               var, varlen,
                                               val->str, val->length)))
      return rc;
  }
  UdmSQLWhereDateParam(&CondParam->timecond, &CondParam->datep);
  return UDM_OK;
}


static udm_rc_t
UdmSQLBuildWhereCondition(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                          char const **dst)
{
  UDM_ENV *Conf= A->Conf;
  size_t nbytes;
  UDM_SQL_CONDITION_PARAM CondParam;
  udm_bool_t need_new= UdmVarListFindBool(&A->Conf->Vars, "delta", UDM_FALSE);
  udm_rc_t rc= UDM_OK;
  const char *need_new_str= need_new ? /* TODO34: remove this */
    "url.rec_id IN (SELECT url_id FROM bdicti WHERE state=1)" : "";
  
  UDM_LOCK_CHECK_OWNER(A, UDM_LOCK_DB);
  if (Query->where)
  {
    *dst= Query->where;
    return UDM_OK;
  }

  UdmSQLConditionParamInit(&CondParam);

  if (UDM_OK != (rc= UdmSQLConditionParamPopulate(&CondParam,
                                                  A, db, &Conf->Vars)))
    goto ret;

  if (!(nbytes= UdmSQLConditionParamTotalLength(&CondParam) + strlen(need_new_str)))
  {
    Query->where= (char*) UdmStrdup("");
    Query->from= (char*) UdmStrdup("");
    goto ret;
  }
  UdmQueryClearLimits(Query);
  Query->where= (char*) UdmMalloc(nbytes + 100);
  Query->where[0]= '\0';
  Query->from= (char*) UdmStrdup(UdmDSTRLength(&CondParam.from) ?
                                 UdmDSTRPtr(&CondParam.from) : "");
  if (UdmDSTRLength(&CondParam.url))
    strcat(Query->where, UdmDSTRPtr(&CondParam.url));

  WhereConditionDSTRAddAnd(Query->where, &CondParam.lang);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.seed);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.site);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.status);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.tag);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.timecond);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.type);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.ue);
  WhereConditionDSTRAddAnd(Query->where, &CondParam.urlinfo);

  if (UdmDSTRLength(&CondParam.server))
  {
    if (!Query->where[0]) strcat(Query->where, " 1=1 ");
    strcat(Query->where, UdmDSTRPtr(&CondParam.server));
  }

  WhereConditionAddAnd(Query->where, need_new_str);

  /* Need this for test purposes */
  UdmVarListReplaceStr(&Conf->Vars, "WhereCondition", Query->where);

ret:
  *dst= Query->where;
  UdmSQLConditionParamFree(&CondParam);
  return rc;
}


static udm_rc_t
UdmVarListSQLEscape(UDM_AGENT *A, UDM_VARLIST *dst, UDM_VARLIST *src, UDM_DB *db)
{
  size_t i, nbytes= 0;
  char *tmp= NULL;
  for (i= 0; i < src->nvars; i++)
  {
    const UDM_VAR *V= UdmVarListFindConstByIndex(src, i);
    UDM_CONST_STR valuebuf, *value= UdmVarGetConstStr(V, &valuebuf);
    if (nbytes < value->length * 2 + 1)
    {
      nbytes= value->length * 2 + 1;
      tmp= (char*) UdmRealloc(tmp, nbytes);
    }
    UdmDBSQLEscStr(A, db, tmp, value->str ? value->str : "", value->length); /* doc Section */
    UdmVarListAddStr(dst, UdmVarName(V), tmp);
  }
  UdmFree(tmp);
  return UDM_OK;
}


/************* Servers ******************************************/

#define UDM_SERVER_TABLE_COLUMNS \
  "rec_id,url,tag,command,weight,ordre,parent,enabled "

static udm_rc_t
UdmServerInitFromRecord(UDM_SERVER *S, UDM_SQLRES *SQLRes, size_t row)
{
  const char *val;
  S->site_id= UDM_ATOI(UdmSQLValue(SQLRes, row, 0));
  val= UdmSQLValue(SQLRes, row, 1);
  if (UDM_OK != UdmMatchSetPattern(&S->Filter.Match, val ? val : ""))
    return UDM_ERROR;
  
  if ((val= UdmSQLValue(SQLRes, row, 2)) && val[0])
    UdmVarListReplaceStr(&S->Vars, "Tag", val);
  
  S->command= *UdmSQLValue(SQLRes, row, 3);
  S->weight= UDM_ATOF(UdmSQLValue(SQLRes, row, 4));
  S->ordre= UDM_ATOI(UdmSQLValue(SQLRes, row, 5));
  S->parent= UDM_ATOI(UdmSQLValue(SQLRes, row, 6));
  S->enabled= UDM_TEST(UDM_ATOI(UdmSQLValue(SQLRes, row, 7)));
  return UDM_OK;
}


static int
UdmServerNeedsUpdate(UDM_SERVER *a, UDM_SERVER *b)
{
  /* Note: we don't check "srvinfo" content */
  if (a->site_id != b->site_id ||
      strcmp(UdmMatchPatternConstStr(&a->Filter.Match),
             UdmMatchPatternConstStr(&b->Filter.Match)) ||
      a->command != b->command ||
      strcmp(UdmVarListFindStr(&a->Vars, "Tag", ""), UdmVarListFindStr(&b->Vars, "Tag", "")) ||
      a->weight != b->weight ||
      a->ordre != b->ordre ||
      a->parent != b->parent ||
      a->enabled != b->enabled)
    return 1;
  return 0;
}


static void
UdmFilterInitFromVars(UDM_FILTER *Filter, const UDM_VARLIST *Vars)
{
  Filter->Match.Param.match_mode= UdmVarListFindInt(Vars, "Match_type", UDM_MATCH_BEGIN);
  UdmMatchParamSetCaseInsensitive(&Filter->Match.Param,
                                  UDM_TEST(UdmVarListFindInt(Vars, "Case_sense",
                                                             UDM_CASE_INSENSITIVE)));
  UdmMatchParamSetNegative(&Filter->Match.Param,
                           UDM_TEST(UdmVarListFindInt(Vars, "Nomatch", 0)));
  Filter->method= UdmMethod(UdmVarListFindStr(Vars, "Method",
                                              UdmMethodStr(UDM_METHOD_DEFAULT)));
}


static void
UdmFilterToVars(const UDM_FILTER *Filter, UDM_VARLIST *Vars)
{
  if (Filter->method != UDM_METHOD_DEFAULT)
    UdmVarListReplaceStr(Vars, "Method",  UdmMethodStr(Filter->method));
  if (Filter->Match.Param.match_mode != UDM_MATCH_BEGIN)
    UdmVarListReplaceInt(Vars, "Match_type",  Filter->Match.Param.match_mode);
  if (!UdmMatchIsCaseInsensitive(&Filter->Match))
    UdmVarListReplaceInt(Vars, "Case_sense",  UDM_CASE_SENSITIVE);
  if (UdmMatchIsNegative(&Filter->Match))
    UdmVarListReplaceInt(Vars, "Nomatch", 1);
}

static udm_rc_t
UdmLoadServerTable(UDM_AGENT * Indexer, UDM_SERVERLIST *S, UDM_DB *db)
{
  size_t    rows, i, j, jrows;
  UDM_SQLRES  SQLRes, SRes;
  char    qbuf[1024];
  const char *filename= UdmVarListFindStr(UdmSQLDBVars(db), "filename", NULL);
  const char *name = (filename && filename[0]) ? filename : "server";
  const char *infoname = UdmVarListFindStr(UdmSQLDBVars(db), "srvinfo", "srvinfo");
  udm_rc_t rc= UDM_OK;
  const char  *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  
  udm_snprintf(qbuf,sizeof(qbuf)-1,"SELECT " UDM_SERVER_TABLE_COLUMNS
                                   "FROM %s "
                                   "WHERE enabled=1 AND parent=%s0%s "
                                   "ORDER BY ordre", name, qu, qu);

  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, qbuf)))
    return rc;
  
  rows= UdmSQLNumRows(&SQLRes);
  for(i= 0; i < rows; i++)
  {
    UDM_SERVER *Server= Indexer->Conf->Cfg_Srv;
    
    if (UDM_OK != (rc= UdmServerInitFromRecord(Server, &SQLRes, i)))
      goto ex;

    sprintf(qbuf,"SELECT sname,sval FROM %s WHERE srv_id=%s%i%s", infoname, qu, Server->site_id, qu);
    if(UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SRes, qbuf)))
      return rc;
    jrows= UdmSQLNumRows(&SRes);
    for(j= 0; j < jrows; j++)
    {
      const char *sname = UdmSQLValue(&SRes, j, 0);
      const char *sval = UdmSQLValue(&SRes, j, 1);
      UdmVarListReplaceStr(&Server->Vars, sname, sval);
    }
    UdmSQLFree(&SRes);

    Server->webspace= (udm_webspace_t) UdmVarListFindInt(&Server->Vars, "Follow", UDM_WEBSPACE_DEFAULT);
    UdmFilterInitFromVars(&Server->Filter, &Server->Vars);

    if (Server->command == 'S')
    {
      UdmServerAdd(Indexer, Server, 0);
      if ((Server->Filter.Match.Param.match_mode == UDM_MATCH_BEGIN) &&
          (Indexer->flags & UDM_FLAG_ADD_SERVURL))
      {
        UDM_HREFPARAM  HrefParam;
        UdmHrefParamInit(&HrefParam);
        HrefParam.server_id= Server->site_id;
        HrefParam.hops= (uint4) UdmVarListFindInt(&Server->Vars, "StartHops", 0);
        HrefParam.link_source= UDM_LINK_SOURCE_HTDB;
        UdmHrefListAddConst(&Indexer->Conf->Hrefs, &HrefParam,
                            UdmMatchPatternConstStr(&Server->Filter.Match));
      }
    }
    else
    {
      char errstr[128];
      UDM_CONST_STR str;
      UdmConstStrSetStr(&str, UdmMatchPatternConstStr(&Server->Filter.Match));
      rc= UdmFilterListAdd(&Indexer->Conf->Filters,
                           &Server->Filter.Match.Param,
                           Server->Filter.method,
                           &str,
                           errstr, sizeof(errstr));
      if (rc != UDM_OK)
      {
        udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db),
                    "Error while loading ServerTable '%s' at row %d: %s",
                    name, (int) i, errstr);
        break;
      }
    }
    UdmMatchFreeAndInit(&Server->Filter.Match);
  }
ex:
  UdmSQLFree(&SQLRes);
  return rc;
}


static udm_rc_t
UdmServerTableFlush(UDM_AGENT *A, UDM_DB *db)
{
  udm_rc_t rc;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  char str[128];
  
  udm_snprintf(str, sizeof(str),  "UPDATE server SET enabled=0 WHERE parent=%s0%s", qu, qu);
  rc= UdmDBSQLQuery(A, db, NULL, str);
  return rc;
}


static udm_rc_t
UdmServerTableUpdateSrvInfo(UDM_AGENT *A, UDM_DB *db, UDM_SERVER *S, char *arg)
{
  udm_rc_t rc;
  size_t i;
  UDM_DSTR d;
  const char *E= (UdmSQLDBType(db) == UDM_DB_PGSQL && UdmSQLDBVersion(db) >= 80101) ? "E" :"";
  UdmDSTRInit(&d, 64);
  
  UDM_ASSERT(UdmSQLDBConnected(db)); /* make sure E is set to a correct value */
  
  UdmDSTRAppendf(&d, "DELETE FROM srvinfo WHERE srv_id=%i", S->site_id);
  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, UdmDSTRPtr(&d))))
    goto ex;

  for (i= 0; i < S->Vars.nvars; i++)
  {
    const UDM_VAR *Sec= UdmVarListFindConstByIndex(&S->Vars, i);
    if (UdmVarStr(Sec) && UdmVarName(Sec) && strcasecmp(UdmVarName(Sec), "Tag"))
    {
      UDM_CONST_STR valuebuf, *value= UdmVarGetConstStr(Sec, &valuebuf);
      UdmDBSQLEscStr(A, db, arg, value->str, value->length); /* srvinfo */
      UdmDSTRReset(&d);
      UdmDSTRAppendf(&d, "INSERT INTO srvinfo (srv_id,sname,sval) "
                         "VALUES (%i,'%s',%s'%s')",
                     S->site_id, UdmVarName(Sec), E, arg);
      if(UDM_OK != (rc = UdmDBSQLQuery(A, db, NULL, UdmDSTRPtr(&d))))
        goto ex;
    }
  }
ex:
  UdmDSTRFree(&d);
  return rc;
}


static udm_rc_t
UdmServerTableUpdateWithLock(UDM_AGENT *A, UDM_DB *db, UDM_SERVER *S,
                             const char *buf, char *arg)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDBSQLLockOrBegin(A, db, "server WRITE, srvinfo WRITE")) ||
      UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, buf)) ||
      UDM_OK != (rc= UdmServerTableUpdateSrvInfo(A, db, S, arg)) ||
      UDM_OK != (rc= UdmDBSQLUnlockOrCommit(A, db)))
    return rc;
  return UDM_OK;
}


static udm_rc_t
UdmServerTableAdd(UDM_AGENT *A, UDM_SERVERLIST *S, UDM_DB *db)
{
  udm_rc_t res= UDM_OK;
  int found;
  const char  *alias=UdmVarListFindStr(&S->Server->Vars,"Alias",NULL);
  const char *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  size_t    i, len= 0;
  char    *buf, *arg;
  UDM_VARLIST  *Vars= &S->Server->Vars;
  UDM_SERVER Old;

  UdmServerInit(&Old);

  S->Server->site_id= UdmStrHash32(UdmMatchPatternConstStr(&S->Server->Filter.Match));
  
  for (i=0; i < Vars->nvars; i++)
  {
    const UDM_VAR *Var= UdmVarListFindConstByIndex(Vars, i);
    len= udm_max(len, UdmVarLength(Var));
  }

  len+= UdmMatchPatternConstStr(&S->Server->Filter.Match) ?
        strlen(UdmMatchPatternConstStr(&S->Server->Filter.Match)) : 0;
  len+= alias ? strlen(alias) : 0;
  len+= 2048;
  
  buf = (char*)UdmMalloc(len);
  arg = (char*)UdmMalloc(len);
  if (buf == NULL || arg == NULL)
  {
    UDM_FREE(buf);
    UDM_FREE(arg);
    strcpy(UdmDBSQLError(db), "Out of memory");
    UdmSQL(db)->errcode = 1;
    return UDM_ERROR;
  }

  for (found= 0; ; S->Server->site_id++)
  {
    UDM_SQLRES SQLRes;
    udm_snprintf(buf, len, "SELECT " UDM_SERVER_TABLE_COLUMNS
                           "FROM server WHERE rec_id=%s%d%s",
                           qu, S->Server->site_id, qu);
    if (UDM_OK != (res= UdmDBSQLQuery(A, db, &SQLRes, buf)))
      goto ex;
    
    if (!UdmSQLNumRows(&SQLRes))
    {
      UdmSQLFree(&SQLRes);
      break; /* Not found */
    }
    
    UdmServerInitFromRecord(&Old, &SQLRes, 0);
    found= !strcasecmp(UdmMatchPatternConstStr(&S->Server->Filter.Match),
                       UdmSQLValue(&SQLRes, 0, 1));
    UdmSQLFree(&SQLRes);
    
    if (found)
      break;
  }

  if (S->Server->webspace != UDM_WEBSPACE_DEFAULT)
    UdmVarListReplaceInt(&S->Server->Vars, "Follow",  S->Server->webspace);
  UdmFilterToVars(&S->Server->Filter, &S->Server->Vars);

  UdmDBSQLEscStr(A, db, arg,  /* Server pattern */
                 UDM_NULL2EMPTY(UdmMatchPatternConstStr(&S->Server->Filter.Match)),
                 strlen(UDM_NULL2EMPTY(UdmMatchPatternConstStr(&S->Server->Filter.Match))));
  
  if (!found)
  {
    udm_snprintf(buf, len,
                 "INSERT INTO server (rec_id, enabled, tag,"
                 " command, parent, ordre, weight, url, pop_weight) "
                 " VALUES "
                 "(%s%d%s, 1, '%s', '%c', %s%d%s, %d, %f, '%s', 0)",
         qu, S->Server->site_id, qu,
         UdmVarListFindStr(&S->Server->Vars, "Tag", ""),
         S->Server->command,
         qu, S->Server->parent, qu,
         S->Server->ordre,
         S->Server->weight,
         arg
     );
    if (UDM_OK != (res= UdmServerTableUpdateWithLock(A, db, S->Server, buf, arg)))
      goto ex;
  }
  else
  {
    if (UdmServerNeedsUpdate(S->Server, &Old))
    {
      udm_snprintf(buf, len,
                   "UPDATE server SET enabled=1, tag='%s',"
                   "command='%c', parent=%s%i%s, ordre=%d, weight=%f "
                   "WHERE rec_id=%s%d%s",
                    UdmVarListFindStr(&S->Server->Vars, "Tag", ""),
                    S->Server->command,
                    qu, S->Server->parent, qu,
                    S->Server->ordre,
                    S->Server->weight,
                    qu, S->Server->site_id, qu);
      if (UDM_OK != (res= UdmServerTableUpdateWithLock(A, db, S->Server, buf, arg)))
        goto ex;
    }
  }
  
  UDM_ASSERT(res == UDM_OK);

ex:
  UDM_FREE(buf);
  UDM_FREE(arg);
  UdmServerFree(&Old);
  return res;
}


static udm_rc_t
UdmSrvActionSQL(UDM_AGENT *A, UDM_DB *db, UDM_SERVERLIST *S, udm_srvcmd_t cmd)
{
  switch(cmd)
  {
    case UDM_SRV_ACTION_TABLE:
      return UdmLoadServerTable(A,S,db);
    case UDM_SRV_ACTION_FLUSH:
      return UdmServerTableFlush(A, db);
    case UDM_SRV_ACTION_ADD:
      return UdmServerTableAdd(A, S, db);
  }
  UdmLog(A, UDM_LOG_ERROR, "Unsupported Srv Action SQL");
  return UDM_ERROR;
}


/********** Searching for URL_ID by various parameters ****************/

static urlid_t
UdmURLIdCacheFind(const UDM_URLID_CACHE *Cache, const char *url)
{
  size_t i;
  for (i= 0; i < UDM_FINDURL_CACHE_SIZE; i++)
  {
    if (Cache->url[i] && !strcmp(url, Cache->url[i]))
      return Cache->id[i];
  }
  return 0;
}


static udm_rc_t
UdmURLIdCacheAdd(UDM_URLID_CACHE *Cache, const char *url, urlid_t id)
{
  UDM_FREE(Cache->url[Cache->pURLCache]);
  if (!(Cache->url[Cache->pURLCache]= (char*) UdmStrdup(url)))
    return UDM_ERROR;
  Cache->id[Cache->pURLCache]= id;
  Cache->pURLCache= (Cache->pURLCache + 1) % UDM_FINDURL_CACHE_SIZE;
  return UDM_OK;
}


static void
UdmURLIdCacheFree(UDM_URLID_CACHE *Cache)
{
  size_t i;
  for(i= 0; i < UDM_FINDURL_CACHE_SIZE; i++)
    UDM_FREE(Cache->url[i]);
}


static udm_rc_t
UdmFindURL(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc, UDM_DB *db)
{
  UDM_SQLRES  SQLRes;
  const char  *url=UdmVarListFindStr(&Doc->Sections,"URL","");
  udmhash32_t  id = 0;
  udm_rc_t rc= UDM_OK;
  udm_bool_t UseCRC32URLId= UdmVarListFindBool(&Indexer->Conf->Vars, "UseCRC32URLId", UDM_FALSE);

  if (UseCRC32URLId)
  {
    /* Auto generation of rec_id */
    /* using CRC32 algorythm     */
    id= UdmStrHash32(url);
  }
  else
  {
    const char *o;
    char *qbuf, *e_url;
    size_t i, l, url_length= strlen(url);
    
    /* Escape URL string */
    if ((e_url = (char*)UdmMalloc(l = (8 * url_length + 1))) == NULL ||
        (qbuf = (char*)UdmMalloc( l + 100 )) == NULL)
    {
      UDM_FREE(e_url);
      UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
      return UDM_ERROR;
    }
    UdmDBSQLEscStr(Indexer, db, e_url, url, url_length);

    if (!(id= UdmURLIdCacheFind(&UdmSQLDB(db)->URLIdCache, e_url)))
    {
      udm_snprintf(qbuf, l + 100, "SELECT rec_id FROM url WHERE url='%s'",e_url);
      if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, qbuf)))
      {
        UDM_FREE(e_url);
        UDM_FREE(qbuf);
        return rc;
      }
      for(i=0;i<UdmSQLNumRows(&SQLRes);i++)
      {
        if((o=UdmSQLValue(&SQLRes,i,0)))
        {
          id=atoi(o);
          break;
        }
      }
      UdmSQLFree(&SQLRes);
      rc= UdmURLIdCacheAdd(&UdmSQLDB(db)->URLIdCache, e_url, id);
    }
    UDM_FREE(e_url);
    UDM_FREE(qbuf);
  }
  UdmVarListReplaceInt(&Doc->Sections, "ID", id);
  return  rc;
}


static udm_rc_t
UdmFindMessage(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc, UDM_DB *db)
{
  size_t     i, len;
  char     *qbuf;
  char     *eid;
  UDM_SQLRES  SQLRes;
  const char  *message_id=UdmVarListFindStr(&Doc->Sections,"Header.Message-ID",NULL);
  udm_rc_t rc;
  
  if(!message_id)
    return UDM_OK;
  
  len = strlen(message_id);
  eid = (char*)UdmMalloc(4 * len + 1);
  if (eid == NULL) return UDM_ERROR;
  qbuf = (char*)UdmMalloc(4 * len + 128);
  if (qbuf == NULL)
  { 
    UDM_FREE(eid);
    return UDM_ERROR;
  }

  /* Escape URL string */
  UdmDBSQLEscStr(Indexer, db, eid, message_id, len); /* Message ID */
  
  udm_snprintf(qbuf, 4 * len + 128, 
     "SELECT rec_id FROM url u, urlinfo i WHERE u.rec_id=i.url_id AND i.sname='Message-ID' AND i.sval='%s'", eid);
  rc= UdmDBSQLQuery(Indexer, db, &SQLRes, qbuf);
  UDM_FREE(qbuf);
  UDM_FREE(eid);
  if (UDM_OK != rc)
    return rc;
  
  for(i=0;i<UdmSQLNumRows(&SQLRes);i++)
  {
    const char * o;
    if((o=UdmSQLValue(&SQLRes,i,0)))
    {
      UdmVarListReplaceInt(&Doc->Sections,"ID", UDM_ATOI(o));
      break;
    }
  }
  UdmSQLFree(&SQLRes);
  return(UDM_OK);
}


/********************* Limits ********************/


udm_rc_t
UdmLoadSlowLimit(UDM_AGENT *A, UDM_DB *db, UDM_URLID_LIST *list, const char *q)
{
  udm_timer_t ticks= UdmStartTimer();
  size_t i;
  udm_rc_t rc;
  UDM_SQLRES SQLRes;
  int exclude= list->exclude;
  bzero((void*) list, sizeof(UDM_URLID_LIST));
  list->exclude= exclude;
  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, q)))
    goto ret;

  if (!(list->nurls= UdmSQLNumRows(&SQLRes)))
    goto sqlfree;

  if (!(list->urls= (urlid_t *) UdmMalloc(sizeof(urlid_t) * list->nurls)))
  {
    rc= UDM_ERROR;
    list->nurls= 0;
    goto ret;
  }
  for (i= 0; i < list->nurls; i++)
  {
    list->urls[i]= atoi(UdmSQLValue(&SQLRes, i, 0));
  }
  
sqlfree:
  UdmLog(A, UDM_LOG_DEBUG, "Limit query retured %d rows: %.2f",
         (int) list->nurls, UdmStopTimer(&ticks));
  UdmSQLFree(&SQLRes);
ret:
  return rc;
}


static udm_rc_t
UdmLoadSlowLimitWithSort(UDM_AGENT *A, UDM_DB *db,
                         UDM_URLID_LIST *list, const char *q)
{
  udm_rc_t rc= UdmLoadSlowLimit(A, db, list, q);
  if (rc == UDM_OK)
    UdmURLIdListSort(list);
  return rc;
}


static udm_rc_t
UdmSlowLimitLoadForConv(UDM_AGENT *A,
                        UDM_DB *db,
                        UDM_URLID_LIST *fl_urls,
                        const char *fl)
{
  udm_rc_t rc= UDM_OK; 
  udm_timer_t ticks= UdmStartTimer();
  char name[64];
  const char *q;

  bzero((void*) fl_urls, sizeof(*fl_urls));

  UdmLog(A, UDM_LOG_INFO, "Loading fast limit '%s'", fl);
  if ((fl_urls->exclude= (fl[0] == '-')))
    fl++;

  udm_snprintf(name, sizeof(name), "Limit.%s", fl);
  if (!(q= UdmVarListFindStr(&A->Conf->Vars, name, NULL)))
  {
    UdmLog(A, UDM_LOG_ERROR, "Limit '%s' not specified", fl);
    return UDM_ERROR;
  }

  if (UDM_OK != (rc= UdmLoadSlowLimitWithSort(A, db, fl_urls, q)))
    return rc;
  UdmLog(A, UDM_LOG_DEBUG, "Limit '%s' loaded%s, %d records, %.2f sec",
         fl, fl_urls->exclude ? " type=excluding" : "", (int) fl_urls->nurls,
         UdmStopTimer(&ticks));
  return rc;
}



/******************** Orders ********************************/

/*
  Apply a sorted UserOrder to UDM_URLDATALIST
*/
static udm_rc_t
UdmApplyFastOrderToURLDataList(UDM_URLDATALIST *Data,
                               UDM_URL_INT4_LIST *Order)
{
  UDM_URLDATA *d= Data->Item;
  UDM_URLDATA *de= Data->Item + Data->nitems;

  if (!Order->nitems)
    return UDM_OK;
  
  for ( ; d < de; d++)
  {
    UDM_URL_INT4 *found;
    if ((found= (UDM_URL_INT4*) UdmBSearch(&d->url_id,
                                           Order->Item,
                                           Order->nitems,
                                           sizeof(UDM_URL_INT4),
                                           (udm_qsort_cmp)UdmCmpURLID)))
    {
      char buf[64];
      sprintf(buf, "%08X", found->param);
      d->section= UdmStrdup(buf);
    }
    else
    {
      d->section= UdmStrdup("00000001");
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmFastOrderLoadAndApplyToURLDataList(UDM_AGENT *Agent,
                                      UDM_DB *db,
                                      UDM_URLDATALIST *Data,
                                      const char *name,
                                      size_t *norder)
{
  UDM_URL_INT4_LIST Order;
  udm_rc_t rc;
  
  if ((UDM_OK != (rc= UdmBlobLoadFastOrder(Agent, db, &Order, name))) ||
      !Order.nitems)
    goto ret;

  rc= UdmApplyFastOrderToURLDataList(Data, &Order);
  
ret:
  *norder= Order.nitems;
  UDM_FREE(Order.Item);
  return rc;
}



/******************** URLData *******************************/

static udm_rc_t
UdmLoadURLDataFromURLForConv(UDM_AGENT *A,
                             UDM_DB *db,
                             UDM_QUERY *Query,
                             const UDM_URLID_LIST *fl_urls)
{
  udm_rc_t rc;
  udm_timer_t ticks= UdmStartTimer();
  char qbuf[4*1024];
  udm_bool_t need_url= UDM_TRUE; /* TODO34: remove this variable */
  /* TODO34: do not load the "url" column if no url.* sections defined */
  const char *urlfield, *urltable;
  size_t nbytes, i, j;
  UDM_SQLRES  SQLres;
  const char *dummy_where;

  bzero((void*) &Query->URLData, sizeof(Query->URLData));
  if (UDM_OK != (rc= UdmSQLBuildWhereCondition(A, db, Query, &dummy_where)))
    return rc;
  
  UdmLog(A, UDM_LOG_INFO, "Loading URL list");

  urlfield= need_url ? ", url.url" : "";
  urltable= (Query->from && Query->from[0]) ? "url." : "";
  udm_snprintf(qbuf, sizeof(qbuf),
              "SELECT %srec_id, last_mod_time%s"
              " FROM url%s%s%s",
              urltable, urlfield, Query->from,
              Query->where[0] ? " WHERE " : "", Query->where);

  if (UDM_OK != (rc = UdmDBSQLQuery(A, db, &SQLres, qbuf)))
    goto fin;
  
  Query->URLData.nitems= UdmSQLNumRows(&SQLres);
  nbytes= Query->URLData.nitems * sizeof(UDM_URLDATA);
  Query->URLData.Item= (UDM_URLDATA*) UdmMalloc(nbytes);

  for (i= 0, j= 0; i < Query->URLData.nitems; i++)
  {
    UDM_URLDATA *Data= &Query->URLData.Item[j];
    urlid_t url_id= UDM_ATOI(UdmSQLValue(&SQLres, i, 0));
    if (fl_urls->nurls)
    {
      void *found= UdmBSearch(&url_id, fl_urls->urls, fl_urls->nurls,
                              sizeof(urlid_t), (udm_qsort_cmp)UdmCmpURLID);
      if (found && fl_urls->exclude)
        continue;
      if (!found && !fl_urls->exclude)
        continue;
    }
    Data->url_id= url_id;
    Data->score= 0;
    Data->per_site= 0;
    Data->pop_rank= 0;
    Data->site_id= 0;
    Data->last_mod_time= UDM_ATOI(UdmSQLValue(&SQLres, i, 1));
    Data->url= need_url ? UdmStrdup(UdmSQLValue(&SQLres, i, 2)) : NULL;
    Data->section= NULL;
    j++;
  }
  Query->URLData.nitems= j;

  UdmURLDataListSort(&Query->URLData);
  UdmSQLFree(&SQLres);

fin:
  UdmLog(A, UDM_LOG_INFO, "URL list loaded: %d documents, %.2f sec",
                           (int) Query->URLData.nitems, UdmStopTimer(&ticks));
  return rc;
}


static udm_rc_t
UdmLoadURLDataFromURLAndSlowLimitForConv(UDM_AGENT *A,
                                         UDM_DB *db,
                                         UDM_QUERY *Query)
{
  udm_rc_t rc= UDM_OK;
  const char *fl= UdmVarListFindStr(&A->Conf->Vars, "fl", NULL);
  UDM_URLID_LIST fl_urls;

  bzero((void*)&fl_urls, sizeof(fl_urls));
  
  if (fl && (UDM_OK != (rc= UdmSlowLimitLoadForConv(A, db, &fl_urls, fl))))
    return rc;

  rc= UdmLoadURLDataFromURLForConv(A, db, Query, &fl_urls);
  
  UDM_FREE(fl_urls.urls);

  return rc;
}


/*
  Load the section with the given name from the table "urlinfo",
  for sorting by section: "s=S&su=section".
*/
static udm_rc_t
UdmLoadURLDataFromURLInfoUsingIN(UDM_AGENT *A,
                                 UDM_DB *db,
                                 UDM_URLDATALIST *DataList,
                                 const char *esu)
{
  udm_rc_t rc= UDM_OK;
  size_t offs;
  char qbuf[4*1024];

  for (offs= 0; offs < DataList->nitems; offs+= 256)
  {  
    size_t nrows, s, i;
    int notfirst= 0;
    UDM_SQLRES SQLres;
    char *end= qbuf + sprintf(qbuf, "SELECT url_id, sval"
                              " FROM urlinfo"
                              " WHERE sname='%s' AND url_id IN (", esu);
    
    for (i= 0; (i < 256) && (offs + i < DataList->nitems); i++)
    {
      end+= sprintf(end, "%s%i", (notfirst) ? "," : "",
                    DataList->Item[offs + i].url_id);
      notfirst= 1;
    }
    end+= sprintf(end, ") ORDER BY url_id");
  
    if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLres, qbuf)))
      goto fin;
  
    nrows= UdmSQLNumRows(&SQLres);

    for(i= 0, s= i + offs; i < nrows; s++)
    {
      if (s == DataList->nitems)
        break;
      if (DataList->Item[s].url_id != (urlid_t) UDM_ATOI(UdmSQLValue(&SQLres, i, 0)))
      {
        DataList->Item[s].section= UdmStrdup("");
      }
      else
      {
        DataList->Item[s].section= UdmStrdup(UdmSQLValue(&SQLres, i, 1));
        i++;
      }
    }
    UdmSQLFree(&SQLres);
  }

fin:
  return rc;
}


/*
  Load URL data from "url" for sorting by:
  last_mod_time
  url
  section
*/
static udm_rc_t
UdmLoadURLDataFromURLUsingIN(UDM_AGENT *A,
                             UDM_DB *db,
                             UDM_URLDATALIST *DataList,
                             int flag)
{
  int need_url= (flag & (UDM_URLDATA_URL | UDM_URLDATA_SITE));
  udm_rc_t rc= UDM_OK;
  char qbuf[4*1024];
  UDM_SQLRES SQLres;
  UDM_STR row[5];
  size_t j;
  const char *hi_priority= UdmSQLDBType(db) == UDM_DB_MYSQL ? "HIGH_PRIORITY " : " ";

  for (j= 0; j < DataList->nitems; j+= 256)
  {
    size_t i;
    int notfirst = 0;
    udm_snprintf(qbuf, sizeof(qbuf),
            "SELECT %srec_id,last_mod_time%s"
            " FROM url"
            " WHERE rec_id IN (",
            hi_priority,
            need_url ? ",url" : "");
    for (i= 0; (i < 256) && (j + i < DataList->nitems); i++)
    {
      sprintf(UDM_STREND(qbuf), "%s%i", (notfirst) ? "," : "", DataList->Item[j + i].url_id);
      notfirst= 1;
    }
    sprintf(UDM_STREND(qbuf), ") ORDER BY rec_id");
    if (UDM_OK != (rc= UdmDBSQLExecDirect(A, db, &SQLres, qbuf)))
      goto fin;
    for (i= 0; UdmDBSQLFetchRow(A, db, &SQLres, row) == UDM_OK; i++)
    {
      UDM_URLDATA *D= &DataList->Item[i + j];
      if (D->url_id != (urlid_t) UDM_ATOI(row[0].str))
      {
        UdmLog(A, UDM_LOG_ERROR, "Dat url_id (%d) != SQL url_id (%d)", 
               D->url_id, UDM_ATOI(row[0].str));
      }
      D->last_mod_time= UDM_ATOI(row[1].str);
      if (need_url)
      {
        size_t sitelen= UdmAbsoluteURLSiteLength(row[2].str);
        D->site_id= UdmHash32(row[2].str, sitelen);
        D->url= UdmStrdup(row[2].str);
      }
      else
      {
        D->url= NULL;
        D->site_id= 0;
      }
      D->pop_rank= 0;
      D->section= NULL;
    }
    UdmSQLFree(&SQLres);

  }

fin:
  return rc;
}


static udm_rc_t
UdmLoadURLDataFromURLUsingLoop(UDM_AGENT *A, UDM_DB *db,
                               UDM_URLDATALIST *DataList, int flag)
{
  udm_rc_t rc= UDM_OK;
  char qbuf[256];
  size_t i;
  int need_url= (flag & UDM_URLDATA_URL);
  const char *hi_priority= UdmSQLDBType(db) == UDM_DB_MYSQL ? "HIGH_PRIORITY" : "";

  for (i = 0; i < DataList->nitems; i++)
  {
    UDM_SQLRES SQLres;
    UDM_URLDATA *D= &DataList->Item[i];
    udm_snprintf(qbuf, sizeof(qbuf),
                 "SELECT %s last_mod_time%s"
                 " FROM url WHERE rec_id=%i",
                 hi_priority,
                 need_url ? ",url" : "",
                 DataList->Item[i].url_id);
    if (UDM_OK != (rc = UdmDBSQLQuery(A, db, &SQLres, qbuf)))
      goto fin;
    if(UdmSQLNumRows(&SQLres))
    {
      D->url_id= DataList->Item[i].url_id;
      D->site_id= 0;
      D->last_mod_time= UDM_ATOI(UdmSQLValue(&SQLres, 0, 0));
      D->url= need_url ? UdmStrdup(UdmSQLValue(&SQLres, 0, 1)) : NULL;
      D->pop_rank= 0;
      D->section= NULL;
    }
    UdmSQLFree(&SQLres);
  }

fin:
  return rc;
}


static udm_rc_t
UdmLoadURLDataFromURL(UDM_AGENT *A, UDM_DB *db,
                      UDM_URLDATALIST *DataList, int flag)
{
  udm_rc_t rc= UDM_OK;
  udm_bool_t use_urlbasicinfo= UdmVarListFindBool(&A->Conf->Vars, "LoadURLBasicInfo", UDM_TRUE);
  const char *su= UdmVarListFindStr(&A->Conf->Vars, "su", NULL);

  /* Check that DataList is not empty and is sorted by url_id */
  UDM_ASSERT(DataList->nitems);
  UDM_ASSERT(DataList->Item[0].url_id <= DataList->Item[DataList->nitems - 1].url_id);

  if (!use_urlbasicinfo)
  {
    UdmLog(A,UDM_LOG_DEBUG,"Not using basic URL data from url");
    UdmURLDataListClearParams(DataList);
  }
  else if (UdmSQLDBHaveIn(db))
  {
    UdmLog(A,UDM_LOG_DEBUG,"Trying to load URL data from url");
    rc= UdmLoadURLDataFromURLUsingIN(A, db, DataList, flag);
  }
  else
  {
    UdmLog(A,UDM_LOG_DEBUG,"Trying to load URL data from url, not using IN");
    rc= UdmLoadURLDataFromURLUsingLoop(A, db, DataList, flag);
  }

  if ((flag & UDM_URLDATA_SU) && su && su[0])
  {
    char *esu=su ? UdmDBSQLEscStrSimple(A, db, NULL, su, strlen(su)) : NULL; /* User sort name */
    rc= UdmLoadURLDataFromURLInfoUsingIN(A, db, DataList, esu);
    UDM_FREE(esu);
  }

  return rc;
}



/****************************** User score **************************/ 

udm_rc_t
UdmUserScoreListLoad(UDM_AGENT *A, UDM_DB *db,
                     UDM_URL_INT4_LIST *List, const char *q)
{
  size_t i;
  udm_rc_t rc;
  UDM_SQLRES SQLRes;
  udm_timer_t ticks= UdmStartTimer();
  
  bzero((void*) List, sizeof(UDM_URL_INT4_LIST));

  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, q)))
    goto ret;

  if (!(List->nitems= UdmSQLNumRows(&SQLRes)))
    goto sqlfree;

  if (2 != UdmSQLNumCols(&SQLRes))
  {
    udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db),
                 "User Score query must return 2 columns, returned %d columns",
                 (int) UdmSQLNumCols(&SQLRes));
    rc= UDM_ERROR;
    UdmSQL(db)->errcode= 1;
    goto sqlfree;
  }

  if (!(List->Item= (UDM_URL_INT4*) UdmMalloc(sizeof(UDM_URL_INT4) * List->nitems)))
  {
    rc= UDM_ERROR;
    List->nitems= 0;
    goto sqlfree;
  }
  for (i= 0; i < List->nitems; i++)
  {
    List->Item[i].url_id= atoi(UdmSQLValue(&SQLRes, i, 0));
    List->Item[i].param= atoi(UdmSQLValue(&SQLRes, i, 1));
  }
  UdmSort(List->Item, List->nitems, sizeof(UDM_URL_INT4), (udm_qsort_cmp)UdmCmpURLID);

  UdmLog(A, UDM_LOG_DEBUG,
         "UserScore query returned %d columns, %d rows: %.2f",
         (int) UdmSQLNumCols(&SQLRes), (int) List->nitems, UdmStopTimer(&ticks));
  

sqlfree:
  UdmSQLFree(&SQLRes);
ret:
  return rc;
}


static udm_rc_t
UdmUserScoreListLoadAndApplyToURLScoreList(UDM_AGENT *Agent,
                                           UDM_URLSCORELIST *List,
                                           UDM_DB *db,
                                           const UDM_QUERY_PARAM *prm)
{
  char name[128];
  const char *us, *query;
  UDM_URL_INT4_LIST UserScoreList;
  udm_rc_t rc;
  udm_timer_t ticks= UdmStartTimer();
  
  if (!prm->UserScoreFactor ||
      !(us= UdmVarListFindStr(&Agent->Conf->Vars, "us", NULL)))
    return UDM_OK;
  
  UdmLog(Agent, UDM_LOG_DEBUG, "Start loading UserScore '%s'", us);
  
  udm_snprintf(name, sizeof(name), "Score.%s", us);
  query= UdmVarListFindStr(&Agent->Conf->Vars, name, NULL);

  if (UDM_OK != (rc= query ?
                 UdmUserScoreListLoad(Agent, db, &UserScoreList, query) :
                 UdmBlobLoadFastScore(Agent, db, &UserScoreList, us)) ||
      !UserScoreList.nitems)
    goto ret;

  rc= UdmUserScoreListApplyToURLScoreList(Agent, List, &UserScoreList, prm);

ret:  
  UdmLog(Agent, UDM_LOG_DEBUG, "%-30s%.2f (%d docs found)",
         "Stop  loading UserScore",  UdmStopTimer(&ticks), (int) UserScoreList.nitems);
  UDM_FREE(UserScoreList.Item);
  return rc;
}


static udm_rc_t
UdmUserSiteScoreListLoadAndApplyToURLDataList(UDM_AGENT *Agent,
                                              UDM_URLDATALIST *List,
                                              UDM_DB *db,
                                              const UDM_QUERY_PARAM *prm)
{
  char name[128];
  const char *us, *query;
  UDM_URL_INT4_LIST UserScoreList;
  udm_rc_t rc;

  if (!prm->UserScoreFactor ||
      !(us= UdmVarListFindStr(&Agent->Conf->Vars, "ss", NULL)))
    return UDM_OK;
  udm_snprintf(name, sizeof(name), "SiteScore.%s", us);
  if (!(query= UdmVarListFindStr(&Agent->Conf->Vars, name, NULL)))
    return UDM_OK;
  
  if ((UDM_OK != (rc= UdmUserScoreListLoad(Agent, db,
                                           &UserScoreList, query))) ||
      !UserScoreList.nitems)
    goto ret;

  rc= UdmUserScoreListApplyToURLDataList(Agent, List, &UserScoreList, prm);

ret:
  UDM_FREE(UserScoreList.Item);
  return rc;
}



/*********************** Creating fast index ******************/

static udm_rc_t
UdmIndexSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmLoadURLDataFromURLAndSlowLimitForConv(Indexer, db, Query)))
    return rc;

  rc= UdmSQLDBModeHandler(db)->QueryAction(Indexer, db, Query, UDM_QUERYCMD_INDEX);

  UdmURLDataListFree(&Query->URLData);
  return rc;
}


static udm_rc_t
UdmStoreWords(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  UDM_ASSERT(UdmSQLDBModeHandler(db)->StoreWords != NULL);
  return UdmSQLDBModeHandler(db)->StoreWords(Indexer, db, Doc);
}


/********************* Inserting/Deleting URLs and Links *******************/

static udm_rc_t
UdmDeleteRedirects(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  char qbuf[128];
  urlid_t  id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  udm_snprintf(qbuf, sizeof(qbuf),
               "DELETE FROM redirect WHERE url_id=%d", id);
  return UdmDBSQLQuery(Indexer, db, NULL, qbuf);
}


static udm_rc_t
UdmStoreRedirects(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  char qbuf[128];
  urlid_t  id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  udm_rc_t rc;
  size_t i;

  if (!Doc->Spider.collect_links_destination)
    return UDM_OK;
  if (UDM_OK != (rc= UdmDeleteRedirects(A, db, Doc)))
    return rc;

  udm_snprintf(qbuf, sizeof(qbuf),
               "INSERT INTO redirect (url_id,seed,url) VALUES(%s,%s,%s)",
                UdmDBSQLParamPlaceHolder(db, 1),
                UdmDBSQLParamPlaceHolder(db, 2),
                UdmDBSQLParamPlaceHolder(db, 3));

  if (UDM_OK != (rc= UdmDBSQLPrepare(A, db, qbuf)))
    return rc;

  /* UdmDocBaseHref() && UdmDocConvertHrefs() is already done here */
  for (i= 0; i < Doc->Hrefs.nhrefs; i++)
  {
    UDM_HREF *H= &Doc->Hrefs.Href[i];
    udmcrc32_t url_seed;

    if (H->Param.link_source != UDM_LINK_SOURCE_REDIRECT)
      continue;
    if (!(H->Param.method_reason & Doc->Spider.collect_links_destination))
      continue;

    url_seed= UdmStrHash32(H->url) & 0xFF;

    /*
      TODO34: Remove duplicate links.
      Some links can be duplicate here,
      because after UdmDocConvertHrefs() different links can become the same:
      http://site/  -> http://site/
      http://site/? -> http://site/
    */
    if (UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 1, &id, (int) sizeof(id),
                                             UDM_SQLTYPE_INT32)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 2, &url_seed, (int) sizeof(url_seed),
                                           UDM_SQLTYPE_INT32)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 3, H->url, (int) strlen(H->url),
                                           UDM_SQLTYPE_VARCHAR)) ||
      UDM_OK != (rc= UdmDBSQLExecute(A, db)))
    {
      break;
    }
  }
  UdmDBSQLStmtFree(A, db);
  return rc;
}


static udm_rc_t
UdmDeleteLinks(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  char  qbuf[128];
  urlid_t  url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  sprintf(qbuf,"DELETE FROM links WHERE url_id=%d", url_id);
  return UdmDBSQLQuery(Indexer, db, NULL, qbuf);
}


static udm_rc_t
UdmStoreLinks(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  char qbuf[128];
  udm_rc_t rc;
  size_t i;
  urlid_t url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  udmcrc32_t url_seed= 0;
  UDM_CHARSET *cs= Doc->lcs;

  if (!Doc->Spider.collect_links_destination)
    return UDM_OK;
  
  if (UdmSQLDBType(db) == UDM_DB_MYSQL &&
      UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, "LOCK TABLE links WRITE")))
    return rc;

  if (UDM_OK != (rc= UdmDeleteLinks(A, db, Doc)))
    return rc;

  udm_snprintf(qbuf, sizeof(qbuf),
               "INSERT INTO links (url_id,url,linktext,src,rel,seed) "
               "VALUES(%s,%s,%s,%s,%s,%s)",
                UdmDBSQLParamPlaceHolder(db, 1),
                UdmDBSQLParamPlaceHolder(db, 2),
                UdmDBSQLParamPlaceHolder(db, 3),
                UdmDBSQLParamPlaceHolder(db, 4),
                UdmDBSQLParamPlaceHolder(db, 5),
                UdmDBSQLParamPlaceHolder(db, 6));

  if (UDM_OK != (rc= UdmDBSQLPrepare(A, db, qbuf)))
    return rc;

  /* UdmDocBaseHref() && UdmDocConvertHrefs() is already done here */
  for (i= 0; i < Doc->Hrefs.nhrefs; i++)
  {
    UDM_HREF *H= &Doc->Hrefs.Href[i];
    UDM_CONST_STR def= {"", 0};
    UDM_CONST_STR txt, rel;
    const char *link_source= UdmLinkSourceStr(H->Param.link_source);

    if (H->Param.link_source == UDM_LINK_SOURCE_REDIRECT)
      continue;   
    /*
      TODO34: Allow to specify link sources in CollectLinks,
      with options similar to FollowLinks.
    */
    if (H->Param.link_source == UDM_LINK_SOURCE_DIR)
      continue;
    if (!(H->Param.method_reason & Doc->Spider.collect_links_destination))
      continue;

    UdmVarListFindConstStr(&txt, &H->HrefVars, "LinkText", &def);
    UdmVarListFindConstStr(&rel, &H->HrefVars, "Rel", &def);
    txt.length= cs->cset->well_formed_length(cs, txt.str, UDM_MIN(1024, txt.length), UDM_RECODE_HTML);
    rel.length= cs->cset->well_formed_length(cs, rel.str, UDM_MIN(32, rel.length), UDM_RECODE_HTML);
    url_seed= UdmStrHash32(H->url) & 0xFF;

    /*
      TODO34: Remove duplicate links.
      Some links can be duplicate here,
      because after UdmDocConvertHrefs() different links can become the same:
      http://site/  -> http://site/
      http://site/? -> http://site/
    */
    if (UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 1, &url_id, (int) sizeof(url_id),
                                             UDM_SQLTYPE_INT32)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 2, H->url, (int) strlen(H->url),
                                           UDM_SQLTYPE_VARCHAR)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 3, txt.str, (int) txt.length,
                                           UDM_SQLTYPE_VARCHAR)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 4, link_source, (int) strlen(link_source),
                                           UDM_SQLTYPE_VARCHAR)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 5, rel.str, (int) rel.length,
                                           UDM_SQLTYPE_VARCHAR)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 6, &url_seed, (int) sizeof(url_seed),
                                           UDM_SQLTYPE_INT32)) ||
      UDM_OK != (rc= UdmDBSQLExecute(A, db)))
    {
      break;
    }
  }
  UdmDBSQLStmtFree(A, db);
  if (rc== UDM_OK && UdmSQLDBType(db) == UDM_DB_MYSQL)
    rc= UdmDBSQLQuery(A, db, NULL, "UNLOCK TABLES");
  return rc;
}


static udm_rc_t
UdmExportURL(UDM_AGENT *Indexer,
             UDM_DB *db,
             UDM_DOCUMENT *Doc,
             const char *sql_export)
{
  udm_rc_t rc= UDM_OK;
  char *part, *lt, *sql_export_copy= UdmStrdup(sql_export);
  UDM_DSTR d;
  UDM_VARLIST Vars;
  UdmVarListInit(&Vars);
  UdmDSTRInit(&d,256);
  
  UdmVarListSQLEscape(Indexer, &Vars, &Doc->Sections, db);
  for (part= udm_strtok_r(sql_export_copy, ";", &lt) ;
       part ;
       part= udm_strtok_r(NULL, ";", &lt))
  {
    UdmDSTRParse(&d, part, &Vars);
    if (UDM_OK!= (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&d))))
      break;
    UdmDSTRReset(&d);
  }
  UdmVarListFree(&Vars);
  UdmDSTRFree(&d);
  UdmFree(sql_export_copy);
  return rc;
}


static udm_rc_t
UdmAddURL(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc,UDM_DB *db)
{
  char    *e_url, *qbuf;
  const char  *url;
  int    url_seed;
  int    use_crc32_url_id;
  int    usehtdburlid;
  udm_rc_t rc= UDM_OK;
  size_t          len;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  urlid_t rec_id = 0;

  url = UdmVarListFindStr(&Doc->Sections,"URL","");
  use_crc32_url_id = !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars, "UseCRC32URLId", "no"), "yes");
  usehtdburlid = UdmVarListFindInt(&Indexer->Conf->Vars, "UseHTDBURLId", 0);

  len = strlen(url);
  e_url = (char*)UdmMalloc(4 * len + 1);
  if (e_url == NULL) return UDM_ERROR;
  qbuf = (char*)UdmMalloc(4 * len + 512);
  if (qbuf == NULL)
  { 
    UDM_FREE(e_url);
    return UDM_ERROR;
  }
  
  url_seed = UdmStrHash32(url) & 0xFF;
  
  /* Escape URL string */
  UdmDBSQLEscStr(Indexer, db, e_url, url, len);
  
  if(use_crc32_url_id || usehtdburlid)
  {
    /* Auto generation of rec_id */
    /* using CRC32 algorythm     */
    if (use_crc32_url_id) rec_id = UdmStrHash32(url);
    else rec_id = UdmVarListFindInt(&Doc->Sections, "HTDB_URL_ID", 0);
    
    udm_snprintf(qbuf, 4 * len + 512, "INSERT INTO url "
                 "(rec_id,url,referrer,hops,crc32,next_index_time,status,seed,bad_since_time,server_id,docsize,last_mod_time,shows) "
                 "VALUES (%s%i%s,'%s',%s%i%s,%d,0,%d,0,%d,%d,%s%i%s,%s%i%s,%li,0)",
           qu, rec_id, qu,
           e_url,
           qu, UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0), qu,
           UdmVarListFindInt(&Doc->Sections,"Hops",0),
           (int)time(NULL),
           url_seed, (int)time(NULL),
           qu, UdmVarListFindInt(&Doc->Sections, "Server_id", 0), qu,
           qu, UdmVarListFindInt(&Doc->Sections, "Content-Length", 0), qu,
           UdmHttpDate2Time_t(UdmVarListFindStrNonEmpty(&Doc->Sections, "Last-Modified",
                              UdmVarListFindStrNonEmpty(&Doc->Sections, "Date", "")))
       );
  }else{
    /* Use dabatase generated rec_id */
    /* It depends on used DBType     */
    switch(UdmSQLDBType(db))
    {
    case UDM_DB_SOLID:
    case UDM_DB_ORACLE8:
    case UDM_DB_SAPDB:
      /* FIXME: Dirty hack for stupid too smart databases 
       Change this for config parameter checking */
/*      if (strlen(e_url)>UDM_URLSIZE)e_url[UDM_URLSIZE]=0;*/
      /* Use sequence next_url_id.nextval */
      udm_snprintf(qbuf, 4 * len + 512, "INSERT INTO url "
                   "(url,referrer,hops,rec_id,crc32,next_index_time,status,seed,bad_since_time,server_id)"
                   " VALUES "
                   "('%s',%i,%d,next_url_id.nextval,0,%d,0,%d,%d,%i)",
        e_url,
        UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
        UdmVarListFindInt(&Doc->Sections,"Hops",0),
        (int)time(NULL),
         url_seed, (int)time(NULL),
         UdmVarListFindInt(&Doc->Sections, "Server_id", 0));
      break;
    case UDM_DB_MIMER:
      udm_snprintf(qbuf, 4 * len + 512, "INSERT INTO url "
                   "(url,referrer,hops,rec_id,crc32,next_index_time,status,seed,bad_since_time,server_id)"
                   " VALUES "
                   "('%s',%i,%d,NEXT_VALUE OF rec_id_GEN,0,%d,0,%d,%d,%i)",
        e_url,
        UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
        UdmVarListFindInt(&Doc->Sections,"Hops",0),
        (int)time(NULL),
         url_seed, (int)time(NULL),
         UdmVarListFindInt(&Doc->Sections, "Server_id", 0));
      break;    
    case UDM_DB_IBASE:
      udm_snprintf(qbuf, 4 * len + 512, "INSERT INTO url "
                   "(url,referrer,hops,rec_id,crc32,next_index_time,status,seed,bad_since_time,server_id)"
                   " VALUES "
                   "('%s',%i,%d,GEN_ID(rec_id_GEN,1),0,%d,0,%d,%d,%i)",
        e_url,
        UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
        UdmVarListFindInt(&Doc->Sections,"Hops",0),
        (int)time(NULL),
         url_seed, (int)time(NULL),
         UdmVarListFindInt(&Doc->Sections, "Server_id", 0));
      break;
    case UDM_DB_PGSQL:
      if (UdmSQLDBVersion(db) > 90100)
      {
        /* Use 9.1 syntax: INSERT INTO t1 SELECT ... WHERE EXISTS (SELECT) */
        udm_snprintf(qbuf, 4 * len + 512, "INSERT INTO url "
                     "(url,referrer,hops,crc32,next_index_time,status,seed,bad_since_time,server_id,docsize,last_mod_time,shows)"
                     " SELECT "
                     "'%s',%s%i%s,%d,0,%d,0,%d,%d,%s%i%s,%s%i%s,%li,0"
                     " WHERE NOT EXISTS "
                     "(SELECT rec_id FROM url WHERE url='%s')",
                     e_url,
                     qu, UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0), qu,
                     UdmVarListFindInt(&Doc->Sections,"Hops",0),
                     (int)time(NULL),
                     url_seed, (int)time(NULL),
                     qu, UdmVarListFindInt(&Doc->Sections, "Server_id", 0), qu,
                     qu, UdmVarListFindInt(&Doc->Sections, "Content-Length", 0), qu,
                     UdmHttpDate2Time_t(UdmVarListFindStrNonEmpty(&Doc->Sections, "Last-Modified",
                     UdmVarListFindStr(&Doc->Sections, "Date", ""))),
                    e_url);
        break;
      }
      /* else fall through */
    case UDM_DB_MYSQL: /* MySQL generates itself */
    case UDM_DB_VIRT:
    case UDM_DB_MSSQL:
    case UDM_DB_DB2:
    case UDM_DB_SQLITE:
    case UDM_DB_ACCESS:
    case UDM_DB_CACHE:
    case UDM_DB_SYBASE:
    case UDM_DB_SQLITE3:
    case UDM_DB_MONETDB:
      udm_snprintf(qbuf, 4 * len + 512, "INSERT INTO url "
                   "(url,referrer,hops,crc32,next_index_time,status,seed,bad_since_time,server_id,docsize,last_mod_time,shows)"
                   " VALUES "
                   "('%s',%s%i%s,%d,0,%d,0,%d,%d,%s%i%s,%s%i%s,%li,0)",
             e_url,
             qu, UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0), qu,
             UdmVarListFindInt(&Doc->Sections,"Hops",0),
             (int)time(NULL),
             url_seed, (int)time(NULL),
             qu, UdmVarListFindInt(&Doc->Sections, "Server_id", 0), qu,
             qu, UdmVarListFindInt(&Doc->Sections, "Content-Length", 0), qu,
             UdmHttpDate2Time_t(UdmVarListFindStrNonEmpty(&Doc->Sections, "Last-Modified",
             UdmVarListFindStr(&Doc->Sections, "Date", "")))
         );
    }
  }

  /* Exec INSERT now */
  if(UDM_OK!=(rc=UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
    goto ex;

ex:

  UDM_FREE(qbuf);
  UDM_FREE(e_url);
  return rc;
}


/******************* Cached Copy *********************/

#define SQLRESTODOC_COLUMNS \
  "rec_id,url,last_mod_time,docsize," \
  "next_index_time,referrer,crc32,status"

static void
SQLResToDoc(UDM_ENV *Conf, UDM_DOCUMENT *D, UDM_SQLRES *sqlres, size_t i)
{
  time_t    last_mod_time;
  char    dbuf[UDM_MAXTIMESTRLEN];
  const char  *format = UdmVarListFindStr(&Conf->Vars, "DateFormat", "%a, %d %b %Y, %X %Z");
  double          pr;
  
  UdmVarListReplaceStr(&D->Sections,"URL",UdmSQLValue(sqlres,i,1));
  UdmVarListReplaceInt(&D->Sections, "URL_ID", UdmStrHash32(UdmSQLValue(sqlres,i,1)));
  last_mod_time=atol(UdmSQLValue(sqlres,i,2));
  UdmVarListReplaceInt(&D->Sections, "Last-Modified-Timestamp", (int) last_mod_time);
  if (strftime(dbuf, sizeof(dbuf), format, localtime(&last_mod_time)) == 0)
  {
    UdmTime_t2HttpStr(last_mod_time, dbuf, sizeof(dbuf));
  }
  UdmVarListReplaceStr(&D->Sections,"Last-Modified",dbuf);
  UdmVarListReplaceStr(&D->Sections,"Content-Length",UdmSQLValue(sqlres,i,3));
  pr= atof(UdmSQLValue(sqlres,i,3)) / 1024;
  sprintf(dbuf, "%.2f", pr);
  UdmVarListReplaceStr(&D->Sections,"Content-Length-K",dbuf);  
  last_mod_time=atol(UdmSQLValue(sqlres,i,4));
  if (strftime(dbuf, sizeof(dbuf), format, localtime(&last_mod_time)) == 0)
  {
    UdmTime_t2HttpStr(last_mod_time, dbuf, sizeof(dbuf));
  }
  UdmVarListReplaceStr(&D->Sections,"Next-Index-Time",dbuf);
  UdmVarListReplaceInt(&D->Sections, "Referrer-ID", UDM_ATOI(UdmSQLValue(sqlres,i,5)));
  UdmVarListReplaceInt(&D->Sections,"crc32",atoi(UdmSQLValue(sqlres,i,6)));

#if BAR_COMMA_PERIOD_ORACLE_PROBLEM
  {
	char *num= UdmSQLValue(sqlres, i, 8);
	char *comma= strchr(num, ',');
	if (comma)
	  *comma= '.';
  }
#endif

  UdmVarListReplaceStr(&D->Sections, "Status", UdmSQLValue(sqlres, i, 7));
}


static udm_rc_t
UdmGetURLInfoOneDoc(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc;
  char buf[64];
  size_t i;
  UDM_SQLRES SQLRes;
  
  udm_snprintf(buf, sizeof(buf), "SELECT sname, sval FROM urlinfo WHERE url_id=%d", UDM_ATOI(UdmVarListFindStr(&Doc->Sections, "ID", "0")));
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, buf)))
    return rc;

  for (i= 0; i < UdmSQLNumRows(&SQLRes); i++)
  {
    const char *sname= UdmSQLValue(&SQLRes, i, 0);
    const char *sval= UdmSQLValue(&SQLRes, i, 1);
    /*size_t l= UdmSQLLen(&SQLRes, i, 1);*/
    
    if (!sname)
      continue;

    UdmVarListReplaceStr(&Doc->Sections, sname, sval ?  sval : "");
  }
  UdmSQLFree(&SQLRes);
  
  return UDM_OK;
}


static udm_rc_t
UdmGetCachedCopyOneDoc(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  urlid_t url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  size_t max_doc_size= UdmVarListFindInt(&Indexer->Conf->Vars, "MaxDocSize", UDM_MAXDOCSIZE);
  udm_rc_t rc;
  char buf[128];
  UDM_SQLRES SQLRes;

  udm_snprintf(buf, sizeof(buf),
               "SELECT content FROM cachedcopy "
               "WHERE url_id=%d",
               (int) url_id);

  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, buf)))
    return rc;

  if (UdmSQLNumRows(&SQLRes) == 1)
  {
    const char *sval= UdmSQLValue(&SQLRes, 0, 0);
    size_t len= UdmSQLLen(&SQLRes, 0, 0);
    udm_timer_t timer= 0;
    UdmDocSetFromCachedHTTPResponse(Doc, sval, len, max_doc_size, &timer);
  }

  UdmSQLFree(&SQLRes);

  /* TODO34: Get cached copy from the original location
  if (unpack_cached_copy && !CachedCopy_found)
  {
    const char *url= UdmVarListFindStr(&Doc->Sections, "url", NULL);
    UdmGetURLSimple(Indexer, Doc, url);
  }
  */

  return UDM_OK;
}


static udm_rc_t
UdmGetCachedCopy(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_DB *db)
{
  UDM_SQLRES SQLRes;
  char buf[1024];
  udm_rc_t rc;
  int url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);

  if (!url_id)
    UdmFindURL(Indexer, Doc, db);
  url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  udm_snprintf(buf, sizeof(buf),
               "SELECT " SQLRESTODOC_COLUMNS
               " FROM url WHERE rec_id=%d", url_id);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, buf)))
    return rc;

  if (!UdmSQLNumRows(&SQLRes))
  {
    UdmSQLFree(&SQLRes);
    return UDM_ERROR;
  }

  SQLResToDoc(Indexer->Conf, Doc, &SQLRes, 0);
  UdmSQLFree(&SQLRes);

  if (UDM_OK != (rc= UdmGetCachedCopyOneDoc(Indexer, db, Doc)))
    return rc;

  if (UDM_OK != (rc= UdmGetURLInfoOneDoc(Indexer, db, Doc)))
    return rc;
 
  return UDM_OK;
}


/********************** Reindexing "indexer -a" *************************/

static udm_rc_t
UdmMarkForReindex(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  char    qbuf[1024];
  const char  *where;
  UDM_SQLRES   SQLRes;
  size_t          i, j;
  udm_rc_t rc;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  UDM_DSTR buf;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DB);
  if (UDM_OK != (rc= UdmSQLBuildWhereCondition(Indexer, db, Query, &where)))
    return rc;
  
  if (UdmSQLDBFlags(db) & UDM_SQL_HAVE_SUBSELECT &&
      UdmSQLDBType(db) != UDM_DB_MYSQL)
  {
    udm_snprintf(qbuf,sizeof(qbuf),"UPDATE url SET next_index_time=%d WHERE rec_id IN (SELECT url.rec_id FROM url%s %s %s)",
       (int)time(NULL), Query->from, (where[0]) ? "WHERE" : "", where);
    return UdmDBSQLQuery(Indexer, db, NULL, qbuf);
  }

  udm_snprintf(qbuf, sizeof(qbuf), "SELECT url.rec_id FROM url%s %s %s",
               Query->from, (where[0]) ? "WHERE" : "", where);
  if(UDM_OK != (rc = UdmDBSQLQuery(Indexer, db, &SQLRes, qbuf))) return rc;

  UdmDSTRInit(&buf, 4096);
  if (UdmSQLDBHaveIn(db)) 
  {
    for (i = 0; i < UdmSQLNumRows(&SQLRes); i += 512) 
    {
      UdmDSTRReset(&buf);
      UdmDSTRAppendf(&buf, "UPDATE url SET next_index_time=%d WHERE rec_id IN (", (int)time(NULL));
      for (j = 0; (j < 512) && (i + j < UdmSQLNumRows(&SQLRes)); j++) 
      {
        UdmDSTRAppendf(&buf, "%s%s%s%s", (j) ? "," : "", qu, UdmSQLValue(&SQLRes, i + j, 0), qu);
      }
      UdmDSTRAppendf(&buf, ")");
      if(UDM_OK != (rc = UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&buf)))) 
      {
        UdmSQLFree(&SQLRes);
	UdmDSTRFree(&buf);
        return rc;
      }
    }
  } else {
    for (i = 0; i < UdmSQLNumRows(&SQLRes); i++) 
    {
      UdmDSTRReset(&buf);
      UdmDSTRAppendf(&buf, "UPDATE url SET next_index_time=%d WHERE rec_id=%s", (int)time(NULL),  UdmSQLValue(&SQLRes, i, 0));
      if(UDM_OK != (rc = UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&buf)))) 
      {
        UdmSQLFree(&SQLRes);
	UdmDSTRFree(&buf);
        return rc;
      }
    }
  }
  UdmDSTRFree(&buf);
  UdmSQLFree(&SQLRes);
  return UDM_OK;
}


/************** Child - for new extensions ****************/

static udm_rc_t
UdmRegisterChild(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db)
{
  char  qbuf[1024];
  urlid_t  url_id = UdmVarListFindInt(&Doc->Sections,"ID",0);
  urlid_t  parent_id = UdmVarListFindInt(&Doc->Sections,"Parent-ID",0);
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  
  udm_snprintf(qbuf,sizeof(qbuf),"insert into links (ot,k,weight) values(%s%i%s,%s%i%s,0.0)", qu, parent_id, qu, qu, url_id, qu);
  return UdmDBSQLQuery(Indexer, db, NULL, qbuf);
}


/*********************** Update URL ***********************/

static udm_rc_t
UdmUpdateUrl(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db)
{
  char qbuf[256];
  urlid_t  url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
  int  status=UdmVarListFindInt(&Doc->Sections,"Status",0);
  int  prevStatus = UdmVarListFindInt(&Doc->Sections, "PrevStatus", 0);
  int  next_index_time=UdmHttpDate2Time_t(UdmVarListFindStr(&Doc->Sections,"Next-Index-Time",""));
  udm_rc_t res;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  
  if (prevStatus != status && status > 300 && status != 304)
    sprintf(qbuf, "UPDATE url SET "
            "status=%d,next_index_time=%d,bad_since_time=%d,server_id=%s%i%s"
            " WHERE rec_id=%s%i%s",
            status, next_index_time, (int)time(NULL),
            qu, UdmVarListFindInt(&Doc->Sections, "Server_id",0),
            qu, qu, url_id, qu);
  else
    sprintf(qbuf,"UPDATE url SET "
            "status=%d,next_index_time=%d,server_id=%s%i%s"
            " WHERE rec_id=%s%i%s",
            status, next_index_time,
            qu, UdmVarListFindInt(&Doc->Sections, "Server_id",0), qu,
            qu, url_id, qu);

  if(UDM_OK!=(res=UdmDBSQLQuery(Indexer, db, NULL, qbuf)))return res;
  
  /* remove all old broken hrefs from this document to avoid broken link collecting */
  return UdmDeleteBadHrefs(Indexer,Doc,db,url_id);
}


static udm_rc_t
UdmDocNormalizeContentLanguage(UDM_DOCUMENT *Doc)
{
  UDM_VAR    *var;
  if ((var= UdmVarListFindVar(&Doc->Sections, "Content-Language")))
  {
    char language[128];
    const char *lang= UdmVarStr(var) ? UdmVarStr(var) :
                      UdmVarListFindStr(&Doc->Sections, "DefaultLang", "en");
    size_t i, len= udm_snprintf(language, sizeof(language), "%s", lang);
    for(i= 0; i < len; i++)
      language[i]= tolower(language[i]);
    return UdmVarListReplaceStr(&Doc->Sections, "Content-Language", language);
  }
  return UDM_OK;
}


static udm_rc_t
UdmUpdateUrlWithLangAndCharset(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db)
{
  char  *qbuf;
  udm_rc_t rc;
  const char  *charset;
  int    status, prevStatus;
  urlid_t         url_id;
  char    qsmall[64];
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  int IndexTime= UdmVarListFindInt(&Indexer->Conf->Vars, "IndexTime", 0);
  
  status = UdmVarListFindInt(&Doc->Sections, "Status", 0);
  prevStatus = UdmVarListFindInt(&Doc->Sections, "PrevStatus", 0);
  url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);

  if (UDM_OK != (rc= UdmDocNormalizeContentLanguage(Doc)))
    return rc;

  charset = UdmVarListFindStr(&Doc->Sections, "Charset", 
            UdmVarListFindStr(&Doc->Sections, "RemoteCharset", "iso-8859-1"));
  charset = UdmCharsetCanonicalName(charset);
  UdmVarListReplaceStr(&Doc->Sections, "Charset", charset);
  
  if (prevStatus != status && status > 300 && status != 304)
    udm_snprintf(qsmall, 64, ", bad_since_time=%d", (int)time(NULL));
  else qsmall[0] = '\0';

  if (IndexTime)
  {
    if (! prevStatus) udm_snprintf(UDM_STREND(qsmall), 64, ",last_mod_time=%li", time(NULL));
  }
  else
  {
    const char *lmsrc= UdmVarListFindStrNonEmpty(&Doc->Sections, "User.Date",
                       UdmVarListFindStrNonEmpty(&Doc->Sections, "Last-Modified",
                       UdmVarListFindStrNonEmpty(&Doc->Sections, "Date", "")));
    udm_snprintf(UDM_STREND(qsmall), 64, ",last_mod_time=%li", UdmHttpDate2Time_t(lmsrc));
  }
  qbuf=(char*)UdmMalloc(1024);
  
  
  udm_snprintf(qbuf, 1023, "\
UPDATE url SET \
status=%d,\
next_index_time=%d,\
docsize=%d,\
crc32=%d%s, server_id=%s%i%s \
WHERE rec_id=%s%i%s",
  status,
  (int) UdmHttpDate2Time_t(UdmVarListFindStr(&Doc->Sections,"Next-Index-Time","")),
  UdmVarListFindInt(&Doc->Sections,"Content-Length",0),
  UdmVarListFindInt(&Doc->Sections,"crc32",0),
  qsmall,
  qu, UdmVarListFindInt(&Doc->Sections, "Server_id",0), qu,
  qu, url_id, qu);
  
  rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf);
  UDM_FREE(qbuf);
  return rc;
}


static udm_bool_t
UdmCachedCopyPrintCharset(udm_content_type_t ct)
{
  switch (ct)
  {
    case UDM_CONTENT_TYPE_TEXT_PLAIN: 
    case UDM_CONTENT_TYPE_TEXT_HTML: 
    case UDM_CONTENT_TYPE_TEXT_XML: 
    case UDM_CONTENT_TYPE_HTDB: 
      return UDM_TRUE;
    case UDM_CONTENT_TYPE_MESSAGE_RFC822: 
    case UDM_CONTENT_TYPE_AUDIO_MPEG: 
    case UDM_CONTENT_TYPE_DOCX: 
    case UDM_CONTENT_TYPE_TEXT_RTF:
    case UDM_CONTENT_TYPE_UNKNOWN: 
      return UDM_FALSE;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_rc_t
UdmDocInsertCachedCopy(UDM_AGENT *Agent, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  char qbuf[256];
  udm_rc_t rc;
  UDM_CONST_STR content;
  urlid_t url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  UDM_HTTPBUF tmp;
  const char *orig_content_type= UdmVarListFindStrNonEmpty(&Doc->Sections, "Content-Type", NULL);
  const char *parser_content_type= UdmVarListFindStrNonEmpty(&Doc->Sections, "Cached.Content-Type", orig_content_type);
  udm_content_type_t orig_ct= orig_content_type ? UdmContentTypeByName(orig_content_type) : UDM_CONTENT_TYPE_UNKNOWN;
  udm_content_type_t parser_ct= parser_content_type ? UdmContentTypeByName(parser_content_type) : UDM_CONTENT_TYPE_UNKNOWN;
  UDM_CHARSET *remote_charset= UdmVarListFindCharset(&Doc->Sections, "RemoteCharset", NULL);
  UDM_CHARSET *orig_charset= UdmVarListFindCharset(&Doc->Sections, "Charset", remote_charset);
  UDM_CHARSET *parser_charset= UdmVarListFindCharset(&Doc->Sections, "Cached.Charset", orig_charset);
  const char *encoding= UdmVarListFindStr(&Agent->Conf->Vars, "CachedCopyEncoding", "deflate");
  urlid_t ts= (urlid_t) time(0);

  if (!parser_content_type || !UdmSQLDBHandler(db)->Prepare)
    return UDM_ERROR;

  /*UdmVarListPrint(&Doc->Sections, stdout);*/

  /*
    Don't store cached copies if neither indexing nor archiving is allowed.
    TODO34: respect noarchive at search time.
  */
  if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content) || !content.length ||
      (!Doc->Spider.robots.archive && !Doc->Spider.robots.index))
    return UDM_OK;

  UdmHTTPBufInit(&tmp);
  if (UDM_OK != (rc= UdmHTTPBufAlloc(&tmp, content.length + 256)))
    return rc;

  UdmHTTPBufPrintf(&tmp, "Content-Type: %s", parser_content_type);
  if (parser_charset && UdmCachedCopyPrintCharset(parser_ct))
    UdmHTTPBufAppendf(&tmp, "; charset=%s", parser_charset->name);
  UdmHTTPBufAppendf(&tmp, "\r\n");

  /* Add the original content type if differs from the parser content type */
  if ((orig_content_type && orig_ct != parser_ct) ||
      (orig_charset != NULL && UdmCachedCopyPrintCharset(orig_ct) &&
       orig_charset != parser_charset))
      
  {
    UdmHTTPBufAppendf(&tmp, "X-Orig.Content-Type: %s", orig_content_type);
    if (orig_charset && UdmCachedCopyPrintCharset(orig_ct))
      UdmHTTPBufAppendf(&tmp, "; charset=%s", orig_charset->name);
    UdmHTTPBufAppendf(&tmp, "\r\n");
  }

  if (encoding && content.length > 128)
  {
    udm_content_encoding_t ce= UdmContentEncodingID(encoding);
    if (ce == UDM_CONTENT_ENCODING_DEFLATE)
    {
      UDM_HTTPBUF tmp2= tmp; /* Backup the buffer state */
      UdmHTTPBufAppendf(&tmp, "Content-Encoding: deflate\r\n\r\n");
      if (UDM_OK == UdmHTTPBufDeflateAppend(&tmp, content.str, content.length))
        goto ins;
      tmp= tmp2; /* Deflate failed, restore the buffer state */
    }
  }

  UdmHTTPBufAppendf(&tmp, "\r\n");
  UdmHTTPBufAppend(&tmp, content.str, content.length);

ins:
  udm_snprintf(qbuf, sizeof(qbuf),
               "INSERT INTO cachedcopy (url_id,ts,content) "
               "VALUES(%s,%s,%s)",
                UdmDBSQLParamPlaceHolder(db, 1),
                UdmDBSQLParamPlaceHolder(db, 2),
                UdmDBSQLParamPlaceHolder(db, 3));

  if (UDM_OK != (rc= UdmDBSQLPrepare(Agent, db, qbuf)))
    goto ex;

  UDM_ASSERT(sizeof(url_id) == 4);
  if (UDM_OK != (rc= UdmDBSQLBindParameter(Agent, db, 1,
                                           &url_id, (int) sizeof(url_id),
                                           UDM_SQLTYPE_INT32)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(Agent, db, 2,
                                           &ts, (int) sizeof(ts),
                                           UDM_SQLTYPE_INT32)) ||
      UDM_OK != (rc= UdmDBSQLBindParameter(Agent, db, 3,
                                           UdmHTTPBufPtr(&tmp),
                                           (int) UdmHTTPBufSize(&tmp),
                                           UDM_SQLTYPE_LONGVARBINARY)) ||
      UDM_OK != (rc= UdmDBSQLExecute(Agent, db)))
  {
    UdmDBSQLStmtFree(Agent, db);
    goto ex;
  }

  rc= UdmDBSQLStmtFree(Agent, db);

ex:
  UdmHTTPBufFree(&tmp);
  return rc;
}


static udm_bool_t
UdmDocSectionToBeStored(const UDM_ENV *Env, const UDM_VAR *Var)
{
  const UDM_SECTION *Sec;
  if (UdmVarValueHandlerType(Var) != UDM_VALUE_HANDLER_TYPE_STR)
    return UDM_FALSE;
  Sec= (const UDM_SECTION *) UdmVarConstDataPtr(Var);
  return 
    UdmEnvSectionMaxLengthEx(Env, Sec->Param.secno, Var->header.name) &&
    UdmSectionPtr(Sec) && UdmSectionLength(Sec);
}


static size_t
UdmDocHaveSectionsToBeStored(const UDM_ENV *Env, const UDM_DOCUMENT *Doc)
{
  size_t i;
  for (i= 0; i < Doc->Sections.nvars; i++)
  {
    if (UdmDocSectionToBeStored(Env, UdmVarListFindByIndex(&Doc->Sections, i)))
      return UDM_TRUE;
  }
  return UDM_FALSE;
}


static udm_rc_t
UdmDocInsertSectionsUsingBind(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  size_t i;
  char qbuf[256];
  urlid_t    url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);

  UDM_ASSERT(UdmSQLDBHandler(db)->Prepare);

  if (!UdmDocHaveSectionsToBeStored(A->Conf, Doc))
    return UDM_OK;

  udm_snprintf(qbuf, sizeof(qbuf),
               "INSERT INTO urlinfo (url_id,sname,sval) "
               "VALUES(%s, %s, %s)",
                UdmDBSQLParamPlaceHolder(db, 1),
                UdmDBSQLParamPlaceHolder(db, 2),
                UdmDBSQLParamPlaceHolder(db, 3));

  if (UDM_OK != (rc= UdmDBSQLPrepare(A, db, qbuf)))
    return rc;
  
  for(i= 0; i< Doc->Sections.nvars; i++)
  {
    const UDM_VAR *Sec= UdmVarListFindConstByIndex(&Doc->Sections, i);
    UDM_CONST_STR valuebuf, *value= UdmVarGetConstStr(Sec, &valuebuf);
    if (UdmDocSectionToBeStored(A->Conf, Sec))
    {
      udm_sqltype_t bindtype= UdmSQLLongVarCharBindType(UdmSQL(db));
      UDM_ASSERT(sizeof(url_id) == 4);
      if (UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 1,
                                               &url_id, (int) sizeof(url_id),
                                               UDM_SQLTYPE_INT32)) ||
          UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 2,
                                               UdmVarName(Sec),
                                               (int) UdmVarNameLength(Sec),
                                               UDM_SQLTYPE_VARCHAR)) ||
          UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 3,
                                               value->str, (int) value->length,
                                               bindtype)) ||
          UDM_OK != (rc= UdmDBSQLExecute(A, db)))
        return rc;
    }
  }

  return UdmDBSQLStmtFree(A, db);
}


static udm_rc_t
UdmDocInsertSectionsUsingEscapeBuildQuery(UDM_AGENT *A, UDM_DB *db,
                                          const char *table,
                                          urlid_t url_id,
                                          const char *extra_column_names,
                                          const UDM_CONST_STR *values,
                                          size_t nvalues,
                                          UDM_DSTR *qbuf)
{
  const char *E= (UdmSQLDBDriver(db) == UDM_DBAPI_PGSQL && UdmSQLDBVersion(db) >= 80101) ? "E" : "";
  size_t i;
  UdmDSTRReset(qbuf);
  UdmDSTRAppendf(qbuf, "INSERT INTO %s (url_id,%s) VALUES(",
                 table, extra_column_names);
  if (url_id)
    UdmDSTRAppendf(qbuf, "%d", url_id);
  else
    UdmDSTRAppendSTR(qbuf, "last_insert_id()");

  for (i= 0; i < nvalues; i++)
  {
    const UDM_CONST_STR *value= &values[i];
    size_t esclen;
    UdmDSTRReserve(qbuf, 2 + value->length * (UdmSQLDBType(db) == UDM_DB_PGSQL ? 4 : 2));
    UdmDSTRAppendf(qbuf, ",%s'", E);
    esclen= UdmDBSQLEscStr(A, db, qbuf->Val.str + qbuf->Val.length,
                           value->str, value->length);
    qbuf->Val.length+= esclen;
    UdmDSTRAppendf(qbuf, "'");
  }
  UdmDSTRAppend(qbuf, ")", 2);
  return UDM_OK;
}


static udm_rc_t
UdmDocInsertSectionsUsingEscape(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc,
                                const char *table,
                                urlid_t url_id)
{
  udm_rc_t rc= UDM_OK;
  size_t i, len, esc_multiply = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? 4 : 2;
  UDM_DSTR qbuf;

  /* Calculate maximum arg length */
  for(len= 0, i= 0; i < Doc->Sections.nvars; i++)
  {
    const UDM_VAR *Var= UdmVarListFindConstByIndex(&Doc->Sections, i);
    size_t l= UdmVarLength(Var) + UdmVarNameLength(Var);
    if (len < l)
      len= l;
  }
  if (!len)
    return UDM_OK;

  UdmDSTRInit(&qbuf, 256);
  UdmDSTRAlloc(&qbuf, esc_multiply * len + 128);
  
  for(i= 0; i< Doc->Sections.nvars; i++)
  {
    const UDM_VAR *Sec= UdmVarListFindConstByIndex(&Doc->Sections, i);
    if (UdmDocSectionToBeStored(A->Conf, Sec))
    {    
      UDM_CONST_STR column[2];
      UDM_CONST_STR valuebuf, *value= UdmVarGetConstStr(Sec, &valuebuf);
      const UDM_CONST_STR *c= (const UDM_CONST_STR*) &column;
      UDM_ASSERT(value);
      UdmConstStrSetStr(&column[0], UdmVarName(Sec));
      column[1]= value[0];
      UdmDocInsertSectionsUsingEscapeBuildQuery(A, db, table,
                                                url_id, "sname,sval",
                                                c, 2, &qbuf);
      if(UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, UdmDSTRPtr(&qbuf))))
        break;
    }
  }
  UdmDSTRFree(&qbuf);
  return rc;
}


static udm_bool_t
UdmLongUpdateURLUseTnx(UDM_AGENT *A, UDM_DB *db)
{
  switch (UdmSQLDBType(db))
  {
    case UDM_DB_VIRT:   /* TODO34: check */
    case UDM_DB_ACCESS: /* TODO34: check */
    case UDM_DB_DB2:    /* TODO34: check */
    case UDM_DB_CACHE:  /* TODO34: check */
      return UDM_FALSE;
    default:
    case UDM_DB_MYSQL:
      return UDM_TEST(UdmSQLDBFlags(db) & UDM_SQL_HAVE_TRANSACT);
  }
  return UDM_FALSE;
}

static udm_rc_t
UdmLongUpdateURL(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db)
{
  udm_rc_t rc= UDM_OK;
  urlid_t url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
  const char *c;
  udm_bool_t use_tnx= UdmLongUpdateURLUseTnx(Indexer, db);
  /*
   TNX works fine: Sybase: ASE-15.0.2 Dev Edition + UnixODBC.
  */
  
  if (use_tnx && UDM_OK != (rc= UdmDBSQLBegin(Indexer, db)))
    return rc;
  
  /* Now store words */
  if(UDM_OK != (rc= UdmStoreWords(Indexer, db, Doc)))
    return rc;

  /* Store links */
  if (UDM_OK != (rc= UdmStoreLinks(Indexer, db, Doc)) ||
      UDM_OK != (rc= UdmStoreRedirects(Indexer, db, Doc)))
    return rc;

  /* Copy default languages, if not given by server and not guessed */
  if (!(c= UdmVarListFindStr(&Doc->Sections,"Content-Language",NULL)))
  {
    if ((c= UdmVarListFindStr(&Doc->Sections,"DefaultLang",NULL)))
      UdmVarListReplaceStr(&Doc->Sections,"Content-Language",c);
  }
  

  if(UDM_OK != (rc= UdmUpdateUrlWithLangAndCharset(Indexer, Doc, db)))
    return rc;
  
  /* remove all old broken hrefs from this document to avoid broken link collecting */
  if (UDM_OK != (rc= UdmDeleteBadHrefs(Indexer,Doc,db,url_id)))
    return rc;
  
  /* Remove old URLInfo only if PrevStatus != 0 */
  if (UdmVarListFindInt(&Doc->Sections, "PrevStatus", 1))
  {
    char    qsmall[128];
    sprintf(qsmall,"DELETE FROM urlinfo WHERE url_id=%i", url_id);
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qsmall)))
      return rc;
    sprintf(qsmall,"DELETE FROM cachedcopy WHERE url_id=%i", url_id);
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qsmall)))
      return rc;
  }

/* No need delete from links here, it has been done before */
  
  if (UdmSQLDBFlags(db) & UDM_SQL_HAVE_BIND_TEXT)
  {
    rc= UdmDocInsertSectionsUsingBind(Indexer, db, Doc);
  }
  else
  {
    rc= UdmDocInsertSectionsUsingEscape(Indexer, db, Doc, "urlinfo", url_id);
  }

  if (rc == UDM_OK)
    rc= UdmDocInsertCachedCopy(Indexer, db, Doc);

  if(use_tnx && rc == UDM_OK)
    rc= UdmDBSQLCommit(Indexer, db);

  if (rc == UDM_OK && UdmSQLDBMode(db) == UDM_SQLDBMODE_MULTI)
  {
    int WordCacheSize= UdmVarListFindInt(&Indexer->Conf->Vars, "WordCacheSize", 0);
    if (WordCacheSize <= 0) WordCacheSize = 0x800000;
    /* UdmWordCacheWrite starts its own transaction */
    rc= UdmWordCacheWrite(Indexer, db, WordCacheSize);
  }
  return rc;
}


static udm_rc_t
UdmUpdateClone(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDeleteWordFromURL(Indexer, Doc, db)))
    return rc;
  rc= UdmUpdateUrlWithLangAndCharset(Indexer, Doc, db);
  return rc;
}



/************************ Clones stuff ***************************/
static udm_rc_t
UdmFindOrigin(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db)
{
  size_t    i=0;
  char    qbuf[256]="";
  UDM_SQLRES  SQLRes;
  urlid_t    origin_id = 0;
  int    scrc32=UdmVarListFindInt(&Doc->Sections,"crc32",0);
  udm_rc_t rc;
  
  if (scrc32==0)return UDM_OK;
  
  if (UdmSQLDBHaveIn(db))
    sprintf(qbuf,"SELECT rec_id FROM url WHERE crc32=%d AND status IN (200,304,206)",scrc32);
  else
    sprintf(qbuf,"SELECT rec_id FROM url WHERE crc32=%d AND (status=200 OR status=304 OR status=206)",scrc32);
  
  if(UDM_OK!=(rc=UdmDBSQLQuery(Indexer,db,&SQLRes,qbuf)))
    return rc;
  
  for(i=0;i<UdmSQLNumRows(&SQLRes);i++)
  {
    const char *o;
    if((o=UdmSQLValue(&SQLRes,i,0)))
      if((!origin_id) || (origin_id > UDM_ATOI(o)))
        origin_id = UDM_ATOI(o);
  }
  UdmSQLFree(&SQLRes);
  UdmVarListReplaceInt(&Doc->Sections, "Origin-ID", origin_id);
  return(UDM_OK);
}


static udm_rc_t
UdmCloneListSQL(UDM_AGENT * Indexer, UDM_DB *db, UDM_DOCUMENT *Doc, UDM_RESULT *Res)
{
  size_t    i, nr, nadd;
  char    qbuf[256];
  UDM_SQLRES  SQLres;
  int    scrc32=UdmVarListFindInt(&Doc->Sections,"crc32",0);
  urlid_t    origin_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
  udm_rc_t rc;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  const char  *format = UdmVarListFindStr(&Indexer->Conf->Vars, "DateFormat", "%a, %d %b %Y, %X %Z");

  if (UdmResultNumRows(Res) > 4) return UDM_OK;
  
  if (!scrc32)
    return UDM_OK;
  
  sprintf(qbuf,"SELECT rec_id,url,last_mod_time,docsize FROM url WHERE crc32=%d AND (status=200 OR status=304 OR status=206) AND rec_id<>%s%i%s", scrc32, qu, origin_id, qu);
  if (UDM_OK!= (rc= UdmDBSQLQuery(Indexer, db, &SQLres, qbuf)))
    return UDM_OK;
  
  nr = UdmSQLNumRows(&SQLres);
  if( nr == 0)
  {
    UdmSQLFree(&SQLres);
    return UDM_OK;
  }
  nadd = 5 - UdmResultNumRows(Res);
  if(nr < nadd) nadd = nr;
  
  Res->Doc= (UDM_DOCUMENT*)UdmRealloc(Res->Doc, 
                                      (UdmResultNumRows(Res) + nadd) *
                                      sizeof(UDM_DOCUMENT));
  for(i = 0; i < nadd; i++)
  {
    time_t    last_mod_time;
    char    buf[UDM_MAXTIMESTRLEN];
    UDM_DOCUMENT  *D = &Res->Doc[Res->num_rows + i];
    
    UdmDocInit(D);
    UdmVarListAddInt(&D->Sections, "ID", UDM_ATOI(UdmSQLValue(&SQLres,i,0)));
    UdmVarListAddStr(&D->Sections,"URL",UdmSQLValue(&SQLres,i,1));
    UdmVarListReplaceInt(&D->Sections, "URL_ID", UdmStrHash32(UdmSQLValue(&SQLres,i,1)));
    last_mod_time=atol(UdmSQLValue(&SQLres,i,2));
    if (strftime(buf, sizeof(buf), format, localtime(&last_mod_time)) == 0)
    {
      UdmTime_t2HttpStr(last_mod_time, buf, sizeof(buf));
    }
    UdmVarListAddStr(&D->Sections,"Last-Modified",buf);
    UdmVarListAddInt(&D->Sections,"Content-Length",atoi(UdmSQLValue(&SQLres,i,3)));
    UdmVarListAddInt(&D->Sections,"crc32",scrc32);
    UdmVarListAddInt(&D->Sections, "Origin-ID", origin_id);
  }
  Res->num_rows += nadd;
  UdmSQLFree(&SQLres);
  return UDM_OK;
}


static udm_rc_t
UdmQueryClones(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  size_t i;
  udm_rc_t rc= UDM_OK;
  for (i= 0; i < UdmResultNumRows(&Query->Res) && rc == UDM_OK; i++)
  {
    UDM_RESULT Cl;
    UdmResultInit(&Cl);
    if (UDM_OK == (rc= UdmCloneListSQL(A, db, &Query->Res.Doc[i], &Cl)))
    {
      size_t c;
      UdmVarListReplaceInt(&Query->Res.Doc[i].Sections, "nclones",
                           UdmResultNumRows(&Cl));
      for (c= 0; c < UdmResultNumRows(&Cl); c++)
      {
        char name[32];
        sprintf(name, "Clone%d", (int) c);
        UdmVarListReplaceLst(&Query->Res.Doc[i].Sections,
                             &Cl.Doc[c].Sections, name, "*");
      }
    }
    UdmResultFree(&Cl);
  }
  return rc;
}


/************** Get Target to be indexed ***********************/



static void
UdmSQLTopInit(UDM_SQL_TOP_CLAUSE *Top)
{
  Top->rownum[0]= 0;
  Top->limit[0]= 0;
  Top->top[0]= 0;
}



void
UdmSQLTopClause(UDM_SQL *db, size_t top_num, UDM_SQL_TOP_CLAUSE *Top)
{
  UdmSQLTopInit(Top);
  if (db->flags & UDM_SQL_HAVE_LIMIT)
  {
    udm_snprintf(Top->limit, UDM_SQL_TOP_BUF_SIZE, " LIMIT %d", (int) top_num);
  }
  else if (db->flags & UDM_SQL_HAVE_TOP)
  {
    udm_snprintf(Top->top, UDM_SQL_TOP_BUF_SIZE, " TOP %d ", (int) top_num);
  }
  else if (db->flags & UDM_SQL_HAVE_FIRST_SKIP)
  {
    udm_snprintf(Top->top, UDM_SQL_TOP_BUF_SIZE, " FIRST %d ", (int) top_num);
  }
  else if (db->DBType == UDM_DB_ORACLE8)
  {
#if HAVE_ORACLE8
    if(db->DBDriver == UDM_DBAPI_ORACLE8)
    {
      udm_snprintf(Top->rownum, UDM_SQL_TOP_BUF_SIZE,
                   " AND ROWNUM<=%d", (int) top_num);
    }
#endif
    if(!Top->rownum[0])
      udm_snprintf(Top->rownum, UDM_SQL_TOP_BUF_SIZE,
                   " AND ROWNUM<=%d", (int) top_num); 
  }
}


static const  char select_url_str[]=
"url.url,url.rec_id,docsize,status,hops,crc32,last_mod_time,seed";

static const char select_url_str_for_dump[]=
"url.url,url.rec_id,docsize,status,hops,crc32,last_mod_time,seed,"
"next_index_time,bad_since_time,server_id"
;

/*
  The columns that are dumped:
  - status
  - docsize
  - last_mod_time
  - hops
  - crc32
  - seed
  - url
  - next_index_time
  - bad_since_time
  - site_id
  - server_id
  
  The columns that don't need to be dumped for restore purposes:
  - rec_id
  - shows
  - sop_rank
  - referrer
*/
static udm_rc_t
UdmTargetSQLResDump(UDM_AGENT *Indexer, UDM_DB *db,
                    UDM_DOCUMENT *Doc,
                    UDM_SQLRES *SQLRes, size_t rownum,
                    UDM_DSTR *eurl)
{
  int seed= UDM_ATOI(UdmSQLValue(SQLRes, rownum, 7));
  UdmVarListAddInt(&Doc->Sections, "ID", UDM_ATOI(UdmSQLValue(SQLRes,rownum,1)));
  printf("--seed=%d\n", seed);
  printf("INSERT INTO url ");
  printf("(url,docsize,status,hops,crc32,last_mod_time,seed,next_index_time,bad_since_time,server_id) VALUES (");
  if (UDM_OK != UdmSQLEscDSTR(UdmSQL(db), eurl,
                              UdmSQLValue(SQLRes, rownum, 0),
                              UdmSQLLen(SQLRes, rownum, 0)))
    return UDM_ERROR;
  printf("'%s',", UdmDSTRPtr(eurl));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 2));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 3));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 4));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 5));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 6));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 7));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 8));
  printf("%s,", UdmSQLValue(SQLRes, rownum, 9));
  printf("%s", UdmSQLValue(SQLRes, rownum, 10));
  printf(");\n");
  return UDM_OK;
}


static void
UdmTargetSQLResToDoc(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,
                     UDM_SQLRES *SQLRes, size_t i)
{
  char buf[UDM_MAXTIMESTRLEN]= "";
  time_t last_mod_time;
  UdmVarListAddStr(&Doc->Sections,"URL",UdmSQLValue(SQLRes,i,0));
  UdmVarListAddInt(&Doc->Sections, "ID", UDM_ATOI(UdmSQLValue(SQLRes,i,1)));
  UdmVarListAddInt(&Doc->Sections,"Content-Length",atoi(UdmSQLValue(SQLRes,i,2)));
  UdmVarListAddInt(&Doc->Sections,"Status",atoi(UdmSQLValue(SQLRes,i,3)));
  UdmVarListAddInt(&Doc->Sections,"Hops",atoi(UdmSQLValue(SQLRes,i,4)));
  UdmVarListAddInt(&Doc->Sections,"crc32",atoi(UdmSQLValue(SQLRes,i,5)));
  last_mod_time= (time_t) atol(UdmSQLValue(SQLRes,i,6));
  UdmTime_t2HttpStr(last_mod_time, buf, sizeof(buf));
  if (last_mod_time != 0 && strlen(buf) > 0)
  {
    UdmVarListReplaceStr(&Doc->Sections, "Last-Modified", buf);
  }
}

/*
  Setting extending sections - only needed for targets
*/
static void
UdmTargetSQLResToDoc_Extra(UDM_AGENT *A, UDM_DOCUMENT *Doc,
                           UDM_SQLRES *SQLRes, size_t i)
{
  UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(UdmSQLValue(SQLRes,i,0)));
  UdmVarListAddInt(&Doc->Sections,"PrevStatus",atoi(UdmSQLValue(SQLRes,i,3)));
}


static udm_rc_t
UdmTargetsSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  char    sortstr[128]= "";
  char    updstr[64]="";
  char    tblhint[64]="";
  UDM_SQL_TOP_CLAUSE Top;
  size_t    i = 0, j, start, nrows, qbuflen;
  UDM_SQLRES   SQLRes;
  char    smallbuf[128];
  udm_rc_t rc= UDM_OK;
  const char  *where;
  char    *qbuf=NULL;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  udm_bool_t skip_lock= UdmVarListFindBool(&Indexer->Conf->Vars, "URLSelectSkipLock", UDM_FALSE);
  size_t url_num= UdmVarListFindInt(&Indexer->Conf->Vars, "URLSelectCacheSize", URL_SELECT_CACHE);

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_TARGETS);
  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DB);

  if (UdmSQLDB(db)->last_notargets_time == time(0))
  {
    /*
      Do not query the database again if it was already queried
      in the same second and returned 0 targets.
    */
    return UDM_OK;
  }

  if (Indexer->Conf->url_number < (int) url_num)
    url_num= Indexer->Conf->url_number;
  if (UDM_OK != (rc= UdmSQLBuildWhereCondition(Indexer, db, Query, &where)))
    return rc;
  qbuflen= 1024 + 4 * strlen(where);
  
  if ((qbuf = (char*)UdmMalloc(qbuflen + 2)) == NULL)
  {
      UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
      return UDM_ERROR;
  }

  if ((Indexer->flags & (UDM_FLAG_SORT_HOPS | UDM_FLAG_SORT_EXPIRED)) ||
      !(Indexer->flags & UDM_FLAG_DONTSORT_SEED))
  {    
    sprintf(sortstr, " ORDER BY %s%s%s", 
      (Indexer->flags & UDM_FLAG_SORT_HOPS) ? "hops" : "",
      (Indexer->flags & UDM_FLAG_DONTSORT_SEED) ? "" : ((Indexer->flags & UDM_FLAG_SORT_HOPS) ? ",seed" : "seed"),
      (Indexer->flags & UDM_FLAG_SORT_EXPIRED) ? 
      ( ((Indexer->flags & UDM_FLAG_SORT_HOPS) || !(Indexer->flags & UDM_FLAG_DONTSORT_SEED)  ) ? 
        ",next_index_time" : "next_index_time") : "");
  }

  UdmDBSQLTopClause(Indexer, db, url_num, &Top);

  if(1)
  {
    switch(UdmSQLDBType(db))
    {
      case UDM_DB_MYSQL:
        udm_snprintf(qbuf, qbuflen,
                     "INSERT INTO udm_url_tmp "
                     "SELECT url.rec_id FROM url%s "
                     "WHERE next_index_time<=%d %s%s%s%s",
                      Query->from,
                      (int)time(NULL), where[0] ? "AND " : "",  where,
                      sortstr, Top.limit);
        if (UDM_OK != (rc= UdmDBSQLDropTableIfExists(Indexer, db, "udm_url_tmp")) ||
            UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, "CREATE TEMPORARY TABLE udm_url_tmp (rec_id int not null) ENGINE=MyISAM")) ||
            (!skip_lock &&
             UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, "LOCK TABLES udm_url_tmp WRITE, url WRITE, urlinfo AS it WRITE, urlinfo AS il WRITE, server AS s WRITE"))) ||
            UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
          return rc;
        break;
      case UDM_DB_PGSQL:
        rc= UdmDBSQLQuery(Indexer, db, NULL,"BEGIN WORK");
        sprintf(updstr, " FOR UPDATE ");
/*        rc=UdmDBSQLQuery(Indexer, db,NULL,"LOCK url");*/
        break;
      case UDM_DB_ORACLE8:
        sprintf(updstr, " FOR UPDATE ");
        break;
      case UDM_DB_MSSQL:
        strcpy(tblhint, " (TABLOCKX)");
        rc= UdmDBSQLBegin(Indexer, db);
        break;
      case UDM_DB_SAPDB:
        sprintf(updstr, " WITH LOCK ");
        break;
      default:
        break;
    }
    if (rc != UDM_OK)
      goto ex;
  }
  
  UdmSQL(db)->res_limit= url_num;
  if (UdmSQLDBType(db) == UDM_DB_MYSQL)
    udm_snprintf(qbuf, qbuflen, "SELECT %s FROM url, udm_url_tmp "
                                "WHERE url.rec_id=udm_url_tmp.rec_id",
                                select_url_str);
  else
    udm_snprintf(qbuf, qbuflen, "SELECT %s%s "
                                "FROM url%s%s "
                                "WHERE next_index_time<=%d %s%s%s"
                                "%s%s%s",
                 Top.top, select_url_str, tblhint, Query->from,
                 (int)time(NULL), where[0] ? "AND " : "",  where, Top.rownum,
                 sortstr, updstr, Top.limit);
  
  if(UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, qbuf)))
    goto commit;
  
  if(!(nrows = UdmSQLNumRows(&SQLRes)))
  {
    UdmSQLFree(&SQLRes);
    UdmSQLDB(db)->last_notargets_time= time(0);
    goto commit;
  }

  start = Indexer->Conf->Targets.num_rows;
  Indexer->Conf->Targets.num_rows += nrows;
  
  Indexer->Conf->Targets.Doc = 
    (UDM_DOCUMENT*)UdmRealloc(Indexer->Conf->Targets.Doc, sizeof(UDM_DOCUMENT)*(Indexer->Conf->Targets.num_rows + 1));
  if (Indexer->Conf->Targets.Doc == NULL)
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory at realloc %s[%d]", __FILE__, __LINE__);
    rc= UDM_ERROR;
    goto commit;
  }
  
  for(i = 0; i < nrows; i++)
  {
    UDM_DOCUMENT  *Doc = &Indexer->Conf->Targets.Doc[start + i];
    UdmDocInit(Doc);
    UdmTargetSQLResToDoc(Indexer, Doc, &SQLRes, i);
    UdmTargetSQLResToDoc_Extra(Indexer, Doc, &SQLRes, i);
  }
  UdmSQLFree(&SQLRes);
  
  
  if (UdmSQLDBHaveIn(db))
  {
    char  *urlin=NULL;
    
    if ( (qbuf = (char*)UdmRealloc(qbuf, qbuflen = qbuflen + 35 * URL_SELECT_CACHE)) == NULL)
    {
      UDM_FREE(qbuf);
      UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
      rc= UDM_ERROR;
      goto commit;
    }
    
    if ( (urlin = (char*)UdmMalloc(35 * URL_SELECT_CACHE)) == NULL)
    {
      UDM_FREE(qbuf);
      UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
      rc = UDM_ERROR;
      goto commit;
    }
    urlin[0]=0;
    
    for(i = 0; i < nrows; i+= URL_SELECT_CACHE)
    {

      urlin[0] = 0;

      for (j = 0; (j < URL_SELECT_CACHE) && (i + j < nrows) ; j++)
      {

      UDM_DOCUMENT  *Doc = &Indexer->Conf->Targets.Doc[start + i + j];
      urlid_t    url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
      
      if(urlin[0])strcat(urlin,",");
      sprintf(urlin+strlen(urlin), "%s%i%s", qu, url_id, qu);
      }
      udm_snprintf(qbuf, qbuflen, "UPDATE url SET next_index_time=%d WHERE rec_id in (%s)",
             (int)(time(NULL) + URL_LOCK_TIME), urlin);
      if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
        goto commit;
    }
    UDM_FREE(urlin);
  }
  else
  {
    for(i = 0; i < nrows; i++)
    {
      UDM_DOCUMENT  *Doc = &Indexer->Conf->Targets.Doc[start + i];
      urlid_t    url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
      
      udm_snprintf(smallbuf, 128, "UPDATE url SET next_index_time=%d WHERE rec_id=%i",
             (int)(time(NULL) + URL_LOCK_TIME), url_id);
      if(UDM_OK!=(rc=UdmDBSQLQuery(Indexer, db, NULL, smallbuf)))
        goto commit;
    }
  }


commit:

  if (rc != UDM_OK)
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "UdmTargetsSQL: DB error: %s", UdmDBSQLError(db));
  }
  if(1)
  {
    switch(UdmSQLDBType(db))
    {
      case UDM_DB_MYSQL:
        if (!skip_lock)
          rc= UdmDBSQLQuery(Indexer, db, NULL, "UNLOCK TABLES");
        break;
      case UDM_DB_PGSQL:
        rc=UdmDBSQLQuery(Indexer, db, NULL, "END WORK");
        break;
      case UDM_DB_MSSQL:
        rc= UdmDBSQLCommit(Indexer, db);
      default:
        break;
    }
  }
ex:
  UDM_FREE(qbuf);
  return rc;
}



/******************* Truncate database ********************/

static udm_rc_t
UdmTruncateURL(UDM_AGENT *Indexer,UDM_DB *db)
{
  udm_rc_t rc;
  
  rc= UdmDBSQLTableTruncateOrDelete(Indexer, db, "url");
  if(rc!=UDM_OK)return rc;
  
  rc= UdmDBSQLTableTruncateOrDelete(Indexer, db, "redirect");
  if(rc != UDM_OK) return rc;

  rc= UdmDBSQLTableTruncateOrDelete(Indexer, db, "links");
  if(rc != UDM_OK) return rc;

  rc= UdmDBSQLTableTruncateOrDelete(Indexer, db, "cachedcopy");
  if(rc != UDM_OK) return rc;
  
  rc= UdmDBSQLTableTruncateOrDelete(Indexer, db, "urlinfo");
  return rc;
}


static udm_rc_t
UdmTruncateDict(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  UDM_ASSERT(UdmSQLDBModeHandler(db)->QueryAction != NULL);
  return UdmSQLDBModeHandler(db)->QueryAction(Indexer, db, Query, UDM_QUERYCMD_CLEAR);
}


static udm_rc_t
UdmTruncateDB(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  if((UDM_OK != (rc= UdmTruncateDict(Indexer, db, Query))) ||
     (UDM_OK != (rc= UdmTruncateURL(Indexer, db))))
    return rc;
  return UDM_OK;
}


/******************* Clear database with condition ********/

static udm_rc_t
UdmDeleteWordsAndLinks(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_DB *db)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDeleteWordFromURL(Indexer,Doc,db)))
    return rc;

  if (Doc->Spider.collect_links_destination)
  {
    int status= UdmVarListFindInt(&Doc->Sections, "Status", 0);
    /* In case of redirect, we store links from the "Location" header */
    if (UDM_OK != (rc= UdmDeleteLinks(Indexer, db, Doc)))
      return rc;

    if (UDM_OK != (rc= (status >= 300 && status <= 303) ?
                   UdmStoreRedirects(Indexer, db, Doc) :
                   UdmDeleteRedirects(Indexer, db, Doc)))
      return rc;
  }

  /* Set status, bad_since_time, etc */
  if (UDM_OK != (rc= UdmUpdateUrl(Indexer, Doc, db)))
    return rc;

  return rc;
}


static udm_rc_t
UdmDeleteWordFromURL(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_DB *db)
{
  urlid_t url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);

  if (!UdmVarListFindInt(&Doc->Sections, "PrevStatus", 0))
    return UDM_OK;

  UDM_ASSERT(UdmSQLDBModeHandler(db)->DeleteWordsFromURL != NULL);
  return UdmSQLDBModeHandler(db)->DeleteWordsFromURL(Indexer, db, url_id);
}


static udm_rc_t
UdmDeleteBadHrefs(UDM_AGENT *Indexer,
                  UDM_DOCUMENT *Doc,
                  UDM_DB *db,
                  urlid_t url_id)
{
  UDM_DOCUMENT  rDoc;
  UDM_SQLRES  SQLRes;
  char    q[256];
  size_t    i;
  size_t    nrows;
  udm_rc_t rc= UDM_OK;
  int    hold_period= UdmVarListFindInt(&Doc->Sections,"HoldBadHrefs",0);
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  
  if (hold_period <= 0)
    return UDM_OK;
  
  udm_snprintf(q, sizeof(q), "SELECT rec_id FROM url WHERE status > 300 AND status<>304 AND referrer=%s%i%s AND bad_since_time<%d",
    qu, url_id, qu, (int)time(NULL) - hold_period);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, q)))
    return rc;
  
  nrows = UdmSQLNumRows(&SQLRes);
  
  UdmDocInit(&rDoc);
  for(i = 0; i < nrows ; i++)
  {
    UdmVarListReplaceStr(&rDoc.Sections,"ID", UdmSQLValue(&SQLRes,i,0));
    if(UDM_OK!=(rc=UdmDeleteURL(Indexer, &rDoc, db)))
      break;
  }
  UdmDocFree(&rDoc);
  UdmSQLFree(&SQLRes);
  return rc;
}


static udm_rc_t
UdmDeleteURL(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db)
{
  char  qbuf[128];
  udm_rc_t rc;
  urlid_t  url_id  =UdmVarListFindInt(&Doc->Sections, "ID", 0);
  const char *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";

  if(UDM_OK!=(rc=UdmDeleteWordFromURL(Indexer,Doc,db)))return(rc);
  
  sprintf(qbuf,"DELETE FROM url WHERE rec_id=%s%i%s", qu, url_id, qu);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
    return rc;
  
  sprintf(qbuf,"DELETE FROM urlinfo WHERE url_id=%s%i%s", qu, url_id, qu);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
    return rc;
  
  sprintf(qbuf,"DELETE FROM cachedcopy WHERE url_id=%s%i%s", qu, url_id, qu);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
    return rc;

  sprintf(qbuf,"DELETE FROM redirect WHERE url_id=%s%i%s", qu, url_id, qu);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
    return rc;

  sprintf(qbuf,"DELETE FROM links WHERE url_id=%s%i%s", qu, url_id, qu);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
    return rc;

  /* remove all old broken hrefs from this document to avoid broken link collecting */
  if (UDM_OK != (rc= UdmDeleteBadHrefs(Indexer, Doc, db, url_id)))
    return rc;

  sprintf(qbuf,"UPDATE url SET referrer=%s0%s WHERE referrer=%s%i%s", qu, qu, qu, url_id, qu);
  return UdmDBSQLQuery(Indexer, db, NULL, qbuf);
}


static udm_rc_t
UdmClearDBUsingIN(UDM_AGENT *Indexer, UDM_DB *db, UDM_URLID_LIST *list)
{
  UDM_DSTR qbuf, urlin;
  udm_rc_t rc= UDM_OK; /* if list if empty */
  size_t part;
  size_t url_num = UdmVarListFindInt(&Indexer->Conf->Vars, "URLSelectCacheSize", URL_DELETE_CACHE);
  
  UdmDSTRInit(&qbuf, 4096);
  UdmDSTRInit(&urlin, 4096);
  
  for (part= 0; part < list->nurls; part+= url_num)
  {
    size_t offs;
    urlid_t *item= &list->urls[part];
    UdmDSTRReset(&urlin);
    for(offs= 0; (offs < url_num) && ((part + offs) < list->nurls); offs++)
    {
      if(offs) UdmDSTRAppend(&urlin,",", 1);
      UdmDSTRAppendf(&urlin, "%d", item[offs]);
    }
    
    if (UDM_OK != (rc= UdmDBSQLBegin(Indexer, db)))
      goto ret;
    
    switch (UdmSQLDBMode(db))
    {
      case UDM_SQLDBMODE_BLOB:
        break;
      
      case UDM_SQLDBMODE_MULTI:
        {
          int dictnum;
          for (dictnum= 0; dictnum <= MULTI_DICTS; dictnum++)
          {
            UdmDSTRReset(&qbuf);
            UdmDSTRAppendf(&qbuf,"DELETE FROM dict%02X WHERE url_id in (%s)",
                           dictnum, UdmDSTRPtr(&urlin));
            if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
              goto ret;
          }
        }
        break;
        
      case UDM_SQLDBMODE_SINGLE:
        UdmDSTRReset(&qbuf);
        UdmDSTRAppendf(&qbuf, "DELETE FROM dict WHERE url_id in (%s)",
                       UdmDSTRPtr(&urlin));
        if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
          goto ret;
        break;
      case UDM_SQLDBMODE_RAWBLOB:
        udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db), "ClearDB is not supported by this DBMode");
        rc= UDM_ERROR;
    }
    
    UdmDSTRReset(&qbuf);
    UdmDSTRAppendf(&qbuf, "DELETE FROM url WHERE rec_id in (%s)",UdmDSTRPtr(&urlin));
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
      goto ret;
    
    UdmDSTRReset(&qbuf);
    UdmDSTRAppendf(&qbuf,"DELETE FROM urlinfo WHERE url_id in (%s)",UdmDSTRPtr(&urlin));
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
      goto ret;
    
    UdmDSTRReset(&qbuf);
    UdmDSTRAppendf(&qbuf,"DELETE FROM cachedcopy WHERE url_id in (%s)",UdmDSTRPtr(&urlin));
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
      goto ret;
    
    UdmDSTRReset(&qbuf);
    UdmDSTRAppendf(&qbuf,"DELETE FROM links WHERE url_id in (%s)",UdmDSTRPtr(&urlin));
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
      goto ret;
    
    UdmDSTRReset(&qbuf);
    UdmDSTRAppendf(&qbuf,"DELETE FROM redirect WHERE url_id in (%s)",UdmDSTRPtr(&urlin));
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
      goto ret;
    
    if (UDM_OK != (rc= UdmDBSQLCommit(Indexer, db)))
      goto ret;
  }
  
ret:
  UdmDSTRFree(&qbuf);
  UdmDSTRFree(&urlin);
  return rc;
}


static udm_rc_t
UdmClearDBUsingLoop(UDM_AGENT *Indexer, UDM_DB *db, UDM_URLID_LIST *list)
{
  udm_rc_t rc= UDM_OK;
  size_t i;
  UDM_DOCUMENT Doc;
  bzero((void*)&Doc, sizeof(Doc));

  for(i=0; i < list->nurls; i++)
  {
    UdmVarListReplaceInt(&Doc.Sections, "ID", list->urls[i]);
    if(UDM_OK != (rc= UdmDeleteURL(Indexer, &Doc, db)))
      break;
  }
  UdmDocFree(&Doc);
  return rc;
}


static udm_rc_t
UdmClearDBSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  const char *where, *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";
  char ClearDBHook[128];

  rc= UdmSQLBuildWhereCondition(Indexer, db, Query, &where);
  udm_snprintf(ClearDBHook, sizeof(ClearDBHook), "%s",
               UdmVarListFindStr(&Indexer->Conf->Vars, "SQLClearDBHook", ""));
  
  if (rc != UDM_OK ||
      (ClearDBHook[0] && (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, ClearDBHook)))))
    return rc;
  
  if(!where[0])
  {
    return UdmTruncateDB(Indexer, db, Query);
  }
  else
  {
    UDM_URLID_LIST urllist;
    UDM_DSTR qbuf;
    UdmDSTRInit(&qbuf, 4096);
    
    bzero((void*) &urllist, sizeof(urllist));
    UdmDSTRAppendf(&qbuf,"SELECT url.rec_id, url.url FROM url%s WHERE url.rec_id<>%s0%s AND %s", 
                   Query->from, qu, qu,  where);
    
    if (UDM_OK != (rc= UdmLoadSlowLimitWithSort(Indexer, db, &urllist, UdmDSTRPtr(&qbuf))))
      goto fin;
    
    rc= UdmSQLDBHaveIn(db) ? UdmClearDBUsingIN(Indexer, db, &urllist) :
                       UdmClearDBUsingLoop(Indexer, db, &urllist);

fin:
    UdmFree(urllist.urls);
    UdmDSTRFree(&qbuf);
  }
  return rc;
}


/******************** Hrefs ****************************/

static udm_bool_t
UdmStoreHrefsUseTnx(UDM_AGENT *A)
{
  if (A->Conf->DBList.nitems > 1)
    return UDM_FALSE;
  if (UdmSQLDBType(&A->Conf->DBList.Item[0]) == UDM_DB_PGSQL &&
      UdmSQLDBVersion(&A->Conf->DBList.Item[0]) > 90100)
    return UDM_TRUE;
  return (UdmSQLDBFlags(&A->Conf->DBList.Item[0]) & UDM_SQL_HAVE_GOOD_COMMIT) ?
         UDM_TRUE : UDM_FALSE;
}


static udm_rc_t
UdmStoreHrefsLock(UDM_AGENT *A, UDM_DB *db)
{
  if (UdmSQLDBType(db) == UDM_DB_MYSQL /*&&
      !UdmVarListFind(&Doc->Sections, "SQLExportHref")*/)
    return UdmDBSQLQuery(A, db, NULL, "LOCK TABLE url WRITE");
  return UDM_OK;
}


static udm_rc_t
UdmStoreHrefsUnlock(UDM_AGENT *A, UDM_DB *db)
{
  if (UdmSQLDBType(db) == UDM_DB_MYSQL)
    return UdmDBSQLQuery(A, db, NULL, "UNLOCK TABLES");
  return UDM_OK;
}


static udm_rc_t
UdmHrefStartBulk(UDM_AGENT *A, UDM_DB *db)
{
  return UdmStoreHrefsUseTnx(A) ? UdmDBSQLBegin(A, db) :
                                  UdmStoreHrefsLock(A, db);
}


static udm_rc_t
UdmHrefStopBulk(UDM_AGENT *A, UDM_DB *db)
{
  return UdmStoreHrefsUseTnx(A) ? UdmDBSQLCommit(A, db) :
                                  UdmStoreHrefsUnlock(A, db);
}


static void
UdmHrefToVarList(UDM_VARLIST *Vars, const UDM_HREF *H)
{
  UdmVarListAddLst(Vars, &H->HrefVars, NULL, "*");
  UdmVarListReplaceInt(Vars, "Referrer-ID", H->Param.referrer);
  UdmVarListReplaceUnsigned(Vars,"Hops", H->Param.hops);
  UdmVarListReplaceStr(Vars,"URL",H->url?H->url:"");
  UdmVarListReplaceInt(Vars, "URL_ID", UdmStrHash32(H->url ? H->url : ""));
  UdmVarListReplaceInt(Vars,"Server_id", H->Param.server_id);
  UdmVarListReplaceInt(Vars, "HTDB_URL_ID", H->Param.rec_id);
}


static udm_rc_t
UdmHrefAdd(UDM_AGENT *A, UDM_DB *db, UDM_HREF *H)
{
  udm_rc_t rc;
  UDM_DOCUMENT Doc;
  UdmDocInit(&Doc);
  UdmHrefToVarList(&Doc.Sections, H);
  rc= UdmAddURL(A, &Doc, db);
  UdmDocFree(&Doc);
  UdmSQLDB(db)->last_notargets_time= 0;
  return rc;
}


static udm_rc_t
UdmHrefExport(UDM_AGENT *A, UDM_DB *db, UDM_HREF *H)
{
  udm_rc_t rc;
  UDM_DOCUMENT Doc;
  const char *sql_export= UdmVarListFindStr(&H->HrefVars, "SQLExportHref", NULL);
  if (!sql_export)
    return UDM_OK;
  UdmDocInit(&Doc);
  UdmHrefToVarList(&Doc.Sections, H);
  rc= UdmExportURL(A, db, &Doc, sql_export);
  UdmDocFree(&Doc);
  UdmSQLDB(db)->last_notargets_time= 0;
  return rc;
}


static udm_rc_t
UdmHrefActionSQL(UDM_AGENT *A, UDM_DB *db, UDM_HREF *Href, udm_hrefcmd_t cmd)
{
  UDM_LOCK_CHECK_OWNER(A, UDM_LOCK_DB);
  switch (cmd)
  {
    case UDM_HREFCMD_START_BULK: return UdmHrefStartBulk(A, db);
    case UDM_HREFCMD_STOP_BULK:  return UdmHrefStopBulk(A, db);
    case UDM_HREFCMD_ADD:        return UdmHrefAdd(A, db, Href);
    case UDM_HREFCMD_EXPORT:     return UdmHrefExport(A, db, Href);
  }
  return UDM_ERROR;
}


/******************* WWList *************************/
static udm_rc_t
UdmWWListExtend(UDM_AGENT *A, UDM_DB *db,
                UDM_WIDEWORDLIST *result,
                const UDM_WIDEWORD *uw,
                const char *sql)
{
  char *word= uw->Word.str;
  char buf[1024];
  size_t i, nrows;
  UDM_SQLRES SQLRes;
  UDM_WIDEWORD_PARAM Param;
  UdmBuildParamStr(buf, sizeof(buf), sql, &word, 1);
  if (UDM_OK != UdmDBSQLQuery(A, db, &SQLRes, buf))
    return UDM_ERROR;
  nrows= UdmSQLNumRows(&SQLRes);
  UdmWideWordParamInit(&Param);
  UdmWideWordParamCopySynonym(&Param, &uw->Param);
  for (i= 0; i < nrows; i++)
  {
    UDM_CONST_STR cstr;
    cstr.length= UdmSQLLen(&SQLRes, i, 0);
    cstr.str= UdmSQLValue(&SQLRes, i, 0);
    UdmWideWordListAddLikeConstStr(result, &Param, cstr);
  }
  UdmSQLFree(&SQLRes);
  return UDM_OK;
}


static udm_rc_t
UdmQueryWordForms(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  UDM_WIDEWORDLIST Tmp;
  udm_rc_t rc= UDM_OK;
  size_t i;
  const char *sql= UdmVarListFindStr(&A->Conf->Vars, "SQLWordForms", NULL);
  if (!sql)
    return UDM_OK;
  UdmWideWordListInit(&Tmp);
  for (i= 0; i < Query->Res.WWList.nwords; i++)
  {
    if (UDM_OK != (rc= UdmWWListExtend(A, db, &Tmp, &Query->Res.WWList.Word[i], sql)))
      break;
  }
  for (i= 0; i < Tmp.nwords; i++)
  {
    /* TODO34: change UdmWideWordListAddXXX() to return udm_rc_t */
    UdmWideWordListAdd(&Query->Res.WWList, &Tmp.Word[i]);
  }
  UdmWideWordListFree(&Tmp);
  return rc;
}


/******************* Search *************************/
static int
cmp_score_urlid(UDM_URL_SCORE *s1, UDM_URL_SCORE *s2)
{
  if (s1->url_id > s2->url_id) return(1);
  if (s1->url_id < s2->url_id) return(-1);
  return 0;
}
      

static void
UdmScoreListToURLData(UDM_URLDATA *D, UDM_URL_SCORE *C, size_t num)
{
  for ( ; num > 0; num--, D++, C++)
  {
    D->url_id= C->url_id;
    D->score= C->score;
  }
}


#ifdef HAVE_DEBUG
static void
UdmURLScoreListPrint(UDM_URLSCORELIST *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
  {
    UDM_URL_SCORE *Item= &List->Item[i];
    fprintf(stderr, "%d:%d\n", Item->url_id, Item->score);
  }
}
#endif


static void
UdmDebugScoreAppendScoreAndRank(UDM_VARLIST *Vars,
                                const UDM_QUERY_PARAM *query_param,
                                const UDM_URLDATALIST *DataList)
{
  UDM_VAR *var;
  if (query_param->DebugURLId &&
      (var= UdmVarListFindVar(Vars, "DebugScore")))
  {
    size_t i;
    for (i= 0; i < DataList->nitems; i++)
    {
      if (DataList->Item[i].url_id == query_param->DebugURLId)
      {
        char tmp[256];
        size_t length= udm_snprintf(tmp, sizeof(tmp), " rank=%d", (int) i + 1);
        UdmVarAppendStrn(var, tmp, length);
        break;
      }
    }
  }
}


static udm_rc_t 
UdmSortAndGroupByURL(UDM_AGENT *A,
                     UDM_QUERY *Query,
                     UDM_SEARCHSECTIONLIST *SectionList,
                     UDM_DB *db)
{
  UDM_QUERY_PARAM query_param;
  UDM_URLSCORELIST ScoreList;
  UDM_URLDATALIST DataList;
  udm_timer_t ticks=UdmStartTimer();
  const char *pattern= UdmVarListFindStr(&A->Conf->Vars, "s", "R");
  size_t nbytes;
  int flags= 0, flags2= 0;
  udm_rc_t rc= UDM_OK;
  const char *p;
  const char *su= UdmVarListFindStr(&A->Conf->Vars, "su", NULL);
  int group_by_site= UdmVarListFindBool(&A->Conf->Vars, "GroupBySite", UDM_FALSE) 
                     && UdmVarListFindStr(&A->Conf->Vars, "site", "")[0] == '\0' ?
                     UDM_URLDATA_SITE : 0;
  int group_by_site_rank= !strcmp(UdmVarListFindStr(&A->Conf->Vars, "GroupBySite", "no"), "rank");
  size_t BdictThreshold= (size_t) UdmVarListFindInt(&A->Conf->Vars,
                                                    "URLDataThreshold", 0);
  size_t MaxResults= (size_t) UdmVarListFindInt(UdmSQLDBVars(db), "MaxResults", 0);
  udm_bool_t use_qcache= UdmVarListFindBool(UdmSQLDBVars(db), "qcache", UDM_FALSE);
  size_t num_best_rows= Query->num_best_rows;

  UdmQueryParamInit(&query_param, A->Conf, UdmSQLDBVars(db));

  flags|= group_by_site ? UDM_URLDATA_SITE : 0;
  flags|= group_by_site_rank ? UDM_URLDATA_SITE_RANK : 0;
  flags|= query_param.DateFactor ? UDM_URLDATA_LM : 0;
  flags2|= query_param.PopularityFactor > 0 ? UDM_URLDATA_POP : 0;

  for (p = pattern; *p; p++)
  {
    if (*p == 'U' || *p == 'u') flags|= UDM_URLDATA_URL;
    if (*p == 'D' || *p == 'd') flags|= UDM_URLDATA_LM;
    if (*p == 'S' || *p == 's') flags|= (su && su[0]) ? UDM_URLDATA_SU : 0;
    if (*p == 'P' || *p == 'p') flags2|= UDM_URLDATA_POP;
  }

  ticks=UdmStartTimer();
  bzero((void*) &ScoreList, sizeof(ScoreList));
  UdmURLDataListInit(&DataList);

  UdmLog(A,UDM_LOG_DEBUG, "Start GroupByURL %d sections", (int) SectionList->nsections);
  UdmGroupByURL2(A, db, Query, &query_param, SectionList, &ScoreList);

  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f (%d docs found)",
         "Stop  GroupByURL", UdmStopTimer(&ticks), (int) ScoreList.nitems);

#ifdef HAVE_DEBUG
  if (UdmVarListFindBool(&A->Conf->Vars, "DebugGroupByURL", UDM_FALSE))
  {
    UdmURLScoreListPrint(&ScoreList);
  }
#endif

  UdmApplyCachedQueryLimit(A, &ScoreList, db);
  if (ScoreList.nitems == 0)
    goto ex;
  
  if (UDM_OK != (rc=  UdmUserScoreListLoadAndApplyToURLScoreList(A, &ScoreList,
                                                                 db,
                                                                 &query_param)))
    goto ex;

  UdmLog(A,UDM_LOG_DEBUG,"Start load url data %d docs (%d best needed)",
         (int) ScoreList.nitems, (int) Query->num_best_rows);
  ticks=UdmStartTimer();

  nbytes= UdmHashSize(ScoreList.nitems) * sizeof(UDM_URLDATA);
  DataList.Item = (UDM_URLDATA*)UdmMalloc(nbytes);
  bzero((void*) DataList.Item, nbytes);
  DataList.nitems= ScoreList.nitems;

  /* Use full sort in case if DebugURLId is specified */
  if (query_param.DebugURLId)
    num_best_rows= ScoreList.nitems;
  
  if (num_best_rows > ScoreList.nitems)
    num_best_rows= ScoreList.nitems;
  
  /* Try fast sorting if sorting is on score */
  if (num_best_rows < 256 && !flags && !flags2 && !use_qcache)
  {
    udm_timer_t ticks1;
    
    Query->stats.total_found= ScoreList.nitems;
    UdmLog(A, UDM_LOG_DEBUG, "Start SortByScore %d docs", (int) ScoreList.nitems);
    ticks1=UdmStartTimer();
    if (ScoreList.nitems > 1000)
    {
      UdmURLScoreListSortByScoreThenURLTop(&ScoreList, 1000);
    }
    else
    {
      UdmURLScoreListSortByScoreThenURL(&ScoreList);
    }
    UdmSort((void*) ScoreList.Item, num_best_rows,
            sizeof(UDM_URL_SCORE), (udm_qsort_cmp) cmp_score_urlid);
    UdmScoreListToURLData(DataList.Item, ScoreList.Item, num_best_rows);
    UdmLog(A,UDM_LOG_DEBUG,"%-30s%.2f", "Stop  SortByScore:", UdmStopTimer(&ticks1));
    DataList.nitems= num_best_rows; /* Put only num_best_rows into DataList */
    goto date_factor;
  }

  UdmScoreListToURLData(DataList.Item, ScoreList.Item, DataList.nitems);
  
  /* Sort by a user defined section, if given */
  if (flags & UDM_URLDATA_SU)
  {
    size_t norder;
    udm_timer_t ticks1= UdmStartTimer();
    UdmLog(A, UDM_LOG_DEBUG, "Trying to load fast section order '%s'", su);
    rc= UdmFastOrderLoadAndApplyToURLDataList(A, db, &DataList, su, &norder);
    UdmLog(A, UDM_LOG_DEBUG, "Loading fast order '%s' done, %d docs found, %.2f sec",
           su, (int) norder, UdmStopTimer(&ticks1));
    if (norder)
      flags^= UDM_URLDATA_SU;
  }

  if (flags)
  {
    if (!(UdmSQLDBMode(db) == UDM_SQLDBMODE_BLOB &&
         !(flags & UDM_URLDATA_URL)    &&
         !(flags & UDM_URLDATA_SU)     &&
           BdictThreshold < ScoreList.nitems) ||
        (UDM_NOTARGET == UdmLoadURLDataFromBdict(A, db, &DataList, flags)))
      rc= UdmLoadURLDataFromURL(A, db, &DataList, flags);
  }

  /* Do this before GroupBySite, while DataList is sorted by url_id */
  UdmLoadURLDataFromBdict(A, db, &DataList, flags2);

  if (group_by_site)
    UdmURLDataListGroupBySiteUsingSort(A, &DataList, db);

  if (UDM_OK != UdmUserSiteScoreListLoadAndApplyToURLDataList(A, &DataList, db,
                                                              &query_param))
  {
    rc= UDM_ERROR;
    goto ex;
  }

  if (flags & UDM_URLDATA_SITE_RANK)
  {
    udm_timer_t ticks1= UdmStartTimer();
    UdmLog(A, UDM_LOG_DEBUG, "Start applying in-site-rank");
    UdmURLDataSortBySite(&DataList);
    UdmURLDataApplySiteRank(A, &DataList, 0);
    UdmLog(A, UDM_LOG_DEBUG, "Stop applying in-site-rank:   %.2f sec", UdmStopTimer(&ticks1));
  }

  Query->stats.total_found= DataList.nitems;

date_factor:
  
  if (rc != UDM_OK)
    goto ex;

  /* TODO: check whether limit by site works fine */
  if (!query_param.RelevancyFactor || query_param.DateFactor)
    UdmURLDataListApplyRelevancyFactors(A, &DataList, &query_param);

  if (query_param.PopularityFactor > 0)
    UdmURLDataListApplyPopularity(A, &DataList, &query_param);

  UdmLog(A,UDM_LOG_DEBUG,"%-30s%.2f", "Stop  load url data:", UdmStopTimer(&ticks));
  
  UdmLog(A, UDM_LOG_DEBUG, "Start SortByPattern %d docs", (int) DataList.nitems);
  ticks=UdmStartTimer();
  if (DataList.nitems)
    UdmURLDataSortByPattern(&DataList, pattern);
  UdmLog(A,UDM_LOG_DEBUG,"%-30s%.2f", "Stop  SortByPattern:", UdmStopTimer(&ticks));

  Query->URLData= DataList;
  bzero((void *) &DataList, sizeof(DataList));
  UdmDebugScoreAppendScoreAndRank(&A->Conf->Vars, &query_param, &Query->URLData);

  if (MaxResults && MaxResults < Query->stats.total_found)
  {
    UdmLog(A, UDM_LOG_DEBUG, "Applying MaxResults=%d, total_found=%d\n",
           (int) MaxResults, (int) Query->stats.total_found);
    Query->stats.total_found= MaxResults;
    if (Query->URLData.nitems > MaxResults)
    {
      /* Free the part of URLData that will not be unused */
      UdmURLDataListFreeItems(&Query->URLData, MaxResults, Query->URLData.nitems);
      Query->URLData.nitems= MaxResults;
    }
  }

ex:
  UdmURLDataListFree(&DataList);
  UdmFree(ScoreList.Item);
  return rc;
}


static udm_rc_t /* WHERE limit */
LoadURL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
        const char *where, UDM_URLID_LIST *buf)
{
  udm_rc_t rc;
  UDM_SQLRES SQLRes;
  char qbuf[1024 * 4];
  size_t nrows;
  urlid_t *tmp;
  size_t i;

  if (!*where)
    return UDM_OK;

  /* TODO: reuse LoadSlowLimitWithSort() here */
  udm_snprintf(qbuf, sizeof(qbuf),
               "SELECT url.rec_id FROM url%s WHERE %s",
               Query->from, where);
  if  (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, qbuf)))
    return rc;

  if (!(nrows= UdmSQLNumRows(&SQLRes)))
  {
    buf->empty= 1;
    UdmSQLFree(&SQLRes);
    return(UDM_OK);
  }

  tmp= (urlid_t*) UdmMalloc(sizeof(urlid_t) * nrows);
  buf->urls= (urlid_t*) UdmMalloc(sizeof(urlid_t) * nrows);
  if (!tmp || !buf->urls)
  {
    UDM_FREE(buf->urls);
    UDM_FREE(tmp);
    goto ex;
  }
  
  for (i= 0; i < nrows; i++)
  {
    tmp[i]= (urlid_t) UDM_ATOI(UdmSQLValue(&SQLRes, i, 0));
  }
  UdmSort(tmp, nrows, sizeof(urlid_t), (udm_qsort_cmp)UdmCmpURLID);
  
  /* Remove duplicates */
  for (i= 0; i < nrows; )
  {
    while (++i < nrows && tmp[i] == tmp[i - 1]);
    buf->urls[buf->nurls++] = tmp[i - 1];
  }
  UDM_FREE(tmp);
  if ((tmp= (urlid_t*) UdmRealloc(buf->urls, sizeof(urlid_t) * buf->nurls)))
    buf->urls = tmp;

ex:
  UdmSQLFree(&SQLRes);
  return UDM_OK;
}


/*
  MySQL: no cast needed
  - SQLite: CAST(word AS INTEGER)
  - PostgreSQL: CASE WHEN a~'^[0-9]*$' THEN a::integer ELSE 0 END
  - MSSQL:    CAST(word AS INTEGER)
    There is a function ISNUMERIC(). However, it returns true
    for things like "0x0123d", Cast does not work for this.

  - Sybase:   returns error when input is non-numeric
  - Oracle:   returns error ...
  - IBM DB2:  returns error ...
  - Firebird: return error ...
  - Mimer:    return error ...
*/
static void
UdmBuildNumericOperatorCondition(UDM_DB *db, char *cmparg, size_t maxlen,
                                 const char *op, int number)
{
  switch (UdmSQLDBType(db))
  {
    case UDM_DB_MYSQL:
      udm_snprintf(cmparg, maxlen, "word%s%d", op, number);
      break;
    case UDM_DB_PGSQL:
      udm_snprintf(cmparg, maxlen, "(word~'^[0-9]*$' AND word::integer%s%d)", op, number);
      break;
    default:
      udm_snprintf(cmparg, maxlen, "(word>='0' AND word <='99999999999' AND CAST(word AS INTEGER)%s%d)", op, number);
  }
}


static udm_rc_t
UdmBuildCmpArgSQL(UDM_AGENT *A, UDM_DB *db, udm_match_mode_t match_mode,
                  const char *word, char *cmparg, size_t maxlen)
{
  const char *left= "", *right= "";
  size_t length= strlen(word);
  char escwrd[1000];

  if (match_mode == UDM_MATCH_RANGE)
  {
    UDM_ASSERT(length > 6);
    if (*word == '[')
      left= ">=";
    else if (*word == '{')
      left= ">";
    if (word[length - 1] == ']')
      right= "<=";
    else if (word[length - 1] == '}')
      right= "<";
    word++;
    length-= 2;
  }

  UdmDBSQLEscStr(A, db, escwrd, word, length); /* Search word */
  switch (match_mode)
  {  
    case UDM_MATCH_BEGIN:
      udm_snprintf(cmparg, maxlen, "word LIKE '%s%%'", escwrd);
      break;
    case UDM_MATCH_END:
      udm_snprintf(cmparg, maxlen, "word LIKE '%%%s'", escwrd);
      break;
    case UDM_MATCH_SUBSTR:
      udm_snprintf(cmparg, maxlen, "word LIKE '%%%s%%'", escwrd);
      break;
    case UDM_MATCH_NUMERIC_LT:
      UdmBuildNumericOperatorCondition(db, cmparg, maxlen, "<", atoi(escwrd));
      break;
    case UDM_MATCH_NUMERIC_GT:
      UdmBuildNumericOperatorCondition(db, cmparg, maxlen, ">", atoi(escwrd));
      break;
    case UDM_MATCH_RANGE:
      {
        char *first= escwrd;
        char *second= strstr(first, " TO ");
        if (!second)
        {
          udm_snprintf(cmparg, maxlen, "word='<ERROR>'");
          return UDM_ERROR;
        }
        *second= '\0';
        second+= 4;
        udm_snprintf(cmparg, maxlen, "word%s'%s' AND word%s'%s'",
                     left, first, right, second);
      }
      break;
    case UDM_MATCH_FULL:
    default:
      udm_snprintf(cmparg, maxlen, "word='%s'", escwrd);
      break;
  }
  return(UDM_OK);
}


static udm_rc_t
UdmFindOneWordSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                  UDM_FINDWORD_ARGS *args)
{
  char cmparg[256];
  UdmBuildCmpArgSQL(A, db,
                    args->Word.Param.match_mode, args->Word.Word.str,
                    cmparg, sizeof(cmparg));
  args->cmparg= cmparg;
  
  UDM_ASSERT(UdmSQLDBModeHandler(db)->FindWord != NULL);
  return UdmSQLDBModeHandler(db)->FindWord(A, db, Query, args);
}


static udm_rc_t
UdmFindMultiWordSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                    UDM_FINDWORD_ARGS *args)
{
  char *lt, *tmp_word, *tok;
  udm_rc_t rc= UDM_OK;
  UDM_SEARCHSECTIONLISTLIST OriginalSearchSectionListList;
  size_t orig_wordnum;
  size_t nparts= 0;
  const char *w;
  char *orig_word= args->Word.Word.str;
  char delim[]= " \r\t_-./";

  /* Check if the word really multi-part */
  for (w= args->Word.Word.str; ; w++)
  {
    if (!*w)
      return UdmFindOneWordSQL(A, db, Query, args); /* No delimiters found */
    
    if (strchr(delim, *w)) /* Delimiter found */
      break;
  }
  
  if (!(tmp_word= UdmStrdup(args->Word.Word.str)))
    return(UDM_ERROR);

  UdmLog(A, UDM_LOG_DEBUG,
         "Start searching for multiword '%s'", args->Word.Word.str);
  OriginalSearchSectionListList= args->SearchSectionListList;
  UdmSearchSectionListListInit(&args->SearchSectionListList);
  orig_wordnum= args->Word.Param.order;
  args->need_coords= 1; /* Force immediate coord unpacking */
  
  for (tok= udm_strtok_r(tmp_word, delim, &lt) ; tok ;
       tok= udm_strtok_r(NULL, delim, &lt))
  {
    udm_timer_t ticks1= UdmStartTimer();
    args->Word.Word.str= tok;
    UdmLog(A, UDM_LOG_DEBUG,
           "Searching for subword '%s'", args->Word.Word.str);
    rc= UdmFindOneWordSQL(A, db, Query, args);
    UdmLog(A, UDM_LOG_DEBUG,
           "Stop searching for subword '%s' %d coords found: %.2f",
           args->Word.Word.str, (int) args->Word.Param.count, UdmStopTimer(&ticks1));
    /* If the next word wasn't found - no need to search for others. */
    if (rc != UDM_OK || !args->Word.Param.count)
      goto ret;
    nparts++;
    args->Word.Param.order++;
  }
  
  /* All parts returned results. Check phrase */
  UdmMultiWordAdd(&args->SearchSectionListList, &Query->Res.WWList, &args->Word,
                  &args->urls, &OriginalSearchSectionListList,
                  orig_wordnum, nparts);
  
  
ret:
  UdmFree(tmp_word);
  UdmSearchSectionListListFree(&args->SearchSectionListList);
  args->SearchSectionListList= OriginalSearchSectionListList;
  args->Word.Word.str= orig_word;
  args->need_coords= 0;
  UdmLog(A, UDM_LOG_DEBUG,
         "Stop searching for multiword '%s'", args->Word.Word.str);
  return rc;
}


static udm_rc_t
UdmFindAlwaysFoundWordSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                          UDM_FINDWORD_ARGS *args)
{
  udm_rc_t rc= UDM_OK;
  UDM_SQLRES SQLRes;
  char qbuf[1024 * 4];
  size_t nrows;
  size_t i;
  UDM_URLCRDLIST CoordList;
  UDM_URL_CRD Coord;

  bzero((void*) &CoordList, sizeof(CoordList));
  bzero((void*) &Coord, sizeof(Coord));
  Coord.urlid_coord.coord.pos= 0x00010100; /* TODO34: why? */
  Coord.num= args->Word.Param.order;

  if (*args->where)
     udm_snprintf(qbuf, sizeof(qbuf), "SELECT url.rec_id FROM url%s WHERE %s",
                  Query->from, args->where);
  else
  {
    if (args->urls.nurls)
    {
      /*
        A fast limit is loaded.
        No needs to do "SELECT FROM url".
        Populate CoordList from the fast limit instead.
      */
      for (i= 0; i < args->urls.nurls; i++)
      {
        Coord.urlid_coord.url_id= args->urls.urls[i];
        if (UDM_OK != (rc= UdmAddOneCoord(&CoordList, &Coord)))
          return UDM_ERROR;
      }
      UdmURLCRDListListAddWithSort2(&args->SearchSectionListList,
                                    &Query->Res.WWList, &args->Word, &CoordList);
      return UDM_OK;
    }
    udm_snprintf(qbuf, sizeof(qbuf), "SELECT url.rec_id FROM url");
  }

  if ((rc= UdmDBSQLQuery(A, db, &SQLRes, qbuf)) != UDM_OK)
    return(rc);
  /* Note that rc is implicitly set to UDM_OK at this point. */
  if (! (nrows= UdmSQLNumRows(&SQLRes)))
    goto err;

  for (i = 0; i < nrows; i++)
  {
    Coord.urlid_coord.url_id= (urlid_t) UDM_ATOI(UdmSQLValue(&SQLRes, i, 0));
    if (UDM_OK != (rc= UdmAddOneCoord(&CoordList, &Coord)))
      break;
  }

  if (args->urls.nurls)
    UdmApplyFastLimit(&CoordList, &args->urls);
  if (CoordList.ncoords)
    UdmURLCRDListListAddWithSort2(&args->SearchSectionListList,
                                  &Query->Res.WWList, &args->Word, &CoordList);

err:
  UdmSQLFree(&SQLRes);
  return(rc);
}


static udm_rc_t
UdmCheckIndex(UDM_AGENT *A, UDM_DB *db)
{
  int tm;
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmBlobReadTimestamp(A, db, &tm, 0)))
    return rc;
  if (tm)
    return UDM_OK;
#ifdef WIN32
  sprintf(A->Conf->errstr, "Inverted word index not found. Probably you forgot to run 'Create fast index'.");
#else
  sprintf(A->Conf->errstr, "Inverted word index not found. Probably you forgot to run 'indexer --index'.");
#endif
  return UDM_ERROR;
}


static udm_rc_t
UdmMergeWords(UDM_AGENT *A, UDM_DB *db,
              UDM_FINDWORD_ARGS *args, UDM_SEARCHSECTIONLIST *SectionList)
{
  udm_timer_t ticks= UdmStartTimer();

  UdmLog(A, UDM_LOG_DEBUG,
         "Start merging %d lists", (int) args->SearchSectionListList.nitems);
  UdmSearchSectionListListMergeSorted(&args->SearchSectionListList, SectionList, 1);
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f (%d sections)",
         "Stop  merging:", UdmStopTimer(&ticks), (int) SectionList->nsections);

  if (!SectionList->nsections &&
      UdmSQLDBMode(db) == UDM_SQLDBMODE_BLOB &&
      !args->live_updates)
    return UdmCheckIndex(A, db);
  return UDM_OK;
}


static udm_rc_t
UdmSearchParamInit(UDM_FINDWORD_ARGS *args,
                   UDM_AGENT *A,
                   UDM_QUERY *Query,
                   UDM_DB *db)
{
  bzero((void*) args, sizeof(*args));
  UdmWideWordListInit(&args->CollationMatches);
  UdmQueryParamInit(&args->query_param, A->Conf, UdmSQLDBVars(db));
  args->Word.Param.match_mode= UdmMatchMode(UdmVarListFindStr(&A->Conf->Vars, "wm", "wrd"));
  args->live_updates= UdmVarListFindBool(UdmSQLDBVars(db), "LiveUpdates", UDM_FALSE);
  return UdmSQLBuildWhereCondition(A, db, Query, &args->where);
}


static void
UdmSearchParamFree(UDM_FINDWORD_ARGS *args)
{
  UDM_FREE(args->urls.urls);
  UDM_FREE(args->live_update_active_urls.urls);
  UDM_FREE(args->live_update_deleted_urls.urls);
  UdmWideWordListFree(&args->CollationMatches);
  UdmSearchSectionListListFree(&args->SearchSectionListList);
  UdmSQLResListFree(&args->SQLResults);
  UdmInvertedIndexCacheFree(&args->IndexCache);
}


static udm_rc_t
UdmFindLoadSlowOrFastLimit(UDM_AGENT *A,
                           UDM_DB *db,
                           UDM_URLID_LIST *list,
                           const char *fl,
                           int count)
{
  udm_rc_t rc;
  char name[64];
  const char *q;
  UDM_URLID_LIST fl_urls;
  bzero((void*) &fl_urls, sizeof(fl_urls));
  if ((fl_urls.exclude= (fl[0] == '-')))
    fl++;
  udm_snprintf(name, sizeof(name), "Limit.%s", fl);
  if (UDM_OK != (rc= ((q= UdmVarListFindStr(&A->Conf->Vars, name, NULL)) ?
                     UdmLoadSlowLimitWithSort(A, db, &fl_urls, q) :
                     UdmBlobLoadFastURLLimit(A, db, fl, &fl_urls))))
    goto ret;
  UdmLog(A,UDM_LOG_DEBUG, "Limit '%s' loaded%s%s %d URLs",
         fl, fl_urls.exclude ? " type=excluding" : "",
         q ? " source=slow":"", (int) fl_urls.nurls);
  if (!count)
  {
    list->exclude= fl_urls.exclude;
    UdmURLIdListUnion(list, &fl_urls);
  }
  else
  {
    if (list->exclude == fl_urls.exclude)
      UdmURLIdListUnion(list, &fl_urls);
    else
      UdmURLIdListJoin(list, &fl_urls);
  }
  UDM_FREE(fl_urls.urls);
ret:
  return rc;
}


/*
  Load WHERE and fl limits from the database at search time.
*/
static udm_rc_t
UdmFindLoadLimits(UDM_AGENT *A,
                  UDM_DB *db,
                  UDM_QUERY *Query,
                  UDM_FINDWORD_ARGS *args,
                  const char *fl)
{
  udm_rc_t rc= UDM_OK;
  udm_timer_t ticks= UdmStartTimer();

  UdmLog(A, UDM_LOG_DEBUG, "Start loading limits");
  ticks= UdmStartTimer();
  if (*args->where)
  {
    LoadURL(A, db, Query, args->where, &args->urls);
    UdmLog(A, UDM_LOG_DEBUG,
           "WHERE limit loaded. %d URLs found", (int) args->urls.nurls);
  }
  if (!args->urls.empty && fl[0])
  {
    UDM_URLID_LIST lim;
    char delim[]= " ,", names[128], *tok, *lt;
    int nlimits= 0;
    bzero(&lim, sizeof(lim));
    udm_snprintf(names, sizeof(names), "%s", fl);
    for (tok= udm_strtok_r(names, delim, &lt) ; tok ;
         tok= udm_strtok_r(NULL, delim, &lt), nlimits++)
    {
      if (UDM_OK != (rc= UdmFindLoadSlowOrFastLimit(A, db, &lim,
                                                    tok, nlimits)))
        goto ret;
    }
    if (nlimits)
      UdmURLIdListMerge(&args->urls, &lim);
    UDM_FREE(lim.urls);
  }
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f (%d URLs found)",
         "Stop  loading limits", UdmStopTimer(&ticks), (int) args->urls.nurls);
ret:
  return rc;
}


/*
  Load word information from the database
*/
static udm_rc_t
UdmFindWordsFetch(UDM_AGENT *A,
                  UDM_DB *db,
                  UDM_QUERY *Query,
                  UDM_FINDWORD_ARGS *args,
                  const char *always_found_word)
{
  size_t wordnum;
  udm_rc_t rc= UDM_OK;
  udm_timer_t ticks0= UdmStartTimer();
  
  UdmLog(A, UDM_LOG_DEBUG, "Start fetching words");
  
  /* Now find each word */
  for(wordnum=0; wordnum < Query->Res.WWList.nwords; wordnum++)
  {
    udm_timer_t ticks= UdmStartTimer();
    UDM_WIDEWORD *W= &Query->Res.WWList.Word[wordnum];
    char quoted_word[64];
    udm_snprintf(quoted_word, sizeof(quoted_word), "'%s'", W->Word.str);

    if (W->Param.origin == UDM_WORD_ORIGIN_STOP) continue;

    UdmLog(A, UDM_LOG_DEBUG, "Start search for %s", quoted_word);

    args->Word.Param.order= wordnum;
    args->Word.Param.count= 0;
    args->Word.Word.str= W->Word.str;
    args->Word.Param.match_mode= W->Param.match_mode;
    args->Word.Param.secno= W->Param.secno;

    /*
       For now SYNONYMs only are treated as a possible multi-word
       origin. Probably it will be changed in future, so we will
       use this feature for phrase search.
     */
    if (always_found_word && !strcmp(W->Word.str, always_found_word))
      rc= UdmFindAlwaysFoundWordSQL(A, db, Query, args);
    else if (W->Param.origin == UDM_WORD_ORIGIN_SYNONYM ||
             W->Param.phrwidth > 0)
      rc= UdmFindMultiWordSQL(A, db, Query, args);
    else
      rc= UdmFindOneWordSQL(A, db, Query, args);

    if (rc != UDM_OK)
      goto ret;

    /*
      If CollationMatches is not empty, then we should skip
      updating word statistics here - it will be updated in
      the loop after UdmSortAndGroupByURL().
     */
    if (!args->CollationMatches.nwords)
      Query->Res.WWList.Word[wordnum].Param.count+= args->Word.Param.count;

    UdmLog(A, UDM_LOG_DEBUG,
           "Stop  search for %-13s%.2f (%u coords found)",
           quoted_word, UdmStopTimer(&ticks), (int) args->Word.Param.count);
  }
  UdmLog(A, UDM_LOG_DEBUG,
         "%-30s%.2f", "Stop  fetching words:", UdmStopTimer(&ticks0));
ret:
  return rc;
}


static udm_rc_t
UdmFindWordsSQLNoCached(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  const char  *always_found_word, *fl;
  udm_rc_t rc= UDM_OK;
  UDM_FINDWORD_ARGS args;

  /* Query->SectionList must be clean */
  UDM_ASSERT(!Query->SectionList.mcoords);
  UDM_ASSERT(!Query->SectionList.ncoords);
  UDM_ASSERT(!Query->SectionList.Coord);
  UDM_ASSERT(!Query->SectionList.msections);
  UDM_ASSERT(!Query->SectionList.nsections);
  UDM_ASSERT(!Query->SectionList.Section);

  UDM_GETLOCK(A, UDM_LOCK_DB);
  {
    if (UDM_OK != (rc= UdmSearchParamInit(&args, A, Query, db)) &&
        UdmDBSQLError(db)[0])
    {
      UdmEnvCopyErrMsgFromDB(A->Conf, db);
    }
    

    always_found_word= UdmVarListFindStr(&A->Conf->Vars, "AlwaysFoundWord", NULL);
    fl= UdmVarListFindStr(&A->Conf->Vars, "fl", UdmVarListFindStr(UdmSQLDBVars(db), "fl", ""));
  }
  UDM_RELEASELOCK(A, UDM_LOCK_DB);
  if (rc != UDM_OK)
    goto ret;

  if ((UdmSQLDBMode(db) == UDM_SQLDBMODE_BLOB && args.where) || fl[0])
  {
    if (UDM_OK != UdmFindLoadLimits(A, db, Query, &args, fl))
      goto ret;

    if (args.urls.empty)
      goto ret;
  }


  if (UDM_OK != (rc= UdmSQLDBModeHandler(db)->InitSearch(A, db, Query, &args)))
    goto ret;

  if (UDM_OK != (rc= UdmFindWordsFetch(A, db, Query, &args, always_found_word)))
    goto ret;

  if (UDM_OK != (rc= UdmMergeWords(A, db, &args, &Query->SectionList)))
    goto ret;
  
  if (UDM_OK != (rc= UdmSortAndGroupByURL(A, Query, &Query->SectionList, db)))
    goto ret;

  /* 
     We cannot add collation matches before
     UdmSortAndGroupByURL - to use optimized groupping
     functions when WWList->nwords==1
  */
  if (args.CollationMatches.nwords)
  {
    size_t i;
    UdmWideWordListSort(&args.CollationMatches);
    for (i= 0; i < args.CollationMatches.nwords; i++)
    {
      UdmWideWordListAdd(&Query->Res.WWList, &args.CollationMatches.Word[i]);
    }
  }

ret:
  UdmSearchParamFree(&args);
  return rc;
}


static udm_rc_t
UdmFindWordsSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmQueryCacheGetSQL(A, db, Query)))
    return rc;
  if (!Query->URLData.nitems)
  {
    rc= UdmFindWordsSQLNoCached(A, db, Query);
    if (rc == UDM_OK && Query->URLData.nitems)
      rc= UdmQueryCachePutSQL(A, db, Query);
  }
  return rc;
}

/****************** Track ***********************************/

static udm_rc_t
UdmTrackSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  UDM_VARLIST Vars;
  char    *qbuf;
  char    *text_escaped;
  udm_bool_t trackquery= UdmVarListFindBool(UdmSQLDBVars(db), "trackquery", UDM_FALSE);
  const char  *words= UdmVarListFindStr(&A->Conf->Vars,"q",""); /* "q-lc" was here */
  const char      *IP = UdmVarListFindStr(&A->Conf->Vars, "IP", "");
  size_t          i, escaped_len, qbuf_len;
  int             qtime, rec_id;
  udm_rc_t rc;
  const char      *qu = (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "";

  if (!trackquery)
    return UDM_OK;

  if (*words == '\0') return UDM_OK; /* do not store empty queries */

  escaped_len = 4 * strlen(words);
  qbuf_len = escaped_len + 4096;

  if ((qbuf = (char*)UdmMalloc(qbuf_len)) == NULL) return UDM_ERROR;
  if ((text_escaped = (char*)UdmMalloc(escaped_len)) == NULL)
  { 
    UDM_FREE(qbuf);
    return UDM_ERROR;
  }
  
  UdmVarListInit(&Vars);
  UdmVarListSQLEscape(A, &Vars, &A->Conf->Vars, db);
  
  /* Escape text to track it  */
  UdmDBSQLEscStr(A, db, text_escaped, words, strlen(words)); /* query for tracking */
  
  if (UdmSQLDBType(db) == UDM_DB_IBASE ||
      UdmSQLDBType(db) == UDM_DB_MIMER ||
      UdmSQLDBType(db) == UDM_DB_ORACLE8)
  {
    const char *next;
    switch (UdmSQLDBType(db))
    {
      case UDM_DB_IBASE: next= "SELECT GEN_ID(qtrack_GEN,1) FROM rdb$database"; break;
      case UDM_DB_MIMER: next= "SELECT NEXT_VALUE OF qtrack_GEN FROM system.onerow"; break;
      case UDM_DB_ORACLE8: next= "SELECT qtrack_seq.nextval FROM dual"; break;
      default: next= NULL; /* Make compiler happy */
    }
    if (UDM_OK != (rc= UdmDBSQLQueryOneRowInt(A, db, &rec_id, next)))
      goto UdmTrack_exit;
    udm_snprintf(qbuf, qbuf_len - 1,
                 "INSERT INTO qtrack (rec_id,ip,qwords,qtime,wtime,nfound) "
                 "VALUES "
                 "(%d,'%s','%s',%d,%d,%d)",
                 rec_id, IP, text_escaped, qtime= (int)time(NULL),
                 UdmVarListFindInt(&Query->Res.Vars, "SearchTime", 0),
                 (int) UdmResultTotalFound(&Query->Res));
    if (UDM_OK != (rc = UdmDBSQLQuery(A, db, NULL, qbuf)))
      goto UdmTrack_exit;
  }
  else
  {
    udm_snprintf(qbuf, qbuf_len - 1,
                 "INSERT INTO qtrack (ip,qwords,qtime,wtime,nfound) "
                 "VALUES "
                 "('%s','%s',%d,%d,%d)",
                 IP, text_escaped, qtime= (int)time(NULL),
                 UdmVarListFindInt(&Query->Res.Vars, "SearchTime", 0),
                 (int) UdmResultTotalFound(&Query->Res));
  
    if (UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, qbuf)))
      goto UdmTrack_exit;
    
    if (UdmSQLDBType(db) == UDM_DB_MYSQL)
      udm_snprintf(qbuf, qbuf_len - 1, "SELECT last_insert_id()");
    else
      udm_snprintf(qbuf, qbuf_len - 1, "SELECT rec_id FROM qtrack WHERE ip='%s' AND qtime=%d", IP, qtime);
    if (UDM_OK != (rc= UdmDBSQLQueryOneRowInt(A, db, &rec_id, qbuf)))
      goto UdmTrack_exit;
  }
  
  for (i = 0; i < Vars.nvars; i++)
  {
    const UDM_VAR *Var= UdmVarListFindConstByIndex(&Vars, i);
    const char *name= UdmVarName(Var);
    if (!strncasecmp(name, "query.",6) &&
        strcasecmp(name, "query.q") &&
        strcasecmp(name, "query.BrowserCharset") &&
        strcasecmp(name, "query.IP") &&
        UdmVarStr(Var) != NULL && UdmVarStr(Var)[0] != '\0') 
    {
      udm_snprintf(qbuf, qbuf_len,
                   "INSERT INTO qinfo (q_id,sname,sval) "
                   "VALUES "
                   "(%s%i%s,'%s','%s')", 
                   qu, rec_id, qu, name + 6, UdmVarStr(Var));
      rc= UdmDBSQLQuery(A, db, NULL, qbuf);
      if (rc != UDM_OK) goto UdmTrack_exit;
    }
  }
UdmTrack_exit:
  UdmVarListFree(&Vars);
  UDM_FREE(text_escaped);
  UDM_FREE(qbuf);
  return rc;
}


/********************* Adding URLInfo to Res *********************/

static udm_rc_t UpdateShows(UDM_AGENT *A, UDM_DB *db, urlid_t url_id)
{
  char qbuf[64];
  udm_snprintf(qbuf, sizeof(qbuf), "UPDATE url SET shows = shows + 1 WHERE rec_id = %s%i%s",
               (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "",
               url_id,
               (UdmSQLDBType(db) == UDM_DB_PGSQL) ? "'" : "");
  return UdmDBSQLQuery(A, db, NULL, qbuf);
}

static void SQLResToSection(UDM_SQLRES *R, UDM_VARLIST *S, size_t row)
{
  const char *sname=UdmSQLValue(R,row,1);
  const char *sval=UdmSQLValue(R,row,2);
  UdmVarListAddStr(S, sname, sval);
}



static void
SQLResToCachedCopy(UDM_SQLRES *R, UDM_DOCUMENT *Doc,
                   size_t max_doc_size, size_t row)
{
  const char *sval= UdmSQLValue(R, row, 1);
  size_t len= UdmSQLLen(R, row, 1);
  udm_timer_t timer= 0;
  UdmDocSetFromCachedHTTPResponse(Doc, sval, len, max_doc_size, &timer);
}


static size_t
UdmDBNum(UDM_QUERY *Query, size_t n)
{
  UDM_URLDATA *Data= &Query->URLData.Item[n + Query->stats.first];
  return UDM_COORD2DBNUM(Data->score);
}


static udm_rc_t
UdmQueryAddURLInfoUsingIN(UDM_AGENT *Agent, UDM_DB *db, UDM_QUERY *Query,
                          const char *qbuf, int urlinfob)
{
  udm_rc_t rc;
  UDM_SQLRES SQLres;
  size_t j, sqlrows;
  size_t max_doc_size= UdmVarListFindInt(&Agent->Conf->Vars, "MaxDocSize", UDM_MAXDOCSIZE);
  
  if (UDM_OK!= (rc= UdmDBSQLQuery(Agent, db, &SQLres, qbuf)))
    return rc;
    
  for (sqlrows= UdmSQLNumRows(&SQLres), j=0;
       j< UdmResultNumRows(&Query->Res);
       j++)
  {
    if (&Agent->Conf->DBList.Item[UdmDBNum(Query, j)] == db)
    {
      size_t i;
      UDM_DOCUMENT *D= &Query->Res.Doc[j];
      urlid_t      url_id = UdmVarListFindInt(&D->Sections, "ID", 0);
      for(i = 0; i < sqlrows; i++)
      {
        if(url_id == UDM_ATOI(UdmSQLValue(&SQLres,i,0)))
        {
          if (urlinfob)
            SQLResToCachedCopy(&SQLres, D, max_doc_size, i);
          else
            SQLResToSection(&SQLres, &D->Sections, i);
        }
      }
    }
  }
  UdmSQLFree(&SQLres);
  return UDM_OK;
}


static udm_rc_t
UdmDocAddURLInfo(UDM_AGENT *A, UDM_DOCUMENT *D, UDM_DB *db, const char *qbuf)
{
  UDM_SQLRES SQLres;
  udm_rc_t rc;
  size_t row;
  
  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLres, qbuf)))
    return rc;
  for(row= 0; row < UdmSQLNumRows(&SQLres); row++)
    SQLResToSection(&SQLres, &D->Sections, row);
  UdmSQLFree(&SQLres);
  return rc;
}


static void
UdmDocSetPopularity(UDM_DOCUMENT *D, double pop)
{
  char tmp[16];
  udm_snprintf(tmp, sizeof(tmp), "%.5f", pop);
  UdmVarListReplaceStr(&D->Sections, "Pop_Rank", tmp);
}


typedef struct
{
  udm_bool_t LoadURLBasicInfo;
  udm_bool_t LoadURLInfo;
  udm_bool_t LoadURLInfoBin;
  udm_bool_t LoadTagInfo;
  udm_bool_t PopRankUseShowCnt;
  double PopRankShowCntRatio;
} UDM_RESINFOPARAM;


static void
UdmResInfoParamInit(UDM_RESINFOPARAM *P, UDM_VARLIST *Vars)
{
  P->PopRankUseShowCnt= UdmVarListFindBool(Vars, "PopRankUseShowCnt", UDM_FALSE);
  P->LoadTagInfo= UdmVarListFindBool(Vars, "LoadTagInfo", UDM_FALSE);
  P->LoadURLInfo= UdmVarListFindBool(Vars, "LoadURLInfo", UDM_TRUE);
  P->PopRankShowCntRatio= UdmVarListFindDouble(Vars, "PopRankShowCntRatio", 25.0);
  P->LoadURLInfoBin= UdmVarListFindBool(Vars, "LoadURLInfoBin", UDM_TRUE);
  P->LoadURLBasicInfo= UdmVarListFindBool(Vars, "LoadURLBasicInfo", UDM_TRUE);
}


static udm_rc_t
UdmQueryAddDocInfoUsingLoop(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  size_t i;
  UDM_SQLRES SQLres;
  UDM_RESINFOPARAM param;
  size_t max_doc_size= UdmVarListFindInt(&A->Conf->Vars, "MaxDocSize", UDM_MAXDOCSIZE);

  UdmResInfoParamInit(&param, &A->Conf->Vars);
  for(i= 0; i < UdmResultNumRows(&Query->Res); i++)
  {
    UDM_DOCUMENT *D= &Query->Res.Doc[i];
    urlid_t  url_id= UdmVarListFindInt(&D->Sections, "ID", 0);
    char  qbuf[128];

    if (&A->Conf->DBList.Item[UdmDBNum(Query, i)] != db)
      continue;

    udm_snprintf(qbuf, sizeof(qbuf), 
                 "SELECT " SQLRESTODOC_COLUMNS
                 " FROM url WHERE rec_id=%d", url_id);
    if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLres, qbuf)))
      return rc;

    if (UdmSQLNumRows(&SQLres))
    {
      double pr;
      SQLResToDoc(A->Conf, D, &SQLres, 0);
      if (param.PopRankUseShowCnt &&
          (pr= atof(UdmVarListFindStr(&D->Sections, "Score", "0.0"))) >= 
          param.PopRankShowCntRatio)
        UpdateShows(A, db, url_id);
    }
    UdmSQLFree(&SQLres);

    if (param.LoadTagInfo)
    {
      udm_snprintf(qbuf, sizeof(qbuf),
                   "SELECT u.rec_id, 'tag', tag FROM url u, server s "
                   "WHERE  u.rec_id=%d AND u.server_id=s.rec_id", url_id); 
      if(UDM_OK != (rc= UdmDocAddURLInfo(A, D, db, qbuf)))
        return rc;
    }

    if (param.LoadURLInfo)
    {
      sprintf(qbuf,"SELECT url_id,sname,sval FROM urlinfo WHERE url_id=%i", url_id);
      if(UDM_OK != (rc= UdmDocAddURLInfo(A, D, db, qbuf)))
        return rc;
    }

    if (param.LoadURLInfoBin)
    {
      sprintf(qbuf,"SELECT url_id,content FROM cachedcopy WHERE url_id=%i", url_id);
      if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLres, qbuf)))
        return rc;
      if (UdmSQLNumRows(&SQLres))
        SQLResToCachedCopy(&SQLres, D, max_doc_size, 0);
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmQueryAddDocInfoUsingIN(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc= UDM_OK;
  size_t i, j, sqlrows;
  UDM_SQLRES SQLres;
  UDM_RESINFOPARAM param;
  UDM_DSTR  in_list, qq;
  const char *hi_priority= UdmSQLDBType(db) == UDM_DB_MYSQL ? "HIGH_PRIORITY" : "";

  UdmResInfoParamInit(&param, &A->Conf->Vars);
    
  UdmDSTRInit(&in_list, 1024);
  UdmDSTRInit(&qq, 1024);
    
  /* Compose IN string and set to zero url_id field */
  for(i=0; i < UdmResultNumRows(&Query->Res); i++)
  {
    if (&A->Conf->DBList.Item[UdmDBNum(Query, i)] == db)
    {
      const char *comma= UdmDSTRLength(&in_list) ? "," : "";
      const char *squot= UdmSQLDBType(db) == UDM_DB_PGSQL ? "'" : "";
      UdmDSTRAppendf(&in_list, "%s%s%i%s", comma, squot,
              UdmVarListFindInt(&Query->Res.Doc[i].Sections, "ID", 0), squot);
    }
  }

  if (!UdmDSTRLength(&in_list))
    goto ret_in;

  if (param.LoadURLBasicInfo)
  {
    UdmDSTRReset(&qq);
    UdmDSTRAppendf(&qq,
                   "SELECT %s " SQLRESTODOC_COLUMNS
                   " FROM url WHERE rec_id IN (%s)",
                   hi_priority, UdmDSTRPtr(&in_list));
    if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLres, UdmDSTRPtr(&qq))))
      goto ret_in;

    for (sqlrows= UdmSQLNumRows(&SQLres), j=0;
         j < UdmResultNumRows(&Query->Res);
         j++)
    {
      if (&A->Conf->DBList.Item[UdmDBNum(Query, j)] == db)
      {
        UDM_DOCUMENT *D= &Query->Res.Doc[j];
        urlid_t      url_id= UdmVarListFindInt(&D->Sections, "ID", 0);
        for(i= 0; i < sqlrows; i++)
        {
          if (url_id == UDM_ATOI(UdmSQLValue(&SQLres,i,0)))
          {
            double pr;
            SQLResToDoc(A->Conf, D, &SQLres, i);
            if (param.PopRankUseShowCnt &&
                (pr= atof(UdmVarListFindStr(&D->Sections, "Score", "0.0"))) >=
                param.PopRankShowCntRatio)
              UpdateShows(A, db, url_id);
            break;
          }
        }
      }
   } 
   UdmSQLFree(&SQLres);
  }
    
  if (param.LoadTagInfo)
  {
    UdmDSTRReset(&qq);
    UdmDSTRAppendf(&qq,
                   "SELECT u.rec_id, 'tag', tag FROM url u, server s "
                   "WHERE  u.rec_id in (%s) AND u.server_id=s.rec_id",
                   UdmDSTRPtr(&in_list)); 
    if (UDM_OK != (rc= UdmQueryAddURLInfoUsingIN(A, db, Query,
                                                 UdmDSTRPtr(&qq), 0)))
      return rc;
  }
    
  if (param.LoadURLInfo)
  {
    UdmDSTRReset(&qq);
    UdmDSTRAppendf(&qq,
                   "SELECT url_id,sname,sval "
                   "FROM urlinfo WHERE url_id IN (%s)",
                   UdmDSTRPtr(&in_list));
    if (UDM_OK != (rc= UdmQueryAddURLInfoUsingIN(A, db, Query,
                                                 UdmDSTRPtr(&qq), 0)))
      return rc;
  }
    
  if (param.LoadURLInfoBin)
  {
    UdmDSTRReset(&qq);
    UdmDSTRAppendf(&qq,
                   "SELECT url_id,content "
                   "FROM cachedcopy WHERE url_id IN (%s)",
                   UdmDSTRPtr(&in_list));
    if (UDM_OK != (rc= UdmQueryAddURLInfoUsingIN(A, db, Query,
                                                 UdmDSTRPtr(&qq), 1)))
      return rc;
  }
    
ret_in:
  UdmDSTRFree(&in_list);
  UdmDSTRFree(&qq);
  return rc;
}


static udm_rc_t
UdmQueryAddDocInfoSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  size_t i;
  udm_rc_t rc;
  UDM_RESINFOPARAM param;

  if (!UdmResultNumRows(&Query->Res))
    return UDM_OK;

  UdmResInfoParamInit(&param, &A->Conf->Vars);
  
  if (param.PopRankUseShowCnt)
    UdmLog(A, UDM_LOG_DEBUG, "use_showcnt: %d  ratio: %f",
           param.PopRankUseShowCnt, param.PopRankShowCntRatio);

  for (i= 0; i < UdmResultNumRows(&Query->Res); i++)
  {
    UDM_URLDATA *Data= &Query->URLData.Item[i + Query->stats.first];
    UdmVarListReplaceInt(&Query->Res.Doc[i].Sections, "id", Data->url_id);
    UdmDocSetPopularity(&Query->Res.Doc[i], Data->pop_rank);
  }
  
  rc= UdmSQLDBHaveIn(db) ?
      UdmQueryAddDocInfoUsingIN(A, db, Query) :
      UdmQueryAddDocInfoUsingLoop(A, db, Query);

  return rc;
}


/************************* Misc *******************************************/


static udm_rc_t
UdmExportSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  UDM_SQLRES SQLRes;
  udm_rc_t rc;
  UDM_STR row[24];

  /* TODO34: add well-formed output */
  printf("<database>\n");
  printf("<urlList>\n");
  rc= UdmDBSQLExecDirect(Indexer, db, &SQLRes,
                         "SELECT rec_id,status,docsize,next_index_time,"
                                 "last_mod_time,referrer,hops,crc32,seed,"
                                 "bad_since_time,server_id,shows,"
                                 "url"
                         " FROM url");
  if (rc != UDM_OK) return(rc);
  while (UdmDBSQLFetchRow(Indexer, db, &SQLRes, row) == UDM_OK)
  {
    printf(
      "<url "
      "rec_id=\"%s\" "
      "status=\"%s\" "
      "docsize=\"%s\" "
      "next_index_time=\"%s\" "
      "last_mod_time=\"%s\" "
      "referrer=\"%s\" "
      "hops=\"%s\" "
      "crc32=\"%s\" "
      "seed=\"%s\" "
      "bad_since_time=\"%s\" "
      "server_id=\"%s\" "
      "shows=\"%s\" "
      "url=\"%s\" "
      "/>\n",
      row[0].str, row[1].str, row[2].str, row[3].str,
      row[4].str, row[5].str, row[6].str, row[7].str,
      row[8].str, row[9].str, row[10].str, row[11].str,
      row[12].str);
  }
  UdmSQLFree(&SQLRes);
  printf("</urlList>\n");

  printf("<linkList>\n");
  if (UDM_OK != (rc= UdmDBSQLExecDirect(Indexer, db, &SQLRes,
                 "SELECT url_id,seed,url,src,rel,linktext FROM links")))
    return rc;
  while (UdmDBSQLFetchRow(Indexer, db, &SQLRes, row) == UDM_OK)
  {
    printf(
      "<link "
      "url_id=\"%s\" "
      "seed=\"%s\" "
      "url=\"%s\" "
      "src=\"%s\" "
      "rel=\"%s\" "
      "linktext=\"%s\" "
      "/>\n",
      row[0].str, row[1].str, row[2].str,
      row[3].str, row[4].str, row[5].str);
  }
  UdmSQLFree(&SQLRes);
  printf("</linkList>\n");

  printf("</database>\n");
  return(0);
}


static udm_rc_t
UdmDocPerSite(UDM_AGENT *A, UDM_DOCUMENT *D, UDM_DB *db)
{
  char qbuf[1024];
  const char *s, *hostinfo= UdmVarListFindStr(&D->Sections, "Hostinfo", NULL);
  udm_rc_t rc;
  int num, prevnum= UdmVarListFindInt(&D->Sections, "DocPerSite", 0);
  UDM_SQLRES SQLRes;
  
  if (!hostinfo)
    return UDM_OK;
  
  for (s= hostinfo; s[0]; s++)
  {
    /*
      Host name good characters: digits, letters, hyphen (-).
      Just check the worst characters.
    */
    if (*s == '\'' || *s == '\"')
    {
      num= 1000000;
      goto ret;
    }
  }
  udm_snprintf(qbuf, sizeof(qbuf),
               "SELECT COUNT(*) FROM url WHERE url LIKE '%s%%'", hostinfo);
  
  if (UDM_OK!= (rc= UdmDBSQLQuery(A, db, &SQLRes, qbuf)))
    return rc;
  num= prevnum + atoi(UdmSQLValue(&SQLRes, 0, 0));
  UdmSQLFree(&SQLRes);
ret:
  UdmVarListReplaceInt(&D->Sections, "DocPerSite", num);
  return UDM_OK;
}

 
static udm_rc_t
UdmImportSection(UDM_AGENT *A, UDM_DOCUMENT *D, UDM_DB *db)
{
  UDM_CONST_TEXTITEM ConstItem;
  UDM_TEXT_PARAM Param;
  UDM_VARLIST Vars;
  UDM_SQLRES SQLRes;
  UDM_DSTR d;
  udm_rc_t rc;
  size_t row, rows, cols;
  const char *fmt= UdmVarListFindStr(&D->Sections, "SQLImportSection", NULL);
  
  if (!fmt)
    return UDM_OK;
  
  UdmDSTRInit(&d, 1024);
  UdmVarListInit(&Vars);
  UdmVarListSQLEscape(A, &Vars, &D->Sections, db);
  UdmDSTRParse(&d, fmt, &Vars);
  UdmVarListFree(&Vars);
  if(UDM_OK!= (rc= UdmDBSQLQuery(A, db, &SQLRes, UdmDSTRPtr(&d))))
    return rc;

  cols= UdmSQLNumCols(&SQLRes);
  UdmConstTextItemInit(&ConstItem);
  UdmTextParamInit(&Param, UDM_TEXTLIST_FLAG_HTML, 0); /* TODO34: indexer.conf: format */
  for (row=0, rows= UdmSQLNumRows(&SQLRes); row < rows; row++)
  {
    size_t col;
    for (col= 0; col + 1 < cols; col+= 2)
    {
      const UDM_VAR *Sec;
      ConstItem.section_name.str= UdmSQLValue(&SQLRes, row, col);
      ConstItem.section_name.length= UdmSQLLen(&SQLRes, row, col);
      if ((Sec= UdmVarListFind(&D->Sections, ConstItem.section_name.str)))
      {
        ConstItem.text.str= UdmSQLValue(&SQLRes, row, col + 1);
        ConstItem.text.length= UdmSQLLen(&SQLRes, row, col + 1);
        Param.secno= UdmVarSecno(Sec);
        UdmTextListAddConst(&D->TextList, &ConstItem, &Param);
      }
    }
  }
  
  UdmDSTRFree(&d);
  UdmSQLFree(&SQLRes);
  return rc;
}


static udm_rc_t
UdmGetReferers(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  size_t    i,j;
  char    qbuf[2048];
  UDM_SQLRES  SQLres;
  const char  *where;
  udm_rc_t rc;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DB);
  if (UDM_OK != (rc= UdmSQLBuildWhereCondition(Indexer, db, Query, &where)))
    return rc;
  
  udm_snprintf(qbuf, sizeof(qbuf),
               "SELECT url.status,url2.url,url.url "
               "FROM url,url url2%s "
               "WHERE url.referrer=url2.rec_id %s %s",
               Query->from, where[0] ? "AND" : "", where);
  
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLres, qbuf)))
    return rc;
  
  j=UdmSQLNumRows(&SQLres);
  for(i=0;i<j;i++)
  {
    if(Indexer->Conf->RefInfo)Indexer->Conf->RefInfo(
      atoi(UdmSQLValue(&SQLres,i,0)),
      UdmSQLValue(&SQLres,i,2),
      UdmSQLValue(&SQLres,i,1)
    );
  }
  UdmSQLFree(&SQLres);
  return rc;
}


static udm_rc_t
UdmGetDocCount(UDM_AGENT * Indexer, UDM_DB *db)
{
  char    qbuf[200]="";
  UDM_SQLRES  SQLres;
  udm_rc_t rc;
  
  sprintf(qbuf,NDOCS_QUERY);
  if (UDM_OK!= (rc= UdmDBSQLQuery(Indexer, db, &SQLres, qbuf)))
    return rc;
  
  if(UdmSQLNumRows(&SQLres))
  {
    const char * s;
    s=UdmSQLValue(&SQLres,0,0);
    if(s)Indexer->doccount += atoi(s);
  }
  UdmSQLFree(&SQLres);
  return(UDM_OK);
}


static udm_rc_t
UdmStatActionSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  size_t    i,j,n;
  char    qbuf[2048];
  UDM_SQLRES  SQLres;
  int    have_group= (UdmSQLDBFlags(db) & UDM_SQL_HAVE_GROUPBY);
  const char  *where;
  udm_rc_t rc= UDM_OK;

  if(UdmSQLDBType(db)==UDM_DB_IBASE)
    have_group=0;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DB);
  if (UDM_OK != (rc= UdmSQLBuildWhereCondition(Indexer, db, Query, &where)))
    return rc;

  if(have_group)
  {
    char func[128];
    int time= (int) Query->StatList.time; /* To use %d on 64bit OSs */
    
    switch(UdmSQLDBType(db))
    {
      case UDM_DB_MYSQL:
        udm_snprintf(func, sizeof(func), "next_index_time<=%d", time);
        break;
    
      case UDM_DB_PGSQL:
      case UDM_DB_MSSQL:
      case UDM_DB_SYBASE:
      case UDM_DB_DB2:
      case UDM_DB_SQLITE:
      case UDM_DB_SQLITE3:
      default:
        udm_snprintf(func, sizeof(func),
                     "case when next_index_time<=%d then 1 else 0 end", time);
        break;

      case UDM_DB_ACCESS:
        udm_snprintf(func, sizeof(func),
                     "IIF(next_index_time<=%d, 1, 0)", time);
        break;

      case UDM_DB_ORACLE8:
      case UDM_DB_SAPDB:
        udm_snprintf(func, sizeof(func),
                     "DECODE(SIGN(%d-next_index_time),-1,0,1,1)", time);
        break;
    }

    udm_snprintf(qbuf, sizeof(qbuf) - 1,
                 "SELECT status, SUM(%s), count(*) FROM url%s %s%s GROUP BY status ORDER BY status",
                 func, Query->from, where[0] ? "WHERE " : "", where);

    if (UDM_OK!= (rc= UdmDBSQLQuery(Indexer, db, &SQLres, qbuf)))
      return rc;
    
    if ((n= UdmSQLNumRows(&SQLres)))
    {
      for (i = 0; i < n; i++)
      {
        for (j= 0; j < Query->StatList.nstats; j++)
        {
          UDM_STAT *Stat= &Query->StatList.Stat[j];
          if (Stat->status== atoi(UdmSQLValue(&SQLres,i,0)))
          {
            Stat->expired += atoi(UdmSQLValue(&SQLres,i,1));
            Stat->total += atoi(UdmSQLValue(&SQLres,i,2));
            break;
          }
        }
        if (j == Query->StatList.nstats)
        {
          UDM_STAT  *S;
          size_t nbytes= (Query->StatList.nstats + 1) * sizeof(Query->StatList.Stat[0]);
          Query->StatList.Stat= (UDM_STAT*) UdmRealloc(Query->StatList.Stat, nbytes);
          S= &Query->StatList.Stat[Query->StatList.nstats];
          S->status= atoi(UdmSQLValue(&SQLres,i,0));
          S->expired= atoi(UdmSQLValue(&SQLres,i,1));
          S->total= atoi(UdmSQLValue(&SQLres,i,2));
          Query->StatList.nstats++;
        }
      }
    }
    UdmSQLFree(&SQLres);  
  }
  else
  {
/*  
  FIXME: learn how to get it from SOLID and IBASE
  (HAVE_IBASE || HAVE_SOLID || HAVE_VIRT )
*/
    
    udm_snprintf(qbuf, sizeof(qbuf) - 1,
                 "SELECT status,next_index_time FROM url%s %s%s ORDER BY status",
                 Query->from, where[0] ? "WHERE " : "", where);

    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLres, qbuf)))
      return rc;
    
    for(i=0;i<UdmSQLNumRows(&SQLres);i++)
    {
      for (j= 0; j< Query->StatList.nstats; j++)
      {
        UDM_STAT *Stat= &Query->StatList.Stat[j];
        if (Stat->status == atoi(UdmSQLValue(&SQLres,i,0)))
        {
          if ((time_t) UDM_ATOU(UdmSQLValue(&SQLres, i, 1)) <=
              Query->StatList.time)
            Stat->expired++;
          Stat->total++;
          break;
        }
      }
      if (j == Query->StatList.nstats)
      {
        size_t nbytes= sizeof(UDM_STAT) * (Query->StatList.nstats + 1);
        Query->StatList.Stat= (UDM_STAT *) UdmRealloc(Query->StatList.Stat, nbytes);
        Query->StatList.Stat[j].status= UDM_ATOI(UdmSQLValue(&SQLres,i,0));
        Query->StatList.Stat[j].expired= 0;
        if ((time_t) UDM_ATOU(UdmSQLValue(&SQLres, i, 1)) <=
            Query->StatList.time)
          Query->StatList.Stat[j].expired++;
        Query->StatList.Stat[j].total=1;
        Query->StatList.nstats++;
      }
    }
    UdmSQLFree(&SQLres);
  }
  return rc;
}


static udm_rc_t
UdmURLInfoDumpDoc(UDM_AGENT *Indexer,
                  UDM_DB *db,
                  UDM_DOCUMENT *Doc,
                  const char *table,
                  const char *column_names)
{
  udm_rc_t rc;
  char buf[64];
  size_t i;
  UDM_SQLRES SQLRes;
  UDM_DSTR dbuf;
  urlid_t url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);

  udm_snprintf(buf, sizeof(buf),
               "SELECT %s FROM %s WHERE url_id=%d", column_names, table, url_id);
  if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, &SQLRes, buf)))
    return rc;
  
  UdmDSTRInit(&dbuf, 256);
  
  for (i= 0; i < UdmSQLNumRows(&SQLRes); i++)
  {
    size_t j;
    UDM_CONST_STR col[2];
    const UDM_CONST_STR *c= (const UDM_CONST_STR*) &col;
    for (j= 0; j < UdmSQLNumCols(&SQLRes); j++)
    {
      UDM_ASSERT(j < 2);
      UdmConstStrSet(&col[j], UdmSQLValue(&SQLRes, i, j), UdmSQLLen(&SQLRes, i, j));
    }
    UdmDocInsertSectionsUsingEscapeBuildQuery(Indexer, db, table,
                                              0, column_names,
                                              c, UdmSQLNumCols(&SQLRes),
                                              &dbuf);
    printf("%s;\n", UdmDSTRPtr(&dbuf));
  }
  UdmSQLFree(&SQLRes);
  UdmDSTRFree(&dbuf);
  return UDM_OK;
}


static udm_rc_t
UdmDumpData(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  char buf[256];
  UDM_SQLRES SQLRes;
  size_t i, nrows;
  udm_rc_t rc;
  const char *where;
  UDM_DSTR eurl;
  
  UDM_ASSERT(UdmSQLDBModeHandler(db)->DumpWordInfo != NULL);
  if (UDM_OK != (rc= UdmSQLBuildWhereCondition(A, db, Query, &where)))
    return rc;
  UdmDSTRInit(&eurl, 256);
  udm_snprintf(buf, sizeof(buf),
               "SELECT %s FROM url%s%s", select_url_str_for_dump,
               where[0] ? " WHERE " : "", where);
  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, buf)))
    goto ret;
  
  nrows= UdmSQLNumRows(&SQLRes);
  
  for(i= 0; i < nrows; i++)
  {
    UDM_DOCUMENT Doc;
    UdmDocInit(&Doc);
    if (UDM_OK != UdmTargetSQLResDump(A, db, &Doc, &SQLRes, i, &eurl))
      goto ret;
    if (UDM_OK != (rc= UdmURLInfoDumpDoc(A, db, &Doc, "urlinfo", "sname,sval")))
      goto ret;
    if (UDM_OK != (rc= UdmURLInfoDumpDoc(A, db, &Doc, "cachedcopy", "sval")))
      goto ret;
    if (UDM_OK != (rc= UdmSQLDBModeHandler(db)->DumpWordInfo(A, db, &Doc)))
      goto ret;
    UdmDocFree(&Doc);
  }
  
ret:
  UdmDSTRFree(&eurl);
  return rc;
}


static udm_rc_t
UdmRestoreData(UDM_AGENT *A, UDM_DOCUMENT *Doc, UDM_DB *db)
{
  size_t i;
  udm_rc_t rc;
  
  for (i= 0; i < Doc->Sections.nvars; i++)
  {
    UDM_VAR *S= UdmVarListFindByIndex(&Doc->Sections, i);
    if (UdmVarValueHandlerType(S) == UDM_VALUE_HANDLER_TYPE_STR)
    {
      UDM_SECTION *Sec= (UDM_SECTION *) UdmVarDataPtr(S);
      printf("%s[%d]=%s\n", UdmVarName(S),
             (int) UdmSectionLength(Sec), UdmSectionPtr(Sec));
    }
  }

  if (UDM_OK != (rc= UdmAddURL(A, Doc, db)))
    goto ex;
  if (UDM_OK != (rc= UdmFindURL(A, Doc, db)))
    goto ex;
  if (UDM_OK != (rc= UdmLongUpdateURL(A, Doc, db)))
    goto ex;

  printf("\n");

ex:
  return rc;
}


/******* "indexer -Ewordstat" - word statistics for suggestions *************/

udm_rc_t
UdmWordStatQuery(UDM_AGENT *A, UDM_DB *db, const char *src)
{
  udm_rc_t rc;
  UDM_SQLRES SQLRes;
  size_t row, rows;
  
  if (UDM_OK!= (rc= UdmDBSQLQuery(A, db, &SQLRes, src)))
    return rc;

  if (UDM_OK != (rc= UdmDBSQLLockOrBegin(A, db, "wrdstat WRITE")))
    return rc;

  rows=UdmSQLNumRows(&SQLRes);
  for(row=0 ; row < rows ; row++)
  {
    const char *word;
    int count;
    size_t wordlen;
    char snd[UDM_MAXWORDSIZE];
    char insert[64 + 2 * UDM_MAXWORDSIZE];
    /*
      Skip words that are longer than UDM_MAXWORDSIZE.
    */
    if ((wordlen= UdmSQLLen(&SQLRes, row, 0)) > sizeof(snd))
      continue;
    word= UdmSQLValue(&SQLRes, row, 0);
    count= UDM_ATOI(UdmSQLValue(&SQLRes, row, 1));
    UdmSoundex(A->Conf->lcs, snd, sizeof(snd), word, wordlen);
    if (snd[0])
    {
      udm_snprintf(insert, sizeof(insert),
                   "INSERT INTO wrdstat (word, snd, cnt) VALUES ('%s','%s',%d)",
                   word, snd, count);
      if (UDM_OK!= (rc= UdmDBSQLQuery(A, db, NULL, insert)))
        return rc;
    }
    if (((row % 1000) == 999) && row + 100 < rows)
    {
      if (UDM_OK != (rc= UdmDBSQLUnlockOrCommit(A, db)))
        return rc;
      if (UDM_OK != (rc= UdmDBSQLLockOrBegin(A, db, "wrdstat WRITE")))
        return rc;
    }
  }
  UdmSQLFree(&SQLRes);
  if (UDM_OK != (rc= UdmDBSQLUnlockOrCommit(A, db)))
    return rc;
  return UDM_OK;
}


static udm_rc_t
UdmWordStatCreate(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDBSQLTableTruncateOrDelete(A, db, "wrdstat")))
    return rc;
  UDM_ASSERT(UdmSQLDB(db)->dbmode_handler->QueryAction != NULL);
  rc= UdmSQLDB(db)->dbmode_handler->QueryAction(A, db, Query, UDM_QUERYCMD_WORDSTAT);
  return rc;
}

/******************* create and drop ***********************/
typedef struct
{
  UDM_AGENT *Agent;
  FILE *infile;
} UDM_CREATE_DROP_PARAM;


static udm_rc_t
sqlmonprompt_create_or_drop(UDM_IOHANDLER *iohandler,
                            udm_msg_t msgtype, const char *msg)
{
  UDM_SQLMON_PARAM *prm= (UDM_SQLMON_PARAM*) iohandler->user_data;
  UDM_CREATE_DROP_PARAM *prm2= (UDM_CREATE_DROP_PARAM *) prm->user_data;
  int level= msgtype == UDM_MSG_ERROR ? UDM_LOG_ERROR : UDM_LOG_EXTRA;
  if (msgtype == UDM_MSG_ERROR)
    UdmLog(prm2->Agent, level, "ERROR at line %d: %s",
           (int) prm->lineno + 1, msg);
  else
    UdmLog(prm2->Agent, level, "%s", msg);
  return UDM_OK;
}


static char *
sqlmongets_create_or_drop(UDM_IOHANDLER *iohandler, char *str, size_t size)
{
  UDM_SQLMON_PARAM *prm= (UDM_SQLMON_PARAM*) iohandler->user_data;
  UDM_CREATE_DROP_PARAM *prm2= (UDM_CREATE_DROP_PARAM *) prm->user_data;
  if (!fgets(str, size, prm2->infile))
    return 0;
  return str;
}


static const char*
UdmCreateOrDropCmdStr(udm_dbcmd_t cmd)
{
  switch(cmd)
  {
    case UDM_DBCMD_CREATE: return "create";
    case UDM_DBCMD_DROP:   return "drop";
    default: return "";
  }
  return "unknown_cmd";
}


static udm_rc_t
UdmCreateOrDropSQL(UDM_AGENT *A, UDM_DB *db, udm_dbcmd_t cmd)
{
  char fname[1024];
  const char *sdir= UdmVarListFindStr(&A->Conf->Vars, "ShareDir", UDM_SHARE_DIR);
#ifdef HAVE_FHS_LAYOUT
  const char *sdir2= "create" UDMSLASHSTR;
#else
  const char *sdir2= "";
#endif
  UDM_SQLMON_PARAM prm;
  UDM_CREATE_DROP_PARAM prm2;

  udm_snprintf(fname,sizeof(fname),"%s%s%s%s%s%s.%s.sql",
               sdir, UDMSLASHSTR, sdir2,
               UdmSQLDBTypeToStr(UdmSQLDBType(db)), UDMSLASHSTR,
               UdmCreateOrDropCmdStr(cmd),
               UdmSQLDBModeToStr(UdmSQLDBMode(db)));
  UdmLog(A, UDM_LOG_ERROR, "Running '%s'", fname);
  prm2.Agent= A;
  if (!(prm2.infile= fopen(fname,"r")))
  {
    sprintf(A->Conf->errstr, "Can't open file '%s'", fname);
    return UDM_ERROR;
  }
  bzero((void*)&prm,sizeof(prm));
  prm.currdbnum= db - A->Conf->DBList.Item;
  prm.flags= UDM_SQLMON_DISPLAY_FIELDS;
  prm.iohandler.gets= sqlmongets_create_or_drop;
  prm.iohandler.prompt= sqlmonprompt_create_or_drop;
  prm.iohandler.user_data= &prm;
  prm.user_data= &prm2;
  UdmSQLMonitor(A, A->Conf,&prm);
  UdmLog(A, UDM_LOG_ERROR, "%d queries sent, %d succeeded, %d failed",
         (int) prm.nqueries, (int) prm.ngood, (int) prm.nbad);
  fclose(prm2.infile);
  return UDM_OK;
}


/******************* URL handlers **************************/

static udm_rc_t
UdmDocActionSQL(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc, udm_doccmd_t cmd)
{
  switch (cmd)
  {
    case UDM_DOCCMD_DELETE:          return UdmDeleteURL(A, Doc, db);
    case UDM_DOCCMD_ADD:             return UdmAddURL(A, Doc, db);
    case UDM_DOCCMD_SUPDATE:         return UdmUpdateUrl(A, Doc, db);
    case UDM_DOCCMD_LUPDATE:         return UdmLongUpdateURL(A, Doc, db);
    case UDM_DOCCMD_DUPDATE:         return UdmDeleteWordsAndLinks(A, Doc, db);
    case UDM_DOCCMD_UPDCLONE:        return UdmUpdateClone(A, Doc, db);
    case UDM_DOCCMD_REGCHILD:        return UdmRegisterChild(A, Doc, db);
    case UDM_DOCCMD_FINDBYURL:       return UdmFindURL(A, Doc, db);
    case UDM_DOCCMD_FINDBYMSG:       return UdmFindMessage(A, Doc, db);
    case UDM_DOCCMD_FINDORIG:        return UdmFindOrigin(A, Doc, db);
    case UDM_DOCCMD_GET_CACHED_COPY: return UdmGetCachedCopy(A, Doc, db);
    case UDM_DOCCMD_DOCPERSITE:      return UdmDocPerSite(A, Doc, db);
    case UDM_DOCCMD_SQLIMPORTSEC:    return UdmImportSection(A, Doc, db);
    case UDM_DOCCMD_RESTOREDATA:     return UdmRestoreData(A, Doc, db);
  }
  return UDM_ERROR;
}


typedef struct udm_sqldb_driver_st
{
  const char *name;
  udm_sqldbtype_t DBType;
  udm_sqldbapi_t  DBDriver;
  int DBSQL_IN;
  int flags;
  const UDM_SQLDB_HANDLER *handler;
} UDM_SQLDB_DRIVER;


static const UDM_SQLDB_DRIVER SQLDriver[]=
{
#if (HAVE_ORACLE8)
  {
    "oracle8", UDM_DB_ORACLE8, UDM_DBAPI_ORACLE8, 1, 
    UDM_SQL_HAVE_GROUPBY   | UDM_SQL_HAVE_TRUNCATE |
    UDM_SQL_HAVE_SUBSELECT | UDM_SQL_HAVE_BIND |
    UDM_SQL_HAVE_ROWNUM    | UDM_SQL_HAVE_GOOD_COMMIT | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_RENAME    | UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_oracle_handler
  },
  {
    "oracle", UDM_DB_ORACLE8, UDM_DBAPI_ORACLE8, 1, 
    UDM_SQL_HAVE_GROUPBY   | UDM_SQL_HAVE_TRUNCATE |
    UDM_SQL_HAVE_SUBSELECT | UDM_SQL_HAVE_BIND |
    UDM_SQL_HAVE_ROWNUM    | UDM_SQL_HAVE_GOOD_COMMIT | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_RENAME    | UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_oracle_handler
  },
#endif
#if (HAVE_CTLIB)
  {
    "mssql", UDM_DB_MSSQL, UDM_DBAPI_CTLIB, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_TRUNCATE |
    UDM_SQL_HAVE_TOP     | UDM_SQL_HAVE_0xHEX | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_RENAME  | UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_ctlib_handler
  },
  {
    "sybase", UDM_DB_SYBASE, UDM_DBAPI_CTLIB, 1,
    UDM_SQL_HAVE_GROUPBY | /*UDM_SQL_HAVE_TRUNCATE |*/
    /*
      Don't use TRUNCATE with Sybase.
      It gives error:
      'TRUNCATE TABLE command not allowed within multi-statement
      transaction.
      TODO: modify the code to use TRUNCATE outside a transaction
    */
    UDM_SQL_HAVE_TOP     | UDM_SQL_HAVE_0xHEX    |
    UDM_SQL_HAVE_GOOD_COMMIT | UDM_SQL_HAVE_TRANSACT
    /*
    Something goes wrong with sp_rename!
    UDM_SQL_HAVE_RENAME  | UDM_SQL_HAVE_CREATE_LIKE
    */
     ,
    &udm_sqldb_ctlib_handler
  },
#endif
#if (HAVE_MYSQL)
  { 
    "mysql", UDM_DB_MYSQL, UDM_DBAPI_MYSQL, 1,
    UDM_SQL_HAVE_BIND  |
    UDM_SQL_HAVE_LIMIT | UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_0xHEX |
    UDM_SQL_HAVE_RENAME | UDM_SQL_HAVE_CREATE_LIKE |
    UDM_SQL_HAVE_DROP_IF_EXISTS,
    &udm_sqldb_mysql_handler
  },
#endif
#if (HAVE_PGSQL)
  {
    "pgsql", UDM_DB_PGSQL, UDM_DBAPI_PGSQL, 1,
    UDM_SQL_HAVE_BIND  |
    UDM_SQL_HAVE_LIMIT | UDM_SQL_HAVE_GROUPBY |
    UDM_SQL_HAVE_SUBSELECT | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_RENAME | UDM_SQL_HAVE_CREATE_LIKE,
    /* UDM_SQL_HAVE_DROP_IF_EXISTS depends on server version */
    &udm_sqldb_pgsql_handler,
  },
#endif
#if (HAVE_IBASE)
  {
    "ibase", UDM_DB_IBASE, UDM_DBAPI_IBASE, 0,
    /* 
    while indexing large sites and using the SQL in statement 
    interbase will fail when the items in the in IN statements
    are more then 1500. We'd better have to fix code to avoid 
    big INs instead of hidding DBSQL_IN.
    */
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_BIND |
    UDM_SQL_HAVE_FIRST_SKIP | UDM_SQL_HAVE_TRANSACT | UDM_SQL_HAVE_GOOD_COMMIT,
    &udm_sqldb_ibase_handler,
  },
#endif
#if (HAVE_SQLITE)
  {
    "sqlite",
    UDM_DB_SQLITE, UDM_DBAPI_SQLITE, 1,
    UDM_SQL_HAVE_BIND  |
    UDM_SQL_HAVE_LIMIT | UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_sqlite_handler,
  },
#endif
#if (HAVE_SQLITE3)
  {
    "sqlite3",
    UDM_DB_SQLITE3, UDM_DBAPI_SQLITE3, 1,
    UDM_SQL_HAVE_BIND   |
    UDM_SQL_HAVE_LIMIT  | UDM_SQL_HAVE_GROUPBY |
    UDM_SQL_HAVE_STDHEX | UDM_SQL_HAVE_GOOD_COMMIT | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_DROP_IF_EXISTS | UDM_SQL_HAVE_RENAME |
    UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_sqlite3_handler,
  },
#endif
#if (HAVE_ODBC)
  {
    "odbc-solid", UDM_DB_SOLID, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_BIND | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-sapdb", UDM_DB_SAPDB, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_BIND | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-db2", UDM_DB_DB2, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_BIND |
    UDM_SQL_HAVE_BIND    | UDM_SQL_HAVE_STDHEX | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_RENAME | UDM_SQL_HAVE_CREATE_LIKE,    
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-access", UDM_DB_ACCESS, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_SUBSELECT |
    UDM_SQL_HAVE_0xHEX   | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-mimer", UDM_DB_MIMER, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_SUBSELECT |
    UDM_SQL_HAVE_BIND | UDM_SQL_HAVE_STDHEX | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-cache", UDM_DB_CACHE, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_SUBSELECT |
    UDM_SQL_HAVE_BIND    | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
  { 
    "odbc-virtuoso", UDM_DB_VIRT, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_SUBSELECT |
    UDM_SQL_HAVE_BIND    | UDM_SQL_HAVE_TRANSACT | UDM_SQL_HAVE_TOP,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-oracle", UDM_DB_ORACLE8, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY   | UDM_SQL_HAVE_TRUNCATE |
    UDM_SQL_HAVE_SUBSELECT | UDM_SQL_HAVE_BIND |
    UDM_SQL_HAVE_ROWNUM    | UDM_SQL_HAVE_GOOD_COMMIT |
    UDM_SQL_HAVE_TRANSACT  |
    UDM_SQL_HAVE_RENAME    | UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-oracle8", UDM_DB_ORACLE8, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY   | UDM_SQL_HAVE_TRUNCATE |
    UDM_SQL_HAVE_SUBSELECT | UDM_SQL_HAVE_BIND | UDM_SQL_HAVE_ROWNUM |
    UDM_SQL_HAVE_TRANSACT  |
    UDM_SQL_HAVE_RENAME    | UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-mssql", UDM_DB_MSSQL, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_TRUNCATE |
    UDM_SQL_HAVE_TOP     | UDM_SQL_HAVE_0xHEX | UDM_SQL_HAVE_TRANSACT |
    UDM_SQL_HAVE_RENAME  | UDM_SQL_HAVE_CREATE_LIKE,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-sybase", UDM_DB_SYBASE, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | /*UDM_SQL_HAVE_TRUNCATE |*/
    UDM_SQL_HAVE_TOP     | UDM_SQL_HAVE_0xHEX |
    UDM_SQL_HAVE_TRANSACT| UDM_SQL_HAVE_GOOD_COMMIT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-mysql", UDM_DB_MYSQL, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_BIND  |
    UDM_SQL_HAVE_LIMIT | UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_0xHEX |
    UDM_SQL_HAVE_RENAME | UDM_SQL_HAVE_CREATE_LIKE |
    UDM_SQL_HAVE_DROP_IF_EXISTS,
    &udm_sqldb_odbc_handler,
  },
  {
    /* Bind does not seem to work with BYTEA in Windows */
    "odbc-pgsql", UDM_DB_PGSQL, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_LIMIT      | UDM_SQL_HAVE_GROUPBY |
    UDM_SQL_HAVE_SUBSELECT  /*| UDM_SQL_HAVE_BIND*/|
    UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-ibase", UDM_DB_IBASE, UDM_DBAPI_ODBC, 0,
    /* 
    while indexing large sites and using the SQL in statement 
    interbase will fail when the items in the in IN statements
    are more then 1500. We'd better have to fix code to avoid 
    big INs instead of hidding DBSQL_IN.
    */
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_BIND |
    UDM_SQL_HAVE_FIRST_SKIP | UDM_SQL_HAVE_TRANSACT | UDM_SQL_HAVE_GOOD_COMMIT,
    &udm_sqldb_odbc_handler,
  },
  {
    "odbc-monetdb", UDM_DB_MONETDB, UDM_DBAPI_ODBC, 1,
    UDM_SQL_HAVE_GROUPBY | UDM_SQL_HAVE_SUBSELECT | UDM_SQL_HAVE_LIMIT |
    UDM_SQL_HAVE_BIND |
    /*UDM_SQL_HAVE_BIND_TEXT | */UDM_SQL_HAVE_BLOB_AS_HEX | UDM_SQL_HAVE_TRANSACT,
    &udm_sqldb_odbc_handler,
  },
#endif
  {
    NULL, UDM_DB_MYSQL, UDM_DBAPI_MYSQL, 0, 0, NULL
  }
};


static const UDM_SQLDB_DRIVER *UdmSQLDriverByName(const char *name)
{
  const UDM_SQLDB_DRIVER *drv;
  for (drv= SQLDriver; drv->name; drv++)
  {
    if (!strcasecmp(name, drv->name))
      return drv;
    if (!strncasecmp(drv->name, "odbc-", 5) &&
        !strcasecmp(drv->name + 5, name))
      return drv;
  }
  return NULL;
}


static const UDM_DBMODE_HANDLER *
UdmSQLDBModeHandlerByID(int DBMode)
{
  switch (DBMode)
  {
    case UDM_SQLDBMODE_BLOB:
      return &udm_dbmode_handler_blob;
    case UDM_SQLDBMODE_SINGLE:
      return &udm_dbmode_handler_single;
    case UDM_SQLDBMODE_MULTI:
      return &udm_dbmode_handler_multi;
    case UDM_SQLDBMODE_RAWBLOB:
      return &udm_dbmode_handler_rawblob;
  }
  UDM_ASSERT(0);
  return NULL;
}


static udm_rc_t
UdmDBSetParam(UDM_DB *db, char *param)
{
  char *tok, *lt;
  
  for(tok = udm_strtok_r(param, "&",&lt) ; tok ; 
      tok = udm_strtok_r(NULL,"&",&lt))
  {
    char * val;
    if((val=strchr(tok,'=')))
    {
      *val++='\0';
      UdmVarListReplaceStr(UdmSQLDBVars(db), tok, val);
    }
    else
    {
      UdmVarListReplaceStr(UdmSQLDBVars(db), tok, "");
    }
  }
  return UDM_OK;
}



static udm_rc_t
UdmDBSetAddrCommon(UDM_DB *db, UDM_URL *addr)
{
  char *s;
  if (addr->auth)
  {
    /*
      Unescape user and password to allow URL specific
      characters like '"<>@#? to be used as user or password part.
      
      It's safe to spoil addr->auth here, as we don't
      need it anymore after setting DBUser and DBPass
    */
    
    if ((s= strchr(addr->auth,':')))
    {
      *s++= 0;
      UdmUnescapeCGIQuery(s, s);
      UdmVarListReplaceStr(UdmSQLDBVars(db), "DBPass", s);
    }
    UdmUnescapeCGIQuery(addr->auth, addr->auth);
    UdmVarListReplaceStr(UdmSQLDBVars(db), "DBUser", addr->auth);
  }
  
  UdmVarListReplaceStr(UdmSQLDBVars(db), "DBHost", addr->hostname);
  if (addr->port)
    UdmVarListReplaceInt(UdmSQLDBVars(db), "DBPort", addr->port);

  if((s = strchr(UDM_NULL2EMPTY(addr->filename), '?')))
  {
    *s++='\0';
    if (UDM_OK != UdmDBSetParam(db, s))
      return UDM_ERROR;
    UdmVarListReplaceStr(UdmSQLDBVars(db), "filename", addr->filename);
  }
  else
  {
    UdmVarListReplaceStr(UdmSQLDBVars(db), "filename", addr->filename);
  }
  
  return UDM_OK;
}


static udm_rc_t
UdmDBSetAddrSQLFromURL(UDM_DB *db, UDM_URL *addr, const char *dbaddr)
{
  const char *v;
  const UDM_SQLDB_DRIVER *drv= UdmSQLDriverByName(addr->schema);  
  if (!drv)
    return UDM_NOTARGET;

  if (!(db->specific= UdmMalloc(sizeof(UDM_SQLDB))))
    return UDM_ERROR;
  bzero(db->specific, sizeof(UDM_SQLDB));

  UdmVarListReplaceStr(UdmSQLDBVars(db), "DBAddr", dbaddr);

  if (UdmDBSetAddrCommon(db, addr))
    return UDM_ERROR;

  UdmSQLDB(db)->DBMode= UDM_SQLDBMODE_BLOB;
  UdmSQL(db)->DBType= drv->DBType;
  UdmSQL(db)->DBDriver= drv->DBDriver;
  UdmSQL(db)->DBSQL_IN= drv->DBSQL_IN;
  UdmSQL(db)->flags= drv->flags;
  UdmSQL(db)->handler= drv->handler[0];
  db->dbhandler= &udm_dbhandler_sql;

  if ((v= UdmVarListFindStr(UdmSQLDBVars(db),"dbmode",NULL)))
  {
    udm_bool_t error;
    UdmSQLDB(db)->DBMode= UdmStr2DBMode(v, &error);
    if (error)
    {
      udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db), "Unknown DBMode");
      UDM_FREE(db->specific);
      return UDM_ERROR;
    }
  }
  
  UdmSQLDB(db)->dbmode_handler= UdmSQLDBModeHandlerByID(UdmSQLDBMode(db));
  
  if ((v= UdmVarListFindStr(UdmSQLDBVars(db),"dbmodesearch",NULL)))
  {
    udm_bool_t error;
    udm_sqldbmode_t DBMode= UdmStr2DBMode(v, &error);
    if (error)
    {
      udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db), "Unknown DBModeSearch");
      UDM_FREE(db->specific);
      return UDM_ERROR;
    }
    if (DBMode == UDM_SQLDBMODE_BLOB  &&
        UdmSQLDBType(db) != UDM_DB_MYSQL &&
        UdmSQLDBType(db) != UDM_DB_SYBASE &&
        UdmSQLDBType(db) != UDM_DB_MSSQL &&
        UdmSQLDBType(db) != UDM_DB_MIMER &&
        UdmSQLDBType(db) != UDM_DB_ORACLE8 &&
        UdmSQLDBType(db) != UDM_DB_DB2 &&
        UdmSQLDBType(db) != UDM_DB_PGSQL &&
        UdmSQLDBType(db) != UDM_DB_IBASE &&
        UdmSQLDBType(db) != UDM_DB_VIRT &&
        UdmSQLDBType(db) != UDM_DB_SQLITE3 &&
        UdmSQLDBType(db) != UDM_DB_MONETDB)
    {
      udm_snprintf(UdmDBSQLError(db), UdmDBSQLErrorSize(db),
                   "This DBMode=blob is not supported with this database");
      UDM_FREE(db->specific);
      return UDM_ERROR;
    }
  }

  if((v= UdmVarListFindStr(UdmSQLDBVars(db), "debugsql", "no")))
  {
    if (!strcasecmp(v, "yes")) 
      UdmSQL(db)->flags|= UDM_SQL_DEBUG_QUERY;
  }

  if(UdmSQLDBDriver(db) == UDM_DBAPI_IBASE ||
     UdmSQLDBDriver(db) == UDM_DBAPI_SQLITE ||
     UdmSQLDBDriver(db) == UDM_DBAPI_SQLITE3)
  {
    /*
      Ibase is a special case:
      It's database name consists of full path and file name        
    */
    UdmVarListReplaceStr(UdmSQLDBVars(db), "DBName", UDM_NULL2EMPTY(addr->path));
  }
  else
  {
    /*
      ODBC Data Source Names may contain space and
      other tricky characters, let's unescape them.
    */
    size_t len= strlen(UDM_NULL2EMPTY(addr->path));
    char *DBName= (char*) UdmMalloc(len + 1);
    char *src= (char*) UdmMalloc(len + 1);
    src[0]= '\0';
    sscanf(UDM_NULL2EMPTY(addr->path), "/%[^/]s", src);
    UdmUnescapeCGIQuery(DBName, src);
    UdmVarListReplaceStr(UdmSQLDBVars(db), "DBName", DBName);
    UdmFree(src);
    UdmFree(DBName);
  }
  
  if (UdmVarListFindInt(UdmSQLDBVars(db), "ps", 0) == 123)
  {
    UdmSQL(db)->handler.Prepare= UdmSQLPrepareGeneric;
    UdmSQL(db)->handler.Bind= UdmSQLBindGeneric; 
    UdmSQL(db)->handler.Exec= UdmSQLExecGeneric;
    UdmSQL(db)->handler.StmtFree= UdmSQLStmtFreeGeneric;
    UdmSQL(db)->flags|= UDM_SQL_HAVE_BIND;
  }
  else if ((UdmSQLDBType(db) == UDM_DB_MSSQL   ||
            UdmSQLDBType(db) == UDM_DB_SYBASE  ||
            UdmSQLDBType(db) == UDM_DB_MYSQL   ||
            UdmSQLDBType(db) == UDM_DB_PGSQL   ||
            UdmSQLDBType(db) == UDM_DB_SQLITE  ||
            UdmSQLDBType(db) == UDM_DB_SQLITE3)&&
            !strcasecmp(UdmVarListFindStr(UdmSQLDBVars(db), "ps", ""), "none"))
  {
    UdmSQL(db)->flags&= (0x7FFFFFFF ^ UDM_SQL_HAVE_BIND);
  }
  else if ((UdmSQLDBDriver(db) == UDM_DBAPI_MYSQL ||
            UdmSQLDBDriver(db) == UDM_DBAPI_PGSQL ||
            UdmSQLDBDriver(db) == UDM_DBAPI_ODBC) &&
            UdmSQL(db)->handler.Exec &&
            UdmVarListFindBool(UdmSQLDBVars(db), "ps", UDM_FALSE))
  {
    UdmSQL(db)->flags|= UDM_SQL_HAVE_BIND;
  }

  return UDM_OK;
}


static udm_rc_t
UdmDBSetAddrSQL(UDM_DB *db, const char *dbaddr)
{
  udm_rc_t rc= UDM_NOTARGET;
  UDM_URL addr;
  UdmURLInit(&addr);
  if (!UdmURLParse(&addr, dbaddr) && addr.schema)
    rc= UdmDBSetAddrSQLFromURL(db, &addr, dbaddr);
  UdmURLFree(&addr);
  return rc;
}


static udm_rc_t
UdmDBCloseSQL(UDM_DB *db)
{
  if (db->specific)
  {
    if (UdmSQLDBConnected(db))
      UdmSQLClose(UdmSQL(db));
    UdmWordCacheInit(&UdmSQLDB(db)->WordCache);
    UdmVarListFree(UdmSQLDBVars(db));
    UdmURLIdCacheFree(&UdmSQLDB(db)->URLIdCache);
    UDM_FREE(db->specific);
  }
  return UDM_OK;
}  


static udm_rc_t
UdmDBActionSQL(UDM_AGENT *A, UDM_DB *db, udm_dbcmd_t cmd)
{
  switch (cmd)
  {
    case UDM_DBCMD_CREATE:    return UdmCreateOrDropSQL(A, db, cmd);
    case UDM_DBCMD_DROP:      return UdmCreateOrDropSQL(A, db, cmd);
    case UDM_DBCMD_DOCCOUNT:  return UdmGetDocCount(A, db);
    case UDM_DBCMD_CRAWLER_FINALIZE:   return UdmWordCacheWrite(A, db, 0);
  }
  return UDM_ERROR;
}


static udm_rc_t
UdmDBInfoSQL(UDM_DB *db, void *dst, size_t dstlen, size_t *bytes_written,
             udm_dbinfo_t info)
{
  switch (info)
  {
    case UDM_DBINFO_IS_THREAD_SAFE:
      if (!dstlen)
        return UDM_ERROR;
      /* TODO: link against libmysqlclient_r */
      ((udm_bool_t *) dst)[0]= UDM_FALSE;
      *bytes_written= 1;
      return UDM_OK;
    case UDM_DBINFO_ERRCODE:
      if (dstlen < sizeof(int))
        return UDM_ERROR;
      ((int *)dst)[0]= UdmSQL(db)->errcode;
      *bytes_written= sizeof(int);
      return UDM_OK;
    case UDM_DBINFO_ADDR:
    {
      const char *dbaddr= UdmVarListFindStr(UdmSQLDBVars(db), "DBAddr", "<noaddr>");
      if (dstlen < 1)
        return UDM_ERROR;
      *bytes_written= udm_snprintf((char*) dst, dstlen, "%s", dbaddr);
      return UDM_OK;
    }
    case UDM_DBINFO_ERRSTR:
    {
      if (dstlen < 1)
        return UDM_ERROR;
      *bytes_written= udm_snprintf((char*) dst, dstlen, "%s", UdmDBSQLError(db));
      return UDM_OK;
    }
  }
  return UDM_ERROR;
}


static udm_rc_t
UdmRewriteURLData(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  UDM_URLID_LIST limit;
  bzero((void*) &limit, sizeof(limit));
  if (UDM_OK != (rc= UdmLoadURLDataFromURLForConv(A, db, Query, &limit)))
    return rc;
  return udm_dbmode_handler_blob.QueryAction(A, db, Query, UDM_QUERYCMD_REWRITE_URLDATA);
}


static udm_rc_t
UdmRewritePopularity(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc;
  UDM_URLID_LIST limit;
  bzero((void*) &limit, sizeof(limit));
  if (UDM_OK != (rc= UdmLoadURLDataFromURLForConv(A, db, Query, &limit)))
    return rc;
  return udm_dbmode_handler_blob.QueryAction(A, db, Query, UDM_QUERYCMD_REWRITE_POPULARITY);
}


static udm_rc_t
UdmQueryActionSQL(UDM_AGENT *A, UDM_DB *db,
                  UDM_QUERY *Query, udm_querycmd_t cmd)
{
  switch (cmd)
  {
    case UDM_QUERYCMD_FINDWORDS: return UdmFindWordsSQL(A, db, Query);
    case UDM_QUERYCMD_SUGGEST:   return UdmQuerySuggest(A, db, Query);
    case UDM_QUERYCMD_DOCINFO:   return UdmQueryAddDocInfoSQL(A, db, Query);
    case UDM_QUERYCMD_TRACK:     return UdmTrackSQL(A, db, Query);
    case UDM_QUERYCMD_CLONES:    return UdmQueryClones(A, db, Query);
    case UDM_QUERYCMD_WORDFORMS: return UdmQueryWordForms(A, db, Query);
    case UDM_QUERYCMD_TARGETS:   return UdmTargetsSQL(A, db, Query);
    case UDM_QUERYCMD_CLEAR:     return UdmClearDBSQL(A, db, Query);
    case UDM_QUERYCMD_INDEX:     return UdmIndexSQL(A, db, Query);
    case UDM_QUERYCMD_EXPORT:    return UdmExportSQL(A, db, Query);
    case UDM_QUERYCMD_EXPIRE:    return UdmMarkForReindex(A, db, Query);
    case UDM_QUERYCMD_REFERERS:  return UdmGetReferers(A, db, Query);
    case UDM_QUERYCMD_WORDSTAT:  return UdmWordStatCreate(A, db, Query);
    case UDM_QUERYCMD_DUMPDATA:  return UdmDumpData(A, db, Query);
    case UDM_QUERYCMD_STATISTICS:return UdmStatActionSQL(A, db, Query);
    case UDM_QUERYCMD_REWRITE_URLDATA:    return UdmRewriteURLData(A, db, Query);
    case UDM_QUERYCMD_REWRITE_POPULARITY: return UdmRewritePopularity(A, db, Query);
    case UDM_QUERYCMD_REWRITE_LIMITS:  
      return udm_dbmode_handler_blob.QueryAction(A, db, Query, cmd);
  }
  return UDM_ERROR;
}


const UDM_DBHANDLER udm_dbhandler_sql=
{
  UdmDBSetAddrSQL,     /*udm_rc_t (*Init)       (UDM_DB*, const char *addr);*/
  UdmDBCloseSQL,       /*udm_rc_t (*Close)      (UDM_DB*);*/
  UdmDBInfoSQL,        /*udm_rc_t (*Info)          (UDM_DB *db, void *dst, size_t dstlen, size_t *bytes_written, udm_dbhandler_info_t info);*/
  UdmQueryActionSQL,   /*udm_rc_t (*QueryAction)   (UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query, udm_querycmd_t cmd);*/
  UdmDBActionSQL,      /*udm_rc_t (*DBAction)      (UDM_AGENT *A, UDM_DB *db, udm_dbcmd_t cmd);*/
  UdmDocActionSQL,     /*udm_rc_t (*DocumentAction)(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *D, udm_doccmt_t cmd);*/
  UdmHrefActionSQL,    /*udm_rc_t (*HrefAction)    (UDM_AGENT *A, UDM_DB *db, UDM_HREF *H, udm_hrefcmd_t cmd)*/
  UdmSrvActionSQL,     /*udm_rc_t (*ServerAction)  (UDM_AGENT *A, UDM_DB *db, UDM_SERVERLIST *Srv, udm_srvcmd_t cmd);*/
};


#endif /* HAVE_SQL */
